(* ElGamal and Hash ElGamal encryptions (an RRF example) *)

module Lib
open Bytes 

type bigint = int (* actually nicely defined in .NET *)
val bigint: x:int -> y:bigint{x=y}
type nat = i:bigint{i >= 0}
type pos = i:bigint{i > 0}
val bytes2bigint : bytes -> bigint
(* let bigint_of_bytes (b:bytes) : bigint = *)
(*   Array.fold (fun x y -> bigint 256 * x + (bigint (int(y:byte)))) (bigint 0) b *)
val bigint2bytes : bigint -> bytes

extern reference BigInt {language="C#";
                         dll="mscorlib";
                         namespace="System";
                         classname="Numerics"}
extern BigInt type bigint 


(*
val xGCD : x:nat -> n:nat -> a:nat * b:nat{ 1 = (a*x + b*n)  }
let rec xGCD x n =
  if n = bigint 0 then (bigint 1, bigint 0)
  else
    let (q, r) = ((x / n), (x % n)) in
    let (s, t) = xGCD n r in
    (t, s - q * t)
*)

type intmod (n:bigint) = i:bigint{0 <= i && i < n}
logic val Modulo: x:bigint -> n:pos -> m:intmod n { exists p. Eq  (x + p*n) m }

(* Default is to not export lets only to export val *)
val invMod : n:bigint -> x:intmod n -> y:intmod n{1 = ((x * y) % n)}
(*
let invMod n x = (* (n:bigint) (x:intmod n) : y:intmod n{1 = (Modulo (x * y) n)} =  *)
  let (a,b) = xGCD x n in
  let r = a % n in
  if r < bigint 0 then r+n else r 
*)
end 

module Group 
open Lib

logic val p : pos
let p:pos = 2*3*5*7*11*13*17 - 1 (* a large prime group *)

type elt = intmod p

val mlt: x:bigint -> y:bigint -> r:bigint {r = ((x*y) % p) }
let mlt x y = (x * y) % p

val div : elt -> elt -> elt
let div x y = x * (invMod p y)

val exp : elt -> nat -> elt
let rec exp x n = 
  if n = 1 then 
    x
  else 
    let d = n / 2 in
    let r = n % 2 in
    let a = exp x d in
    let a2 = mlt a a in
    if r = 0 then a2 else mlt a2 x

val g:elt
let g = 17 

 end

module DDH 
open Lib
open Group

(* abstracting over the use of DH in ElGamal,  *)
(* so that we can apply DDH *)

val q:nat
let q = 2*3*5*7*11*13 - 1 
  
type exn = intmod q 
    
(* single instance,  *)
(* we need x, y, z abstract & linearly generated *)
(* Maybe we don't implement private abbrevs ... should add *)
(*  three distinct abstract types *)
private type x = exn 
private type y = exn
private type z = exn

(* May be nice to have these be functions, but they must be affine 
   and top-level lets cannot be affine *)
val sample_x : x 
let sample_x = sample q 

val sample_y : y
let sample_y = sample q 

val sample_z : z
let sample_z = sample q 
 
(*
  private function val Exp(e,x) = exp(e,x) // an abstract logical function

  // needed only for the correctness of decryption
  theorem !e,x,y. Exp(Exp(e,x),y) = Exp(e,Prod(x,y)) 


  // a function to abstract over the g^xy vs g^z in ElGamal encryption
  val exp_ddh: e:group -> y:y -> z:z -> e':elt { !x,y,z. e = Exp(g,x) => e' = < Exp(gx,y) , Exp(g,z) > } 
*) 

logic (* private *) val ExpX : elt -> x -> elt
logic (* private *) val ExpY : elt -> y -> elt
logic (* private *) val ExpZ : elt -> z -> elt
(* private assume forall e x. (ExpX e x) = (exp e x) *)
(* private assume forall e y. (ExpY e y) = (exp e y) *)
(* private assume forall e z. (ExpZ e z) = (exp e z) *)

val exp_x: e:elt -> x:x -> r:elt { r = (ExpX e x) } 
let exp_x = exp 

val exp_y: e:elt -> y:y -> r:elt { r = (ExpY e y) } 
let exp_y = exp 

val exp_z: e:elt -> z:z -> r:elt { r = (ExpZ e z) } 
let exp_z = exp 

(* let exp_ddh e y z  = exp_z e z  *)

(*
module SampleZ = 
  // typing the sampling of z in ElGamal
  // no need for z to be abstract? 

  private type z = int 

  val sample_z: g:group -> z:z { Prod(Exp(g,z0)*m0 = Prod(Exp(g,z1)*m1 } 
  // this follows from sampling, using fun z0 -> z0 + Log(m1) - Log(m0)

  val exp_z: g:group -> z:z -> e:group { e = exp(g,z) } 

*)

 end

module ElG
open Lib
open Group
open DDH

(* 
let qH: int   = failwith "spec"
let g : group = failwith "spec" 
let k : int   = failwith "spec"
type kbytes  = bytes // of size k
let zero : kbytes = failwith "spec"
type key        = secret_key * public_key
type message    = kbytes 
type cipher     = group * kbytes 

*) 

(* // later, try getting plain abstraction *)
type plain = elt 
type cipher = elt * elt  
type secret_key = exn
type public_key = elt

(*
  // a logical specification of El Gamal encryptions.
  function val Enc(gx,y,m) = (Exp(g,y), Prod(Exp(gx,y),m))
  predicate Encrypted(gx,m,c) = ?y. c = Enc(gx,y,m) 
*)

val KeyGen: unit ->  (secret_key * public_key)
let KeyGen () : secret_key * public_key = 
  let x = sample_x in
  (x, exp_x g x)

val Enc: pk:public_key -> m:plain ->  cipher 
let Enc pk m = 
  let y = sample_y in
  let z = sample_z in
  (exp_y g y, (exp_ddh pk y) * m)

val Dec: secret_key -> cipher -> plain 
let Dec x c = 
  let (gy,v) = c in 
  let gxy = exp_x gy x in
    div v gxy
    
(*   // !m,c. Encrypted(gx,m,c) => r = m  *)

module CPA 
open ElG 

(* Two adversaries, implicitly stateful *)
type A1 = public_key -> plain * plain 
type A2 = cipher -> bool

let game (a1:A1) (a2:A2) b = 
  let (sk,pk) = KeyGen() in
  let (m0,m1,s) = a1 pk in 
  let m = if b then m0 else m1 in
  let c = Enc(pk, m) in
  a2 c


(* module RO *)

(*   val l: nat *)
(*   type msg = b:bytes { Length(b) = l }  *)

(*   ty[e 'a eqBad = x:'a { Bad() \/ x0 = x1 }  *)

(*   val hash: e:elt eqBad -> h:msg eqBad *)

(*   // could use a single flagged function *)

(*   type entry =  *)
(*     | Public of h:msg eqBad *)
(*     | Secret of h:msg (\* we abstract over the encrypted messages *\) *)
(*   let hash e =  *)
(*     match List.find sampled e with  *)
(*     | Public(h) -> h *)
(*     | Secret(h) -> collision:= e::collision; h  *)
(*     | None      -> let h = sample() *)
(*                    sampled := (e,Public(v))::sampled *)
(*                    h *)

(*   val hashElG: m:msg (\*ghost*\) -> elt -> n:msg { n0 * m0 = n1 * m1 }  *)

(*   let hashElg m e =  *)
(*     match List.find sampled e with  *)
(*     | Public(h) -> collision:= e::collision; h  *)
(*     | Secret(h) -> collision:= e::collision; h  *)
(*     | None      -> let h = sample' m  *)
(*                    sampled := (e,Secret(v))::sampled *)
(*                    h *)

(*   // What happens if I sample twice the same secret mask?  *)

(*   // internally, we maintain some mutable state *)
(*   // where the boolean tracks whether the first call was via  *)

(*   definition Bad state <=> hashDom  *)

(*   type state =  *)
(*   { sample:    (elt * msg ) list; *)
(*     domOracle: elt list  *)
(*     domElg:    elt list  *)
(*  * elg *)

(*   val sampled: (elt * msg) list ref // actually a map *)
(*   val *)

(*   / *)
(* a table elt -> msg  *)
(*   // and two sets H, for the sampled domains of hash and hashElg *)
(*   // elt -> Oracle( Oracle msg  *)
(*   //  *)



(* module HElGamal *)

(*   // idem, except for Enc and Dec: *)

(*  let Enc (pk:public_key) (m:message) *)


(* module Game = *)

(* let cpa (KG,enc) (A1:A1) (A2:A2) b =  *)
(*   let (sk,pk) = KG() *)
(*   let (m0,m1,s) = A1 pk *)
(*   let m = if b then m0 else m1 *)
(*   let c = Enc(pk, m) *)
(*   A2 c *)

(* // 1. reverse-inline to type against DDH.fs7 ; apply DDH assumption ; inline *)
(* // 2. reverse-inline to type against sampleZ, giving eqtypes to everything else; conclude by relational typing *)



(*   // main (after inlining KG & Enc) *)
(*   let x = sample q // [0..q-1] *)
(*   let pk = g^x *)
(*   let (m0,m1) = A1 (Hash_adv,pk) *)
(*   let b = (sample 2 = 1) *)
(*   let m = if b then m0 else m1 *)
(*   let y = sample q  *)
(*   let c = (g^y, pk^y * m) *)
(*   A2 c *)

(*   // can we use abstract types for sampled values?  *)
(*   // we need x,y,z abstract & gx gy  *)



(*   // main0: after inlining Enc (we may also want to swap lets) *)
(*   let (sk,pk) = KG() *)
(*   let (m0,m1) = A1 (Hash_adv,pk) *)
(*   let b = (sample 2 = 1) *)
(*   let c =  *)
(*     let y = sample q *)
(*     let y' = pk^y  *)
(*     let hy = Hash(y') *)
(*     (g^y, hy ^^ if b then m0 else m1) *)
(*   let b' = A2(Hash_adv,pk, c) *)
(*   b = b' *)

(*   // main1: after fixing hy & applying up-to-bad  *)
(*   let (sk,pk) = KG() *)
(*   let (m0,m1) = A1 (Hash_adv,pk) *)
(*   let b = (sample 2 = 1) *)
(*   let c =  *)
(*     let y = sample q *)
(*     let y' = pk^y  *)
(*     let hy = randomBytes k // Hash(y') *)
(*     (g^y, hy ^^ if b then m0 else m1) *)
(*   let b' = A2(Hash_adv,pk, c) *)
(*   b = b' *)

(* // We need to show that A2 matches the following spec: *)

(* Pre = *)
