#monadic(RDST, iRetTx, iBindTx)
(* base concrete implementation: AES *)

(* in F#: 

open System.Security.Cryptography
open System.Text

let aes = Aes.Create() (* no managed "raw" AES block, so we use 1-block ECB *)
aes.Mode <- CipherMode.ECB 
let encrypt = aes.CreateEncryptor()

let f i = 
  let inBlock = System.BitConverter.GetBytes(12) in
  let encrypt = aes.CreateEncryptor()
  encrypt.TransformFinalBlock (inBlock, 0, inBlock.Length)

*)

module PRF 
open Bytes 
(* pattern predicates *)
type INST2 :: bytes => bytes => E
type INST4 :: bytes => bytes => bytes => bytes => E

(* treating bytes as an abstract type whose representation is 128 bits, 
   rather than as a refinement of a byte array. *)
type index = bytes
type block = bytes
logic val XOR : block -> block -> block
assume XOR_commut: forall x y. XOR x y = XOR y x
assume XOR_id: forall x y. XOR (XOR x y) x = y
type Mask :: _ = fun (p0:block) (p1:block) (m0:block) (m1:block) => 
    (Eq (XOR p0 m0) (XOR p1 m1))

(* TODO: implement in terms of the sample in RO ...show it is injective etc. *)
val sample: p:bytes -> (iR block 
                          (fun h0 h1 => True)
                          (fun h0 h1 m0 m1 h0' h1' => 
                              (h0=h0' && h1=h1' && Mask (L p) (R p) m0 m1)))

logic data type entry = 
  | Entry : i:index -> p:block (*ghost*) -> m:block -> entry

logic val log: ref (list entry)
assume InHeap_log:(|forall h0 h1.HeapInv h0 h1 ==> (InHeap h0 (L log) && InHeap h1 (R log))|)
(* let log = ref [] *)

logic val n : ref int
assume InHeap_n:(|forall h0 h1. HeapInv h0 h1 ==> (InHeap h0 (L n) && InHeap h1 (R n))|)
(* let n = ref 0  *)

logic val Domain: list entry -> list block
(* define *) assume Domain_nil: Domain [] = []
(* define *) assume Domain_cons: forall hd tl. Eq (Domain (hd::tl)) (((Entry_proj_0 hd):block)::Domain tl)
(* query *)  assume Lemma_domain_in: forall i l. In i (Domain l) <==> (exists p m. In (Entry i p m) l)

(* A log is good if it has the same spine on either side;
                       and the same domain;
                       and the points are related by the Mask predicate *)
type GoodEntry::entry => entry => E
type GoodLog:: list entry => list entry => E
assume GoodEntry: forall e0 e1. (GoodEntry e0 e1) 
                          <==> ((Entry_proj_0 e0 = Entry_proj_0 e1) 
                                && Mask (Entry_proj_1 e0) (Entry_proj_1 e1) (Entry_proj_2 e0) (Entry_proj_2 e1))
assume GoodLog_nil: GoodLog [] [] 
assume GoodLog_cons: forall e0 e1 tl0 tl1. (GoodLog (e0::tl0) (e1::tl1)
                                      <==>  (GoodLog tl0 tl1 && GoodEntry e0 e1))
assume GoodLog_inv: forall l0 l1. GoodLog l0 l1 ==> (l0 = [] <==> l1 = [])
assume HeapInv_def: 
(|forall h0 h1. HeapInv h0 h1 <==> 
    (GoodLog (SelHeap h0 (L log)) (SelHeap h1 (R log))
     && (forall i. (In i (Domain (SelHeap h0 (L log))) ==> (Seqn i < (SelHeap h0 (L n))))) (* filled up to n *)
     && (SelHeap h0 (L n) = SelHeap h1 (R n)) )|) (* n is low *)

(* let f i = match assoc i !log with  *)
(*   | Some (Entry _ _ m) -> m *)
(*   | None ->  *)
(*       let m = randomBytes 16 in *)
(*         log := ((Entry i p m)::!log); *)
(*         m *)


val assoc: i:index
        -> l:list entry 
        -> (iR (option entry)
              (fun h0 h1 => (L i = R i && GoodLog (L l) (R l)))
              (fun h0 h1 eopt0 eopt1 h0' h1' => 
                  ((h0=h0') 
                   && (h1=h1')
                   && (exists e0 e1. (eopt0 = Some e0) <==> (eopt1 = Some e1))
                   && (eopt0 = None <==> eopt1 = None)
                   && (eopt0 = None ==> not (In (L i) (Domain (L l))))
                   && ((exists e. eopt0 = Some e)
                       ==> (GoodEntry (Some_proj_1 eopt0) (Some_proj_1 eopt1)
                            && (Entry_proj_0 (Some_proj_1 eopt0) = L i) 
                            && (Entry_proj_0 (Some_proj_1 eopt1) = R i))))))
let rec assoc i l = match l with 
  | [] -> None
  | e::tl -> 
      if Entry_proj_0 e = i 
      then Some e 
      else assoc i tl

(* two refinements of f, our implicitly-keyed prf function *)
val f_new: 
  i:index
  -> p:block
  -> (iR block
        (fun h0 h1 =>    (* Requires: i not in the log yet *)
            (not (In (L i) (Domain (SelHeap h0 (L log))))
             && (L i = R i)
             && (Seqn (L i) < SelHeap h0 (L n))))
        (fun h0 h1 m0 m1 h0' h1' =>   (* Ensures: the log is extended with (i,p,m) *)
            (Mask (L p) (R p) m0 m1
             && (h0' = UpdHeap h0 (L log) ((Entry (L i) (L p) m0)::(SelHeap h0 (L log))))
             && (h1' = UpdHeap h1 (R log) ((Entry (R i) (R p) m1)::(SelHeap h1 (R log)))))))
let f_new i p = match assoc i (!log) with
  (* | Some _ -> unreachable ()  *) (*$ TODO *)
  | None ->
      let m = sample p in
      let _ = (log := ((Entry i p m)::!log)) in
        m

val f_old: 
  i:index
  -> (iR block
        (fun h0 h1 =>   (* Requires: i is already in the log *)
            (In (L i) (Domain (SelHeap h0 (L log)))
             && (L i = R i))) (* maybe get rid of this clause *)
        (fun h0 h1 m0 m1 h0' h1' =>   (* Ensures: the log contains some (i,p,m) *)
            ((h0=h0') 
             && (h1=h1')
             && (exists (p0:block) (p1:block). INST2 p0 p1 && Mask p0 p1 m0 m1))))
let f_old i = match assoc i (!log) with 
  | Some (Entry _ p m) -> 
      (assume (| INST2 (L p) (R p) |);  
       m)
  | None -> unreachable()


(* provided the encryptor and decryptor are synchronized, 
   we can use an increasing counter as index *)
type Cipher :: _ = fun (i0:index) (i1:index) (c0:block) (c1:block) (h0:heap) (h1:heap) => 
    (exists p0 p1 m0 m1.
        (INST4 p0 p1 m0 m1
         && Mask p0 p1 m0 m1
         && (i0 = i1)
         (* && (c0 = c1) *)
         && (c0 = (XOR p0 m0))
         && (c1 = (XOR p1 m1))
         && (In (Entry i0 p0 m0) (SelHeap h0 (L log)))
         && (In (Entry i1 p1 m1) (SelHeap h1 (R log)))) )
val enc: 
  i:index 
  -> p:block 
  -> (iR block
        (fun h0 h1 => (* Requires: i not in the log yet *) 
            ((not (In (L i) (Domain (SelHeap h0 (L log)))))
             && (L i = R i)
             && (Seqn (L i) < SelHeap h0 (L n))))
        (fun h0 h1 c0 c1 h0' h1' =>  (* Ensures: the log is extended with some (i,p,m) *)
            (Cipher (L i) (R i) c0 c1 h0' h1')))
let enc i p =
  let m = f_new i p in (* p is ghostly *) 
    (assume (| INST4 (L p) (R p) (L m) (R m) |);
     XOR p m)

val dec: i:index 
  -> c:block 
  -> (iR block
        (fun h0 h1 => (Cipher (L i) (R i) (L c) (R c) h0 h1))
        (fun h0 h1 p0 p1 h0' h1' => 
            ((h0 = h0') 
             && (h1 = h1')
             && (exists (m0:block) (m1:block). Mask p0 p1 m0 m1))))
let dec i c = XOR (f_old i) c


(* informally, encrypt assumes Msg(i,p); 
   we could more generally take a 'Msg parameter *)
val encrypt: p:block -> (iR (index * block) (* add modifies clause *)
                           (fun h0 h1 => True)
                           (fun h0 h1 u0 u1 h0' h1' => 
                               (Cipher (fst u0) (fst u1) (snd u0) (snd u1) h0' h1')))
let encrypt (p:block) = 
  let i = toBytes (!n) in 
  let _ = n := ((!n) + 1) in
  let res = (i, enc i p) in 
  res

val decrypt: i:index -> c:block -> (iR block 
                                      (fun h0 h1 => Cipher (L i) (R i) (L c) (R c) h0 h1)
                                      (fun h0 h1 p0 p1 h0' h1' => (h0=h0' && h1=h1')))
let decrypt i c = dec i c 
