#monadic(RDST, iRetTx, iBindTx)
(* #monadic(RDST, iRetTx, (iBindTx HeapInv)) *)
module PRF 
open Bytes 

(** 12-10-11 counter-mode encryption example
             to be completed as in randomOracle-2.fst and ../misc/Auth.fst **) 
type fbytes (n:int) = b:bytes{n=Length b}
type index = eq (fbytes 16)
type block = fbytes 16
type mask (p:block) = m:block {| Eq (L p + L m) (R p + R m) |}  
(*$ using + for XOR; any injective function would do. *)
val sample: p:block -> mask p 
(* TODO: implement in terms of the sample in RO ...show it is injective etc. *)

logic data type entry = 
  | Entry : index -> p:block -> mask p -> entry

val log: ref (list entry) 
logic val Domain: list entry -> list index
assume Domain_nil: Domain [] = []
assume Domain_cons: forall hd tl. Domain (hd::tl) = (Entry_proj_0 hd)::(Domain tl)

(* we also need the two maps to have the same domain *)
assume HeapInv_def: (forall h0 h1. HeapInv h0 h1 <==> 
                       (| forall i. ((In i (Domain (SelHeap h0 (L log)))
                                      <==> In i (Domain (SelHeap h1 (R log))))
                        |)))

let f i = match assoc i !log with 
  | Some (Entry _ _ m) -> m
  | None -> 
      let m = randomBytes 16 in
        log := ((Entry i p m)::!log);
        m

(* two refinements of f, our implicitly-keyed prf function *)
val f_new: 
  i:eq index 
  -> p:block
  -> (iR (mask p)
        (fun h0 h1 =>    (* Requires: i not in the log yet *)
            (not (In (L i) (Domain (SelHeap h0 (L log))))))
        (fun h0 h1 m0 m1 h0' h1' =>   (* Ensures: the log is extended with (i,p,m) *)
            ((h0' = UpdHeap h0 (L log) (Entry (L i) (L p) m0 (SelHeap h0 (L log))))
          && (h1' = UpdHeap h1 (R log) (Entry (R i) (R p) m1 (SelHeap h1 (R log))))))
let f_new i p = match assoc i !log with 
  | Some (Entry _ _ m) -> m
  | None -> 
      let m = sample p in 
        log := ((Entry i p m)::!log);
        m

val f_old: 
  i:eq index
  -> (iR block 
        (fun h0 h1 =>   (* Requires: i is already in the log *)
            (In (L i) (Domain (SelM h0 (L log)))))
        (fun h0 h1 p0 p1 h0' h1' =>   (* Ensures: the log contains some (i,p,m) *)
            ((h0=h0') 
             && (h1=h1')
             && (exists m0 m1. 
                   (In (L i) (Entry (L i) p0 m0) (SelM h0 (L log))
                    && In (R i) (Entry (R i) p1 m1) (SelM h1 (R log)))))))
let f_old i = match assoc i !log with 
  | Some (Entry _ _ m) -> m
  | None -> unreachable()


end

(* base concrete implementation: AES *)

(* in F#: 

open System.Security.Cryptography
open System.Text

let aes = Aes.Create() (* no managed "raw" AES block, so we use 1-block ECB *)
aes.Mode <- CipherMode.ECB 
let encrypt = aes.CreateEncryptor()

let f i = 
  let inBlock = System.BitConverter.GetBytes(12) in
  let encrypt = aes.CreateEncryptor()
  encrypt.TransformFinalBlock (inBlock, 0, inBlock.Length)

 *)

module Fresh
open PRF

(* mref 'a 'P    (defined in prims-rrf.fst)
     is a monotonic reference to an 'a with update condition 'P

   MonoBoolT is an update condition for monotonic boolean flags.
*)
val bad : mref bool MonoBoolT

type block = fbytes 16
val sample: unit -> EnsuresUpTo bad block (fun h0 h1 u0 u1 h0' h1' => 
    ((h0=h0')
     && (h1=h1')
     && (u0=u1)
     && not (In u0 (Domain (SelHeap h0 (L log))))))
let sample () =
  let u = randomBytes 16 in
  if mem u (Domain !log) 
  then (mwrite bad true; u)
  else u

(* The probability of Bad is bounded by \sum_{sample} (length S)/2^l, or
   - if length S < N, #sample * L / 2^l
   - if S is the set of previously-fresh values, #sample^2 / 2^{l+1}
 *)
end

module CounterMode
open Fresh

type Cipher :: _ = fun (i0:index) (i1:index) (c0:block) (c1:block) (h0:heap) (h1:heap) => 
    (| exists p0 p1 m0 m1. 
        ((c0 = p0 + m0) 
         && (c1 = p1 + m1) 
         && (In (Entry i p0 m0) (SelM h0 (L log)))
         && (In (Entry i p1 m1) (SelM h1 (R log)))) |)

type cipher (i:index) = c:eq block { exists i p m. (i,p,m) In log && c = p + m }

val enc: 
  i:index 
  -> p:block 
  -> (iR (eq block)
        (fun h0 h1 => (* Requires: i not in the log yet *) 
            (not (In (L i) (Domain (SelM h0 (L log))))))
        (fun h0 h1 c0 c1 h0' h1' =>  (* Ensures: the log is extended with some (i,p,m) *)
            (Cipher (L i) (R i) c0 c1 h0' h1'
             && (h0' = UpdM h0 (L log) (Entry (L i) (L p) m0 (SelM h0 (L log))))
             && (h1' = UpdM h1 (R log) (Entry (R i) (R p) m1 (SelM h1 (R log)))))))
let enc i (p:block) = 
  let m = f_new i p in (* p is ghostly *) 
  m + p 

val dec: i:index 
  -> c:eq block 
  -> (iR block
        (fun h0 h1 => (Cipher (L i) (R i) (L c) (R c) h0 h1))
        (fun h0 h1 p0 p1 h0' h1' => 
            ((h0 = h0') 
             && (h1 = h1')
             && (exists m0 m1. (In (Entry (L i) p0 m0) (SelM h0 (L log))
                                && In (Entry (R i) p1 m1) (SelM h1 (R log)))))))
let dec i c = (f_old i) + c


(* we need an invariant telling us the log is filled up to n *) 

(* informally, encrypt assumes Msg(i,p); 
   we could more generally take a 'Msg parameter 
 *)

val encrypt: p:block -> (i:index * c:cipher i { Msg(i,p) })


val encrypt: p:block -> (EnsuresUpTo bad (index * block)
                           (fun h0 h1 u0 u1 h0' h1' => 
                               (u0 = u1
                                && Cipher (fst u0) (fst u1) (snd u0) (snd u1) h0' h1')))(* { Msg(i,p) }) *)
let encrypt (p:block) = 
  let l = !log in 
  let n = Fresh.sample (Domain l) in 
    (n,enc n p)

val decrypt: i:index -> c:block -> Req
cipher i -> p:block  { Msg(i,p) })
let decrypt i c = dec i c 

(* otherwise, the encryptor can generate an i, and
   we need up-to-bad collisions. *)
