#monadic(RDST, iRetTx, iBindTx)
(* #monadic(RDST, iRetTx, (iBindTx HeapInv)) *)
module PRF 
open Bytes 

(** 12-10-11 counter-mode encryption example
             to be completed as in randomOracle-2.fst and ../misc/Auth.fst **) 
type fbytes (n:int) = b:bytes{n=Length b}
type index = eq (fbytes 16)
type block = fbytes 16
logic val XOR : block -> block -> block
type Mask :: _ = fun (p0:block) (p1:block) (m0:block) (m1:block) => 
    (Eq (XOR p0 m0) (XOR p1 m1))
type mask (p:block) = m:block {| Eq (XOR (L p) (L m)) (XOR (R p) (R m)) |}  
(*$ using + for XOR; any injective function would do. *)
val sample: p:bytes{DoubleSided (Eq (Length p) 16)} -> (iR block 
                                         (fun h0 h1 => True)
                                         (fun h0 h1 m0 m1 h0' h1' => 
                                             (h0=h0' && h1=h1' && Mask (L p) (R p) m0 m1)))
(* TODO: implement in terms of the sample in RO ...show it is injective etc. *)

logic data type entry = 
  | Entry : (* i:bytes{DoubleSided ((Relational (Eq (L i) (R i)) && (Eq (Length i) 16)))}  *)
            i:bytes{DoubleSided (Eq (Length i) 16)} 
         -> p:bytes{DoubleSided (Eq (Length p) 16)} 
         -> m:bytes{DoubleSided (Eq (Length m) 16)} -> entry

(* logic data type entry =  *)
(*   | Entry : index -> p:bytes{Length p=16} -> n:(m:bytes{(16=Length m)}){| Mask (L p) (R p) (L n) (R n) |} -> entry *)

logic val log: ref (list entry) 
(* let log = ref [] *)

logic val n : ref int 
(* let n = ref 0  *)
logic val Domain: list entry -> list block
assume Domain_nil: Domain [] = []
assume Domain_cons: forall hd tl. Eq (Domain (hd::tl)) (((Entry_proj_0 hd):block)::Domain tl)

(* we also need the two maps to have the same domain *)
assume HeapInv_def: (|forall h0 h1. HeapInv h0 h1 <==> 
                       (forall (i:block). ((In i (Domain (SelHeap h0 (L log)))
                                            <==> In i (Domain (SelHeap h1 (R log)))))
                        && (In i (Domain (SelHeap h0 (L log)))
                            ==> (Seqn i < (SelHeap h0 (L n))))
                        && (SelHeap h0 (L n) = SelHeap h1 (R n))
                       )|)


(* let f i = match assoc i !log with  *)
(*   | Some (Entry _ _ m) -> m *)
(*   | None ->  *)
(*       let m = randomBytes 16 in *)
(*         log := ((Entry i p m)::!log); *)
(*         m *)

(* val assoc: i:index -> l:list entry -> option entry *)
val assoc: i:bytes 
        -> l:list entry 
        -> (iR (option entry)
              (fun h0 h1 => True)
              (fun h0 h1 eopt0 eopt1 h0' h1' => 
                  ((h0=h0') 
                   && (h1=h1')
                   && ((exists e0. (eopt0 = Some e0)) ==> (In (Some_proj_1 eopt0) (L l)
                                                           && (L i = Entry_proj_0 (Some_proj_1 eopt0))))
                   && ((exists e1. (eopt1 = Some e1)) ==> (In (Some_proj_1 eopt1) (R l)
                                                           && (R i = Entry_proj_0 (Some_proj_1 eopt1)))))))
(* two refinements of f, our implicitly-keyed prf function *)
val f_new: 
  i:index 
  -> p:block
  -> (iR block
        (fun h0 h1 =>    (* Requires: i not in the log yet *)
            (not (In (L i : block) (Domain (SelHeap h0 (L log))))
             && ((Seqn (L i)) < SelHeap h0 (L n))))
        (fun h0 h1 m0 m1 h0' h1' =>   (* Ensures: the log is extended with (i,p,m) *)
            (Mask (L p) (R p) m0 m1
             && (h0' = UpdHeap h0 (L log) ((Entry (L i) (L p) m0)::(SelHeap h0 (L log))))
             && (h1' = UpdHeap h1 (R log) ((Entry (R i) (R p) m1)::(SelHeap h1 (R log)))))))
let f_new i p = match assoc i (!log) with
  (* | Some _ -> unreachable () *)
  | None ->
      let m = sample p in
      let _ = (log := ((Entry i p m)::!log)) in
        m

val f_old: 
  i:eq index
  -> (iR block
        (fun h0 h1 =>   (* Requires: i is already in the log *)
            (In (L i : block) (Domain (SelHeap h0 (L log)))))
        (fun h0 h1 m0 m1 h0' h1' =>   (* Ensures: the log contains some (i,p,m) *)
            ((h0=h0') 
             && (h1=h1')
             (*  && (Length p0 = 16) *)
             (*  && (Length p1 = 16) *)
             && (exists (p0:block) (p1:block).
                   (In (Entry (L i) p0 m0) (SelHeap h0 (L log))
                    && In (Entry (R i) p1 m1) (SelHeap h1 (R log)))))))
let f_old i = 
  (* let l = (!log) in *)
    match assoc i (!log) with 
      | Some (Entry _ _ m) -> m
          (* assert (In e l); *)
          (* (Entry _ _ m) -> m *)
          (* | None -> unreachable() *)



(* base concrete implementation: AES *)

(* in F#: 

open System.Security.Cryptography
open System.Text

let aes = Aes.Create() (* no managed "raw" AES block, so we use 1-block ECB *)
aes.Mode <- CipherMode.ECB 
let encrypt = aes.CreateEncryptor()

let f i = 
  let inBlock = System.BitConverter.GetBytes(12) in
  let encrypt = aes.CreateEncryptor()
  encrypt.TransformFinalBlock (inBlock, 0, inBlock.Length)

 *)



(* provided the encryptor and decryptor are synchronized, 
   we can use an increasing counter as index *)

type Cipher :: _ = fun (i0:index) (i1:index) (c0:block) (c1:block) (h0:heap) (h1:heap) => 
    (| exists p0 p1 m0 m1.
        (Mask p0 p1 m0 m1
         && (c0 = (XOR p0 m0))
         && (c1 = (XOR p1 m1))
         && (In (Entry i0 p0 m0) (SelHeap h0 (L log)))
         && (In (Entry i1 p1 m1) (SelHeap h1 (R log)))) |)

(* type cipher (i:index) = c:eq block (\* { exists i p m. In (Entry i p m) log && c = p + m } *\) *)

val enc: 
  i:eq index 
  -> p:block 
  -> (iR (eq block)
        (fun h0 h1 => (* Requires: i not in the log yet *) 
            ((not (In (L i : block) (Domain (SelHeap h0 (L log)))))
             && (Seqn (L i) < SelHeap h0 (L n))))
        (fun h0 h1 c0 c1 h0' h1' =>  (* Ensures: the log is extended with some (i,p,m) *)
            (Cipher (L i) (R i) c0 c1 h0' h1'
             && (exists m0. h0' = UpdHeap h0 (L log) (Entry (L i) (L p) m0 :: (SelHeap h0 (L log))))
             && (exists m1. h1' = UpdHeap h1 (R log) (Entry (R i) (R p) m1 :: (SelHeap h1 (R log)))))))
let enc i (p:block) = 
  let m = f_new i p in (* p is ghostly *) 
  XOR m p 

val dec: i:index 
  -> c:eq block 
  -> (iR block
        (fun h0 h1 => (Cipher (L i) (R i) (L c) (R c) h0 h1))
        (fun h0 h1 p0 p1 h0' h1' => 
            ((h0 = h0') 
             && (h1 = h1')
             && (exists (m0:block) (m1:block). 
                   (In (Entry (L i) p0 m0) (SelHeap h0 (L log))
                    && In (Entry (R i) p1 m1) (SelHeap h1 (R log)))))))
let dec i c = XOR (f_old i) c


(* we need an invariant telling us the log is filled up to n *) 

(* informally, encrypt assumes Msg(i,p); 
   we could more generally take a 'Msg parameter 
 *)

val encrypt: p:block -> (iR (index * eq block) (* add modifies clause *)
                           (fun h0 h1 => True)
                           (fun h0 h1 u0 u1 h0' h1' => 
                               (Cipher (fst u0) (fst u1) (snd u0) (snd u1) h0' h1')))(* { Msg(i,p) }) *)
let encrypt (p:block) = 
  let _ = (n := ((!n) + 1)) in
  let i = toBytes (!n) in 
    (i,enc i p)

val decrypt: i:index -> c:eq block -> (iR block 
                                         (fun h0 h1 => Cipher (L i) (R i) (L c) (R c) h0 h1)
                                         (fun h0 h1 p0 p1 h0' h1' => (h0=h0' && h1=h1')))
let decrypt i c = dec i c 

(* otherwise, the encryptor can generate an i, and
   we need up-to-bad collisions. *)
