module Nat

(* type eq 'a = x:'a{|L x = R x|} *)
type eq 'a = 'a 

(* using int instead of bigint for now *) 

(*
extern reference BigInt {language="C#";
                         dll="mscorlib";
                         namespace="System";
                         classname="Numerics"}
extern BigInt type bigint
extern BigInt val op_Addition : bigint -> bigint -> bigint 
...
*)

type num = int 
type nat = n:num { 0 <= n } 

type mod:: _ = fun (n:nat) => ( m:nat { m < n } )

(* // module Group *)

(** multiplicative group **) 

type elt (p:nat) = x:mod p{ 0 <> x }

logic val MulE: p:nat -> elt p -> elt p -> elt p
val mulE: p:nat -> x:elt p -> y:elt p -> r:elt p { r = (MulE p x y) }

logic val ExpE: p:nat -> elt p -> num -> elt p
val expE: p:nat -> x:elt p -> n:num -> r:elt p { r = (ExpE p x n) }

assume forall p x y. MulE p x y = MulE p y x
assume forall p x y z. MulE p (MulE p x y) z = MulE p x (MulE p y z)
assume forall p e x y. ExpE p e (Add x y) = MulE p (ExpE p e x) (ExpE p e y)
assume forall p e (x:num) (y:num). ExpE p e (Mul x y) = ExpE p (ExpE p e x) y

type gen p = g:elt p { forall (q: elt p). exists n. ExpE p g n = q }

type pp = PP : n:nat -> gen n -> gen n -> pp
logic val ppp: p:pp -> nat
logic val ppg: p:pp -> gen (ppp p)
logic val pph: p:pp -> gen (ppp p)


val commit: p:pp -> x:num -> o:num -> r:elt (ppp p)
  { r = MulE (ppp p) (ExpE (ppp p) (ppg p) x) (ExpE (ppp p) (pph p) o) 
(* this triggers an unexpected exception:
  && | ((L p) = (R p) && (L x) = (R x) && (L o) = (R o)) ==> (L r) = (R r) | *)
  }


let commit p x o = mulE (ppp p) (expE (ppp p) (ppg p) x) (expE (ppp p) (pph p) o)

val test: p:pp -> x:num -> o:num -> r:elt (ppp p)
  { r = MulE (ppp p) (ExpE (ppp p) (ppg p) (Mul 2 x)) (ExpE (ppp p) (pph p) (Add o o)) 
  }

let test p x o = commit p (x+x) (o+o)
