
(** 12-10-10 Cedric: this is an outline for LRE, written with Santiago. **)

module Guess

(* generates a value, keeps it private,   
   and checks if the adversary can guess it, triggering a bad event. 
   
   this would work for arbitrary distribution, although for now
   we just sample uniformally a fixed-sized bitstring. *) 

type t: b:bytes { Length b = 16 } 

val bad = ref bool
let bad = ref false

val r:t
let r = sample 16 

val check (r':t) -> unit { r = r' => bad = true}

let check (r':t) = if r = r' then bad:=true 

(* We'll use the probability that any user of this interface 
   given access just to check sets bad to express security bounds. *)


(* some other, more general outline: 

(* we inport some key sampling function: *)

val n : nat
val Key : coins:bytes {length coins = n } -> key

type repr = bytes

type key

let private k = let coins = sample n in Key n 
 (* insert here your favorite key sampling function *)

val check (v:repr): unit { v <> k } 

let check v = if v = k then bad:=true
 *)

module Leak

type t: b:bytes { Length b = 32 } 

val bad = ref bool
let bad = ref false

val r:t
let r = sample 32

val leak (f: bytes -> bool (*pure*) ) 
let leak f = incr nleak; f r

val check (r':t) -> unit { r = r' => bad = true}

let check (r':t) = if r = r' then bad:=true 


module Leak' (* for multiple keys  *)

type repr = bytes  (* abstract *) 
type key 
type leakage = b:bytes {Length b = 16} 

let Gen (f:repr -> leakage) = 
  let k = Key() in 
  (k, f (repr k))

let Gen0 (f:repr -> leakage) = (* unrelated leakage *)
  let k  = Key() in 
  let b = sample 16 
  (k, b)

let Check (Key k) v = if repr k = v then bad := true

(* lemma: Pr[ bad with Gen ] < 2^16 * Pr[ bad with Gen0 ] *) 

let k = Key 


let test f Adv = Adv(f k) = k 
  
let test0 f Adv = let k' = Key() in Adv(f k) = k 


(*** We program LRE on top of a CPA encryptiomn scheme, 
     for a single implicit key and a single encryption ***)

module LRE 

open Bytes 

(* initial game *) 

(* pre-sampling *) 
let r = Guess.r
let k = Leak.k
let h = Enc.keyGen (* just an alias for uniform sampling withing Enc *) 

let log :=  
  List.init sample_tag ... 
  (* pre-sampling on the full domain of H; up to inlining, this is just 
     a plain, eagerly-sample random oracle *)

let H (k',r') = List.assoc log (k',r')  
let H'(k',r') = if (k,r) = (k',r') then h else H (k',r')

  (* the two are equivalent provided that (k','r) <> (k,r) *)
 
let H0 (k',r') = Guess.check r'; H(k',r') 
let H1 (k',r') = Leak.check k' ; H(k',r') 

let s0       = Leak.leak (Adv0 (*pure*) H0)
let m0,m1,s1 = Adv1 s0 H1
let b        = sample{0,1}
let e        = r, Enc h (if b then m0 else m1)
let b'       = Adv2 s1 H1
(b = b') 



(* by inlining, and being parametric in H again, we get the actual code: *)

let k     = sample 32
let leak  = Adv0 (*pure*) H k : b:bytes{ Length b = 16 }
let m0,m1 = Adv1 H leak
let b     = sample{0,1}
let r     = sample 16
let h     = H (k,r)
let e     = r, Enc h (if b then m0 else m1)
let b'    = Adv2 H e
(b = b') 

(* by typing, we get that, up to bad, h remains entirely abstract,
   and we conclude by conformance to the CPA interface. 

   Adv[LRE] = Adv[CPA] + Pr[Guess r] + Pr[Guess k]*2^leak
            = Adv[CPA] + 2^-128 * (2*#qH)
 *)



(* a more abvanced variant, stateful & with key refresh: *)

let encrypt k m =
  let r = sample 16
  let ke, k' = H(k,r)
  let e = r, Enc ke m
  (k',e) 





























let encrypt m = 
  let R = randomBytes rsize
  let P = H(K||R) 
  R || cheat(xor M P) (* this rightfully fails to typecheck. *) 




















(*
type bytes = byte array
#if concrete
let randomBytes n = Lib.randomBytes n
let zeroCreate n = Lib.zeroCreate n
let xor x y : bytes = Array.map2 (^^^) x y 
#else
type 't eq = 't
let randomBytes (n:int) = failwith "spec" : bytes 
let zeroCreate (n:int) = failwith "spec" : bytes 
let (||) (b0:bytes) (b1:bytes) = failwith "spec" : bytes
let split (n:int) (b:bytes) = failwith "spec" : bytes * bytes
let xor (x:bytes) (y:bytes) = failwith "spec" : bytes
#endif
 *)

let keysize = 16
let plainsize = 128
let rsize = keysize


type hashed = bytes
type key = bytes
type plain = bytes  // {Length(b)=plainsize}
type cipher = bytes // {Length(b)=rsize+plainsize}

(* standard lists, for retyping *)
let rec mem xs (x:bytes) = 
  match xs with 
  | y::_ when x = y -> true
  | _::xs           -> mem xs x
  | []              -> false 
let rec assoc xzs (x:bytes) =
  match xzs with 
  | (y,z)::xzs when x = y -> Some (z:bytes)
  | _::xzs                -> assoc xzs x
  | []                    -> None 
let cons (x:hashed) (h:plain) xhs = (x,h)::xhs

let n = keysize + rsize // size of the domain of H
let HT = ref []         // sampled definition of H so far

let bad = ref false
let chi = ref []     // domain of H used secretly by the encryption 
let H x = 
  // we add this line in H1: 
  if mem !chi x then bad:=true  
  match assoc !HT x with 
  | Some(h) -> h
  | None -> 
      let h:plain = randomBytes plainsize
      HT:= cons x h !HT
      h

let gen() = randomBytes keysize
let K = gen()

let decrypt E = 
  let (R,C) = split plainsize E
  let P = H(K||R)
  xor C P

(*** the kdm main game ***)

type innerAdv = (plain -> plain) -> key -> plain 

let zero = zeroCreate plainsize

(* this is NOT typable, as it requires Sample *)
let randomMask (M:plain) = failwith "not yet": plain // randomBytes plainsize 

type leaker = key (* noref *) -> plain  
//take2
// type leaker = key{true} -> plain{true}  

let game (GEN,ENC) b adv =
  let k = GEN() 
  let log = ref [] // empty_log k 
  let enc (g:innerAdv) = // g must be typed as pure 
    let x0 = g H k       // added explicit access to H
    let x = if b then x0 else zero 
    let v = ENC k x in
    log := (v,x):: !log; v
  adv enc   
 
// Adv[a enc] =def= | Pr[ game true a = true ] - Pr[ game false a = true ] | 
// GOAL Adv[a enc] <= (Length !HT)^2 / 2^(rsize*8-2) 
// assuming that keysize >= rsize 

(*** concrete construction ***)

let cheat (M:plain) = failwith "not typable otherwise!": plain eq

let encrypt g = // (g: htype -> key -> plain {Pure(g)})  
  let M = g H K // g implicitly gets to call RO
  let R = randomBytes rsize
  let P = H(K||R) 
  R || cheat(xor M P) (* this rightfully fails to typecheck. *) 

(* we can typecheck that decrypt;encrypt(...) is identity on plaintexts *)

(*1* obtained by inlining H, and if we don't observe bad and chi,
     we give H1 to the external attacker but still H to the internal one 
     this is perfectly equivalent to: ***)

let encrypt1 g = // (g: htype -> key -> plain {Pure(g)}) 
  let M = g H K // g implicitly gets to call RO
  let R = randomBytes rsize
  let x = K||R 
  let C: plain eq = 
    match assoc !HT x with 
    | Some(P) -> 
        bad := true; cheat (xor M P) (*2*)
    | None -> 
        let P = randomMask M  
        HT:= (x,P)::!HT
        chi:= x::!chi
        xor M P in
   R || C

// guess: the fake game is obtained from the real one by using
// let fake g H K = let plain = g H K in zero

// we can bound the probability of bad happening from collisions, e.g.
// Pr[ ... | bad ] < length !HT 

(*2: encrypt2 is encrypt1 with zero instead of xor M P (*2*); 
     using an alias of sample with type

       val sample_shift: M:bytes{Length(M)=plainsize} -> 
                         P:bytes{Length(P)=plainsize /\ xor M0 P0 = xor M1 P1 }  

     we can give C the type

       C:bytes{Length(C)=plainsize /\ C0=C1 /\ (Bad \/ ...) }   *)
