#monadic(RDST, iRetTx, iBindTx)

module Passport
open Bytes

type nonce = bytes
type key = bytes
type plain = bytes
type cipher = bytes

type classicPure :: _ = fun ('a::*) (res:'a) ('Post::'a => heap => E) h => ('Post res h)

type concatClassic :: _ = fun (b1:bytes) (b2:bytes) ('Post::bytes => heap => E) h => 
    ('Post (Concat b1 b1) h)

val concat': b1:bytes -> b2:bytes -> RDST bytes (composeClassic bytes 
                                                   (classicPure bytes (Concat (L b1) (L b2)))
                                                   (classicPure bytes (Concat (R b1) (R b2))))
logic val NT : plain -> nonce
assume NT_def: forall nr nt. NT (Concat nr nt) = nt
logic val NR : plain -> nonce
assume NR_def: forall nr nt. NR (Concat nr nt) = nr
val getNR : p:plain -> RDST nonce (composeClassic nonce 
                                     (classicPure nonce (NR (L p)))
                                     (classicPure nonce (NR (R p))))
val getNT : p:plain -> RDST nonce (composeClassic nonce 
                                     (classicPure nonce (NT (L p)))
                                     (classicPure nonce (NT (R p))))

type eqClassicTx :: _ = fun ('a::*) (x:'a) (y:'a) ('Post::bool => heap => E) h => 
    (forall b.  ((b=true) <==> (x = y)) ==> 'Post b h)
val eqClassic: x:'a -> y:'a -> RDST bool (composeClassic bool 
                                            (eqClassicTx 'a (L x) (L y))
                                            (eqClassicTx 'a (R x) (R y)))

(* A log associating nonces with keys *)
logic data type ntk = 
             | NTK: nonce -> key -> ntk
val nts : ref (list ntk)
type GoodNTK :: _ = fun ntk0 ntk1 => (NTK_proj_0 ntk0 = NTK_proj_0 ntk1)
type GoodNTS :: list ntk => list ntk => E
assume GoodNTS_nil: GoodNTS [] []
assume GoodNTS_cons: forall ntk0 ntk1 tl0 tl1. (GoodNTS (ntk0::tl0) (ntk1::tl1)
                                                <==>  (GoodNTS tl0 tl1 && GoodNTK ntk0 ntk1))
assume GoodNTS_inv: forall l0 l1. GoodNTS l0 l1 ==> (l0 = [] <==> l1 = [])

(* A log for decryption, indexed by cipher- and plain-texts *)
logic data type entry = 
             | Entry : cipher -> plain -> key -> entry
logic val mkNTK : entry -> ntk
define Def_mkNTK: forall e. mkNTK e = (NTK (NT (Entry_proj_1 e)) (Entry_proj_2 e))

val log : ref (list entry)
type GoodEntry::entry => entry => E
type GoodLog:: list entry => list entry => E
assume GoodEntry: forall e0 e1. (GoodEntry e0 e1) 
                          <==> ((Entry_proj_0 e0 = Entry_proj_0 e1) (* ciphers are the same *)
                                && (Entry_proj_1 e0 = Entry_proj_1 e1)) (* plain texts are the same *)
assume GoodLog_nil: GoodLog [] [] 
assume GoodLog_cons: forall e0 e1 tl0 tl1. (GoodLog (e0::tl0) (e1::tl1)
                                      <==>  (GoodLog tl0 tl1 && GoodEntry e0 e1))
assume GoodLog_inv: forall l0 l1. GoodLog l0 l1 ==> (l0 = [] <==> l1 = [])

type LogInjectiveP :: heap => heap => ref (list entry) => ref (list entry) => E

assume LogInjectiveP_elim: (forall h0 h1 log0 log1. 
                              (LogInjectiveP h0 h1 log0 log1 
                               ==> (forall e0 e1. (In e0 (SelHeap h0 log0)
                                                   && In e1 (SelHeap h1 log1)
                                                   && NT (Entry_proj_1 e0) = NT (Entry_proj_1 e1))
                                    ==> (Entry_proj_0 e0 = Entry_proj_0 e1))))
assume LogInjectiveP_intro: (forall h0 h1 log0 log1. 
                               LogInjectiveP h0 h1 log0 log1
                             ==> (forall e0 e1. 
                                    ((NT (Entry_proj_1 e0) = NT (Entry_proj_1 e1)) ==> (Entry_proj_0 e0 = Entry_proj_0 e1))
                                    ==> (LogInjectiveP 
                                           (UpdHeap h0 log0 (e0::(SelHeap h0 log0)))
                                           (UpdHeap h1 log1 (e1::(SelHeap h1 log1)))
                                           log0 log1)))
                            
type RelateLogNts :: heap => heap => ref (list entry) => ref (list entry) => ref (list ntk) => ref (list ntk) => E
assume RelateLogNts_elim: (forall h0 h1 log0 log1 nts0 nts1.
                             RelateLogNts h0 h1 log0 log1 nts0 nts1
                           ==> (forall e0. (In e0 (SelHeap h0 log0)
                                            && In (mkNTK e0) (SelHeap h0 nts0))
                                ==> (forall ntk1. (In ntk1 (SelHeap h1 nts1)
                                                   && (NTK_proj_0 ntk1 = (NTK_proj_0 (mkNTK e0))))
                                     ==> ((NTK_proj_1 ntk1 = NTK_proj_1 (mkNTK e0))
                                          && In e0 (SelHeap h1 log1)))))

type RelateLogNts_intro :: _ = fun h0 h1 log0 log1 nts0 nts1 e => 
    ((RelateLogNts h0 h1 log0 log1 nts0 nts1
      && In (mkNTK e) (SelHeap h0 nts0) 
      && In (mkNTK e) (SelHeap h1 nts1))
     ==> (RelateLogNts
            (UpdHeap h0 log0 (e::SelHeap h0 log0))
            (UpdHeap h1 log1 (e::SelHeap h1 log1))
            log0 log1 nts0 nts1))

(* assume RelateLogNts_intro: (forall h0 h1 log0 log1 nts0 nts1. *)
(*                               RelateLogNts h0 h1 log0 log1 nts0 nts1 *)
(*                             ==> (forall e0. (In (mkNTK e0) (SelHeap h0 nts0)  *)
(*                                              && In (mkNTK e0) (SelHeap h1 nts1)) *)
(*                                  ==> (RelateLogNts  *)
(*                                         (UpdHeap h0 log0 (e0::SelHeap h0 log0)) *)
(*                                         (UpdHeap h1 log1 (e0::SelHeap h1 log1)) *)
(*                                         log0 log1 nts0 nts1))) *)
    
assume HeapInv_def: (| forall h0 h1. (HeapInv h0 h1  <==>
                                        (GoodNTS (SelHeap h0 (L nts)) (SelHeap h1 (R nts))
                                         && GoodLog (SelHeap h0 (L log)) (SelHeap h1 (R log))
                                         && LogInjectiveP h0 h1 (L log) (R log)
                                         && RelateLogNts h0 h1 (L log) (R log) (L nts) (R nts)
                                         && RelateLogNts h1 h0 (R log) (L log) (R nts) (L nts)
                                        ))|)


val sample: k:key -> p:plain -> (iReqEns cipher
                                   (fun h0 h1 => (L p = R p))
                                   (fun h0 h1 c0 c1 h0' h1' => (h0=h0' && h1=h1' && c0=c1)))

(* samples a cipher c based on p and k. bad if collision, but ignoring up to bad, for now *)
val encrypt: k:key -> p:plain -> (iReqEns cipher 
                                    (fun h0 h1 => ((L p = R p)
                                                   && (L k = R k) (* yes, because of the nonce check *)
                                                   && In (NTK (NT (L p)) (L k)) (SelHeap h0 (L nts))
                                                   && In (NTK (NT (R p)) (R k)) (SelHeap h1 (R nts))))                                     
                                    (fun h0 h1 c0 c1 h0' h1' => (c0 = c1))) (* say that it's in the log, or will the invariant just do *)
let encrypt k p =
  let c = sample k p in
  let l = !log in
  let h = get () in
  let e = Entry c p k in 
    assume (| RelateLogNts_intro (L h) (R h) (L log) (R log) (L nts) (R nts) (L (Entry c p k)) |); (* provide instantiations directly *)
    assume (| RelateLogNts_intro (R h) (L h) (R log) (L log) (R nts) (L nts) (L (Entry c p k)) |);
    (log := (e::l));
    c
  
logic data type optbytes = 
             | NoneB : optbytes
             | SomeB : bytes -> optbytes

(* easy ... todo lift this from counter.assoc *)
val decrypt: k:key -> c:cipher -> (iReqEns optbytes 
                                     (fun h0 h1 => (L c = R c))
                                     (fun h0 h1 opt0 opt1 h0' h1' => 
                                         ((h0=h0' && h1=h1')
                                          && ((SomeB_is opt0 && SomeB_is opt1) ==> (SomeB_proj_0 opt0 = SomeB_proj_0 opt1))
                                          && (opt0 = NoneB <==> (forall e. In e (SelHeap h0 (L log)) ==> (neq (Entry_proj_0 e) (L c) || neq (Entry_proj_2 e) (L k))))
                                          && (opt1 = NoneB <==> (forall e. In e (SelHeap h1 (R log)) ==> (neq (Entry_proj_0 e) (R c) || neq (Entry_proj_2 e) (R k))))
                                          && (SomeB_is opt0  <==> (In (Entry (L c) (SomeB_proj_0 opt0) (L k)) (SelHeap h0 (L log))))
                                          && (SomeB_is opt1  <==> (In (Entry (L c) (SomeB_proj_0 opt1) (L k)) (SelHeap h0 (L log))))
                                         )))
(* let decrypt k c =  *)
(*   match find (fun (c', _, k') -> k=k' && c=c') !log with  *)
(*     | Some (_, p,_) -> Some p *)
(*     | _ -> None *)

val errorMAC:cipher
val errorNonce:bytes
assume MacNonceEQ: errorMAC = errorNonce
assume MacNonceEQ2: (| (L errorMAC = R errorMAC) && (L errorNonce = R errorNonce) |)

val accept: k:key -> nt:nonce -> c:cipher -> (iReqEns cipher
                                                (fun h0 h1 => ((L nt = R nt) 
                                                               && (L c = R c)
                                                               && In (NTK (L nt) (L k)) (SelHeap h0 (L nts))
                                                               && In (NTK (R nt) (R k)) (SelHeap h1 (R nts))))
                                                (fun h0 h1 c0' c1' h0' h1' => (c0' = c1')))
let accept k nt c = 
  match decrypt k c with
    | SomeB plain -> 
      if eqClassic (NT plain) nt
      then encrypt k (concat' (getNR plain) (getNT plain))
      else errorNonce
    | NoneB -> errorMAC

(* cross-cases: 

     Easy: emac=enonce
       L-decryption-fails: errorMac
       R-dec-succeeds, nonce-check fails: errorNonce

     ----------------------------------------     
       L-decryption-fails: errorMac
       R-dec-succeeds, nonce-check succeeds: encrypt
                 1.   (c, _, L k) not-in L log          (L-dec fails)
                 2.   (c, R plain, R k) in R log        (R-dec succeeds)
                 3.   NT (R plain) = R nt               (nonce-check succeeds)
                 4.   NTK (R nt) (R k) in R nts         (pre-condition)
                 5.   mkNTK (c, R plain, R k) in R log  (2 + def)
                 6.   NTK (L nt) (L k) in L nts         (pre-condition)
                 7.   L nt = R nt                       (pre-condition)
                 8.   L k = R k                         (relateLogNts)
                 9.   (c, L plain, kk) in L log         (GoodLog)
                10.?  kk = L k?
       L-decrypt-succeeds, nonce-fails: enonce
       R-decrypt-succeeds, nonce-succeeds: encrypt
*)      
