#monadic(M, retTx, bindTx)
(*
  A tiny demo for monadic relational inference. 

  The #monadic directive above specifies:
     M the monad to use (see relmon.fst)
     retTx, the specification of the monadic return
     bindTx, the specification of the monadic bind

  Aside from the single-sided rules that are specific to the
  relational monad, inference is parametric with regard to the choice
  of the monad.

  Limitations: 

  -- At the moment, the monad M is for pure code. Extending to state
  is straightforward. Maybe tomorrow.

  -- I don't implement a single-sided rule yet. So, currently the
  relational inference for programs with branches is unsound. But, see
  lang-sec/eflow/relational-iflow.txt for a nice single-sided rule
  based on self-composition. Will implement this soon.

*)

module InfoFlowPure
open Relational

(* 
   ll_and_hh is the type of function 
   that maps low inputs to low outputs 
         and on high inputs does whatever
*)

type ll_and_hh = x:int -> M int (fun ('Post::int => int => E) => 
    (forall y0 y1. ((L x) = (R x) ==> y0 = y1) ==> 'Post y0 y1))

(* You can write down a let binding and leave its type out. 
   We'll infer a precise type for it. 
   If you write down a val decl with a type, we'll check that the inferred
   type is a subtype of the annotated type. 

   For now, all except the last annotation is commented, but you can
   uncomment them if you like.
*)

(* val id : ll_and_hh *)
(* inferred: *)
(* val id : x:int -> M int (fun ('Post::int => int => E) => 'Post (L x) (R x)) *)
let id x = x


val incr : ll_and_hh
(* inferred *)
(* val incr: x:int -> M int (fun ('Post::int => int => E) => *)
(*     'Post ((L x) + 1) ((R x) + 1)) *)
let incr x = x + 1

(* Here's the type of a function that always returns a low result *)
type any_to_low = x:int -> M int (fun ('Post::int => int => E) =>
    (forall y. 'Post y y))

(* val zero : any_to_low *)
let zero x = x - x

(* The main interesting part of the demo is this:
   We don't need to annotate the types of zero, incr, id etc.
   But, we can write the function foo below,
   infer a type for it, and check that it always returns low *)

val foo : any_to_low
let foo x = (zero (incr x) - 1)

(* (\* val bar : any_to_low  *\) (\* should fail if you use this type *\) *)
(* let bar x = (zero (incr (id x))) + x *)
  
(* val sample: x:int -> M int (fun ('Post::int => int => E) =>  *)
(*   (forall y0 y1. ((Eq ((L x) + y0) ((R x) + y1)) ==> 'Post y0 y1))) *)

(* val noise_to_declassify: any_to_low *)
(* let noise_to_declassify x = (sample x) + x *)

(* (\* ?  pre x0 x1 /\ forall y0 y1. (post x0 x1 y0 y1 => 'Post y0 y1) *\) *)

