#monadic(RDST, iRetTx, iBindTx)
module RandomOracle
open Bytes
type key = bytes
type value = int
logic data type entry = 
  | H : key -> tag:value -> entry 
  | A : key -> tag:value -> entry
logic val Key : entry -> key
assume(* define *) Key_projection: forall e. ((A_is e ==> (Key e = A_proj_0 e))
                                  && (H_is e ==> (Key e = H_proj_0 e)))
logic val Value : entry -> value
assume (* define *) Value_projection: forall e. ((A_is e ==> (Value e = A_proj_1 e))
                                 && (H_is e ==> (Value e = H_proj_1 e)))
type log = list entry
val logref : mref log (ListInclusion entry) (* a monotonic log *)
(* let logref = newmref [] *) (* TODO: add initialization ... *)

(* Sampling: can sample a value providing a key and 
             get a pair of values related by any injective function *)
type Function :: _ = fun ('P::(value => value => E)) =>
    ((forall v. exists v'. 'P v v') && (forall v v1 v2. ('P v v1 && 'P v v2) ==> v1=v2))
type Injective :: _ = fun ('P::(value => value => E)) =>
    ((Function 'P
      && (forall v1 v1' v2 v2'. ('P v1 v1' && 'P v2 v2')
          ==> (v1 = v2 <==> v1' = v2'))))

val sample: 'P::value => value => E 
           -> u:unit{Injective 'P}
           -> tag:value {| 'P (L tag) (R tag) |}

(* A relational invariant on the log: 
   H entries are related by an injective function; 
   A entries are equal *)
type GoodEntry :: entry => entry => E
assume GoodEntry_H: (forall k v0 v1. exists ('P::value => value => E).
                       ((Injective 'P && 'P v0 v1) ==> GoodEntry (H k v0) (H k v1)))
assume GoodEntry_A: forall k v. GoodEntry (A k v) (A k v)
assume GoodEntry_Inv: forall e0 e1. GoodEntry e0 e1 ==> ((A_is e0 <==> A_is e1)
                                                         && (H_is e0 <==> H_is e1)
                                                         && (A_is e0 ==> e0=e1)
                                                         && (H_is e0 ==> 
                                                               ((Key e0 = Key e1)
                                                                && (exists ('P::value => value => E). 
                                                                      Injective 'P && 'P (Value e0) (Value e1)))))
type GoodLog :: log => log => E
assume GoodNil: GoodLog [] []
assume GoodConsH: forall l0 l1 e0 e1. (GoodEntry e0 e1 && GoodLog l0 l1) <==> GoodLog (e0::l0) (e1::l1)
assume HeapInv_definition: |forall h0 h1. (HeapInv h0 h1 <==> GoodLog (SelM h0 (L logref)) (SelM h1 (R logref)))|


val assoc: 
     k:key 
  -> l:log
  -> iRDST (option entry) (fun ('Post::option entry => option entry => heap => heap => E) h0 h1 => 
      (forall r0 r1. 
         ((r0 = None ==> (forall e. In e (L l) ==> (Key e <> (L k)))) 
          && (r1 = None ==> (forall e. In e (R l) ==> (Key e <> (R k))))
          && ((exists v. r0 = Some v) ==> (Key (Some_proj_1 r0) = (L k) && In (Some_proj_1 r0) (L l)))
          && ((exists v. r1 = Some v) ==> (Key (Some_proj_1 r1) = (R k) && In (Some_proj_1 r1) (R l)))
          && ((GoodLog (L l) (R l) && (exists v. r0 = Some v) && (exists v. r1 = Some v)) ==> GoodEntry (Some_proj_1 r0) (Some_proj_1 r1)))
       ==> 'Post r0 r1 h0 h1))
let rec assoc k l = match l with 
  | [] -> None
  | hd::tl -> 
      if Key hd = k then Some hd 
      else assoc k tl

                   
val hashH:
     'P::value => value => E
  -> k:eq key{Injective 'P} 
  -> iRDST value (fun ('Post::value => value => heap => heap => E) h0 h1 => 
      (forall r0 r1 h0' h1'.
         ((exists e0 e1. ((Key e0 = L k) && In e0 (SelM h0 (L logref)) && (r0 = Value e0)
                          && (Key e1 = R k) && In e1 (SelM h1 (R logref)) && (r1 = Value e1))) (* collision *)
          || ('P r0 r1
              && In (H (L k) r0) (SelM h0' (L logref))
              && In (H (R k) r1) (SelM h1' (R logref))
              && ModifiesM (L logref) h0 h0'
              && ModifiesM (R logref) h1 h1'))
       ==> 'Post r0 r1 h0' h1'))
let hashH ('P::value => value => E) k =
  let l = mread logref in 
    match assoc k l with
      | Some e -> Value e
      | None ->
          let t = sample<'P> () in
            mwrite logref ((H k t)::l);
            t
              
val hashA: k:eq key 
  -> iRDST value (fun ('Post::value => value => heap => heap => E) h0 h1 => 
      (forall r0 r1 h0' h1'.
         ((exists e0 e1. ((Key e0 = L k) && H_is e0 && In e0 (SelM h0 (L logref)) && (r0 = Value e0)
                          && (Key e1 = R k) && H_is e1 && In e1 (SelM h1 (R logref)) && (r1 = Value e1))) (* H collision (bad) *)
          || (r0 = r1
              && In (A (L k) r0) (SelM h0' (L logref))
              && In (A (R k) r1) (SelM h1' (R logref))
              && ModifiesM (L logref) h0 h0'
              && ModifiesM (R logref) h1 h1'))
       ==> 'Post r0 r1 h0' h1'))
let hashA k = 
  let l = mread logref in 
    match assoc k l with 
      | Some (H _ t) -> t
      | Some (A _ t) -> t
      | None ->
          let t = sample<Eq value> () in
            mwrite logref ((A k t)::l);
            t
