(* This module defines a relational monad for use with inference *)
module Relational

(* (M 'a 'Tx) is a type of a computation that produces an 'a result x
   such that, for any relation ('Post :: 'a => 'a => E)
   the weakest pre-condition of the computation to establish it
   is ('Tx 'Post). 

   That is, 'Tx is a predicate transformer, 
   from relational post-conditions to relational pre-conditions.

   Notice that the definition of M uses a relational refinement x:t{| phi |}
   on both its argument and result.
*)

type M :: _ = fun ('a::*) ('Tx::('a => 'a => E) => E) =>
      'Post::('a => 'a => E) 
    -> u:unit{|'Tx 'Post|} -> (x:'a{|'Post (L x) (R x)|})

(* u:unit so far because the monad is pure *) 

type retTx :: _ = fun ('a::*) 
                      (x:'a) 
                      ('Post::'a => 'a => E) => 'Post (L x) (R x)
(* val returnM : x:'a -> M 'a (retTx 'a x) *)
(* Commenting out implementations for now, would need to add it to exec *)
(* let returnM x = fun () -> x *)

type bindTx :: _ = fun ('a::*) ('b::*) 
                       ('Tx1::('a => 'a => E) => E)
                       ('Tx2::'a => ('b => 'b => E) => E)
                       ('Post::'b => 'b => E) => 
    ('Tx1 (fun (y0:'a) (y1:'a) => forall (x:'a). ((L x)=y0 && (R x)=y1) ==> 'Tx2 x 'Post))

(* val bindM : 'a::* -> 'b::* *)
(*         -> 'Tx1::('a => 'a => E) => E *)
(*         -> 'Tx2::('a => ('b => 'b => E) => E) *)
(*         -> c1:M 'a 'Tx1 *)
(*         -> c2:(x:'a -> M 'b ('Tx2 x)) *)
(*         -> M 'b (bindTx 'a 'b 'Tx1 'Tx2) *)
(*
let bindM c1 c2 =
 (* 'Tx1 (\y0 y1. forall x. (L x)=y0 && (R x)=y1 ==> 'Tx2 x 'Post) *)
  let x = c1 () in
 (* 'Tx2 x 'Post *)
    c2 x ()
 (* 'Post *)
*)
