#monadic(RDST, iRetTx, bindTx)

module SingleSided

(* classic specs for some functions, so they can be called single-sidedly *)
(* DD lifts the classic spec double-sidedly into RDST *)
val zero : x:int -> DD int (fun ('Post::int => heap => E) h => 'Post 0 h)
let zero x = x - x

type addTx :: _ = fun (x:int) (y:int) ('Post::int => heap => E) h => ('Post (Add x y) h)
val incr : x:int -> RDST int (composeClassic int (addTx (L x) 1) (addTx (R x) 1))
let incr x = x + 1

type subTx :: _ = fun (x:int) (y:int) ('Post::int => heap => E) h => ('Post (Sub x y) h)
val subtract : x:int -> y:int -> RDST int (composeClassic int (subTx (L x) (L y)) (subTx (R x) (R y)))
let subtract x y = x - y

(* single-sided examples *)
type any_to_low 'a 'b =  x:'a -> Ensures 'b (fun h0 h1 b0 b1 h0' h1' => (b0=b1))

val ex1 : any_to_low bool int
let ex1 x = if x then 1 else 1

val ex2 : any_to_low int int
let ex2 x = if x = 0 then x else 0

val ex3 : x:int -> (Ensures int (fun h0 h1 i0 i1 h0' h1' => (i0 <= 0 && i1 <= 0)))
let ex3 x = if x > 0 then 0 else x

val ex4: x:ref int -> y:ref int -> b:bool -> (Ensures unit (fun h0 h1 u0 u1 h0' h1' => (SelHeap h0' (L y) = SelHeap h1' (R y))))
let ex4 x y b = 
  if b 
  then ((x := 1); (y := 1))
  else ((x := 0); (y := 1)) (* two double-sided calls on either side *)

val ex5: x:ref int -> y:ref int -> b:bool -> (ReqEns unit 
                                                 (fun h0 h1 => (DoubleSided (neq x y))) (* y is low after only if x<>y *)
                                                 (fun h0 h1 u0 u1 h0' h1' => (SelHeap h0' (L y) = SelHeap h1' (R y))))
let ex5 x y b = 
  if b 
  then ((x := 1); (y := 1))
  else ((y := 1); (x := 0)) (* two double-sided calls (to write) on either side ... the arguments don't have to match *)

val ex6 : any_to_low int int
let ex6 x = if x > 0 then 0 else zero x (* single-sided call to zero *)

val ex7 : any_to_low bool int
let ex7 x = if x then 0 else subtract 1 1  (* single-sided call to subtract *)

val ex8 : x:int -> (Ensures int 
                      (fun h0 h0 i0 i1 h0' h1' => (i0 = L x && i1 = R x)))
let ex8 x = 
  if x = 0 
  then 
    let tmp = incr x in 
      subtract tmp 1  
  else x (* LHS has lets and single-sided calls *)

val ex9: x:ref int -> b:bool -> (ReqEns unit 
                                   (fun h0 h1 => (h0=h1 && (L x = R x)))
                                   (fun h0 h1 u0 u1 h0' h1' => (h0'=h1')))
let ex9 x b = 
  if b 
  then 
    let tmp0 = subtract 1 1 in 
      x := tmp0
  else x := 0 (* LHS has lets and single-sided calls, and a double-sided call *)

val ex10: x:ref int -> b:bool -> (Ensures unit (fun h0 h1 u0 u1 h0' h1' => 
                                      ((SelHeap h0' (L x) > SelHeap h0 (L x))
                                       && (SelHeap h1' (R x) > SelHeap h1 (R x))))) (* h'[x] > h[x] on both sides *)
let ex10 x b = 
  let v = !x in 
    if b 
    then 
      let tmp = incr v in 
        write' x tmp
    else
      let v' = incr v in
      let v'' = incr v in 
        write' x v'' (* single-sided assignments ... *)
          

assume HeapInv_defL: forall h0 h1 ('a::*) (x:ref 'a) (v:'a). 
          (HeapInv h0 h1 ==> 
             (HeapInv (UpdHeap h0 x v) h1))

assume HeapInv_defR: forall h0 h1 ('a::*) (x:ref 'a) (v:'a). 
          (HeapInv h0 h1 ==> 
             (HeapInv h0 (UpdHeap h1 x v)))

val ex11: 'a::* -> option 'a -> x:ref 'a -> (iReqEns 'a
                                               (fun h0 h1 => (InHeap h0 (L x) && InHeap h1 (R x)))
                                               (fun h0 h1 u0 u1 h0' h1' => ( (* (u0 = u1) && ... fails, as it should, if you add this *)
                                                  (forall ('b::*) (y:ref 'b). SelHeap h0 y = SelHeap h0' y)
                                                  && (forall ('b::*) (y:ref 'b). SelHeap h1 y = SelHeap h1' y))))
let ex11 ('a::*) (o:option 'a) (x:ref 'a) = 
  match o with 
    | None -> read' x
    | Some i -> 
        let tmp = read' x in (* single-sided read ... and write below *)
          write' x i; 
          write' x tmp;
          tmp


(* Note, fails with this type because of whacky heap typing *)
(* val ex11: option int -> x:ref int -> (iReqEns int *)
(*                                                (fun h0 h1 => (InHeap h0 (L x) && InHeap h1 (R x))) *)
(*                                                (fun h0 h1 u0 u1 h0' h1' => ( (\* (u0 = u1) && ... fails, as it should, if you add this *\) *)
(*                                                   (forall ('b::*\) (y:ref 'b). SelHeap h0 y = SelHeap h0' y) *)
(*                                                   && (forall ('b::*\) (y:ref 'b). SelHeap h1 y = SelHeap h1' y)))) *)
