module Util

(* TODO: Implementation of exhaustiveness checks is incomplete. 
         I document infeasible cases that we can't currently prove 
         with calls to these functions. *)
val impos : string -> 'a
let impos x = raise (strcat "Impossible: " x)

val __todo : string -> 'a
let __todo x = raise (strcat "TODO: " x)

type Eq :: 'a::* => 'a => 'a => P = 
  | Refl_eq : 'a::* -> x:'a -> Eq 'a x x

type Disj :: P => P => P = 
  | Inl : 'a::P -> 'b::P -> 'a -> Disj 'a 'b
  | Inr : 'a::P -> 'b::P -> 'b -> Disj 'a 'b

type EqReln :: 'ctx::* => 'a::* => ('ctx => 'a => 'a => P) => 'ctx => 'a => 'a => P =
  | EQ_Refl : 'ctx::* -> 'a::* -> 'Base::('ctx => 'a => 'a => P)
             -> G:'ctx -> x:'a 
             -> EqReln 'ctx 'a 'Base G x x 

  | EQ_Sym : 'ctx::* -> 'a::* -> 'Base::('ctx => 'a => 'a => P)
             -> G:'ctx -> x:'a -> y:'a
             -> EqReln 'ctx 'a 'Base G x y
             -> EqReln 'ctx 'a 'Base G y x

  | EQ_Trans : 'ctx::* -> 'a::* -> 'Base::('ctx => 'a => 'a => P)
             -> G:'ctx -> x:'a -> y:'a -> z:'a
             -> EqReln 'ctx 'a 'Base G x y
             -> EqReln 'ctx 'a 'Base G y z
             -> EqReln 'ctx 'a 'Base G x z

  | EQ_Base : 'ctx::* -> 'a::* -> 'Base::('ctx => 'a => 'a => P)
             -> G:'ctx -> x:'a -> y:'a 
             -> 'Base G x y 
             -> EqReln 'ctx 'a 'Base G x y

type mu :: 'a::* => ('a => *) => * =
   | MkMu : 'a::* -> 'b::('a => *)
         -> (mu 'a 'b -> x:'a -> 'b x)
         -> mu 'a 'b

val nyi : string -> 'a
let nyi x = raise (strcat "not yet implemented: " x)

(* TODO: We should really use a list type in the P universe *)
type seq 'a = list 'a

type Zip :: 'a::* => 'b::* => ('a => 'b => P) => list 'a => list 'b => P = 
   | Zip_Nil : 'a::* -> 'b::* -> 'Q::('a => 'b => P) 
             -> Zip 'a 'b 'Q [] []

   | Zip_Cons: 'a::* -> 'b::* -> 'Q::('a => 'b => P)
             -> l1:list 'a -> l2:list 'b -> Zip 'a 'b 'Q l1 l2
             -> x:'a -> y:'b -> 'Q x y
             -> Zip 'a 'b 'Q (x::l1) (y::l2)

(* TODO: Remove the let rec as below? *)
val zip_p: 'a::* -> 'b::* -> 'Q::('a => 'b => P)
      -> f:(x:'a -> y:'b -> option ('Q x y))
      -> l1:list 'a
      -> l2:list 'b
      -> option (Zip 'a 'b 'Q l1 l2)
(*let rec zip_p f l1 l2 = match l1, l2 with
  | [],[] -> Some (Zip_Nil<'a,'b,'Q>)
  | (x1::tl1), (x2::tl2) ->
      match zip_p<'a,'b,'Q> f tl1 tl2, f x1 x2 with
        | (Some pf_tl), Some pf_hd -> Some (Zip_Cons<'a,'b,'Q> tl1 tl2 pf_tl x1 x2 pf_hd)
        | _ -> None
*)
(*val map_p: 'a::* -> 'b::* -> 'Q::('a => 'b => P)
      -> f:(x:'a -> (y:'b * 'Q x y))
      -> l1:list 'a
      -> (l2:list 'b * Zip 'a 'b 'Q l1 l2)
let rec map_p f l1 = match l1 with
  | [] -> [], Zip_Nil<'a,'b,'Q>
  | x::tlx ->
      let y, pfHd = f x in
      let tly, pfTl = map_p<'a,'b,'Q> f tlx in
        (y::tly), Zip_Cons<'a,'b,'Q> tlx tly pfTl x y pfHd *)
end


module FiniteSet
open Util

type Mem :: 'a::* => 'a => seq 'a => P = 
   | Mem_hd : 'a::* -> x:'a -> tl:seq 'a -> Mem x (x::tl)
   | Mem_tl : 'a::* -> x:'a -> y:'a -> tl:seq 'a 
            -> Mem x tl -> Mem x (y::tl)

(*val find : 'a::* -> 'P::('a => P) 
         -> f:(x:'a -> option ('P x))
         -> l:seq 'a 
         -> option (x:'a * 'P x * Mem x l)
let rec find f l = match l with 
  | [] -> None
  | hd::tl -> 
      match f hd with 
        | Some p -> Some (hd, p, Mem_hd hd tl)
        | None -> 
            match find f tl with 
              | None -> None
              | Some ((x, p, mtl)) -> Some(x, p, Mem_tl x hd tl mtl)*)

val append : seq 'a -> seq 'a -> seq 'a
let rec append l1 l2 = match l1 with 
  | [] -> l2
  | hd::tl -> hd::(append tl l2)

type Append :: 'a::* => seq 'a => seq 'a => seq 'a => P =
  | Append_Nil : 'a::* -> l:seq 'a -> Append 'a [] l l
  | Append_Cons : 'a::* -> x:'a -> tl:seq 'a -> l2:seq 'a -> tl_l2:seq 'a
               -> Append 'a tl l2 tl_l2
               -> Append 'a (x::tl) l2 (x::tl_l2)

type Partition :: 'a::* => seq 'a => seq 'a => seq 'a => P =
  | Part_Admit : l1:seq 'a -> l2:seq 'a -> l3:seq 'a -> Partition l1 l2 l3

and Perm :: 'a::* => seq 'a => seq 'a => P =
  | Perm_Refl : 'a::* -> l:seq 'a -> Perm 'a l l

  | Perm_Sym  : 'a::* -> l1:seq 'a -> l2:seq 'a -> Perm 'a l1 l2 -> Perm 'a l2 l1

  | Perm_Trans : 'a::* -> l1:seq 'a -> l2:seq 'a -> l3:seq 'a 
               -> Perm 'a l1 l2 -> Perm 'a l2 l3
               -> Perm 'a l1 l3

  | Perm_Cons : 'a::* -> x:'a -> tl1:seq 'a -> tl2:seq 'a
              -> Perm 'a tl1 tl2
              -> tl2_pfx:seq 'a -> tl2_sfx:seq 'a -> Partition 'a tl2 tl2_pfx tl2_sfx
              -> tl2x:seq 'a -> Append 'a tl2_pfx (x::tl2_sfx) tl2x 
              -> Perm 'a (x::tl1) tl2x

type W :: P => * = 
  | MkW : 'a::P -> 'a -> W 'a

val partitionUnconsL : x:'a -> a:seq 'a -> a1:seq 'a -> a2:seq 'a
                    -> Partition (x::a) (x::a1) a2
                    -> Partition a a1 a2
let partitionUnconsL x a a1 a2 p = Part_Admit a a1 a2

val partitionUnconsR : x:'a -> a:seq 'a -> a1:seq 'a -> a2:seq 'a
                    -> Partition (x::a) a1 (x::a2)
                    -> Partition a a1 a2
let partitionUnconsR x a a1 a2 p = Part_Admit a a1 a2


val permNil1: l:seq 'a -> Perm [] l -> W (Eq l [])
val permNil2: l:seq 'a -> Perm l [] -> W (Eq l [])
let rec permNil1 l p = match p with 
  | Perm_Refl _ -> MkW (Refl_eq l)
  | Perm_Sym _l _nil p' -> permNil2 l p'
  | Perm_Trans l1 l2 l3 p12 p23 -> 
      (match permNil1 l2 p12 with
         | MkW (Refl_eq _) -> permNil1 l3 p23)
and permNil2 l p = match p with 
  | Perm_Refl _ -> MkW (Refl_eq l)
  | Perm_Sym _nil _nil p' -> permNil1 l p'
  | Perm_Trans l1 l2 l3 p12 p23 -> 
      (match permNil2 l2 p23 with
         | MkW (Refl_eq _) -> permNil2 l1 p12)


(*val partitionUnNil : l1:seq 'a -> l2:seq 'a -> Partition [] l1 l2 
                   -> (Eq l1 [] * Eq l2 [])
let rec partitionUnNil l1 l2 pnil = 
  PAssume<(l1=[])> 0;
  PAssume<(l2=[])> 0;
  (Refl_eq l1, Refl_eq l2)  


val findPartition : x:'a -> l:seq 'a -> option (pfx:seq 'a * sfx:seq 'a * Partition 'a l pfx (x::sfx))
let rec findPartition x l = match l with 
  | [] -> None
  | hd::tl -> 
      if hd=x 
      then 
        let pf = Part_Admit l [] (x::tl) in 
          Some ([], tl, pf)
      else 
        match findPartition x tl with 
          | None -> None
          | Some ((tl_pfx, tl_sfx, pf)) -> 
              Some ((hd::tl_pfx), tl_sfx, Part_Admit l (hd::tl_pfx) (x::tl_sfx))

val findPerm' : x:'a -> l:seq 'a -> option (l':seq 'a * Perm l (x::l'))
let rec findPerm' x l = match l with
  | [] -> None
  | hd::tl ->
      if hd=x
      then Some (tl, Perm_Refl l)
      else
        match findPerm' x tl with
          | None -> None
          | Some ((tl', perm_tl_xtl')) ->
              let l' = hd::tl' in
              let part_tl'_nil = Part_Admit tl' [] tl' in
              let part_xtl' = Part_Admit (x::tl') [x] tl' in 

              let hd_tl' = (hd::tl') in
              let pf_append = Append_Cons x [] hd_tl' hd_tl' (Append_Nil hd_tl') in
              let x_hd_tl' = x::hd_tl' in
              let pf_perm = Perm_Cons hd tl (x::tl') perm_tl_xtl' [x] tl' part_xtl' x_hd_tl' pf_append in

                Some (hd_tl', pf_perm)

val findPerm : x:'a -> l:seq 'a -> option (l':seq 'a * Partition l [x] l')
let rec findPerm x l =
  match findPerm' x l with
    | None -> None
    | Some((l', pf_perm)) -> 
        let goal = Part_Admit l [x] l' in 
          Some (l', goal)
*)

val nilPartition: l1:seq 'a -> l2:seq 'a -> Partition 'a l1 [] l2 -> Eq l1 l2
let rec nilPartition l1 l2 pf = 
  PAssume<(l1=l2)> 0;
  Refl_eq l1 

(*val mergePartition: l:seq 'a -> l1:seq 'a -> l2:seq 'a -> Partition 'a l l1 l2
                 -> l2_1:seq 'a -> l2_2:seq 'a -> Partition 'a l2 l2_1 l2_2
                 -> (l1_l2_1:seq 'a * Partition l l1_l2_1 l2_2 * Partition l1_l2_1 l1 l2_1)
let rec mergePartition l l1 l2 pf12 l2_1 l2_2 pf2_12 =
  let l1_l2_1 = append l1 l2_1 in 
    (l1_l2_1, Part_Admit l l1_l2_1 l2_2, Part_Admit l1_l2_1 l1 l2_1)
*)

val partAllL: l:seq 'a -> Partition 'a l l []
let partAllL l = Part_Admit l l []

val partAllR: l:seq 'a -> Partition 'a l [] l
let partAllR l = Part_Admit l [] l

val partConsL: x:'a -> tl:seq 'a -> l:seq 'a -> r:seq 'a
              -> Partition 'a tl l r
              -> Partition 'a (x::tl) (x::l) r
let partConsL x tl l r p = Part_Admit (x::tl) (x::l) r

val partConsR : 'a::* -> x:'a -> tl:seq 'a -> l:seq 'a -> r:seq 'a
              -> Partition 'a tl l r
              -> Partition 'a (x::tl) l (x::r)
let partConsR x tl l r p = Part_Admit (x::tl) l (x::r)

val partPerm1 : 'a::* -> l1:seq 'a -> l2:seq 'a -> pfx:seq 'a -> sfx:seq 'a
              -> Perm 'a l1 l2
              -> Partition 'a l1 pfx sfx
              -> Partition 'a l2 pfx sfx

let partPerm1 l1 l2 pfx sfx perm p = Part_Admit l2 pfx sfx

val partPerm2 : 'a::* -> l:seq 'a -> pfx1:seq 'a -> pfx2:seq 'a -> sfx:seq 'a
              -> Perm 'a pfx1 pfx2
              -> Partition 'a l pfx1 sfx
              -> Partition 'a l pfx2 sfx

let partPerm2 l pfx1 pfx2 sfx perm p = Part_Admit l pfx2 sfx

val partPerm3 : 'a::* -> l:seq 'a -> pfx:seq 'a -> sfx1:seq 'a -> sfx2:seq 'a
              -> Perm 'a sfx1 sfx2
              -> Partition 'a l pfx sfx1
              -> Partition 'a l pfx sfx2
let partPerm3 l pfx sfx1 sfx2 perm p = Part_Admit l pfx sfx2

val partUnconsL : x:'a -> a:seq 'a -> a1:seq 'a -> a2:seq 'a
                  -> Partition (x::a) (x::a1) a2
                  -> Partition a a1 a2
let partUnconsL x a a1 a2 p = Part_Admit a a1 a2

val partUnconsR : x:'a -> a:seq 'a -> a1:seq 'a -> a2:seq 'a
                  -> Partition (x::a) a1 (x::a2)
                  -> Partition a a1 a2
let partUnconsR x a a1 a2 p = Part_Admit a a1 a2
end


module Terms
open Util

type bvar    = string
type tyname  = string
type vlname  = string 
type cname   = string 
type iname   = string
type boxname = string
type name    = string 

type gvar :: * => P = 
   | GV_Bound : bvar -> gvar 'a
   | GV_Free : 'a::* -> 'a -> gvar 'a

type pattern :: P = 
   | MkPattern : cname -> seq bvar -> seq bvar -> pattern

type basekind :: P = 
   | BK_Comp  : basekind
   | BK_Prop  : basekind
   | BK_Erase : basekind
   | BK_Afn   : basekind

type kind :: P = 
  | K_Base  : basekind -> kind
  | K_ProdK : bvar -> kind -> kind -> kind
  | K_ProdT : bvar -> typ -> kind -> kind

and typ :: P  =
  | T_Var   : gvar tyname -> typ
  | T_Ind   : iname -> typ
  | T_VApp  : typ -> value -> typ
  | T_App   : typ -> typ -> typ
  | T_Prod  : bvar -> typ -> typ -> typ
  | T_ProdK : bvar -> kind -> typ -> typ
  | T_Ref   : bvar -> typ -> typ -> typ
  | T_Lam   : bvar -> typ -> typ -> typ
  | T_Afn   : typ -> typ
  | T_Ascribe : typ -> kind -> typ

and value :: P =
  | V_Var   : gvar vlname -> value
  | V_Fun   : bvar  -> typ -> expr -> value
  | V_FunT  : bvar  -> kind -> expr -> value
  | V_Boxed : boxname -> value -> value
  | V_Const : cname -> seq typ -> seq value -> value
  | V_Ascribe : value -> typ -> value
    
and expr :: P = 
  | E_Value  : value -> expr
  | E_App    : expr -> value -> expr
  | E_TApp   : expr -> typ -> expr
  | E_LetIn  : bvar -> expr -> expr -> expr
  | E_Match  : value -> pattern -> expr -> expr -> expr
  | E_Assume : typ -> expr
  | E_Ascribe : expr -> typ -> expr


type fix :: (kind => *) => (typ => *) => (value => *) => (expr => *) => * =
  | MkFix : 'a::(kind  => *)
         -> 'b::(typ   => *)
         -> 'c::(value => *)
         -> 'd::(expr  => *)
         -> (fix 'a 'b 'c 'd -> k:kind  -> 'a k)
         -> (fix 'a 'b 'c 'd -> t:typ   -> 'b t)
         -> (fix 'a 'b 'c 'd -> v:value -> 'c v)
         -> (fix 'a 'b 'c 'd -> e:expr  -> 'd e)
         -> fix 'a 'b 'c 'd

type CurryT :: expr => seq typ => expr => P =
  | CurryT_nil  : e:expr -> CurryT e [] e
  | CurryT_cons : e:expr -> t:typ -> ts:seq typ -> e':expr
                -> CurryT (E_TApp e t) ts e'
                -> CurryT e (t::ts) e'

type CurryV :: expr => seq value => expr => P =
  | CurryV_nil  : e:expr -> CurryV e [] e
  | CurryV_cons : e:expr -> v:value -> vs:seq value -> e':expr
                -> CurryV (E_App e v) vs e'
                -> CurryV e (v::vs) e'

(*
type ct :: _ = fun (e:expr) => (ts:seq typ   -> (e':expr * CurryT e ts e'))
type cv :: _ = fun (e:expr) => (vs:seq value -> (e':expr * CurryV e vs e'))

private val curryt' : mu expr ct -> e:expr -> ct e 
let curryt' f e ts =
  match f with MkMu curryt' ->
    match ts with
      | [] -> e, CurryT_nil e
      | hd::tl ->
          let eres, pf_tl = curryt' f (E_TApp e hd) tl  in
            eres, CurryT_cons e hd tl eres pf_tl

private val curryv' : mu expr cv -> e:expr -> cv e 
let curryv' f e vs =
  match f with MkMu curryv' ->
    match vs with
      | [] -> e, CurryV_nil e
      | hd::tl ->
          let eres, pf_tl = curryv' f (E_App e hd) tl  in
            eres, CurryV_cons e hd tl eres pf_tl

val curryt : e:expr -> ct e
let curryt e ts = curryt' (MkMu curryt') e ts

val curryv : e:expr -> cv e
let curryv e vs = curryv' (MkMu curryv') e vs
*)
val pr : string -> unit
let pr s = print_string s; ()

val print_list: sep:string -> ('a -> unit) -> list 'a -> unit
let print_list sep f l = match l with 
  | [] -> ()
  | hd::tl -> f hd; pr sep; print_list sep f tl

let print_pattern p = match p with 
  | MkPattern x ts xs -> 
      pr x; 
      print_list " " (fun x -> pr "'"; pr x; ()) ts; 
      print_list " " pr xs

val print_typ : typ -> unit
val print_kind : kind -> unit
val print_value : value -> unit
val print_expr : expr -> unit
let rec print_typ t = match t with 
  | T_Var (GV_Bound f) -> pr "'"; pr f
  | T_Var (GV_Free f) -> pr f
  | T_Ind x -> pr x
  | T_VApp t v -> pr "("; print_typ t; print_value v; pr ")"
  | T_App t1 t2 -> pr "("; print_typ t1; pr " "; print_typ t2; pr ")"
  | T_Prod x t1 t2 -> 
      pr x; 
      pr ":"; 
      print_typ t1;
      pr " -> ";
      print_typ t2 
 | T_ProdK a k t -> 
      pr "'"; 
      pr a; 
      pr "::"; 
      print_kind k;
      pr " -> ";
      print_typ t

  | T_Ref x t phi -> 
      pr x; 
      pr ":"; 
      print_typ t;
      pr "{";
      print_typ phi;
      pr "}"

  | T_Lam x t t' -> 
      pr "fun ("; 
      pr x;
      pr ":"; 
      print_typ t;
      pr ")"; 
      pr "->";
      print_typ t'

  | T_Afn t -> pr "!"; print_typ t

  | T_Ascribe t k -> pr "("; print_typ t; pr "::"; print_kind k; pr ")"

and print_kind k = match k with 
  | K_Base (BK_Comp)  -> pr "*"
  | K_Base (BK_Prop)  -> pr "P"
  | K_Base (BK_Afn)   -> pr "A"
  | K_Base (BK_Erase) -> pr "E"
  | K_ProdK a k k' -> 
      pr "'"; 
      pr a; 
      pr "::"; 
      print_kind k;
      pr " => ";
      print_kind k'
  | K_ProdT x t k -> 
      pr x; 
      pr ":"; 
      print_typ t;
      pr " => ";
      print_kind k

and print_value v = match v with 
  | V_Var (GV_Bound f) -> pr f
  | V_Var (GV_Free f) -> pr f
  | V_Fun x t e -> 
      pr "fun ("; 
      pr x;
      pr ":"; 
      print_typ t;
      pr ")"; 
      pr "->";
      print_expr e


  | V_FunT a k e ->
      pr "fun ('"; 
      pr a;
      pr "::"; 
      print_kind k;
      pr ")"; 
      pr "->";
      print_expr e

  | V_Boxed b v -> pr "box("; pr b; pr ", "; print_value v; pr ")"

  | V_Const c ts vs -> 
      pr c; 
      print_list ", " print_typ ts; 
      print_list ", " print_value vs

  | V_Ascribe v t -> pr "("; print_value v; pr ":"; print_typ t; pr ")"
    
and print_expr e =  match e with
  | E_Value v -> print_value v
  | E_App e v -> pr "("; print_expr e; pr " "; print_value v; pr ")"
  | E_TApp e t -> pr "("; print_expr e; pr " "; print_typ t; pr ")"
  | E_LetIn x e1 e2 -> pr "let "; pr x; pr " = "; print_expr e1; pr " in "; print_expr e2
  | E_Match v p e1 e2 -> 
      pr "match "; print_value v; pr " with "; print_pattern p; pr " -> "; print_expr e1; pr " else "; print_expr e2
  | E_Assume t -> 
      pr "assume "; print_typ t
  | E_Ascribe e t -> pr "("; print_expr e; pr ":"; print_typ t; pr ")"

end


(* ********************************************************************************  *)

module Subst
open Util
open Terms

type vt :: _ = Disj value typ

type SubstK :: kind => bvar => vt => kind => P =
   | SK_Base : x:bvar -> r:vt -> b:basekind
              -> SubstK (K_Base b) x r (K_Base b)

   | SK_ProdK  : x:bvar -> r:vt -> y:bvar{y<>x}
               -> k1:kind -> k1':kind -> SubstK k1 x r k1'
               -> k2:kind -> k2':kind -> SubstK k2 x r k2'
               -> SubstK (K_ProdK y k1 k2) x r (K_ProdK y k1' k2')

   | SK_ProdT : x:bvar -> r:vt -> y:bvar{y<>x}
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> k2:kind -> k2':kind -> SubstK k2 x r k2'
               -> SubstK (K_ProdT y t1 k2) x r (K_ProdT y t1' k2')

and SubstT :: typ => bvar => vt => typ => P =
   | ST_Var0 : x:bvar -> r:vt -> y:tyname
              -> SubstT (T_Var (GV_Free y)) x r (T_Var (GV_Free y))

   | ST_Var1 : x:bvar -> r:vt -> y:bvar{y<>x} 
              -> SubstT (T_Var (GV_Bound y)) x r (T_Var (GV_Bound y))

   | ST_Var2 : x:bvar -> t:typ
              -> SubstT (T_Var (GV_Bound x)) x (Inr t) t

   | ST_Ind : x:bvar -> r:vt -> i:iname
              -> SubstT (T_Ind i) x r (T_Ind i)

   | ST_VApp  : x:bvar -> r:vt
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> v1:value -> v1':value -> SubstV v1 x r v1'
               -> SubstT (T_VApp t1 v1) x r (T_VApp t1' v1')

   | ST_App   : x:bvar -> r:vt
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> t2:typ -> t2':typ -> SubstT t2 x r t2'
               -> SubstT (T_App t1 t2) x r (T_App t1' t2')

   | ST_Prod  : x:bvar -> r:vt -> y:bvar{y<>x}
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> t2:typ -> t2':typ -> SubstT t2 x r t2'
               -> SubstT (T_Prod y t1 t2) x r (T_Prod y t1' t2')

   | ST_ProdK : x:bvar -> r:vt -> y:bvar{y<>x}
               -> k1:kind -> k1':kind -> SubstK k1 x r k1'
               -> t2:typ -> t2':typ -> SubstT t2 x r t2'
               -> SubstT (T_ProdK y k1 t2) x r (T_ProdK y k1' t2')

   | ST_Ref : x:bvar -> r:vt -> y:bvar{y<>x}
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> t2:typ -> t2':typ -> SubstT t2 x r t2'
               -> SubstT (T_Ref y t1 t2) x r (T_Ref y t1' t2')

   | ST_Lam : x:bvar -> r:vt -> y:bvar{y<>x}
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> t2:typ -> t2':typ -> SubstT t2 x r t2'
               -> SubstT (T_Lam y t1 t2) x r (T_Lam y t1' t2')

   | ST_Afn : x:bvar -> r:vt
               -> t1:typ -> t1':typ -> SubstT t1 x r t1'
               -> SubstT (T_Afn t1) x r (T_Afn t1')

   | ST_Asc : x:bvar -> r:vt 
               -> t:typ -> t':typ -> SubstT t x r t'
               -> k:kind -> k':kind -> SubstK k x r k'
               -> SubstT (T_Ascribe t k) x r (T_Ascribe t' k')

and SubstV :: value => bvar => vt => value => P =
   | SV_Var0 : x:bvar -> r:vt -> y:vlname
              -> SubstV (V_Var (GV_Free y)) x r (V_Var (GV_Free y))

   | SV_Var1 : x:bvar -> r:vt -> y:bvar{y<>x}
              -> SubstV (V_Var (GV_Bound y)) x r (V_Var (GV_Bound y))

   | SV_Var2 : x:bvar -> v:value
              -> SubstV (V_Var (GV_Bound x)) x (Inl v) v

   | SV_Fun: x:bvar -> r:vt -> y:bvar{y<>x}
            -> t:typ -> t':typ -> SubstT t x r t'
            -> e:expr -> e':expr -> SubstE e x r e'
            -> SubstV (V_Fun y t e) x r (V_Fun y t' e')

   | SV_FunT: x:bvar -> r:vt -> y:bvar{y<>x}
            -> k:kind -> k':kind -> SubstK k x r k'
            -> e:expr -> e':expr -> SubstE e x r e'
            -> SubstV (V_FunT y k e) x r (V_FunT y k' e')

   | SV_Boxed: x:bvar -> r:vt -> b:boxname
            -> v1:value -> v1':value -> SubstV v1 x r v1'
            -> SubstV (V_Boxed b v1) x r (V_Boxed b v1')

   | SV_Const: x:bvar -> r:vt -> d:cname
              -> tl:seq typ -> tl':seq typ
              -> Zip typ typ (fun (t1:typ) (t2:typ) => SubstT t1 x r t2) tl tl'
              -> vl:seq value  -> vl':seq value
              -> Zip value value (fun (v1:value) (v2:value) => SubstV v1 x r v2) vl vl'
              -> SubstV (V_Const d tl vl) x r (V_Const d tl' vl')

   | SV_Asc : x:bvar -> r:vt 
              -> v1:value -> v1':value -> SubstV v1 x r v1'
              -> t:typ -> t':typ -> SubstT t x r t'
              -> SubstV (V_Ascribe v1 t) x r (V_Ascribe v1' t')

and SubstE :: expr => bvar => vt => expr => P =
  | SE_Val : x:bvar -> r:vt
            -> v1:value -> v1':value -> SubstV v1 x r v1'
            -> SubstE (E_Value v1) x r (E_Value v1')

  | SE_App : x:bvar -> r:vt
            -> e:expr -> e':expr -> SubstE e x r e'
            -> v1:value -> v1':value -> SubstV v1 x r v1'
            -> SubstE (E_App e v1) x r (E_App e' v1')

  | SE_TApp : x:bvar -> r:vt
            -> e:expr -> e':expr -> SubstE e x r e'
            -> t:typ -> t':typ -> SubstT t x r t'
            -> SubstE (E_TApp e t) x r (E_TApp e' t')

  | SE_LetIn : x:bvar -> r:vt -> y:bvar{y<>x}
           -> e1:expr -> e1':expr -> SubstE e1 x r e1'
           -> e2:expr -> e2':expr -> SubstE e2 x r e2'
           -> SubstE (E_LetIn y e1 e2) x r (E_LetIn y e1' e2')

  | SE_Match : x:bvar -> r:vt
           -> v1:value -> v1':value -> SubstV v1 x r v1'
           -> p:pattern
           -> e1:expr -> e1':expr -> SubstE e1 x r e1'
           -> e2:expr -> e2':expr -> SubstE e2 x r e2'
           -> SubstE (E_Match v1 p e1 e2) x r (E_Match v1' p e1' e2')

  | SE_Assume : x:bvar -> r:vt 
           -> t:typ -> t':typ -> SubstT t x r t'
           -> SubstE (E_Assume t) x r (E_Assume t')

  | SE_Asc : x:bvar -> r:vt 
              -> e1:expr -> e1':expr -> SubstE e1 x r e1'
              -> t:typ -> t':typ -> SubstT t x r t'
              -> SubstE (E_Ascribe e1 t) x r (E_Ascribe e1' t')

(*type sk :: _ = fun (k:kind)   => (x:bvar -> r:vt -> (k':kind   * SubstK k x r k'))
type st :: _ = fun (t:typ)    => (x:bvar -> r:vt -> (t':typ    * SubstT t x r t'))
type sv :: _ = fun (u:value)  => (x:bvar -> r:vt -> (u':value  * SubstV u x r u'))
type se :: _ = fun (e:expr)   => (x:bvar -> r:vt -> (e':expr   * SubstE e x r e'))

private val subst_k': fix sk st sv se -> k:kind -> sk k
let subst_k' f k x r =
  match f with MkFix f_k f_t f_v f_e ->
    match k with
      | K_Base b -> k, SK_Base x r b

      | K_ProdK a k1 k2 ->
          if x=a
          then raise "non-unique names"
          else
            let k1', pf1 = f_k f k1 x r in
            let k2', pf2 = f_k f k2 x r in
              K_ProdK a k1' k2', SK_ProdK x r a k1 k1' pf1 k2 k2' pf2
               
      | K_ProdT y t1 k2 ->
          if x=y
          then raise "non-unique names"
          else
            let t1', pf1 = f_t f t1 x r in
            let k2', pf2 = f_k f k2 x r in
              K_ProdT y t1' k2', SK_ProdT x r y t1 t1' pf1 k2 k2' pf2

private val subst_t': fix sk st sv se -> t:typ -> st t
let subst_t' f t x r =
  match f with MkFix f_k f_t f_v f_e ->
    match t with
      | T_Var (GV_Free y) -> t, ST_Var0 x r y

      | T_Var (GV_Bound y) ->
          if x=y
          then match r with 
            | Inl v -> raise "ill-typed substitution"
            | Inr t -> t, ST_Var2 x t 
          else t, ST_Var1 x r y

      | T_Ind i -> t, ST_Ind x r i

      | T_VApp t1 v1 ->
          let t1', pf1 = f_t f t1 x r in
          let v1', pf2 = f_v f v1 x r in
            T_VApp t1' v1', ST_VApp x r t1 t1' pf1 v1 v1' pf2

      | T_App t1 t2 ->
          let t1', pf1 = f_t f t1 x r in
          let t2', pf2 = f_t f t2 x r in
            T_App t1' t2', ST_App x r t1 t1' pf1 t2 t2' pf2
          
      | T_Prod y t1 t2 ->
          if x=y
          then raise "non-unique names"
          else
            let t1', pf1 = f_t f t1 x r in
            let t2', pf2 = f_t f t2 x r in
              T_Prod y t1' t2', ST_Prod x r y t1 t1' pf1 t2 t2' pf2

      | T_ProdK y k1 t2 ->
          if x=y
          then raise "non-unique names"
          else
            let k1', pf1 = f_k f k1 x r in
            let t2', pf2 = f_t f t2 x r in
              T_ProdK y k1' t2', ST_ProdK x r y k1 k1' pf1 t2 t2' pf2

      | T_Ref y t1 t2 ->
          if x=y
          then raise "non-unique names"
          else
            let t1', pf1 = f_t f t1 x r in
            let t2', pf2 = f_t f t2 x r in
              T_Ref y t1' t2', ST_Ref x r y t1 t1' pf1 t2 t2' pf2

      | T_Lam y t1 t2 ->
          if x=y
          then raise "non-unique names"
          else
            let t1', pf1 = f_t f t1 x r in
            let t2', pf2 = f_t f t2 x r in
              T_Lam y t1' t2', ST_Lam x r y t1 t1' pf1 t2 t2' pf2

      | T_Afn t1 ->
          let t1', pf1 = f_t f t1 x r in
            T_Afn t1', ST_Afn x r t1 t1' pf1

      | T_Ascribe t1 k -> 
          let t1', pf1 = f_t f t1 x r in
          let k', pf2 = f_k f k x r in
            T_Ascribe t1' k', ST_Asc x r t1 t1' pf1 k k' pf2

private val subst_v': fix sk st sv se -> v:value -> sv v
let subst_v' f v1 x r =
  match f with MkFix f_k f_t f_v f_e ->
    match v1 with
      | V_Var (GV_Free y) -> v1, SV_Var0 x r y

      | V_Var (GV_Bound y) ->
          if x=y
          then match r with 
            | Inl v -> v, SV_Var2 x v
            | Inr _ -> raise "Ill-typed substitution"
          else v1, SV_Var1 x r y

      | V_Fun y t e ->
          if x=y
          then raise "non-unique names"
          else
            let t', pf1 = f_t f t x r in
            let e', pf2 = f_e f e x r in
              V_Fun y t' e', SV_Fun x r y t t' pf1 e e' pf2

      | V_FunT y k e ->
          if x=y
          then raise "non-unique names"
          else
            let k', pf1 = f_k f k x r in
            let e', pf2 = f_e f e x r in
              V_FunT y k' e', SV_FunT x r y k k' pf1 e e' pf2

      | V_Boxed b v1 ->
          let v1', pf1 = f_v f v1 x r in
            V_Boxed b v1', SV_Boxed x r b v1 v1' pf1
            
      | V_Const d tl vl ->
          let tl', pf1 =
            map_p<typ, typ, (fun (t:typ) (t':typ) => SubstT t x r t')> (fun t -> f_t f t x r) tl in
          let vl', pf2 =
            map_p<value, value, (fun (v1:value) (v1':value) => SubstV v1 x r v1')> (fun v1 -> f_v f v1 x r) vl in
            V_Const d tl' vl', SV_Const x r d tl tl' pf1 vl vl' pf2

      | V_Ascribe v1 t -> 
          let v1', pf1 = f_v f v1 x r in
          let t', pf2 = f_t f t x r in
            V_Ascribe v1' t', SV_Asc x r v1 v1' pf1 t t' pf2

private val subst_e': fix sk st sv se -> e:expr -> se e
let subst_e' f e x r =
  match f with MkFix f_k f_t f_v f_e ->
    match e with
      | E_Value v1 ->
          let v1', pf = f_v f v1 x r in
            E_Value v1', SE_Val x r v1 v1' pf

      | E_App e1 v1 ->
          let e1', pf1 = f_e f e1 x r in
          let v1', pf2 = f_v f v1 x r in
            E_App e1' v1', SE_App x r e1 e1' pf1 v1 v1' pf2

      | E_TApp e1 t ->
          let e1', pf1 = f_e f e1 x r in
          let t', pf2 = f_t f t x r in
            E_TApp e1' t', SE_TApp x r e1 e1' pf1 t t' pf2
            
      | E_LetIn y e1 e2 ->
          if x=y
          then raise "non-unique names"
          else
            let e1', pf1 = f_e f e1 x r in
            let e2', pf2 = f_e f e2 x r in
              E_LetIn y e1' e2', SE_LetIn x r y e1 e1' pf1 e2 e2' pf2

      | E_Match v1 p e1 e2 ->
          let v1', pf0 = f_v f v1 x r in
          let e1', pf1 = f_e f e1 x r in
          let e2', pf2 = f_e f e2 x r in
            E_Match v1' p e1' e2', SE_Match x r v1 v1' pf0 p e1 e1' pf1 e2 e2' pf2

      | E_Ascribe e1 t -> 
          let e1', pf1 = f_e f e1 x r in
          let t', pf2 = f_t f t x r in
            E_Ascribe e1' t', SE_Asc x r e1 e1' pf1 t t' pf2
*)
type SubstKV :: _ = (fun (k:kind) (x:bvar) (v:value) (k':kind)   => SubstK k  x (Inl v) k')
type SubstTV :: _ = (fun (t:typ)  (x:bvar) (v:value) (t':typ)    => SubstT t  x (Inl v) t')
type SubstKT :: _ = (fun (k:kind) (x:bvar) (t:typ)   (k':kind)   => SubstK k  x (Inr t) k')
type SubstTT :: _ = (fun (t:typ)  (x:bvar) (t1:typ)  (t':typ)    => SubstT t  x (Inr t1) t')

(*val subst_kv : k:kind -> x:bvar -> v:value -> (k':kind * SubstKV k x v k')
let subst_kv k x v =
  subst_k' (MkFix subst_k' subst_t' subst_v' subst_e') k x (Inl v)

val subst_tv : t:typ -> x:bvar -> v:value -> (t':typ * SubstTV t x v t')
let subst_tv t x v =
  subst_t' (MkFix subst_k' subst_t' subst_v' subst_e') t x (Inl v)

val subst_kt : k:kind -> x:bvar -> t:typ -> (k':kind * SubstKT k x t k')
let subst_kt k x t =
  subst_k' (MkFix subst_k' subst_t' subst_v' subst_e') k x (Inr t)

val subst_tt : t:typ -> x:bvar -> t1:typ -> (t':typ * SubstTT t x t1 t')
let subst_tt t x t1 =
  subst_t' (MkFix subst_k' subst_t' subst_v' subst_e') t x (Inr t1)
*)
end

(* ******************************************************************************** *)

module Env
open Util
open FiniteSet
open Terms

type constructor :: P = 
  | MkConstructor : cname -> typ -> constructor
  
type inductive :: P = 
  | MkIndType : iname -> kind -> seq constructor -> inductive

type icompartment = seq inductive

type fvbinding :: P =     
  | FV_VlName : vlname -> typ -> fvbinding
  | FV_TyName : tyname -> kind -> fvbinding

type environment = seq fvbinding

type locbinding :: P = 
  | LB_BvarTy   : bvar -> typ -> locbinding
  | LB_BvarKind : bvar -> kind -> locbinding
  | LB_VlEq     : value -> value -> locbinding
  | LB_TyEq     : typ -> typ -> locbinding

type bindings = seq locbinding    

type ICompBinds_I :: icompartment => iname => kind => P = 
  | ICompBinds_ind : I:icompartment -> i:iname -> k:kind -> ds:seq constructor
                   -> Mem (MkIndType i k ds) I
                   -> ICompBinds_I I i k

type ICompBinds_C :: icompartment => cname => typ => P =
  | ICompBinds_dcon : I:icompartment -> d:cname -> t:typ -> i:iname -> k:kind -> ds:seq constructor
                     -> Mem (MkIndType i k ds) I
                     -> Mem (MkConstructor d t) ds
                     -> ICompBinds_C I d t


type wfEnv :: icompartment => environment => bindings => P = 
  | WFENV_Admit : i:icompartment -> f:environment -> g:bindings -> wfEnv i f g

val checkWF : I:icompartment -> G:environment -> B:bindings -> option (wfEnv I G B)
let checkWF i g b = Some (WFENV_Admit i g b) 

(*val lookup_iname : i:icompartment -> j:name -> option (k:kind * ICompBinds_I i j k)
let rec lookup_iname i j = match i with 
  | [] -> None
  | hd::tl -> 
      match hd with
        | MkIndType j' k ds when j=j' ->
            let m = Mem_hd (MkIndType j k ds) tl in
            let binds = ICompBinds_ind i j k ds m in
              Some (k, binds)
        | _ ->
            match lookup_iname tl j with
              | Some ((k, (ICompBinds_ind _tl _j _k ds mtl))) -> 
                  let m = Mem_tl (MkIndType j k ds) hd tl mtl in
                  let binds = ICompBinds_ind i j k ds m in
                    Some (k, binds)
              | None ->None
              

let lookup_constr s d = match s with 
  | [] -> None
  | hd::tl -> 
      match hd with 
        | MkConstructor d' t when d=d' -> Some (t, Mem_hd hd tl)
        | _ -> 
            match lookup_constr tl d with 
              | None -> None
              | Some ((t, pf)) -> Some (t, Mem_tl (MkConstructor d t) hd tl pf)

val lookup_dcon : I:icompartment -> d:cname -> option (t:typ * ICompBinds_C I d t)
let rec lookup_dcon i d = match i with 
  | [] -> None
  | (MkIndType j k constrs)::tl -> 
      match lookup_constr constrs d with 
        | None -> 
            (match lookup_dcon tl d with 
               | None -> None
               | Some ((t, ICompBinds_dcon _tl _d _t jj kk ds mem_jj mem_ds)) -> 
                   Some (t, ICompBinds_dcon i d t jj kk ds (Mem_tl (MkIndType jj kk ds) (MkIndType j k constrs) tl mem_jj) mem_ds))
        | Some ((t, pf)) -> Some (t, ICompBinds_dcon i d t j k constrs (Mem_hd (MkIndType j k constrs) tl) pf)

val lookup_vlname : F:environment -> x:vlname -> option (t:typ * Mem (FV_VlName x t) F)
let lookup_vlname f x = raise "NYI lookup_dcon"

val lookup_tyname : F:environment -> x:tyname -> option (k:kind * Mem (FV_TyName x k) F)
let lookup_tyname f x = raise "NYI lookup_tyname"

val lookup_ty   : G:bindings -> x:bvar -> option (t:typ * Mem (LB_BvarTy x t) G)
let rec lookup_ty g x = match g with 
  | [] -> None
  | hd::tl -> 
      match hd with 
        | LB_BvarTy(x', t) when x=x' -> 
            Some (t, Mem_hd hd tl)
        | _ -> 
            match lookup_ty tl x with 
              | None -> None
              | Some ((t, mtl)) -> Some (t, Mem_tl (LB_BvarTy x t) hd tl mtl)


val lookup_kind : G:bindings -> x:bvar -> option (k:kind * Mem (LB_BvarKind x k) G)
let rec lookup_kind g x = match g with 
  | [] -> None
  | hd::tl -> 
      match hd with 
        | LB_BvarKind(x', k) when x=x' -> 
            Some (k, Mem_hd hd tl)
        | _ -> 
            match lookup_kind tl x with 
              | None -> None
              | Some ((k, mtl)) -> Some (k, Mem_tl (LB_BvarKind x k) hd tl mtl)

*)
type ValueEq :: _ = 
    EqReln bindings value (fun (G:bindings) (x:value) (y:value) => Mem (LB_VlEq x y) G) 

type TypeEq :: _ = 
    EqReln bindings typ (fun (G:bindings) (x:typ) (y:typ) => Mem (LB_TyEq x y) G) 

val is_value_eq : G:bindings -> v1:value -> v2:value -> option (ValueEq G v1 v2)
let is_value_eq g v1 v2 = raise "NYI is_value_eq"

val is_typ_eq   : G:bindings -> t1:typ -> t2:typ -> option (TypeEq G t1 t2)
let is_typ_eq g t1 t2 = raise "NYI is_value_eq"

end

(* ****************************************************************************** *)

module Z3Interface
open Util
open Terms
open Subst
open Env

private type LogicEntails :: icompartment => environment => bindings => typ => P = 
    | Admit : I:icompartment -> F:environment -> G:bindings -> phi:typ 
           -> LogicEntails I F G phi


val query : I:icompartment -> F:environment -> G:bindings -> phi:typ -> option (LogicEntails I F G phi)
let query i f g phi = Some (Admit i f g phi)

end

(* ****************************************************************************** *)

module Typing
open Util
open FiniteSet
open Terms
open Subst
open Env
open Z3Interface 

type acontext = seq bvar
type AContextSplit :: _ = (fun (a:acontext) (a1:acontext) (a2:acontext) => Partition a a1 a2)

type mode = 
  | Term_level : mode
  | Type_level : mode

type ConcreteKind :: basekind => P =
  | CK_Star : ConcreteKind BK_Comp
  | CK_Afn  : ConcreteKind BK_Afn
  | CK_Prop : ConcreteKind BK_Prop

type KindCompat :: basekind => basekind => P 

type VIndexKindCompat :: basekind => basekind => P 
type Subtyping :: icompartment => environment => bindings => typ => typ => P 
type Subkinding :: icompartment => environment => bindings => kind => kind => P 

type AQual :: acontext => typ => typ => P =
  | AQual_Nil : t:typ -> AQual [] t t 
  | AQual_Q : a:acontext -> t:typ -> AQual a t (T_Afn t)

type StripQual :: typ => typ => P =
  | StripQual_A : t:typ -> StripQual (T_Afn t) t
  | StripQual_Id : t:typ -> StripQual t t

(*val strip_qual : q:typ -> (t:typ * StripQual q t)
let strip_qual q = match q with 
  | T_Afn t -> (t, StripQual_A t)
  | _ -> (q, StripQual_Id q)

val aqual : a:acontext -> t:typ -> (q:typ * AQual a t q) 
let aqual a t = match a with 
  | [] -> t, AQual_Nil t
  | _ -> (T_Afn t), AQual_Q a t
*)  
val kind_compat: b1:basekind -> b2:basekind -> option (KindCompat b1 b2)
let kind_compat b1 b2 = raise "NYI kind_compat"

val vindex_kind_compat: b1:basekind -> b2:basekind -> option (VIndexKindCompat b1 b2)
let vindex_kind_compat b1 b2 = raise "NYI vindex_kind_compat"



type wfK :: icompartment => environment => bindings => kind => basekind => P =
   | WFK_Base  : I:icompartment -> G:environment -> B:bindings
                 -> b:basekind
                 -> wfEnv I G B
                 -> wfK I G B (K_Base b) b

   | WFK_ProdK : I:icompartment -> G:environment -> B:bindings
                 -> a:bvar -> k1:kind -> b1:basekind -> k2:kind -> b2:basekind
                 -> KindCompat b1 b2
                 -> wfK I G B k1 b1
                 -> wfK I G ((LB_BvarKind a k1)::B) k2 b2
                 -> wfK I G B (K_ProdK a k1 k2) b2

   | WFK_ProdT : I:icompartment -> G:environment -> B:bindings
                 -> x:bvar -> t:typ -> tb:basekind -> k:kind -> b:basekind
                 -> VIndexKindCompat tb b
                 -> wfT I G B t (K_Base tb)
                 -> wfK I G ((LB_BvarTy x t)::B) k b
                 -> wfK I G B (K_ProdT x t k) b

and wfT :: icompartment => environment => bindings => typ => kind => P =

   | WFT_Var : I:icompartment -> G:environment -> B:bindings
                -> a:bvar -> k:kind
                -> wfEnv I G B
                -> Mem (LB_BvarKind a k) B
                -> wfT I G B (T_Var (GV_Bound a)) k

   | WFT_FVar : I:icompartment -> G:environment -> B:bindings
                -> a:tyname -> k:kind
                -> wfEnv I G B
                -> Mem (FV_TyName a k) G
                -> wfT I G B (T_Var (GV_Free a)) k

   | WFT_Ind : I:icompartment -> G:environment -> B:bindings
                -> i:name -> k:kind
                -> wfEnv I G B
                -> ICompBinds_I I i k
                -> wfT I G B (T_Ind i) k


   | WFT_VApp : I:icompartment -> G:environment -> B:bindings
                -> u:typ -> v:value -> x:bvar -> t:typ -> k:kind -> kres:kind
                -> wfT I G B u (K_ProdT x t k)
                -> wfV I G B [] Type_level v t
                -> SubstKV k x v kres
                -> wfT I G B (T_VApp u v) kres

   | WFT_App : I:icompartment -> G:environment -> B:bindings
                -> t:typ -> u:typ -> a:bvar -> k:kind -> k':kind -> kres:kind
                -> wfT I G B t (K_ProdK a k k')
                -> wfT I G B u k
                -> SubstKT k' a u kres
                -> wfT I G B (T_App t u) kres

   | WFT_Prod : I:icompartment -> G:environment -> B:bindings
                -> x:bvar -> t:typ -> bt:basekind
                -> u:typ -> bu:basekind
                -> ConcreteKind bt
                -> ConcreteKind bu
                -> wfT I G B t (K_Base bt)
                -> wfT I G ((LB_BvarTy x t)::B) u (K_Base bu)
                -> wfT I G B (T_Prod x t u) (K_Base bu)

   | WFT_ProdK : I:icompartment -> G:environment -> B:bindings
                -> a:bvar -> k:kind -> b:basekind
                -> u:typ -> bu:basekind
                -> ConcreteKind bu
                -> wfK I G B k b
                -> wfT I G ((LB_BvarKind a k)::B) u (K_Base bu)
                -> wfT I G B (T_ProdK a k u) (K_Base bu)

   | WFT_Ref : I:icompartment -> G:environment -> B:bindings
                -> x:bvar -> t:typ -> phi:typ -> bt:basekind
                -> ConcreteKind bt
                -> wfT I G B t (K_Base bt)
                -> wfT I G ((LB_BvarTy x t)::B) phi (K_Base BK_Erase)
                -> wfT I G B (T_Ref x t phi) (K_Base bt)

   | WFT_Lam : I:icompartment -> G:environment -> B:bindings
                -> x:bvar -> t:typ -> bt:basekind -> u:typ -> ku:kind
                -> ConcreteKind bt
                -> wfT I G B t (K_Base bt)
                -> wfT I G ((LB_BvarTy x t)::B) u ku
                -> wfT I G B (T_Lam x t u) (K_ProdT x t ku)

   | WFT_Afn : I:icompartment -> G:environment -> B:bindings
                -> t:typ
                -> wfT I G B t (K_Base BK_Comp)
                -> wfT I G B (T_Afn t) (K_Base BK_Afn)

   | WFT_Ascribe : I:icompartment -> G:environment -> B:bindings
                     -> t:typ -> k:kind
                     -> wfT I G B t k
                     -> wfT I G B (T_Ascribe t k) k

   | WFT_Sub : I:icompartment -> G:environment -> B:bindings
                 -> t:typ -> k:kind -> k':kind -> b:basekind
                 -> wfT I G B t k
                 -> Subkinding I G B k k'
                 -> wfK I G B k' b
                 -> wfT I G B t k'

and wfV :: icompartment => environment => bindings => acontext => mode => value => typ => P =
   | WFV_Var :  I:icompartment -> G:environment -> B:bindings -> m:mode
                  -> x:bvar -> t:typ
                  -> wfEnv I G B
                  -> Mem (LB_BvarTy x t) B
                  -> wfT I G B t (K_Base BK_Comp)
                  -> wfV I G B [] m (V_Var (GV_Bound x)) t

   | WFV_VarT :  I:icompartment -> G:environment -> B:bindings
                  -> x:bvar -> t:typ
                  -> wfEnv I G B
                  -> Mem (LB_BvarTy x t) B
                  -> wfV I G B [] Type_level (V_Var (GV_Bound x)) t

   | WFV_VarA :  I:icompartment -> G:environment -> B:bindings -> m:mode
                  -> x:bvar -> t:typ
                  -> wfEnv I G B
                  -> Mem (LB_BvarTy x t) B
                  -> wfV I G B [x] m (V_Var (GV_Bound x)) t

   | WFV_FVar :  I:icompartment -> G:environment -> B:bindings -> m:mode
                  -> x:vlname -> t:typ
                  -> wfEnv I G B
                  -> Mem (FV_VlName x t) G
                  -> wfV I G B [] m (V_Var (GV_Free x)) t


   | WFV_Const :  I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                   -> c:cname -> ts:seq typ -> vs:seq value
                   -> tc:typ -> u:typ -> k:kind -> x:bvar -> cmon:expr -> e:expr
                   -> CurryT (E_Value (V_Var (GV_Bound x))) ts cmon
                   -> CurryV cmon vs e
                   -> ICompBinds_C I c tc
                   -> wfT I G B u k
                   -> wfE I G ((LB_BvarTy x tc)::B) X m e u
                   -> wfV I G B X m (V_Const c ts vs) u

   | WFV_Fun : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> t:typ -> bt:basekind -> u:typ -> x:bvar -> body:expr -> Q:typ
                 -> AQual X (T_Prod x t u) Q
                 -> wfT I G B t (K_Base bt)
                 -> wfE I G ((LB_BvarTy x t)::B) (x::X) m body u
                 -> wfV I G B X m (V_Fun x t body) Q

   | WFV_FunT :  I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> k:kind -> bk:basekind -> u:typ -> a:bvar -> body:expr -> Q:typ
                 -> AQual X (T_ProdK a k u) Q
                 -> wfK I G B k bk
                 -> wfE I G ((LB_BvarKind a k)::B) X m body u
                 -> wfV I G B X m (V_FunT a k body) Q

   | WFV_Ref :  I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> v:value -> x:bvar -> T:typ -> phi:typ
                 -> LogicEntails I G ((LB_VlEq v (V_Var (GV_Bound x)))::(LB_BvarTy x T)::B) phi
                 -> wfV I G B X m v T
                 -> wfV I G B X m v (T_Ref x T phi)

   | WFV_WeakenX : I:icompartment -> G:environment -> B:bindings -> X:acontext -> X1:acontext -> X2:acontext -> m:mode
                 -> v:value -> T:typ
                 -> AContextSplit X X1 X2
                 -> wfV I G B X1 m v T
                 -> wfV I G B X m v T

   | WFV_Ascribe : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> v:value -> t:typ
                 -> wfV I G B X m v t
                 -> wfV I G B X m (V_Ascribe v t) t

   | WFV_Sub : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> v:value -> t:typ -> u:typ -> k:kind
                 -> wfV I G B X m v t
                 -> Subtyping I G B t u
                 -> wfT I G B u k
                 -> wfV I G B X m v u

and wfE :: icompartment => environment => bindings => acontext => mode => expr => typ => P =
   | WFE_Val : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
             -> v:value -> T:typ
             -> wfV I G B X m v T
             -> wfE I G B X m (E_Value v) T

   | WFE_App : I:icompartment -> G:environment -> B:bindings -> X:acontext -> X1:acontext -> X2:acontext -> m:mode
             -> e:expr -> v:value -> x:bvar -> t:typ -> u:typ -> tv:typ -> tres:typ
             -> AContextSplit X X1 X2
             -> StripQual tv (T_Prod x t u)
             -> SubstTV u x v tres
             -> wfE I G B X1 m e tv
             -> wfV I G B X2 m v t
             -> wfE I G B X m (E_App e v) tres

   | WFE_TApp : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                  -> a:bvar -> k:kind -> t:typ -> e:expr -> te:typ -> u:typ -> tres:typ
                  -> StripQual te (T_ProdK a k u)
                  -> SubstTT t a u tres
                  -> wfE I G B X m e te
                  -> wfE I G B X m (E_TApp e t) tres

   | WFE_LetIn : I:icompartment -> G:environment -> B:bindings -> X:acontext -> X1:acontext -> X2:acontext -> m:mode
                -> x:bvar -> e:expr -> body:expr -> t:typ -> u:typ -> bu:basekind
                -> AContextSplit X X1 X2
                -> wfT I G B u (K_Base bu)
                -> wfE I G B X1 m e t
                -> wfE I G ((LB_BvarTy x t)::B) (x::X2) m body u
                -> wfE I G B X m (E_LetIn x e body) u


   | WFE_Assume : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                   -> x:bvar -> phi:typ
                   -> wfT I G B phi (K_Base BK_Erase)
                   -> wfE I G B X m (E_Assume phi) (T_Ref x (T_Ind "unit") phi)

   | WFE_WeakenX : I:icompartment -> G:environment -> B:bindings -> X:acontext -> X1:acontext -> X2:acontext -> m:mode
                    -> e:expr -> t:typ
                    -> AContextSplit X X1 X2
                    -> wfE I G B X1 m e t
                    -> wfE I G B X m e t

   | WFE_Ascribe : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                     -> e:expr -> t:typ
                     -> wfE I G B X m e t
                     -> wfE I G B X m (E_Ascribe e t) t

   | WFE_Sub : I:icompartment -> G:environment -> B:bindings -> X:acontext -> m:mode
                 -> e:expr -> t:typ -> u:typ -> k:kind
                 -> wfE I G B X m e t
                 -> Subtyping I G B t u
                 -> wfT I G B u k
                 -> wfE I G B X m e u


(*type texpr :: _ = fun (e:expr)  => (I:icompartment -> G:environment -> B:bindings -> A:acontext 
                                     -> m:mode -> (t:typ * A1:acontext * A2:acontext * AContextSplit A A1 A2 * 
                                                            wfE I G B A1 m e t))
type tval :: _  = fun (v:value) => (I:icompartment -> G:environment -> B:bindings -> A:acontext 
                                     -> m:mode -> (t:typ * A1:acontext * A2:acontext * AContextSplit A A1 A2 * 
                                                            wfV I G B A1 m v t))
type ktyp :: _  = fun (t:typ)   => (I:icompartment -> G:environment -> B:bindings -> (k:kind * wfT I G B t k))
type okk  :: _  = fun (k:kind)  => (I:icompartment -> G:environment -> B:bindings -> (b:basekind * wfK I G B k b))
*)

end

(* ****************************************************************************** *)

