module STLC

type bvar = string 
type seq 'a = list 'a

type Eq :: 'a::* => 'a => 'a => P = 
  | Refl_eq : 'a::* -> x:'a -> Eq 'a x x

type ltyp :: P =
  | LTyp_unit : ltyp
  | LTyp_fun : ltyp -> ltyp -> ltyp

type lexpr :: P =
  | LExpr_unit : lexpr
  | LExpr_var : bvar -> lexpr
  | LExpr_fun : bvar -> ltyp -> lexpr -> lexpr
  | LExpr_app : lexpr -> lexpr -> lexpr
  
type lenv = seq (bvar * ltyp)

type lBinds :: lenv => bvar => ltyp => P =
  | LBinds_hd : 
      tl:lenv -> x:bvar -> t:ltyp 
    -> lBinds ((x,t)::tl) x t

  | LBinds_tl : 
      y:bvar -> t':ltyp -> tl:lenv -> x:bvar -> t:ltyp 
    -> lBinds tl x t 
    -> dummy:(Eq true true){y<>x}
    -> lBinds ((y, t')::tl) x t

type lTyping :: lenv => lexpr => ltyp => P =
   | LTyping_unit : g:lenv -> lTyping g LExpr_unit LTyp_unit

   | LTyping_var : 
       g:lenv -> x:bvar -> t:ltyp 
      -> lBinds g x t
      -> lTyping g (LExpr_var x) t
    
   | LTyping_fun : 
       g:lenv -> x:bvar -> t:ltyp -> e:lexpr -> t':ltyp
    -> lTyping ((x,t)::g) e t'
    -> lTyping g (LExpr_fun x t e) (LTyp_fun t t')

   | LTyping_app : 
        g:lenv -> e1:lexpr -> e2:lexpr -> t1:ltyp -> t2:ltyp 
     -> lTyping g e1 (LTyp_fun t1 t2) 
     -> lTyping g e2 t1
     -> lTyping g (LExpr_app e1 e2) t2


(* *************************************************************** *)

val lookup : x:bvar -> g:lenv -> option (t:ltyp * lBinds g x t)
let rec lookup x g  = match g with 
  | [] -> None
  | (y,ty)::tl -> 
      if x=y then Some (ty, LBinds_hd tl x ty)
      else 
        (match lookup x tl with 
           | Some ((tx, pf)) -> Some (tx, LBinds_tl y ty tl x tx pf (Refl_eq true))
           | None -> None)

val stlc_checker : g:lenv -> e:lexpr -> (t:ltyp * lTyping g e t)
let rec stlc_checker g e = match e with 
  | LExpr_unit -> (LTyp_unit, (LTyping_unit g))
  | LExpr_var x -> 
      (match lookup x g with 
         | Some ((t, binds)) -> (t, LTyping_var g x t binds)
         | None -> raise "fail")
  | LExpr_fun x t e -> 
      (match stlc_checker ((x,t)::g) e with 
         | (t', pf) -> (LTyp_fun t t'), (LTyping_fun g x t e t' pf)
         | _ -> raise "fail")
  | LExpr_app e1 e2 -> 
      match stlc_checker g e1 with 
        | (LTyp_fun t1 t2, pf1) -> 
            (match stlc_checker g e2 with 
               | (t1', pf2) when t1=t1' -> (t2, LTyping_app g e1 e2 t1 t2 pf1 pf2)
               | _ -> raise "fail")
        | _ -> raise "fail"
            

