#light "off"

module Microsoft.FStar.ProofState
open Absyn
open ProofCommon

#if PARALLEL
 type ctx = Microsoft.Z3V3.Context
#else
type ctx = Microsoft.Z3.Context
#endif

type z3={ctx:Microsoft.Z3.Context; solver:Microsoft.Z3.Solver}
val z3: z3 ref
val init : unit -> unit
val push : unit -> unit
val pop : unit -> unit
val clear : unit -> unit
val setCurMod:lident -> unit
val curMod:unit -> string 

type name = list<string>
type polyinst = {plid:lident; pkind:kind; targs:list<typ>; mkind:kind; mlid:lident;} 
module Ops : sig
    val storeOps : ops -> unit
    val storeDelayedOps : ops -> unit
    val getOps : unit -> ops
    val getDelayedOps : unit -> ops
    val storePolyInsts : list<polyinst> -> unit
    val getPolyInsts : unit -> list<polyinst>
end

module StringConstants : sig
    val stringOfId : int -> string
    val idOptOfString : string -> option<int>
    val addString: string -> int
end 

module Props : sig 
    val addProp : name -> propid
    val nameOfId: propid -> name
    val idOfName: name -> propid
    val idOptOfName: name -> option<propid>
end


module TypeConstructors : sig
  val addTyp : name -> kind -> sort
  val nameOfSort : sort -> name
  val sortOfName : name -> sort
  val typOfSort : sort -> typ
  val sortOptOfTyp : typ -> option<sort>
  val dconIdsOfSort : sort -> list<dconid>
  val tryAddInstantiation : sort -> list<sort> -> bool
  val isMonomorphic : sort -> bool
  val kindOfSort : sort -> kind
  val isUniform : sort -> bool
  val tagAsNonUniform : sort -> unit
end

module DataConstructors : sig
  type instantiation = {monoAx: option<var<typ>>;
                        retSort: sort}
  val addConstructor : name -> typ -> tconid -> dconid
  val constructorsOfTyp : tconid -> list<dconid * list<typ>>
  val idOfName: name -> dconid
  val idOptOfName: name -> option<dconid>
  val nameOfDconid: dconid -> name 
  val monoAxiom: dconinst -> option<var<typ>>
  val typOfDconid : dconid -> typ
  val destructType: dconid -> (list<typ> * typ * list<btvdef>)
  val addInstantiation: dconinst -> instantiation -> unit
  val retSortOfDconinst: dconinst -> sort
  val isUniform : dconid -> bool
end 

module Records : sig   
  type inst = {sort:sort;
               fnames:list<fieldname>;
               fsorts:list<sort>;
               projAxs:list<fieldid * var<typ>>;
               eqAxs:list<fieldid * var<typ>>;
               monoAx:var<typ>}
  val addRecord : name -> list<fieldname*typ> -> bool -> recdconid
  val idOfName: name -> recdconid
  val typOfSort : sort -> typ
  val sortOptOfTyp : typ -> option<sort>
  val projAxiom: sort -> fieldid -> var<typ>
  val eqAxiom: sort -> fieldid -> var<typ> 
  val monoAxiom: sort -> var<typ>
  val fieldName: fieldid -> name
  val fieldId: name -> fieldid
  val addInstantiation: inst -> unit 
  val fieldNamesAndTypsOfSort : sort -> list<fieldname*typ>
end

module Tuples : sig 
  val addInstantiation : sort -> (sort * sort) -> unit
  val argSorts: sort -> (sort*sort)
end

module BoundTvars : sig
  val addBvar : btvar -> sort
  val typOfSort : sort -> typ
  val sortOptOfTyp : typ -> option<sort>
end

module Uvars : sig
  val addUvar : Absyn.uvar -> sort 
  val typOfSort : sort -> typ
  val sortOptOfTyp : typ -> option<sort>
  val opsForUvars : unit -> ops
end

module TypInsts : sig
  val addTypinst : typinst -> sort
  val moduleOfSort : sort -> string
  val typinstOfSort : sort -> typinst
  val sortOptOfTypinst : typinst -> option<sort>
end

module Disequalities : sig
  val setNeqBool: var<typ> -> unit
  val lookupNeqBool: unit -> var<typ>
  val addDiseqAx: (int -> var<typ>) -> int
end

module Z3Symbols : sig
  val symOfPropid : (propid -> string)
  val propidOfSym : (string -> propid option)

  val symOfDconinst : (dconinst -> string)
  val dconinstOfSym : (string -> dconinst option)

  val symOfTuple : (sort -> string)
  val symOfTupleProj1 : (sort -> string)
  val symOfTupleProj2 : (sort -> string)
  val tupleCtorOfSym : (string -> sort option)
  val tupleProj1OfSym : (string -> sort option)
  val tupleProj2OfSym : (string -> sort option)

  val symOfRecd : (sort -> string)
  val symOfRecdProj : (sort -> fieldid -> string)
  val recdCtorOfSym : (string -> sort option)
  val recdProjOfSym : (string -> (sort * fieldid) option)  
 
  val symOfIdent : (ident -> sort -> string)
  val identSortOptOfSym : (string -> sort option)

  val symOfString : (string -> string)
  val stringidOfSym : (string -> int option)
  
  val qvarOfBvd: bvvdef -> typ -> sort -> string
  val binderOptOfSym : string -> option<bvvdef*typ>
end

#if PARALLEL
open Microsoft.Z3V3
#else
open Microsoft.Z3
#endif
