// Implementation file for parser generated by fsyacc
module Microsoft.FStar.Pars
#nowarn "64";; // turn off warnings that type variables used in production annotations are instantiated to concrete type
open Microsoft.FSharp.Text.Lexing
open Microsoft.FSharp.Text.Parsing.ParseHelpers
# 1 "parser/pars.fsy"

// Adapted from sources of F#
// See LICENSE-fsharp.txt for terms and conditions
// (c) Microsoft Corporation. All rights reserved
open Microsoft.FSharp.Compatibility
open Microsoft.FSharp.Compatibility.OCaml.Pervasives

(* open Microsoft.FSharp.Compiler.AbstractIL *)
(* open Microsoft.FSharp.Compiler.AbstractIL.Internal *)
open Microsoft.FSharp.Compiler
open Microsoft.FSharp.Text.Parsing

open System
open Microsoft.FStar
open Microsoft.FStar.Util
open Microsoft.FStar.Range
open Microsoft.FStar.Sugar
open Microsoft.FStar.Const
// open Microsoft.FSharp.Compiler.Lib


(* error recovery*)
let arbExpr(parseState) = Expr_const(Const_unit,lhs(parseState))

(* let mksyn_anon_constraint ty m = Type_anon_constraint(ty,m)  *)
let is_upper_string (s:string) = 
    s.Length >= 1 && System.Char.IsUpper s.[0] && not (System.Char.IsLower s.[0])
 
let invalid_arg2 arg msg = raise (new System.ArgumentException((msg:string),(arg:string)))        
let FrontAndBack l = 
    let rec loop acc l = 
        match l with
        | [] -> 
            System.Diagnostics.Debug.Assert(false, "empty list")
            invalid_arg2 "l" "empty list" 
        | [h] -> List.rev acc,h
        | h::t -> loop  (h::acc) t
    loop [] l

(* This function is called by the generated parser code. Returning initiates error recovery *)
let parse_error_rich = Some (fun (ctxt: ParseErrorContext<_>) -> 
    errorR(SyntaxError(box ctxt,ctxt.Message, GetLexerRange (GetParserLexbuf ctxt.ParseState))))

let report_parse_error_at m s = errorR(Error(s,m))

let report_parse_warning_at m s = warning(Error(s,m))

(* OCaml generated parsers seem to create corrupted terms if parse error are raised when an "error" symbol is being shifted *)
let raise_parse_error_at m s = 
    report_parse_error_at m s; 
    // This initiates error recovery
    raise RecoverableParseError 

let check_eof_error_USE_ONLY_IF_NOT_IN_ERROR_RECOVERY res t = 
  match t with 
  | AT_hashif(_,m) -> raise_parse_error_at m "end of file in #if section begun at or after here"
  | AT_ifdef_skip(_,_,m) -> raise_parse_error_at m "end of file in #if section begun at or after here"
  | AT_string (_,m) ->  raise_parse_error_at m "end of file in string begun at or before here"
  | AT_vstring (_,m) ->  raise_parse_error_at m "end of file in verbatim string begun at or before here"
  | AT_comment (_,_,m) ->  raise_parse_error_at m "end of file in comment begun at or before here"
  | AT_comment_string (_,_,m) -> raise_parse_error_at m "end of file in string embedded in comment begun at or before here"
  | AT_camlonly (_,m) -> raise_parse_error_at m "end of file in IF-OCAML section begun at or before here" 
  | AT_token(_) -> res

type BindingSet = BindingSet of range * bool * bool * binding list

let mkLocalBindings wholem (BindingSet(_,isRec,isUse,decls)) body = 
   Expr_let (isRec,isUse,decls,body,wholem) 

let mkDefnBindings wholem (BindingSet(_,isRec,isUse,decls))  = 
  if isUse then warning(Error("'use' bindings are treated as 'let' bindings in modules",wholem));
  [ Def_let (isRec,decls,wholem) ] 

let id_of_pat m p = match p with Pat_as (Pat_wild _,id,false,_) -> id | _ -> raise_parse_error_at m "A integer for loop must use a simple identifier"

let prefix s t = s^t : string 

let irrelevant_name_binding : option<ident> -> unit = function
    None -> ()
  | Some x ->  report_parse_warning_at x.idRange "Ignoring irrelevant name binding"


let string_to_lang s r = match s with
  | "F#" -> FSharp 
  | "C#" -> CSharp 
  | _ -> report_parse_error_at r ("unexpected extern language: "^s); FSharp in

let failExtern m = raise_parse_error_at m "Path to extern value should be of the form Assembly.Class1.Class2...Classn.Methodname"
  

# 97 "parser\pars.fs"
// This type is the type of tokens accepted by the parser
type token = 
  | COMMENT of (Sugar.lexcont)
  | WHITESPACE of (Sugar.lexcont)
  | HASH_LINE of (Sugar.lexcont)
  | HASH_LIGHT of (Sugar.lexcont)
  | HASH_IF of (Sugar.lexcont)
  | HASH_ELSE of (Sugar.lexcont)
  | HASH_ENDIF of (Sugar.lexcont)
  | INACTIVECODE of (Sugar.lexcont)
  | LINE_COMMENT of (Sugar.lexcont)
  | STRING_TEXT of (Sugar.lexcont)
  | EOF of (Sugar.lexcont)
  | LEX_FAILURE of (string)
  | ODUMMY of (token)
  | OINTERFACE_MEMBER
  | OBLOCKEND
  | ORIGHT_BLOCK_END
  | ODECLEND
  | OEND
  | OBLOCKSEP
  | OBLOCKBEGIN
  | ORESET
  | OFUN
  | OFUNCTION
  | OWITH
  | OELSE
  | OTHEN
  | ODO_BANG
  | ODO
  | OBINDER of (string)
  | OLET of (bool)
  | HIGH_PRECEDENCE_TYAPP
  | HIGH_PRECEDENCE_APP
  | EXTERN
  | REFERENCE
  | VOID
  | PUBLIC
  | PRIVATE
  | INTERNAL
  | STATIC
  | MEMBER
  | CLASS
  | VIRTUAL
  | ABSTRACT
  | OVERRIDE
  | DEFAULT
  | CONSTRUCTOR
  | INHERIT
  | GREATER_RBRACK
  | STRUCT
  | SIG
  | BAR
  | RBRACK
  | RBRACE
  | MINUS
  | DOLLAR
  | LBRACE_LESS
  | BAR_RBRACK
  | GREATER_RBRACE
  | UNDERSCORE
  | BANG
  | SEMICOLON_SEMICOLON
  | LARROW
  | EQUALS
  | LBRACK
  | LBRACK_BAR
  | LBRACK_LESS
  | LBRACE
  | BACKSLASH
  | QMARK
  | QMARK_QMARK
  | DOT
  | COLON
  | COLON_COLON
  | COLON_GREATER
  | COLON_QMARK_GREATER
  | COLON_QMARK
  | COLON_EQUALS
  | SEMICOLON
  | WHEN
  | WHILE
  | WITH
  | HASH
  | AMP
  | AMP_AMP
  | QUOTE
  | LPAREN
  | RPAREN
  | STAR
  | COMMA
  | RARROW
  | RARROW2
  | RRARROW
  | IFF
  | GREATER_DOT
  | GREATER_BAR_RBRACK
  | LPAREN_STAR_RPAREN
  | OPEN
  | OR
  | PROP
  | REC
  | THEN
  | TO
  | TRUE
  | TRY
  | TYPE
  | VAL
  | INLINE
  | INTERFACE
  | INSTANCE
  | LAZY
  | MATCH
  | METHOD
  | MUTABLE
  | NEW
  | OF
  | EXCEPTION
  | FALSE
  | FOR
  | FUN
  | FUNCTION
  | IF
  | IN
  | FINALLY
  | DO_BANG
  | AND
  | AS
  | ASSERT
  | ASR
  | BEGIN
  | DO
  | DONE
  | DOWNTO
  | ELSE
  | ELIF
  | END
  | DOT_DOT
  | BAR_BAR
  | LESS
  | GREATER
  | UPCAST
  | DOWNCAST
  | NULL
  | RESERVED
  | MODULE
  | DELEGATE
  | CONSTRAINT
  | BASE
  | FORALL
  | EXISTS
  | NOT
  | ASSUME
  | LQUOTE of (string * bool)
  | RQUOTE of (string * bool)
  | SPLICE_SYMBOL of (string)
  | PERCENT_OP of (string)
  | BINDER of (string)
  | LET of (bool)
  | YIELD of (bool)
  | YIELD_BANG of (bool)
  | BIGINT of (byte[])
  | BIGNUM of (byte[])
  | DECIMAL of (System.Decimal)
  | CHAR of (char)
  | IEEE64 of (double)
  | IEEE32 of (single)
  | NATIVEINT of (int64)
  | UNATIVEINT of (uint64)
  | UINT64 of (uint64)
  | UINT32 of (uint32)
  | UINT16 of (uint16)
  | UINT8 of (byte)
  | INT64 of (int64 * bool)
  | INT32 of (int32 * bool)
  | INT32_DOT_DOT of (int32 * bool)
  | INT16 of (int16 * bool)
  | INT8 of (sbyte * bool)
  | FUNKY_OPERATOR_NAME of (string)
  | ADJACENT_PREFIX_PLUS_MINUS_OP of (string)
  | PLUS_MINUS_OP of (string)
  | INFIX_AMP_OP of (string)
  | INFIX_STAR_DIV_MOD_OP of (string)
  | PREFIX_OP of (string)
  | INFIX_BAR_OP of (string)
  | INFIX_AT_HAT_OP of (string)
  | INFIX_COMPARE_OP of (string)
  | INFIX_STAR_STAR_OP of (string)
  | LANG of (string)
  | BASEKIND of (string)
  | IDENT of (string)
  | STRING of (byte[])
  | BYTEARRAY of (byte[])
// This type is used to give symbolic names to token indexes, useful for error messages
type tokenId = 
    | TOKEN_COMMENT
    | TOKEN_WHITESPACE
    | TOKEN_HASH_LINE
    | TOKEN_HASH_LIGHT
    | TOKEN_HASH_IF
    | TOKEN_HASH_ELSE
    | TOKEN_HASH_ENDIF
    | TOKEN_INACTIVECODE
    | TOKEN_LINE_COMMENT
    | TOKEN_STRING_TEXT
    | TOKEN_EOF
    | TOKEN_LEX_FAILURE
    | TOKEN_ODUMMY
    | TOKEN_OINTERFACE_MEMBER
    | TOKEN_OBLOCKEND
    | TOKEN_ORIGHT_BLOCK_END
    | TOKEN_ODECLEND
    | TOKEN_OEND
    | TOKEN_OBLOCKSEP
    | TOKEN_OBLOCKBEGIN
    | TOKEN_ORESET
    | TOKEN_OFUN
    | TOKEN_OFUNCTION
    | TOKEN_OWITH
    | TOKEN_OELSE
    | TOKEN_OTHEN
    | TOKEN_ODO_BANG
    | TOKEN_ODO
    | TOKEN_OBINDER
    | TOKEN_OLET
    | TOKEN_HIGH_PRECEDENCE_TYAPP
    | TOKEN_HIGH_PRECEDENCE_APP
    | TOKEN_EXTERN
    | TOKEN_REFERENCE
    | TOKEN_VOID
    | TOKEN_PUBLIC
    | TOKEN_PRIVATE
    | TOKEN_INTERNAL
    | TOKEN_STATIC
    | TOKEN_MEMBER
    | TOKEN_CLASS
    | TOKEN_VIRTUAL
    | TOKEN_ABSTRACT
    | TOKEN_OVERRIDE
    | TOKEN_DEFAULT
    | TOKEN_CONSTRUCTOR
    | TOKEN_INHERIT
    | TOKEN_GREATER_RBRACK
    | TOKEN_STRUCT
    | TOKEN_SIG
    | TOKEN_BAR
    | TOKEN_RBRACK
    | TOKEN_RBRACE
    | TOKEN_MINUS
    | TOKEN_DOLLAR
    | TOKEN_LBRACE_LESS
    | TOKEN_BAR_RBRACK
    | TOKEN_GREATER_RBRACE
    | TOKEN_UNDERSCORE
    | TOKEN_BANG
    | TOKEN_SEMICOLON_SEMICOLON
    | TOKEN_LARROW
    | TOKEN_EQUALS
    | TOKEN_LBRACK
    | TOKEN_LBRACK_BAR
    | TOKEN_LBRACK_LESS
    | TOKEN_LBRACE
    | TOKEN_BACKSLASH
    | TOKEN_QMARK
    | TOKEN_QMARK_QMARK
    | TOKEN_DOT
    | TOKEN_COLON
    | TOKEN_COLON_COLON
    | TOKEN_COLON_GREATER
    | TOKEN_COLON_QMARK_GREATER
    | TOKEN_COLON_QMARK
    | TOKEN_COLON_EQUALS
    | TOKEN_SEMICOLON
    | TOKEN_WHEN
    | TOKEN_WHILE
    | TOKEN_WITH
    | TOKEN_HASH
    | TOKEN_AMP
    | TOKEN_AMP_AMP
    | TOKEN_QUOTE
    | TOKEN_LPAREN
    | TOKEN_RPAREN
    | TOKEN_STAR
    | TOKEN_COMMA
    | TOKEN_RARROW
    | TOKEN_RARROW2
    | TOKEN_RRARROW
    | TOKEN_IFF
    | TOKEN_GREATER_DOT
    | TOKEN_GREATER_BAR_RBRACK
    | TOKEN_LPAREN_STAR_RPAREN
    | TOKEN_OPEN
    | TOKEN_OR
    | TOKEN_PROP
    | TOKEN_REC
    | TOKEN_THEN
    | TOKEN_TO
    | TOKEN_TRUE
    | TOKEN_TRY
    | TOKEN_TYPE
    | TOKEN_VAL
    | TOKEN_INLINE
    | TOKEN_INTERFACE
    | TOKEN_INSTANCE
    | TOKEN_LAZY
    | TOKEN_MATCH
    | TOKEN_METHOD
    | TOKEN_MUTABLE
    | TOKEN_NEW
    | TOKEN_OF
    | TOKEN_EXCEPTION
    | TOKEN_FALSE
    | TOKEN_FOR
    | TOKEN_FUN
    | TOKEN_FUNCTION
    | TOKEN_IF
    | TOKEN_IN
    | TOKEN_FINALLY
    | TOKEN_DO_BANG
    | TOKEN_AND
    | TOKEN_AS
    | TOKEN_ASSERT
    | TOKEN_ASR
    | TOKEN_BEGIN
    | TOKEN_DO
    | TOKEN_DONE
    | TOKEN_DOWNTO
    | TOKEN_ELSE
    | TOKEN_ELIF
    | TOKEN_END
    | TOKEN_DOT_DOT
    | TOKEN_BAR_BAR
    | TOKEN_LESS
    | TOKEN_GREATER
    | TOKEN_UPCAST
    | TOKEN_DOWNCAST
    | TOKEN_NULL
    | TOKEN_RESERVED
    | TOKEN_MODULE
    | TOKEN_DELEGATE
    | TOKEN_CONSTRAINT
    | TOKEN_BASE
    | TOKEN_FORALL
    | TOKEN_EXISTS
    | TOKEN_NOT
    | TOKEN_ASSUME
    | TOKEN_LQUOTE
    | TOKEN_RQUOTE
    | TOKEN_SPLICE_SYMBOL
    | TOKEN_PERCENT_OP
    | TOKEN_BINDER
    | TOKEN_LET
    | TOKEN_YIELD
    | TOKEN_YIELD_BANG
    | TOKEN_BIGINT
    | TOKEN_BIGNUM
    | TOKEN_DECIMAL
    | TOKEN_CHAR
    | TOKEN_IEEE64
    | TOKEN_IEEE32
    | TOKEN_NATIVEINT
    | TOKEN_UNATIVEINT
    | TOKEN_UINT64
    | TOKEN_UINT32
    | TOKEN_UINT16
    | TOKEN_UINT8
    | TOKEN_INT64
    | TOKEN_INT32
    | TOKEN_INT32_DOT_DOT
    | TOKEN_INT16
    | TOKEN_INT8
    | TOKEN_FUNKY_OPERATOR_NAME
    | TOKEN_ADJACENT_PREFIX_PLUS_MINUS_OP
    | TOKEN_PLUS_MINUS_OP
    | TOKEN_INFIX_AMP_OP
    | TOKEN_INFIX_STAR_DIV_MOD_OP
    | TOKEN_PREFIX_OP
    | TOKEN_INFIX_BAR_OP
    | TOKEN_INFIX_AT_HAT_OP
    | TOKEN_INFIX_COMPARE_OP
    | TOKEN_INFIX_STAR_STAR_OP
    | TOKEN_LANG
    | TOKEN_BASEKIND
    | TOKEN_IDENT
    | TOKEN_STRING
    | TOKEN_BYTEARRAY
    | TOKEN_end_of_input
    | TOKEN_error
// This type is used to give symbolic names to token indexes, useful for error messages
type nonTerminalId = 
    | NONTERM__startimplementationFile
    | NONTERM_implementationFile
    | NONTERM_fileModuleImpls
    | NONTERM_fileModuleImpl
    | NONTERM_moduleIntro
    | NONTERM_moduleDefnsOrExprPossiblyEmpty
    | NONTERM_moduleDefnsOrExpr
    | NONTERM_moduleDefns
    | NONTERM_moduleDefn
    | NONTERM_extern_ref
    | NONTERM_maybenullary
    | NONTERM_defnBindings
    | NONTERM_hardwhiteLetBindings
    | NONTERM_hardwhiteDefnBindingsTerminator
    | NONTERM_localBindings
    | NONTERM_more_localBindings
    | NONTERM_localBinding
    | NONTERM_bindingPattern
    | NONTERM_headBindingPattern
    | NONTERM_tuplePatternElements
    | NONTERM_constrPattern
    | NONTERM_atomicPatterns
    | NONTERM_atomicPattern
    | NONTERM_atomicPatternLongIdent
    | NONTERM_recordPatternElements
    | NONTERM_moreRecordPatternElements
    | NONTERM_listPatternElements
    | NONTERM_moreListPatternElements
    | NONTERM_parenPatternBody
    | NONTERM_parenPattern
    | NONTERM_tupleParenPatternElements
    | NONTERM_constant
    | NONTERM_typedSeqExprBlock
    | NONTERM_declExprBlock
    | NONTERM_typedSeqExpr
    | NONTERM_seqExpr
    | NONTERM_recover
    | NONTERM_declExpr
    | NONTERM_withClauses
    | NONTERM_withPatternClauses
    | NONTERM_patternAndGuard
    | NONTERM_patternClauses
    | NONTERM_patternGuard
    | NONTERM_patternResult
    | NONTERM_typedSeqExprBlockR
    | NONTERM_ifExprCases
    | NONTERM_ifExprThen
    | NONTERM_ifExprElifs
    | NONTERM_tupleExpr
    | NONTERM_minusExpr
    | NONTERM_appExpr
    | NONTERM_argExpr
    | NONTERM_atomicExpr
    | NONTERM_listExprElements
    | NONTERM_atomicExprQualification
    | NONTERM_atomicExprAfterType
    | NONTERM_beginEndExpr
    | NONTERM_parenExpr
    | NONTERM_parenExprBody
    | NONTERM_braceExpr
    | NONTERM_braceExprBody
    | NONTERM_opt_atomicExprAfterType
    | NONTERM_recdExpr
    | NONTERM_recdExprBindings
    | NONTERM_anonLambdaExpr
    | NONTERM_anonMatchingExpr
    | NONTERM_tyconDefnList
    | NONTERM_tyconDefn
    | NONTERM_typeNameInfo
    | NONTERM_tyconNameAndTyparDecls
    | NONTERM_opt_kind
    | NONTERM_kind
    | NONTERM_atomic_kind_with_binding
    | NONTERM_typ_with_binding
    | NONTERM_base_kind
    | NONTERM_postfixTyparDecls
    | NONTERM_typeparDeclsActual
    | NONTERM_onlyTypars
    | NONTERM_onlyTyparList
    | NONTERM_typarIndices
    | NONTERM_termparIndices
    | NONTERM_moreTermParIndices
    | NONTERM_termparDecl
    | NONTERM_tyconDefnRhs
    | NONTERM_unionRepr
    | NONTERM_unionCaseDecls
    | NONTERM_unionCaseDecl
    | NONTERM_braceFieldDeclList
    | NONTERM_recdFieldDeclList
    | NONTERM_recdFieldDecl
    | NONTERM_opt_topReturnType
    | NONTERM_typ
    | NONTERM_tfun
    | NONTERM_appType
    | NONTERM_typars
    | NONTERM_atomType
    | NONTERM_tupleType
    | NONTERM_dtupleTypeTail
    | NONTERM_tupleTypeTail
    | NONTERM_typeArgs
    | NONTERM_typeArgsActual
    | NONTERM_typeIndices
    | NONTERM_termIndices
    | NONTERM_appTypeCon
    | NONTERM_typar
    | NONTERM_atomic_kind
    | NONTERM_formula
    | NONTERM_Form
    | NONTERM_Form0
    | NONTERM_Form0a
    | NONTERM_Form1
    | NONTERM_Form2
    | NONTERM_Form3
    | NONTERM_Form3a
    | NONTERM_Form4
    | NONTERM_Forall
    | NONTERM_Exists
    | NONTERM_fbinder
    | NONTERM_fbinders
    | NONTERM_atom
    | NONTERM_termIndexSeq
    | NONTERM_value
    | NONTERM_valueSeq
    | NONTERM_atomicValue
    | NONTERM_valueConst
    | NONTERM_parenValue
    | NONTERM_parenValBody
    | NONTERM_tupleVal
    | NONTERM_ident
    | NONTERM_opt_ident_colon
    | NONTERM_path
    | NONTERM_opname
    | NONTERM_operatorName
    | NONTERM_barName
    | NONTERM_barNames
    | NONTERM_top_sep
    | NONTERM_top_seps
    | NONTERM_seps
    | NONTERM_opt_OBLOCKSEP
    | NONTERM_opt_seps
    | NONTERM_opt_rec
    | NONTERM_opt_bar
    | NONTERM_opt_HIGH_PRECEDENCE_TYAPP

// This function maps tokens to integers indexes
let tagOfToken (t:token) = 
  match t with
  | COMMENT _ -> 0 
  | WHITESPACE _ -> 1 
  | HASH_LINE _ -> 2 
  | HASH_LIGHT _ -> 3 
  | HASH_IF _ -> 4 
  | HASH_ELSE _ -> 5 
  | HASH_ENDIF _ -> 6 
  | INACTIVECODE _ -> 7 
  | LINE_COMMENT _ -> 8 
  | STRING_TEXT _ -> 9 
  | EOF _ -> 10 
  | LEX_FAILURE _ -> 11 
  | ODUMMY _ -> 12 
  | OINTERFACE_MEMBER  -> 13 
  | OBLOCKEND  -> 14 
  | ORIGHT_BLOCK_END  -> 15 
  | ODECLEND  -> 16 
  | OEND  -> 17 
  | OBLOCKSEP  -> 18 
  | OBLOCKBEGIN  -> 19 
  | ORESET  -> 20 
  | OFUN  -> 21 
  | OFUNCTION  -> 22 
  | OWITH  -> 23 
  | OELSE  -> 24 
  | OTHEN  -> 25 
  | ODO_BANG  -> 26 
  | ODO  -> 27 
  | OBINDER _ -> 28 
  | OLET _ -> 29 
  | HIGH_PRECEDENCE_TYAPP  -> 30 
  | HIGH_PRECEDENCE_APP  -> 31 
  | EXTERN  -> 32 
  | REFERENCE  -> 33 
  | VOID  -> 34 
  | PUBLIC  -> 35 
  | PRIVATE  -> 36 
  | INTERNAL  -> 37 
  | STATIC  -> 38 
  | MEMBER  -> 39 
  | CLASS  -> 40 
  | VIRTUAL  -> 41 
  | ABSTRACT  -> 42 
  | OVERRIDE  -> 43 
  | DEFAULT  -> 44 
  | CONSTRUCTOR  -> 45 
  | INHERIT  -> 46 
  | GREATER_RBRACK  -> 47 
  | STRUCT  -> 48 
  | SIG  -> 49 
  | BAR  -> 50 
  | RBRACK  -> 51 
  | RBRACE  -> 52 
  | MINUS  -> 53 
  | DOLLAR  -> 54 
  | LBRACE_LESS  -> 55 
  | BAR_RBRACK  -> 56 
  | GREATER_RBRACE  -> 57 
  | UNDERSCORE  -> 58 
  | BANG  -> 59 
  | SEMICOLON_SEMICOLON  -> 60 
  | LARROW  -> 61 
  | EQUALS  -> 62 
  | LBRACK  -> 63 
  | LBRACK_BAR  -> 64 
  | LBRACK_LESS  -> 65 
  | LBRACE  -> 66 
  | BACKSLASH  -> 67 
  | QMARK  -> 68 
  | QMARK_QMARK  -> 69 
  | DOT  -> 70 
  | COLON  -> 71 
  | COLON_COLON  -> 72 
  | COLON_GREATER  -> 73 
  | COLON_QMARK_GREATER  -> 74 
  | COLON_QMARK  -> 75 
  | COLON_EQUALS  -> 76 
  | SEMICOLON  -> 77 
  | WHEN  -> 78 
  | WHILE  -> 79 
  | WITH  -> 80 
  | HASH  -> 81 
  | AMP  -> 82 
  | AMP_AMP  -> 83 
  | QUOTE  -> 84 
  | LPAREN  -> 85 
  | RPAREN  -> 86 
  | STAR  -> 87 
  | COMMA  -> 88 
  | RARROW  -> 89 
  | RARROW2  -> 90 
  | RRARROW  -> 91 
  | IFF  -> 92 
  | GREATER_DOT  -> 93 
  | GREATER_BAR_RBRACK  -> 94 
  | LPAREN_STAR_RPAREN  -> 95 
  | OPEN  -> 96 
  | OR  -> 97 
  | PROP  -> 98 
  | REC  -> 99 
  | THEN  -> 100 
  | TO  -> 101 
  | TRUE  -> 102 
  | TRY  -> 103 
  | TYPE  -> 104 
  | VAL  -> 105 
  | INLINE  -> 106 
  | INTERFACE  -> 107 
  | INSTANCE  -> 108 
  | LAZY  -> 109 
  | MATCH  -> 110 
  | METHOD  -> 111 
  | MUTABLE  -> 112 
  | NEW  -> 113 
  | OF  -> 114 
  | EXCEPTION  -> 115 
  | FALSE  -> 116 
  | FOR  -> 117 
  | FUN  -> 118 
  | FUNCTION  -> 119 
  | IF  -> 120 
  | IN  -> 121 
  | FINALLY  -> 122 
  | DO_BANG  -> 123 
  | AND  -> 124 
  | AS  -> 125 
  | ASSERT  -> 126 
  | ASR  -> 127 
  | BEGIN  -> 128 
  | DO  -> 129 
  | DONE  -> 130 
  | DOWNTO  -> 131 
  | ELSE  -> 132 
  | ELIF  -> 133 
  | END  -> 134 
  | DOT_DOT  -> 135 
  | BAR_BAR  -> 136 
  | LESS  -> 137 
  | GREATER  -> 138 
  | UPCAST  -> 139 
  | DOWNCAST  -> 140 
  | NULL  -> 141 
  | RESERVED  -> 142 
  | MODULE  -> 143 
  | DELEGATE  -> 144 
  | CONSTRAINT  -> 145 
  | BASE  -> 146 
  | FORALL  -> 147 
  | EXISTS  -> 148 
  | NOT  -> 149 
  | ASSUME  -> 150 
  | LQUOTE _ -> 151 
  | RQUOTE _ -> 152 
  | SPLICE_SYMBOL _ -> 153 
  | PERCENT_OP _ -> 154 
  | BINDER _ -> 155 
  | LET _ -> 156 
  | YIELD _ -> 157 
  | YIELD_BANG _ -> 158 
  | BIGINT _ -> 159 
  | BIGNUM _ -> 160 
  | DECIMAL _ -> 161 
  | CHAR _ -> 162 
  | IEEE64 _ -> 163 
  | IEEE32 _ -> 164 
  | NATIVEINT _ -> 165 
  | UNATIVEINT _ -> 166 
  | UINT64 _ -> 167 
  | UINT32 _ -> 168 
  | UINT16 _ -> 169 
  | UINT8 _ -> 170 
  | INT64 _ -> 171 
  | INT32 _ -> 172 
  | INT32_DOT_DOT _ -> 173 
  | INT16 _ -> 174 
  | INT8 _ -> 175 
  | FUNKY_OPERATOR_NAME _ -> 176 
  | ADJACENT_PREFIX_PLUS_MINUS_OP _ -> 177 
  | PLUS_MINUS_OP _ -> 178 
  | INFIX_AMP_OP _ -> 179 
  | INFIX_STAR_DIV_MOD_OP _ -> 180 
  | PREFIX_OP _ -> 181 
  | INFIX_BAR_OP _ -> 182 
  | INFIX_AT_HAT_OP _ -> 183 
  | INFIX_COMPARE_OP _ -> 184 
  | INFIX_STAR_STAR_OP _ -> 185 
  | LANG _ -> 186 
  | BASEKIND _ -> 187 
  | IDENT _ -> 188 
  | STRING _ -> 189 
  | BYTEARRAY _ -> 190 

// This function maps integers indexes to symbolic token ids
let tokenTagToTokenId (tokenIdx:int) = 
  match tokenIdx with
  | 0 -> TOKEN_COMMENT 
  | 1 -> TOKEN_WHITESPACE 
  | 2 -> TOKEN_HASH_LINE 
  | 3 -> TOKEN_HASH_LIGHT 
  | 4 -> TOKEN_HASH_IF 
  | 5 -> TOKEN_HASH_ELSE 
  | 6 -> TOKEN_HASH_ENDIF 
  | 7 -> TOKEN_INACTIVECODE 
  | 8 -> TOKEN_LINE_COMMENT 
  | 9 -> TOKEN_STRING_TEXT 
  | 10 -> TOKEN_EOF 
  | 11 -> TOKEN_LEX_FAILURE 
  | 12 -> TOKEN_ODUMMY 
  | 13 -> TOKEN_OINTERFACE_MEMBER 
  | 14 -> TOKEN_OBLOCKEND 
  | 15 -> TOKEN_ORIGHT_BLOCK_END 
  | 16 -> TOKEN_ODECLEND 
  | 17 -> TOKEN_OEND 
  | 18 -> TOKEN_OBLOCKSEP 
  | 19 -> TOKEN_OBLOCKBEGIN 
  | 20 -> TOKEN_ORESET 
  | 21 -> TOKEN_OFUN 
  | 22 -> TOKEN_OFUNCTION 
  | 23 -> TOKEN_OWITH 
  | 24 -> TOKEN_OELSE 
  | 25 -> TOKEN_OTHEN 
  | 26 -> TOKEN_ODO_BANG 
  | 27 -> TOKEN_ODO 
  | 28 -> TOKEN_OBINDER 
  | 29 -> TOKEN_OLET 
  | 30 -> TOKEN_HIGH_PRECEDENCE_TYAPP 
  | 31 -> TOKEN_HIGH_PRECEDENCE_APP 
  | 32 -> TOKEN_EXTERN 
  | 33 -> TOKEN_REFERENCE 
  | 34 -> TOKEN_VOID 
  | 35 -> TOKEN_PUBLIC 
  | 36 -> TOKEN_PRIVATE 
  | 37 -> TOKEN_INTERNAL 
  | 38 -> TOKEN_STATIC 
  | 39 -> TOKEN_MEMBER 
  | 40 -> TOKEN_CLASS 
  | 41 -> TOKEN_VIRTUAL 
  | 42 -> TOKEN_ABSTRACT 
  | 43 -> TOKEN_OVERRIDE 
  | 44 -> TOKEN_DEFAULT 
  | 45 -> TOKEN_CONSTRUCTOR 
  | 46 -> TOKEN_INHERIT 
  | 47 -> TOKEN_GREATER_RBRACK 
  | 48 -> TOKEN_STRUCT 
  | 49 -> TOKEN_SIG 
  | 50 -> TOKEN_BAR 
  | 51 -> TOKEN_RBRACK 
  | 52 -> TOKEN_RBRACE 
  | 53 -> TOKEN_MINUS 
  | 54 -> TOKEN_DOLLAR 
  | 55 -> TOKEN_LBRACE_LESS 
  | 56 -> TOKEN_BAR_RBRACK 
  | 57 -> TOKEN_GREATER_RBRACE 
  | 58 -> TOKEN_UNDERSCORE 
  | 59 -> TOKEN_BANG 
  | 60 -> TOKEN_SEMICOLON_SEMICOLON 
  | 61 -> TOKEN_LARROW 
  | 62 -> TOKEN_EQUALS 
  | 63 -> TOKEN_LBRACK 
  | 64 -> TOKEN_LBRACK_BAR 
  | 65 -> TOKEN_LBRACK_LESS 
  | 66 -> TOKEN_LBRACE 
  | 67 -> TOKEN_BACKSLASH 
  | 68 -> TOKEN_QMARK 
  | 69 -> TOKEN_QMARK_QMARK 
  | 70 -> TOKEN_DOT 
  | 71 -> TOKEN_COLON 
  | 72 -> TOKEN_COLON_COLON 
  | 73 -> TOKEN_COLON_GREATER 
  | 74 -> TOKEN_COLON_QMARK_GREATER 
  | 75 -> TOKEN_COLON_QMARK 
  | 76 -> TOKEN_COLON_EQUALS 
  | 77 -> TOKEN_SEMICOLON 
  | 78 -> TOKEN_WHEN 
  | 79 -> TOKEN_WHILE 
  | 80 -> TOKEN_WITH 
  | 81 -> TOKEN_HASH 
  | 82 -> TOKEN_AMP 
  | 83 -> TOKEN_AMP_AMP 
  | 84 -> TOKEN_QUOTE 
  | 85 -> TOKEN_LPAREN 
  | 86 -> TOKEN_RPAREN 
  | 87 -> TOKEN_STAR 
  | 88 -> TOKEN_COMMA 
  | 89 -> TOKEN_RARROW 
  | 90 -> TOKEN_RARROW2 
  | 91 -> TOKEN_RRARROW 
  | 92 -> TOKEN_IFF 
  | 93 -> TOKEN_GREATER_DOT 
  | 94 -> TOKEN_GREATER_BAR_RBRACK 
  | 95 -> TOKEN_LPAREN_STAR_RPAREN 
  | 96 -> TOKEN_OPEN 
  | 97 -> TOKEN_OR 
  | 98 -> TOKEN_PROP 
  | 99 -> TOKEN_REC 
  | 100 -> TOKEN_THEN 
  | 101 -> TOKEN_TO 
  | 102 -> TOKEN_TRUE 
  | 103 -> TOKEN_TRY 
  | 104 -> TOKEN_TYPE 
  | 105 -> TOKEN_VAL 
  | 106 -> TOKEN_INLINE 
  | 107 -> TOKEN_INTERFACE 
  | 108 -> TOKEN_INSTANCE 
  | 109 -> TOKEN_LAZY 
  | 110 -> TOKEN_MATCH 
  | 111 -> TOKEN_METHOD 
  | 112 -> TOKEN_MUTABLE 
  | 113 -> TOKEN_NEW 
  | 114 -> TOKEN_OF 
  | 115 -> TOKEN_EXCEPTION 
  | 116 -> TOKEN_FALSE 
  | 117 -> TOKEN_FOR 
  | 118 -> TOKEN_FUN 
  | 119 -> TOKEN_FUNCTION 
  | 120 -> TOKEN_IF 
  | 121 -> TOKEN_IN 
  | 122 -> TOKEN_FINALLY 
  | 123 -> TOKEN_DO_BANG 
  | 124 -> TOKEN_AND 
  | 125 -> TOKEN_AS 
  | 126 -> TOKEN_ASSERT 
  | 127 -> TOKEN_ASR 
  | 128 -> TOKEN_BEGIN 
  | 129 -> TOKEN_DO 
  | 130 -> TOKEN_DONE 
  | 131 -> TOKEN_DOWNTO 
  | 132 -> TOKEN_ELSE 
  | 133 -> TOKEN_ELIF 
  | 134 -> TOKEN_END 
  | 135 -> TOKEN_DOT_DOT 
  | 136 -> TOKEN_BAR_BAR 
  | 137 -> TOKEN_LESS 
  | 138 -> TOKEN_GREATER 
  | 139 -> TOKEN_UPCAST 
  | 140 -> TOKEN_DOWNCAST 
  | 141 -> TOKEN_NULL 
  | 142 -> TOKEN_RESERVED 
  | 143 -> TOKEN_MODULE 
  | 144 -> TOKEN_DELEGATE 
  | 145 -> TOKEN_CONSTRAINT 
  | 146 -> TOKEN_BASE 
  | 147 -> TOKEN_FORALL 
  | 148 -> TOKEN_EXISTS 
  | 149 -> TOKEN_NOT 
  | 150 -> TOKEN_ASSUME 
  | 151 -> TOKEN_LQUOTE 
  | 152 -> TOKEN_RQUOTE 
  | 153 -> TOKEN_SPLICE_SYMBOL 
  | 154 -> TOKEN_PERCENT_OP 
  | 155 -> TOKEN_BINDER 
  | 156 -> TOKEN_LET 
  | 157 -> TOKEN_YIELD 
  | 158 -> TOKEN_YIELD_BANG 
  | 159 -> TOKEN_BIGINT 
  | 160 -> TOKEN_BIGNUM 
  | 161 -> TOKEN_DECIMAL 
  | 162 -> TOKEN_CHAR 
  | 163 -> TOKEN_IEEE64 
  | 164 -> TOKEN_IEEE32 
  | 165 -> TOKEN_NATIVEINT 
  | 166 -> TOKEN_UNATIVEINT 
  | 167 -> TOKEN_UINT64 
  | 168 -> TOKEN_UINT32 
  | 169 -> TOKEN_UINT16 
  | 170 -> TOKEN_UINT8 
  | 171 -> TOKEN_INT64 
  | 172 -> TOKEN_INT32 
  | 173 -> TOKEN_INT32_DOT_DOT 
  | 174 -> TOKEN_INT16 
  | 175 -> TOKEN_INT8 
  | 176 -> TOKEN_FUNKY_OPERATOR_NAME 
  | 177 -> TOKEN_ADJACENT_PREFIX_PLUS_MINUS_OP 
  | 178 -> TOKEN_PLUS_MINUS_OP 
  | 179 -> TOKEN_INFIX_AMP_OP 
  | 180 -> TOKEN_INFIX_STAR_DIV_MOD_OP 
  | 181 -> TOKEN_PREFIX_OP 
  | 182 -> TOKEN_INFIX_BAR_OP 
  | 183 -> TOKEN_INFIX_AT_HAT_OP 
  | 184 -> TOKEN_INFIX_COMPARE_OP 
  | 185 -> TOKEN_INFIX_STAR_STAR_OP 
  | 186 -> TOKEN_LANG 
  | 187 -> TOKEN_BASEKIND 
  | 188 -> TOKEN_IDENT 
  | 189 -> TOKEN_STRING 
  | 190 -> TOKEN_BYTEARRAY 
  | 193 -> TOKEN_end_of_input
  | 191 -> TOKEN_error
  | _ -> failwith "tokenTagToTokenId: bad token"

/// This function maps production indexes returned in syntax errors to strings representing the non terminal that would be produced by that production
let prodIdxToNonTerminal (prodIdx:int) = 
  match prodIdx with
    | 0 -> NONTERM__startimplementationFile 
    | 1 -> NONTERM_implementationFile 
    | 2 -> NONTERM_fileModuleImpls 
    | 3 -> NONTERM_fileModuleImpls 
    | 4 -> NONTERM_fileModuleImpls 
    | 5 -> NONTERM_fileModuleImpl 
    | 6 -> NONTERM_moduleIntro 
    | 7 -> NONTERM_moduleIntro 
    | 8 -> NONTERM_moduleDefnsOrExprPossiblyEmpty 
    | 9 -> NONTERM_moduleDefnsOrExprPossiblyEmpty 
    | 10 -> NONTERM_moduleDefnsOrExpr 
    | 11 -> NONTERM_moduleDefnsOrExpr 
    | 12 -> NONTERM_moduleDefnsOrExpr 
    | 13 -> NONTERM_moduleDefnsOrExpr 
    | 14 -> NONTERM_moduleDefns 
    | 15 -> NONTERM_moduleDefns 
    | 16 -> NONTERM_moduleDefns 
    | 17 -> NONTERM_moduleDefns 
    | 18 -> NONTERM_moduleDefn 
    | 19 -> NONTERM_moduleDefn 
    | 20 -> NONTERM_moduleDefn 
    | 21 -> NONTERM_moduleDefn 
    | 22 -> NONTERM_moduleDefn 
    | 23 -> NONTERM_moduleDefn 
    | 24 -> NONTERM_moduleDefn 
    | 25 -> NONTERM_moduleDefn 
    | 26 -> NONTERM_moduleDefn 
    | 27 -> NONTERM_moduleDefn 
    | 28 -> NONTERM_moduleDefn 
    | 29 -> NONTERM_moduleDefn 
    | 30 -> NONTERM_moduleDefn 
    | 31 -> NONTERM_moduleDefn 
    | 32 -> NONTERM_moduleDefn 
    | 33 -> NONTERM_moduleDefn 
    | 34 -> NONTERM_extern_ref 
    | 35 -> NONTERM_maybenullary 
    | 36 -> NONTERM_maybenullary 
    | 37 -> NONTERM_defnBindings 
    | 38 -> NONTERM_hardwhiteLetBindings 
    | 39 -> NONTERM_hardwhiteDefnBindingsTerminator 
    | 40 -> NONTERM_localBindings 
    | 41 -> NONTERM_more_localBindings 
    | 42 -> NONTERM_more_localBindings 
    | 43 -> NONTERM_localBinding 
    | 44 -> NONTERM_bindingPattern 
    | 45 -> NONTERM_headBindingPattern 
    | 46 -> NONTERM_headBindingPattern 
    | 47 -> NONTERM_headBindingPattern 
    | 48 -> NONTERM_headBindingPattern 
    | 49 -> NONTERM_tuplePatternElements 
    | 50 -> NONTERM_tuplePatternElements 
    | 51 -> NONTERM_constrPattern 
    | 52 -> NONTERM_constrPattern 
    | 53 -> NONTERM_constrPattern 
    | 54 -> NONTERM_atomicPatterns 
    | 55 -> NONTERM_atomicPatterns 
    | 56 -> NONTERM_atomicPatterns 
    | 57 -> NONTERM_atomicPattern 
    | 58 -> NONTERM_atomicPattern 
    | 59 -> NONTERM_atomicPattern 
    | 60 -> NONTERM_atomicPattern 
    | 61 -> NONTERM_atomicPattern 
    | 62 -> NONTERM_atomicPattern 
    | 63 -> NONTERM_atomicPattern 
    | 64 -> NONTERM_atomicPattern 
    | 65 -> NONTERM_atomicPatternLongIdent 
    | 66 -> NONTERM_recordPatternElements 
    | 67 -> NONTERM_moreRecordPatternElements 
    | 68 -> NONTERM_moreRecordPatternElements 
    | 69 -> NONTERM_listPatternElements 
    | 70 -> NONTERM_listPatternElements 
    | 71 -> NONTERM_moreListPatternElements 
    | 72 -> NONTERM_moreListPatternElements 
    | 73 -> NONTERM_parenPatternBody 
    | 74 -> NONTERM_parenPatternBody 
    | 75 -> NONTERM_parenPattern 
    | 76 -> NONTERM_parenPattern 
    | 77 -> NONTERM_parenPattern 
    | 78 -> NONTERM_parenPattern 
    | 79 -> NONTERM_parenPattern 
    | 80 -> NONTERM_tupleParenPatternElements 
    | 81 -> NONTERM_tupleParenPatternElements 
    | 82 -> NONTERM_constant 
    | 83 -> NONTERM_constant 
    | 84 -> NONTERM_constant 
    | 85 -> NONTERM_constant 
    | 86 -> NONTERM_constant 
    | 87 -> NONTERM_constant 
    | 88 -> NONTERM_constant 
    | 89 -> NONTERM_constant 
    | 90 -> NONTERM_constant 
    | 91 -> NONTERM_constant 
    | 92 -> NONTERM_constant 
    | 93 -> NONTERM_constant 
    | 94 -> NONTERM_constant 
    | 95 -> NONTERM_constant 
    | 96 -> NONTERM_constant 
    | 97 -> NONTERM_constant 
    | 98 -> NONTERM_constant 
    | 99 -> NONTERM_constant 
    | 100 -> NONTERM_typedSeqExprBlock 
    | 101 -> NONTERM_typedSeqExprBlock 
    | 102 -> NONTERM_declExprBlock 
    | 103 -> NONTERM_declExprBlock 
    | 104 -> NONTERM_typedSeqExpr 
    | 105 -> NONTERM_typedSeqExpr 
    | 106 -> NONTERM_seqExpr 
    | 107 -> NONTERM_seqExpr 
    | 108 -> NONTERM_seqExpr 
    | 109 -> NONTERM_recover 
    | 110 -> NONTERM_recover 
    | 111 -> NONTERM_declExpr 
    | 112 -> NONTERM_declExpr 
    | 113 -> NONTERM_declExpr 
    | 114 -> NONTERM_declExpr 
    | 115 -> NONTERM_declExpr 
    | 116 -> NONTERM_declExpr 
    | 117 -> NONTERM_declExpr 
    | 118 -> NONTERM_declExpr 
    | 119 -> NONTERM_declExpr 
    | 120 -> NONTERM_declExpr 
    | 121 -> NONTERM_declExpr 
    | 122 -> NONTERM_declExpr 
    | 123 -> NONTERM_declExpr 
    | 124 -> NONTERM_declExpr 
    | 125 -> NONTERM_declExpr 
    | 126 -> NONTERM_declExpr 
    | 127 -> NONTERM_declExpr 
    | 128 -> NONTERM_declExpr 
    | 129 -> NONTERM_declExpr 
    | 130 -> NONTERM_declExpr 
    | 131 -> NONTERM_declExpr 
    | 132 -> NONTERM_declExpr 
    | 133 -> NONTERM_declExpr 
    | 134 -> NONTERM_declExpr 
    | 135 -> NONTERM_declExpr 
    | 136 -> NONTERM_declExpr 
    | 137 -> NONTERM_withClauses 
    | 138 -> NONTERM_withClauses 
    | 139 -> NONTERM_withPatternClauses 
    | 140 -> NONTERM_withPatternClauses 
    | 141 -> NONTERM_patternAndGuard 
    | 142 -> NONTERM_patternClauses 
    | 143 -> NONTERM_patternClauses 
    | 144 -> NONTERM_patternGuard 
    | 145 -> NONTERM_patternGuard 
    | 146 -> NONTERM_patternResult 
    | 147 -> NONTERM_typedSeqExprBlockR 
    | 148 -> NONTERM_typedSeqExprBlockR 
    | 149 -> NONTERM_ifExprCases 
    | 150 -> NONTERM_ifExprThen 
    | 151 -> NONTERM_ifExprThen 
    | 152 -> NONTERM_ifExprElifs 
    | 153 -> NONTERM_ifExprElifs 
    | 154 -> NONTERM_ifExprElifs 
    | 155 -> NONTERM_ifExprElifs 
    | 156 -> NONTERM_tupleExpr 
    | 157 -> NONTERM_tupleExpr 
    | 158 -> NONTERM_minusExpr 
    | 159 -> NONTERM_minusExpr 
    | 160 -> NONTERM_minusExpr 
    | 161 -> NONTERM_minusExpr 
    | 162 -> NONTERM_minusExpr 
    | 163 -> NONTERM_minusExpr 
    | 164 -> NONTERM_appExpr 
    | 165 -> NONTERM_appExpr 
    | 166 -> NONTERM_argExpr 
    | 167 -> NONTERM_argExpr 
    | 168 -> NONTERM_atomicExpr 
    | 169 -> NONTERM_atomicExpr 
    | 170 -> NONTERM_atomicExpr 
    | 171 -> NONTERM_atomicExpr 
    | 172 -> NONTERM_atomicExpr 
    | 173 -> NONTERM_atomicExpr 
    | 174 -> NONTERM_atomicExpr 
    | 175 -> NONTERM_listExprElements 
    | 176 -> NONTERM_listExprElements 
    | 177 -> NONTERM_listExprElements 
    | 178 -> NONTERM_atomicExprQualification 
    | 179 -> NONTERM_atomicExprAfterType 
    | 180 -> NONTERM_atomicExprAfterType 
    | 181 -> NONTERM_atomicExprAfterType 
    | 182 -> NONTERM_atomicExprAfterType 
    | 183 -> NONTERM_atomicExprAfterType 
    | 184 -> NONTERM_atomicExprAfterType 
    | 185 -> NONTERM_beginEndExpr 
    | 186 -> NONTERM_parenExpr 
    | 187 -> NONTERM_parenExpr 
    | 188 -> NONTERM_parenExprBody 
    | 189 -> NONTERM_parenExprBody 
    | 190 -> NONTERM_braceExpr 
    | 191 -> NONTERM_braceExprBody 
    | 192 -> NONTERM_opt_atomicExprAfterType 
    | 193 -> NONTERM_opt_atomicExprAfterType 
    | 194 -> NONTERM_recdExpr 
    | 195 -> NONTERM_recdExpr 
    | 196 -> NONTERM_recdExpr 
    | 197 -> NONTERM_recdExpr 
    | 198 -> NONTERM_recdExprBindings 
    | 199 -> NONTERM_recdExprBindings 
    | 200 -> NONTERM_anonLambdaExpr 
    | 201 -> NONTERM_anonLambdaExpr 
    | 202 -> NONTERM_anonLambdaExpr 
    | 203 -> NONTERM_anonMatchingExpr 
    | 204 -> NONTERM_anonMatchingExpr 
    | 205 -> NONTERM_tyconDefnList 
    | 206 -> NONTERM_tyconDefnList 
    | 207 -> NONTERM_tyconDefn 
    | 208 -> NONTERM_tyconDefn 
    | 209 -> NONTERM_typeNameInfo 
    | 210 -> NONTERM_tyconNameAndTyparDecls 
    | 211 -> NONTERM_tyconNameAndTyparDecls 
    | 212 -> NONTERM_opt_kind 
    | 213 -> NONTERM_opt_kind 
    | 214 -> NONTERM_kind 
    | 215 -> NONTERM_kind 
    | 216 -> NONTERM_kind 
    | 217 -> NONTERM_kind 
    | 218 -> NONTERM_atomic_kind_with_binding 
    | 219 -> NONTERM_atomic_kind_with_binding 
    | 220 -> NONTERM_typ_with_binding 
    | 221 -> NONTERM_typ_with_binding 
    | 222 -> NONTERM_base_kind 
    | 223 -> NONTERM_base_kind 
    | 224 -> NONTERM_base_kind 
    | 225 -> NONTERM_postfixTyparDecls 
    | 226 -> NONTERM_typeparDeclsActual 
    | 227 -> NONTERM_typeparDeclsActual 
    | 228 -> NONTERM_typeparDeclsActual 
    | 229 -> NONTERM_typeparDeclsActual 
    | 230 -> NONTERM_onlyTypars 
    | 231 -> NONTERM_onlyTypars 
    | 232 -> NONTERM_onlyTyparList 
    | 233 -> NONTERM_onlyTyparList 
    | 234 -> NONTERM_typarIndices 
    | 235 -> NONTERM_typarIndices 
    | 236 -> NONTERM_typarIndices 
    | 237 -> NONTERM_termparIndices 
    | 238 -> NONTERM_termparIndices 
    | 239 -> NONTERM_moreTermParIndices 
    | 240 -> NONTERM_moreTermParIndices 
    | 241 -> NONTERM_termparDecl 
    | 242 -> NONTERM_tyconDefnRhs 
    | 243 -> NONTERM_tyconDefnRhs 
    | 244 -> NONTERM_tyconDefnRhs 
    | 245 -> NONTERM_unionRepr 
    | 246 -> NONTERM_unionCaseDecls 
    | 247 -> NONTERM_unionCaseDecls 
    | 248 -> NONTERM_unionCaseDecl 
    | 249 -> NONTERM_unionCaseDecl 
    | 250 -> NONTERM_unionCaseDecl 
    | 251 -> NONTERM_unionCaseDecl 
    | 252 -> NONTERM_braceFieldDeclList 
    | 253 -> NONTERM_braceFieldDeclList 
    | 254 -> NONTERM_braceFieldDeclList 
    | 255 -> NONTERM_recdFieldDeclList 
    | 256 -> NONTERM_recdFieldDeclList 
    | 257 -> NONTERM_recdFieldDecl 
    | 258 -> NONTERM_opt_topReturnType 
    | 259 -> NONTERM_opt_topReturnType 
    | 260 -> NONTERM_typ 
    | 261 -> NONTERM_typ 
    | 262 -> NONTERM_typ 
    | 263 -> NONTERM_typ 
    | 264 -> NONTERM_typ 
    | 265 -> NONTERM_typ 
    | 266 -> NONTERM_typ 
    | 267 -> NONTERM_tfun 
    | 268 -> NONTERM_tfun 
    | 269 -> NONTERM_tfun 
    | 270 -> NONTERM_appType 
    | 271 -> NONTERM_appType 
    | 272 -> NONTERM_appType 
    | 273 -> NONTERM_typars 
    | 274 -> NONTERM_typars 
    | 275 -> NONTERM_atomType 
    | 276 -> NONTERM_atomType 
    | 277 -> NONTERM_atomType 
    | 278 -> NONTERM_tupleType 
    | 279 -> NONTERM_tupleType 
    | 280 -> NONTERM_tupleType 
    | 281 -> NONTERM_dtupleTypeTail 
    | 282 -> NONTERM_dtupleTypeTail 
    | 283 -> NONTERM_tupleTypeTail 
    | 284 -> NONTERM_tupleTypeTail 
    | 285 -> NONTERM_typeArgs 
    | 286 -> NONTERM_typeArgs 
    | 287 -> NONTERM_typeArgsActual 
    | 288 -> NONTERM_typeArgsActual 
    | 289 -> NONTERM_typeArgsActual 
    | 290 -> NONTERM_typeArgsActual 
    | 291 -> NONTERM_typeIndices 
    | 292 -> NONTERM_typeIndices 
    | 293 -> NONTERM_typeIndices 
    | 294 -> NONTERM_termIndices 
    | 295 -> NONTERM_termIndices 
    | 296 -> NONTERM_appTypeCon 
    | 297 -> NONTERM_appTypeCon 
    | 298 -> NONTERM_typar 
    | 299 -> NONTERM_typar 
    | 300 -> NONTERM_typar 
    | 301 -> NONTERM_atomic_kind 
    | 302 -> NONTERM_formula 
    | 303 -> NONTERM_Form 
    | 304 -> NONTERM_Form 
    | 305 -> NONTERM_Form 
    | 306 -> NONTERM_Form0 
    | 307 -> NONTERM_Form0 
    | 308 -> NONTERM_Form0a 
    | 309 -> NONTERM_Form0a 
    | 310 -> NONTERM_Form1 
    | 311 -> NONTERM_Form1 
    | 312 -> NONTERM_Form2 
    | 313 -> NONTERM_Form2 
    | 314 -> NONTERM_Form3 
    | 315 -> NONTERM_Form3 
    | 316 -> NONTERM_Form3a 
    | 317 -> NONTERM_Form3a 
    | 318 -> NONTERM_Form3a 
    | 319 -> NONTERM_Form4 
    | 320 -> NONTERM_Form4 
    | 321 -> NONTERM_Forall 
    | 322 -> NONTERM_Exists 
    | 323 -> NONTERM_fbinder 
    | 324 -> NONTERM_fbinders 
    | 325 -> NONTERM_fbinders 
    | 326 -> NONTERM_atom 
    | 327 -> NONTERM_atom 
    | 328 -> NONTERM_atom 
    | 329 -> NONTERM_termIndexSeq 
    | 330 -> NONTERM_termIndexSeq 
    | 331 -> NONTERM_value 
    | 332 -> NONTERM_value 
    | 333 -> NONTERM_valueSeq 
    | 334 -> NONTERM_valueSeq 
    | 335 -> NONTERM_atomicValue 
    | 336 -> NONTERM_atomicValue 
    | 337 -> NONTERM_valueConst 
    | 338 -> NONTERM_valueConst 
    | 339 -> NONTERM_valueConst 
    | 340 -> NONTERM_valueConst 
    | 341 -> NONTERM_valueConst 
    | 342 -> NONTERM_parenValue 
    | 343 -> NONTERM_parenValBody 
    | 344 -> NONTERM_parenValBody 
    | 345 -> NONTERM_parenValBody 
    | 346 -> NONTERM_tupleVal 
    | 347 -> NONTERM_tupleVal 
    | 348 -> NONTERM_ident 
    | 349 -> NONTERM_ident 
    | 350 -> NONTERM_opt_ident_colon 
    | 351 -> NONTERM_opt_ident_colon 
    | 352 -> NONTERM_path 
    | 353 -> NONTERM_path 
    | 354 -> NONTERM_opname 
    | 355 -> NONTERM_opname 
    | 356 -> NONTERM_operatorName 
    | 357 -> NONTERM_operatorName 
    | 358 -> NONTERM_operatorName 
    | 359 -> NONTERM_operatorName 
    | 360 -> NONTERM_operatorName 
    | 361 -> NONTERM_operatorName 
    | 362 -> NONTERM_operatorName 
    | 363 -> NONTERM_operatorName 
    | 364 -> NONTERM_operatorName 
    | 365 -> NONTERM_operatorName 
    | 366 -> NONTERM_operatorName 
    | 367 -> NONTERM_operatorName 
    | 368 -> NONTERM_operatorName 
    | 369 -> NONTERM_operatorName 
    | 370 -> NONTERM_operatorName 
    | 371 -> NONTERM_operatorName 
    | 372 -> NONTERM_operatorName 
    | 373 -> NONTERM_operatorName 
    | 374 -> NONTERM_operatorName 
    | 375 -> NONTERM_operatorName 
    | 376 -> NONTERM_operatorName 
    | 377 -> NONTERM_operatorName 
    | 378 -> NONTERM_operatorName 
    | 379 -> NONTERM_operatorName 
    | 380 -> NONTERM_operatorName 
    | 381 -> NONTERM_operatorName 
    | 382 -> NONTERM_operatorName 
    | 383 -> NONTERM_barName 
    | 384 -> NONTERM_barNames 
    | 385 -> NONTERM_barNames 
    | 386 -> NONTERM_top_sep 
    | 387 -> NONTERM_top_sep 
    | 388 -> NONTERM_top_sep 
    | 389 -> NONTERM_top_seps 
    | 390 -> NONTERM_top_seps 
    | 391 -> NONTERM_seps 
    | 392 -> NONTERM_seps 
    | 393 -> NONTERM_seps 
    | 394 -> NONTERM_seps 
    | 395 -> NONTERM_opt_OBLOCKSEP 
    | 396 -> NONTERM_opt_OBLOCKSEP 
    | 397 -> NONTERM_opt_seps 
    | 398 -> NONTERM_opt_seps 
    | 399 -> NONTERM_opt_rec 
    | 400 -> NONTERM_opt_rec 
    | 401 -> NONTERM_opt_bar 
    | 402 -> NONTERM_opt_bar 
    | 403 -> NONTERM_opt_HIGH_PRECEDENCE_TYAPP 
    | 404 -> NONTERM_opt_HIGH_PRECEDENCE_TYAPP 
    | _ -> failwith "prodIdxToNonTerminal: bad production index"

let _fsyacc_endOfInputTag = 193 
let _fsyacc_tagOfErrorTerminal = 191

// This function gets the name of a token as a string
let token_to_string (t:token) = 
  match t with 
  | COMMENT _ -> "COMMENT" 
  | WHITESPACE _ -> "WHITESPACE" 
  | HASH_LINE _ -> "HASH_LINE" 
  | HASH_LIGHT _ -> "HASH_LIGHT" 
  | HASH_IF _ -> "HASH_IF" 
  | HASH_ELSE _ -> "HASH_ELSE" 
  | HASH_ENDIF _ -> "HASH_ENDIF" 
  | INACTIVECODE _ -> "INACTIVECODE" 
  | LINE_COMMENT _ -> "LINE_COMMENT" 
  | STRING_TEXT _ -> "STRING_TEXT" 
  | EOF _ -> "EOF" 
  | LEX_FAILURE _ -> "LEX_FAILURE" 
  | ODUMMY _ -> "ODUMMY" 
  | OINTERFACE_MEMBER  -> "OINTERFACE_MEMBER" 
  | OBLOCKEND  -> "OBLOCKEND" 
  | ORIGHT_BLOCK_END  -> "ORIGHT_BLOCK_END" 
  | ODECLEND  -> "ODECLEND" 
  | OEND  -> "OEND" 
  | OBLOCKSEP  -> "OBLOCKSEP" 
  | OBLOCKBEGIN  -> "OBLOCKBEGIN" 
  | ORESET  -> "ORESET" 
  | OFUN  -> "OFUN" 
  | OFUNCTION  -> "OFUNCTION" 
  | OWITH  -> "OWITH" 
  | OELSE  -> "OELSE" 
  | OTHEN  -> "OTHEN" 
  | ODO_BANG  -> "ODO_BANG" 
  | ODO  -> "ODO" 
  | OBINDER _ -> "OBINDER" 
  | OLET _ -> "OLET" 
  | HIGH_PRECEDENCE_TYAPP  -> "HIGH_PRECEDENCE_TYAPP" 
  | HIGH_PRECEDENCE_APP  -> "HIGH_PRECEDENCE_APP" 
  | EXTERN  -> "EXTERN" 
  | REFERENCE  -> "REFERENCE" 
  | VOID  -> "VOID" 
  | PUBLIC  -> "PUBLIC" 
  | PRIVATE  -> "PRIVATE" 
  | INTERNAL  -> "INTERNAL" 
  | STATIC  -> "STATIC" 
  | MEMBER  -> "MEMBER" 
  | CLASS  -> "CLASS" 
  | VIRTUAL  -> "VIRTUAL" 
  | ABSTRACT  -> "ABSTRACT" 
  | OVERRIDE  -> "OVERRIDE" 
  | DEFAULT  -> "DEFAULT" 
  | CONSTRUCTOR  -> "CONSTRUCTOR" 
  | INHERIT  -> "INHERIT" 
  | GREATER_RBRACK  -> "GREATER_RBRACK" 
  | STRUCT  -> "STRUCT" 
  | SIG  -> "SIG" 
  | BAR  -> "BAR" 
  | RBRACK  -> "RBRACK" 
  | RBRACE  -> "RBRACE" 
  | MINUS  -> "MINUS" 
  | DOLLAR  -> "DOLLAR" 
  | LBRACE_LESS  -> "LBRACE_LESS" 
  | BAR_RBRACK  -> "BAR_RBRACK" 
  | GREATER_RBRACE  -> "GREATER_RBRACE" 
  | UNDERSCORE  -> "UNDERSCORE" 
  | BANG  -> "BANG" 
  | SEMICOLON_SEMICOLON  -> "SEMICOLON_SEMICOLON" 
  | LARROW  -> "LARROW" 
  | EQUALS  -> "EQUALS" 
  | LBRACK  -> "LBRACK" 
  | LBRACK_BAR  -> "LBRACK_BAR" 
  | LBRACK_LESS  -> "LBRACK_LESS" 
  | LBRACE  -> "LBRACE" 
  | BACKSLASH  -> "BACKSLASH" 
  | QMARK  -> "QMARK" 
  | QMARK_QMARK  -> "QMARK_QMARK" 
  | DOT  -> "DOT" 
  | COLON  -> "COLON" 
  | COLON_COLON  -> "COLON_COLON" 
  | COLON_GREATER  -> "COLON_GREATER" 
  | COLON_QMARK_GREATER  -> "COLON_QMARK_GREATER" 
  | COLON_QMARK  -> "COLON_QMARK" 
  | COLON_EQUALS  -> "COLON_EQUALS" 
  | SEMICOLON  -> "SEMICOLON" 
  | WHEN  -> "WHEN" 
  | WHILE  -> "WHILE" 
  | WITH  -> "WITH" 
  | HASH  -> "HASH" 
  | AMP  -> "AMP" 
  | AMP_AMP  -> "AMP_AMP" 
  | QUOTE  -> "QUOTE" 
  | LPAREN  -> "LPAREN" 
  | RPAREN  -> "RPAREN" 
  | STAR  -> "STAR" 
  | COMMA  -> "COMMA" 
  | RARROW  -> "RARROW" 
  | RARROW2  -> "RARROW2" 
  | RRARROW  -> "RRARROW" 
  | IFF  -> "IFF" 
  | GREATER_DOT  -> "GREATER_DOT" 
  | GREATER_BAR_RBRACK  -> "GREATER_BAR_RBRACK" 
  | LPAREN_STAR_RPAREN  -> "LPAREN_STAR_RPAREN" 
  | OPEN  -> "OPEN" 
  | OR  -> "OR" 
  | PROP  -> "PROP" 
  | REC  -> "REC" 
  | THEN  -> "THEN" 
  | TO  -> "TO" 
  | TRUE  -> "TRUE" 
  | TRY  -> "TRY" 
  | TYPE  -> "TYPE" 
  | VAL  -> "VAL" 
  | INLINE  -> "INLINE" 
  | INTERFACE  -> "INTERFACE" 
  | INSTANCE  -> "INSTANCE" 
  | LAZY  -> "LAZY" 
  | MATCH  -> "MATCH" 
  | METHOD  -> "METHOD" 
  | MUTABLE  -> "MUTABLE" 
  | NEW  -> "NEW" 
  | OF  -> "OF" 
  | EXCEPTION  -> "EXCEPTION" 
  | FALSE  -> "FALSE" 
  | FOR  -> "FOR" 
  | FUN  -> "FUN" 
  | FUNCTION  -> "FUNCTION" 
  | IF  -> "IF" 
  | IN  -> "IN" 
  | FINALLY  -> "FINALLY" 
  | DO_BANG  -> "DO_BANG" 
  | AND  -> "AND" 
  | AS  -> "AS" 
  | ASSERT  -> "ASSERT" 
  | ASR  -> "ASR" 
  | BEGIN  -> "BEGIN" 
  | DO  -> "DO" 
  | DONE  -> "DONE" 
  | DOWNTO  -> "DOWNTO" 
  | ELSE  -> "ELSE" 
  | ELIF  -> "ELIF" 
  | END  -> "END" 
  | DOT_DOT  -> "DOT_DOT" 
  | BAR_BAR  -> "BAR_BAR" 
  | LESS  -> "LESS" 
  | GREATER  -> "GREATER" 
  | UPCAST  -> "UPCAST" 
  | DOWNCAST  -> "DOWNCAST" 
  | NULL  -> "NULL" 
  | RESERVED  -> "RESERVED" 
  | MODULE  -> "MODULE" 
  | DELEGATE  -> "DELEGATE" 
  | CONSTRAINT  -> "CONSTRAINT" 
  | BASE  -> "BASE" 
  | FORALL  -> "FORALL" 
  | EXISTS  -> "EXISTS" 
  | NOT  -> "NOT" 
  | ASSUME  -> "ASSUME" 
  | LQUOTE _ -> "LQUOTE" 
  | RQUOTE _ -> "RQUOTE" 
  | SPLICE_SYMBOL _ -> "SPLICE_SYMBOL" 
  | PERCENT_OP _ -> "PERCENT_OP" 
  | BINDER _ -> "BINDER" 
  | LET _ -> "LET" 
  | YIELD _ -> "YIELD" 
  | YIELD_BANG _ -> "YIELD_BANG" 
  | BIGINT _ -> "BIGINT" 
  | BIGNUM _ -> "BIGNUM" 
  | DECIMAL _ -> "DECIMAL" 
  | CHAR _ -> "CHAR" 
  | IEEE64 _ -> "IEEE64" 
  | IEEE32 _ -> "IEEE32" 
  | NATIVEINT _ -> "NATIVEINT" 
  | UNATIVEINT _ -> "UNATIVEINT" 
  | UINT64 _ -> "UINT64" 
  | UINT32 _ -> "UINT32" 
  | UINT16 _ -> "UINT16" 
  | UINT8 _ -> "UINT8" 
  | INT64 _ -> "INT64" 
  | INT32 _ -> "INT32" 
  | INT32_DOT_DOT _ -> "INT32_DOT_DOT" 
  | INT16 _ -> "INT16" 
  | INT8 _ -> "INT8" 
  | FUNKY_OPERATOR_NAME _ -> "FUNKY_OPERATOR_NAME" 
  | ADJACENT_PREFIX_PLUS_MINUS_OP _ -> "ADJACENT_PREFIX_PLUS_MINUS_OP" 
  | PLUS_MINUS_OP _ -> "PLUS_MINUS_OP" 
  | INFIX_AMP_OP _ -> "INFIX_AMP_OP" 
  | INFIX_STAR_DIV_MOD_OP _ -> "INFIX_STAR_DIV_MOD_OP" 
  | PREFIX_OP _ -> "PREFIX_OP" 
  | INFIX_BAR_OP _ -> "INFIX_BAR_OP" 
  | INFIX_AT_HAT_OP _ -> "INFIX_AT_HAT_OP" 
  | INFIX_COMPARE_OP _ -> "INFIX_COMPARE_OP" 
  | INFIX_STAR_STAR_OP _ -> "INFIX_STAR_STAR_OP" 
  | LANG _ -> "LANG" 
  | BASEKIND _ -> "BASEKIND" 
  | IDENT _ -> "IDENT" 
  | STRING _ -> "STRING" 
  | BYTEARRAY _ -> "BYTEARRAY" 

// This function gets the data carried by a token as an object
let _fsyacc_dataOfToken (t:token) = 
  match t with 
  | COMMENT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | WHITESPACE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HASH_LINE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HASH_LIGHT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HASH_IF _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HASH_ELSE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HASH_ENDIF _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INACTIVECODE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | LINE_COMMENT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | STRING_TEXT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | EOF _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | LEX_FAILURE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | ODUMMY _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | OINTERFACE_MEMBER  -> (null : System.Object) 
  | OBLOCKEND  -> (null : System.Object) 
  | ORIGHT_BLOCK_END  -> (null : System.Object) 
  | ODECLEND  -> (null : System.Object) 
  | OEND  -> (null : System.Object) 
  | OBLOCKSEP  -> (null : System.Object) 
  | OBLOCKBEGIN  -> (null : System.Object) 
  | ORESET  -> (null : System.Object) 
  | OFUN  -> (null : System.Object) 
  | OFUNCTION  -> (null : System.Object) 
  | OWITH  -> (null : System.Object) 
  | OELSE  -> (null : System.Object) 
  | OTHEN  -> (null : System.Object) 
  | ODO_BANG  -> (null : System.Object) 
  | ODO  -> (null : System.Object) 
  | OBINDER _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | OLET _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | HIGH_PRECEDENCE_TYAPP  -> (null : System.Object) 
  | HIGH_PRECEDENCE_APP  -> (null : System.Object) 
  | EXTERN  -> (null : System.Object) 
  | REFERENCE  -> (null : System.Object) 
  | VOID  -> (null : System.Object) 
  | PUBLIC  -> (null : System.Object) 
  | PRIVATE  -> (null : System.Object) 
  | INTERNAL  -> (null : System.Object) 
  | STATIC  -> (null : System.Object) 
  | MEMBER  -> (null : System.Object) 
  | CLASS  -> (null : System.Object) 
  | VIRTUAL  -> (null : System.Object) 
  | ABSTRACT  -> (null : System.Object) 
  | OVERRIDE  -> (null : System.Object) 
  | DEFAULT  -> (null : System.Object) 
  | CONSTRUCTOR  -> (null : System.Object) 
  | INHERIT  -> (null : System.Object) 
  | GREATER_RBRACK  -> (null : System.Object) 
  | STRUCT  -> (null : System.Object) 
  | SIG  -> (null : System.Object) 
  | BAR  -> (null : System.Object) 
  | RBRACK  -> (null : System.Object) 
  | RBRACE  -> (null : System.Object) 
  | MINUS  -> (null : System.Object) 
  | DOLLAR  -> (null : System.Object) 
  | LBRACE_LESS  -> (null : System.Object) 
  | BAR_RBRACK  -> (null : System.Object) 
  | GREATER_RBRACE  -> (null : System.Object) 
  | UNDERSCORE  -> (null : System.Object) 
  | BANG  -> (null : System.Object) 
  | SEMICOLON_SEMICOLON  -> (null : System.Object) 
  | LARROW  -> (null : System.Object) 
  | EQUALS  -> (null : System.Object) 
  | LBRACK  -> (null : System.Object) 
  | LBRACK_BAR  -> (null : System.Object) 
  | LBRACK_LESS  -> (null : System.Object) 
  | LBRACE  -> (null : System.Object) 
  | BACKSLASH  -> (null : System.Object) 
  | QMARK  -> (null : System.Object) 
  | QMARK_QMARK  -> (null : System.Object) 
  | DOT  -> (null : System.Object) 
  | COLON  -> (null : System.Object) 
  | COLON_COLON  -> (null : System.Object) 
  | COLON_GREATER  -> (null : System.Object) 
  | COLON_QMARK_GREATER  -> (null : System.Object) 
  | COLON_QMARK  -> (null : System.Object) 
  | COLON_EQUALS  -> (null : System.Object) 
  | SEMICOLON  -> (null : System.Object) 
  | WHEN  -> (null : System.Object) 
  | WHILE  -> (null : System.Object) 
  | WITH  -> (null : System.Object) 
  | HASH  -> (null : System.Object) 
  | AMP  -> (null : System.Object) 
  | AMP_AMP  -> (null : System.Object) 
  | QUOTE  -> (null : System.Object) 
  | LPAREN  -> (null : System.Object) 
  | RPAREN  -> (null : System.Object) 
  | STAR  -> (null : System.Object) 
  | COMMA  -> (null : System.Object) 
  | RARROW  -> (null : System.Object) 
  | RARROW2  -> (null : System.Object) 
  | RRARROW  -> (null : System.Object) 
  | IFF  -> (null : System.Object) 
  | GREATER_DOT  -> (null : System.Object) 
  | GREATER_BAR_RBRACK  -> (null : System.Object) 
  | LPAREN_STAR_RPAREN  -> (null : System.Object) 
  | OPEN  -> (null : System.Object) 
  | OR  -> (null : System.Object) 
  | PROP  -> (null : System.Object) 
  | REC  -> (null : System.Object) 
  | THEN  -> (null : System.Object) 
  | TO  -> (null : System.Object) 
  | TRUE  -> (null : System.Object) 
  | TRY  -> (null : System.Object) 
  | TYPE  -> (null : System.Object) 
  | VAL  -> (null : System.Object) 
  | INLINE  -> (null : System.Object) 
  | INTERFACE  -> (null : System.Object) 
  | INSTANCE  -> (null : System.Object) 
  | LAZY  -> (null : System.Object) 
  | MATCH  -> (null : System.Object) 
  | METHOD  -> (null : System.Object) 
  | MUTABLE  -> (null : System.Object) 
  | NEW  -> (null : System.Object) 
  | OF  -> (null : System.Object) 
  | EXCEPTION  -> (null : System.Object) 
  | FALSE  -> (null : System.Object) 
  | FOR  -> (null : System.Object) 
  | FUN  -> (null : System.Object) 
  | FUNCTION  -> (null : System.Object) 
  | IF  -> (null : System.Object) 
  | IN  -> (null : System.Object) 
  | FINALLY  -> (null : System.Object) 
  | DO_BANG  -> (null : System.Object) 
  | AND  -> (null : System.Object) 
  | AS  -> (null : System.Object) 
  | ASSERT  -> (null : System.Object) 
  | ASR  -> (null : System.Object) 
  | BEGIN  -> (null : System.Object) 
  | DO  -> (null : System.Object) 
  | DONE  -> (null : System.Object) 
  | DOWNTO  -> (null : System.Object) 
  | ELSE  -> (null : System.Object) 
  | ELIF  -> (null : System.Object) 
  | END  -> (null : System.Object) 
  | DOT_DOT  -> (null : System.Object) 
  | BAR_BAR  -> (null : System.Object) 
  | LESS  -> (null : System.Object) 
  | GREATER  -> (null : System.Object) 
  | UPCAST  -> (null : System.Object) 
  | DOWNCAST  -> (null : System.Object) 
  | NULL  -> (null : System.Object) 
  | RESERVED  -> (null : System.Object) 
  | MODULE  -> (null : System.Object) 
  | DELEGATE  -> (null : System.Object) 
  | CONSTRAINT  -> (null : System.Object) 
  | BASE  -> (null : System.Object) 
  | FORALL  -> (null : System.Object) 
  | EXISTS  -> (null : System.Object) 
  | NOT  -> (null : System.Object) 
  | ASSUME  -> (null : System.Object) 
  | LQUOTE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | RQUOTE _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | SPLICE_SYMBOL _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | PERCENT_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | BINDER _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | LET _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | YIELD _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | YIELD_BANG _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | BIGINT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | BIGNUM _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | DECIMAL _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | CHAR _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | IEEE64 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | IEEE32 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | NATIVEINT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | UNATIVEINT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | UINT64 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | UINT32 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | UINT16 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | UINT8 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INT64 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INT32 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INT32_DOT_DOT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INT16 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INT8 _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | FUNKY_OPERATOR_NAME _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | ADJACENT_PREFIX_PLUS_MINUS_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | PLUS_MINUS_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_AMP_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_STAR_DIV_MOD_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | PREFIX_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_BAR_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_AT_HAT_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_COMPARE_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | INFIX_STAR_STAR_OP _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | LANG _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | BASEKIND _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | IDENT _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | STRING _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
  | BYTEARRAY _fsyacc_x -> Microsoft.FSharp.Core.Operators.box _fsyacc_x 
let _fsyacc_gotos = [| 0us; 65535us; 1us; 65535us; 0us; 1us; 2us; 65535us; 2us; 3us; 6us; 7us; 2us; 65535us; 0us; 2us; 5us; 6us; 2us; 65535us; 0us; 8us; 5us; 8us; 1us; 65535us; 8us; 9us; 3us; 65535us; 8us; 14us; 16us; 17us; 21us; 22us; 4us; 65535us; 8us; 18us; 16us; 18us; 19us; 20us; 21us; 18us; 4us; 65535us; 8us; 19us; 16us; 19us; 19us; 19us; 21us; 19us; 1us; 65535us; 41us; 42us; 1us; 65535us; 32us; 33us; 47us; 65535us; 8us; 24us; 16us; 24us; 19us; 23us; 21us; 24us; 26us; 191us; 98us; 191us; 179us; 191us; 187us; 191us; 192us; 191us; 194us; 191us; 196us; 191us; 200us; 191us; 203us; 191us; 234us; 191us; 235us; 191us; 236us; 191us; 237us; 191us; 238us; 191us; 239us; 191us; 240us; 191us; 241us; 191us; 242us; 191us; 243us; 191us; 244us; 191us; 245us; 191us; 246us; 191us; 247us; 191us; 248us; 191us; 249us; 191us; 250us; 191us; 265us; 191us; 266us; 191us; 272us; 191us; 274us; 191us; 277us; 191us; 279us; 191us; 282us; 191us; 284us; 191us; 285us; 191us; 326us; 191us; 329us; 191us; 338us; 191us; 343us; 191us; 348us; 191us; 354us; 191us; 357us; 191us; 362us; 191us; 47us; 65535us; 8us; 26us; 16us; 26us; 19us; 25us; 21us; 26us; 26us; 194us; 98us; 194us; 179us; 194us; 187us; 194us; 192us; 194us; 194us; 194us; 196us; 194us; 200us; 194us; 203us; 194us; 234us; 194us; 235us; 194us; 236us; 194us; 237us; 194us; 238us; 194us; 239us; 194us; 240us; 194us; 241us; 194us; 242us; 194us; 243us; 194us; 244us; 194us; 245us; 194us; 246us; 194us; 247us; 194us; 248us; 194us; 249us; 194us; 250us; 194us; 265us; 194us; 266us; 194us; 272us; 194us; 274us; 194us; 277us; 194us; 279us; 194us; 282us; 194us; 284us; 194us; 285us; 194us; 326us; 194us; 329us; 194us; 338us; 194us; 343us; 194us; 348us; 194us; 354us; 194us; 357us; 194us; 362us; 194us; 1us; 65535us; 88us; 89us; 2us; 65535us; 84us; 85us; 87us; 88us; 2us; 65535us; 91us; 92us; 94us; 95us; 3us; 65535us; 84us; 91us; 87us; 91us; 93us; 94us; 3us; 65535us; 84us; 96us; 87us; 96us; 93us; 96us; 6us; 65535us; 84us; 100us; 87us; 100us; 93us; 100us; 106us; 101us; 109us; 102us; 110us; 103us; 6us; 65535us; 84us; 107us; 87us; 107us; 93us; 107us; 106us; 107us; 109us; 107us; 110us; 107us; 20us; 65535us; 84us; 108us; 87us; 108us; 93us; 108us; 106us; 108us; 109us; 108us; 110us; 108us; 123us; 158us; 131us; 158us; 136us; 158us; 141us; 158us; 145us; 158us; 155us; 158us; 159us; 158us; 160us; 158us; 252us; 158us; 254us; 158us; 258us; 158us; 263us; 158us; 366us; 158us; 369us; 158us; 6us; 65535us; 111us; 112us; 113us; 114us; 116us; 117us; 118us; 119us; 355us; 356us; 360us; 361us; 26us; 65535us; 84us; 115us; 87us; 115us; 93us; 115us; 106us; 115us; 109us; 115us; 110us; 115us; 111us; 116us; 113us; 116us; 116us; 116us; 118us; 116us; 123us; 115us; 131us; 115us; 136us; 115us; 141us; 115us; 145us; 115us; 155us; 115us; 159us; 115us; 160us; 115us; 252us; 115us; 254us; 115us; 258us; 115us; 263us; 115us; 355us; 116us; 360us; 116us; 366us; 115us; 369us; 115us; 26us; 65535us; 84us; 111us; 87us; 111us; 93us; 111us; 106us; 111us; 109us; 111us; 110us; 111us; 111us; 127us; 113us; 127us; 116us; 127us; 118us; 127us; 123us; 111us; 131us; 111us; 136us; 111us; 141us; 111us; 145us; 111us; 155us; 111us; 159us; 111us; 160us; 111us; 252us; 111us; 254us; 111us; 258us; 111us; 263us; 111us; 355us; 127us; 360us; 127us; 366us; 111us; 369us; 111us; 1us; 65535us; 120us; 121us; 1us; 65535us; 137us; 138us; 1us; 65535us; 123us; 124us; 1us; 65535us; 143us; 144us; 1us; 65535us; 131us; 132us; 14us; 65535us; 123us; 143us; 131us; 147us; 136us; 137us; 141us; 142us; 145us; 146us; 155us; 148us; 159us; 149us; 160us; 150us; 252us; 151us; 254us; 151us; 258us; 151us; 263us; 151us; 366us; 151us; 369us; 151us; 14us; 65535us; 123us; 154us; 131us; 154us; 136us; 154us; 141us; 154us; 145us; 154us; 155us; 154us; 159us; 154us; 160us; 154us; 252us; 154us; 254us; 154us; 258us; 154us; 263us; 154us; 366us; 154us; 369us; 154us; 113us; 65535us; 8us; 320us; 16us; 320us; 21us; 320us; 26us; 320us; 51us; 656us; 79us; 320us; 84us; 128us; 87us; 128us; 93us; 128us; 98us; 320us; 106us; 128us; 109us; 128us; 110us; 128us; 111us; 128us; 113us; 128us; 116us; 128us; 118us; 128us; 123us; 128us; 131us; 128us; 136us; 128us; 141us; 128us; 145us; 128us; 155us; 128us; 159us; 128us; 160us; 128us; 179us; 320us; 187us; 320us; 192us; 320us; 194us; 320us; 196us; 320us; 200us; 320us; 203us; 320us; 234us; 320us; 235us; 320us; 236us; 320us; 237us; 320us; 238us; 320us; 239us; 320us; 240us; 320us; 241us; 320us; 242us; 320us; 243us; 320us; 244us; 320us; 245us; 320us; 246us; 320us; 247us; 320us; 248us; 320us; 249us; 320us; 250us; 320us; 252us; 128us; 254us; 128us; 258us; 128us; 263us; 128us; 265us; 320us; 266us; 320us; 272us; 320us; 274us; 320us; 277us; 320us; 279us; 320us; 282us; 320us; 284us; 320us; 285us; 320us; 286us; 320us; 288us; 320us; 290us; 320us; 292us; 320us; 294us; 320us; 296us; 320us; 297us; 320us; 300us; 320us; 307us; 320us; 309us; 320us; 313us; 320us; 317us; 320us; 326us; 320us; 329us; 320us; 334us; 320us; 338us; 320us; 343us; 320us; 348us; 320us; 354us; 320us; 355us; 128us; 357us; 320us; 360us; 128us; 362us; 320us; 366us; 128us; 369us; 128us; 508us; 656us; 514us; 656us; 531us; 656us; 576us; 320us; 579us; 320us; 584us; 320us; 586us; 320us; 602us; 656us; 605us; 656us; 608us; 656us; 611us; 656us; 613us; 656us; 617us; 656us; 618us; 656us; 620us; 656us; 621us; 656us; 625us; 656us; 629us; 656us; 632us; 656us; 642us; 656us; 644us; 656us; 647us; 656us; 649us; 656us; 651us; 656us; 660us; 656us; 665us; 656us; 4us; 65535us; 26us; 195us; 194us; 195us; 196us; 197us; 357us; 358us; 0us; 65535us; 14us; 65535us; 26us; 182us; 98us; 99us; 179us; 180us; 192us; 193us; 194us; 182us; 196us; 182us; 200us; 201us; 266us; 268us; 274us; 275us; 279us; 280us; 326us; 327us; 329us; 333us; 357us; 182us; 362us; 268us; 15us; 65535us; 26us; 183us; 98us; 183us; 179us; 183us; 187us; 188us; 192us; 183us; 194us; 183us; 196us; 183us; 200us; 183us; 266us; 183us; 274us; 183us; 279us; 183us; 326us; 183us; 329us; 183us; 357us; 183us; 362us; 183us; 2us; 65535us; 330us; 332us; 471us; 473us; 46us; 65535us; 8us; 15us; 16us; 15us; 21us; 15us; 26us; 186us; 98us; 186us; 179us; 186us; 187us; 186us; 192us; 186us; 194us; 186us; 196us; 186us; 200us; 186us; 203us; 204us; 234us; 207us; 235us; 208us; 236us; 209us; 237us; 210us; 238us; 211us; 239us; 212us; 240us; 213us; 241us; 214us; 242us; 215us; 243us; 216us; 244us; 217us; 245us; 218us; 246us; 219us; 247us; 220us; 248us; 221us; 249us; 222us; 250us; 223us; 265us; 224us; 266us; 186us; 272us; 225us; 274us; 186us; 277us; 226us; 279us; 186us; 282us; 227us; 284us; 228us; 285us; 229us; 326us; 186us; 329us; 186us; 338us; 230us; 343us; 231us; 348us; 232us; 354us; 233us; 357us; 186us; 362us; 186us; 1us; 65535us; 201us; 202us; 2us; 65535us; 252us; 253us; 254us; 255us; 6us; 65535us; 252us; 261us; 254us; 261us; 258us; 261us; 263us; 261us; 366us; 261us; 369us; 261us; 6us; 65535us; 252us; 257us; 254us; 257us; 258us; 259us; 263us; 264us; 366us; 367us; 369us; 370us; 1us; 65535us; 151us; 260us; 1us; 65535us; 261us; 262us; 2us; 65535us; 266us; 267us; 362us; 363us; 2us; 65535us; 204us; 205us; 227us; 283us; 2us; 65535us; 204us; 270us; 227us; 270us; 1us; 65535us; 270us; 271us; 46us; 65535us; 8us; 206us; 16us; 206us; 21us; 206us; 26us; 206us; 98us; 206us; 179us; 206us; 187us; 206us; 192us; 206us; 194us; 206us; 196us; 206us; 200us; 206us; 203us; 206us; 234us; 206us; 235us; 206us; 236us; 206us; 237us; 206us; 238us; 206us; 239us; 206us; 240us; 206us; 241us; 206us; 242us; 206us; 243us; 206us; 244us; 206us; 245us; 206us; 246us; 206us; 247us; 206us; 248us; 206us; 249us; 206us; 250us; 206us; 265us; 206us; 266us; 206us; 272us; 206us; 274us; 206us; 277us; 206us; 279us; 206us; 282us; 206us; 284us; 206us; 285us; 206us; 326us; 206us; 329us; 206us; 338us; 206us; 343us; 206us; 348us; 206us; 354us; 206us; 357us; 206us; 362us; 206us; 51us; 65535us; 8us; 251us; 16us; 251us; 21us; 251us; 26us; 251us; 98us; 251us; 179us; 251us; 187us; 251us; 192us; 251us; 194us; 251us; 196us; 251us; 200us; 251us; 203us; 251us; 234us; 251us; 235us; 251us; 236us; 251us; 237us; 251us; 238us; 251us; 239us; 251us; 240us; 251us; 241us; 251us; 242us; 251us; 243us; 251us; 244us; 251us; 245us; 251us; 246us; 251us; 247us; 251us; 248us; 251us; 249us; 251us; 250us; 251us; 265us; 251us; 266us; 251us; 272us; 251us; 274us; 251us; 277us; 251us; 279us; 251us; 282us; 251us; 284us; 251us; 285us; 251us; 286us; 287us; 288us; 289us; 290us; 291us; 292us; 293us; 294us; 295us; 326us; 251us; 329us; 251us; 338us; 251us; 343us; 251us; 348us; 251us; 354us; 251us; 357us; 251us; 362us; 251us; 53us; 65535us; 8us; 296us; 16us; 296us; 21us; 296us; 26us; 296us; 79us; 297us; 98us; 296us; 179us; 296us; 187us; 296us; 192us; 296us; 194us; 296us; 196us; 296us; 200us; 296us; 203us; 296us; 234us; 296us; 235us; 296us; 236us; 296us; 237us; 296us; 238us; 296us; 239us; 296us; 240us; 296us; 241us; 296us; 242us; 296us; 243us; 296us; 244us; 296us; 245us; 296us; 246us; 296us; 247us; 296us; 248us; 296us; 249us; 296us; 250us; 296us; 265us; 296us; 266us; 296us; 272us; 296us; 274us; 296us; 277us; 296us; 279us; 296us; 282us; 296us; 284us; 296us; 285us; 296us; 286us; 296us; 288us; 296us; 290us; 296us; 292us; 296us; 294us; 296us; 326us; 296us; 329us; 296us; 334us; 297us; 338us; 296us; 343us; 296us; 348us; 296us; 354us; 296us; 357us; 296us; 362us; 296us; 2us; 65535us; 296us; 298us; 297us; 298us; 64us; 65535us; 8us; 299us; 16us; 299us; 21us; 299us; 26us; 299us; 79us; 299us; 98us; 299us; 179us; 299us; 187us; 299us; 192us; 299us; 194us; 299us; 196us; 299us; 200us; 299us; 203us; 299us; 234us; 299us; 235us; 299us; 236us; 299us; 237us; 299us; 238us; 299us; 239us; 299us; 240us; 299us; 241us; 299us; 242us; 299us; 243us; 299us; 244us; 299us; 245us; 299us; 246us; 299us; 247us; 299us; 248us; 299us; 249us; 299us; 250us; 299us; 265us; 299us; 266us; 299us; 272us; 299us; 274us; 299us; 277us; 299us; 279us; 299us; 282us; 299us; 284us; 299us; 285us; 299us; 286us; 299us; 288us; 299us; 290us; 299us; 292us; 299us; 294us; 299us; 296us; 302us; 297us; 302us; 300us; 301us; 307us; 303us; 309us; 304us; 313us; 305us; 317us; 305us; 326us; 299us; 329us; 299us; 334us; 299us; 338us; 299us; 343us; 299us; 348us; 299us; 354us; 299us; 357us; 299us; 362us; 299us; 576us; 306us; 579us; 306us; 584us; 306us; 586us; 306us; 2us; 65535us; 313us; 314us; 317us; 318us; 1us; 65535us; 310us; 311us; 64us; 65535us; 8us; 316us; 16us; 316us; 21us; 316us; 26us; 316us; 79us; 316us; 98us; 316us; 179us; 316us; 187us; 316us; 192us; 316us; 194us; 316us; 196us; 316us; 200us; 316us; 203us; 316us; 234us; 316us; 235us; 316us; 236us; 316us; 237us; 316us; 238us; 316us; 239us; 316us; 240us; 316us; 241us; 316us; 242us; 316us; 243us; 316us; 244us; 316us; 245us; 316us; 246us; 316us; 247us; 316us; 248us; 316us; 249us; 316us; 250us; 316us; 265us; 316us; 266us; 316us; 272us; 316us; 274us; 316us; 277us; 316us; 279us; 316us; 282us; 316us; 284us; 316us; 285us; 316us; 286us; 316us; 288us; 316us; 290us; 316us; 292us; 316us; 294us; 316us; 296us; 316us; 297us; 316us; 300us; 316us; 307us; 316us; 309us; 316us; 313us; 316us; 317us; 316us; 326us; 316us; 329us; 316us; 334us; 316us; 338us; 316us; 343us; 316us; 348us; 316us; 354us; 316us; 357us; 316us; 362us; 316us; 576us; 316us; 579us; 316us; 584us; 316us; 586us; 316us; 64us; 65535us; 8us; 325us; 16us; 325us; 21us; 325us; 26us; 325us; 79us; 325us; 98us; 325us; 179us; 325us; 187us; 325us; 192us; 325us; 194us; 325us; 196us; 325us; 200us; 325us; 203us; 325us; 234us; 325us; 235us; 325us; 236us; 325us; 237us; 325us; 238us; 325us; 239us; 325us; 240us; 325us; 241us; 325us; 242us; 325us; 243us; 325us; 244us; 325us; 245us; 325us; 246us; 325us; 247us; 325us; 248us; 325us; 249us; 325us; 250us; 325us; 265us; 325us; 266us; 325us; 272us; 325us; 274us; 325us; 277us; 325us; 279us; 325us; 282us; 325us; 284us; 325us; 285us; 325us; 286us; 325us; 288us; 325us; 290us; 325us; 292us; 325us; 294us; 325us; 296us; 325us; 297us; 325us; 300us; 325us; 307us; 325us; 309us; 325us; 313us; 325us; 317us; 325us; 326us; 325us; 329us; 325us; 334us; 325us; 338us; 325us; 343us; 325us; 348us; 325us; 354us; 325us; 357us; 325us; 362us; 325us; 576us; 325us; 579us; 325us; 584us; 325us; 586us; 325us; 64us; 65535us; 8us; 321us; 16us; 321us; 21us; 321us; 26us; 321us; 79us; 321us; 98us; 321us; 179us; 321us; 187us; 321us; 192us; 321us; 194us; 321us; 196us; 321us; 200us; 321us; 203us; 321us; 234us; 321us; 235us; 321us; 236us; 321us; 237us; 321us; 238us; 321us; 239us; 321us; 240us; 321us; 241us; 321us; 242us; 321us; 243us; 321us; 244us; 321us; 245us; 321us; 246us; 321us; 247us; 321us; 248us; 321us; 249us; 321us; 250us; 321us; 265us; 321us; 266us; 321us; 272us; 321us; 274us; 321us; 277us; 321us; 279us; 321us; 282us; 321us; 284us; 321us; 285us; 321us; 286us; 321us; 288us; 321us; 290us; 321us; 292us; 321us; 294us; 321us; 296us; 321us; 297us; 321us; 300us; 321us; 307us; 321us; 309us; 321us; 313us; 321us; 317us; 321us; 326us; 321us; 329us; 321us; 334us; 321us; 338us; 321us; 343us; 321us; 348us; 321us; 354us; 321us; 357us; 321us; 362us; 321us; 576us; 321us; 579us; 321us; 584us; 321us; 586us; 321us; 1us; 65535us; 329us; 330us; 87us; 65535us; 8us; 322us; 16us; 322us; 21us; 322us; 26us; 322us; 51us; 657us; 79us; 322us; 98us; 322us; 179us; 322us; 187us; 322us; 192us; 322us; 194us; 322us; 196us; 322us; 200us; 322us; 203us; 322us; 234us; 322us; 235us; 322us; 236us; 322us; 237us; 322us; 238us; 322us; 239us; 322us; 240us; 322us; 241us; 322us; 242us; 322us; 243us; 322us; 244us; 322us; 245us; 322us; 246us; 322us; 247us; 322us; 248us; 322us; 249us; 322us; 250us; 322us; 265us; 322us; 266us; 322us; 272us; 322us; 274us; 322us; 277us; 322us; 279us; 322us; 282us; 322us; 284us; 322us; 285us; 322us; 286us; 322us; 288us; 322us; 290us; 322us; 292us; 322us; 294us; 322us; 296us; 322us; 297us; 322us; 300us; 322us; 307us; 322us; 309us; 322us; 313us; 322us; 317us; 322us; 326us; 322us; 329us; 322us; 334us; 322us; 338us; 322us; 343us; 322us; 348us; 322us; 354us; 322us; 357us; 322us; 362us; 322us; 508us; 657us; 514us; 657us; 531us; 657us; 576us; 322us; 579us; 322us; 584us; 322us; 586us; 322us; 602us; 657us; 605us; 657us; 608us; 657us; 611us; 657us; 613us; 657us; 617us; 657us; 618us; 657us; 620us; 657us; 621us; 657us; 625us; 657us; 629us; 657us; 632us; 657us; 642us; 657us; 644us; 657us; 647us; 657us; 649us; 657us; 651us; 657us; 660us; 657us; 665us; 657us; 2us; 65535us; 79us; 80us; 334us; 335us; 0us; 65535us; 2us; 65535us; 79us; 337us; 334us; 337us; 3us; 65535us; 230us; 339us; 231us; 344us; 232us; 349us; 46us; 65535us; 8us; 199us; 16us; 199us; 21us; 199us; 26us; 199us; 98us; 199us; 179us; 199us; 187us; 199us; 192us; 199us; 194us; 199us; 196us; 199us; 200us; 199us; 203us; 199us; 234us; 199us; 235us; 199us; 236us; 199us; 237us; 199us; 238us; 199us; 239us; 199us; 240us; 199us; 241us; 199us; 242us; 199us; 243us; 199us; 244us; 199us; 245us; 199us; 246us; 199us; 247us; 199us; 248us; 199us; 249us; 199us; 250us; 199us; 265us; 199us; 266us; 199us; 272us; 199us; 274us; 199us; 277us; 199us; 279us; 199us; 282us; 199us; 284us; 199us; 285us; 199us; 326us; 199us; 329us; 199us; 338us; 199us; 343us; 199us; 348us; 199us; 354us; 199us; 357us; 199us; 362us; 199us; 46us; 65535us; 8us; 198us; 16us; 198us; 21us; 198us; 26us; 198us; 98us; 198us; 179us; 198us; 187us; 198us; 192us; 198us; 194us; 198us; 196us; 198us; 200us; 198us; 203us; 198us; 234us; 198us; 235us; 198us; 236us; 198us; 237us; 198us; 238us; 198us; 239us; 198us; 240us; 198us; 241us; 198us; 242us; 198us; 243us; 198us; 244us; 198us; 245us; 198us; 246us; 198us; 247us; 198us; 248us; 198us; 249us; 198us; 250us; 198us; 265us; 198us; 266us; 198us; 272us; 198us; 274us; 198us; 277us; 198us; 279us; 198us; 282us; 198us; 284us; 198us; 285us; 198us; 326us; 198us; 329us; 198us; 338us; 198us; 343us; 198us; 348us; 198us; 354us; 198us; 357us; 198us; 362us; 198us; 3us; 65535us; 29us; 30us; 44us; 45us; 373us; 374us; 6us; 65535us; 28us; 29us; 35us; 36us; 38us; 39us; 43us; 44us; 46us; 47us; 372us; 373us; 6us; 65535us; 28us; 375us; 35us; 375us; 38us; 375us; 43us; 375us; 46us; 375us; 372us; 375us; 6us; 65535us; 28us; 378us; 35us; 378us; 38us; 378us; 43us; 378us; 46us; 378us; 372us; 378us; 2us; 65535us; 379us; 380us; 381us; 382us; 8us; 65535us; 383us; 384us; 387us; 388us; 390us; 391us; 393us; 394us; 395us; 401us; 396us; 402us; 399us; 593us; 592us; 593us; 8us; 65535us; 383us; 392us; 387us; 392us; 390us; 392us; 393us; 392us; 395us; 392us; 396us; 392us; 399us; 392us; 592us; 392us; 8us; 65535us; 383us; 389us; 387us; 389us; 390us; 389us; 393us; 389us; 395us; 389us; 396us; 389us; 399us; 389us; 592us; 389us; 8us; 65535us; 383us; 385us; 387us; 385us; 390us; 385us; 393us; 385us; 395us; 385us; 396us; 385us; 399us; 386us; 592us; 385us; 1us; 65535us; 379us; 381us; 1us; 65535us; 416us; 417us; 4us; 65535us; 61us; 62us; 74us; 75us; 463us; 464us; 539us; 540us; 2us; 65535us; 433us; 434us; 436us; 437us; 2us; 65535us; 422us; 423us; 439us; 440us; 3us; 65535us; 425us; 426us; 428us; 429us; 441us; 442us; 2us; 65535us; 443us; 444us; 446us; 447us; 4us; 65535us; 425us; 443us; 428us; 443us; 441us; 443us; 445us; 446us; 1us; 65535us; 376us; 377us; 1us; 65535us; 376us; 452us; 2us; 65535us; 454us; 455us; 457us; 458us; 2us; 65535us; 454us; 456us; 457us; 456us; 1us; 65535us; 376us; 453us; 2us; 65535us; 470us; 471us; 477us; 478us; 2us; 65535us; 470us; 476us; 477us; 476us; 1us; 65535us; 96us; 97us; 46us; 65535us; 55us; 56us; 59us; 60us; 64us; 65us; 68us; 69us; 72us; 73us; 77us; 78us; 156us; 157us; 184us; 185us; 376us; 451us; 383us; 405us; 387us; 405us; 390us; 405us; 393us; 405us; 395us; 406us; 396us; 406us; 399us; 405us; 409us; 411us; 410us; 412us; 449us; 450us; 461us; 462us; 466us; 467us; 468us; 469us; 482us; 485us; 483us; 484us; 486us; 487us; 488us; 489us; 494us; 497us; 495us; 496us; 501us; 511us; 502us; 512us; 503us; 513us; 506us; 507us; 520us; 545us; 527us; 545us; 534us; 535us; 542us; 543us; 553us; 546us; 560us; 548us; 562us; 547us; 564us; 547us; 566us; 548us; 571us; 549us; 573us; 550us; 582us; 551us; 592us; 405us; 636us; 552us; 2us; 65535us; 522us; 523us; 529us; 530us; 46us; 65535us; 55us; 492us; 59us; 492us; 64us; 492us; 68us; 492us; 72us; 492us; 77us; 492us; 156us; 492us; 184us; 492us; 376us; 492us; 383us; 492us; 387us; 492us; 390us; 492us; 393us; 492us; 395us; 492us; 396us; 492us; 399us; 492us; 409us; 490us; 410us; 490us; 449us; 492us; 461us; 492us; 466us; 492us; 468us; 492us; 482us; 492us; 483us; 492us; 486us; 492us; 488us; 492us; 494us; 492us; 495us; 492us; 501us; 490us; 502us; 490us; 503us; 490us; 506us; 492us; 520us; 491us; 527us; 493us; 534us; 492us; 542us; 492us; 553us; 492us; 560us; 492us; 562us; 492us; 564us; 492us; 566us; 492us; 571us; 492us; 573us; 492us; 582us; 492us; 592us; 492us; 636us; 492us; 46us; 65535us; 55us; 537us; 59us; 537us; 64us; 537us; 68us; 537us; 72us; 537us; 77us; 537us; 156us; 537us; 184us; 537us; 376us; 537us; 383us; 537us; 387us; 537us; 390us; 537us; 393us; 537us; 395us; 537us; 396us; 537us; 399us; 537us; 409us; 537us; 410us; 537us; 449us; 537us; 461us; 537us; 466us; 537us; 468us; 537us; 482us; 537us; 483us; 537us; 486us; 537us; 488us; 537us; 494us; 537us; 495us; 537us; 501us; 537us; 502us; 537us; 503us; 537us; 506us; 537us; 520us; 537us; 527us; 537us; 534us; 537us; 542us; 537us; 553us; 537us; 560us; 537us; 562us; 537us; 564us; 537us; 566us; 537us; 571us; 537us; 573us; 537us; 582us; 537us; 592us; 537us; 636us; 537us; 46us; 65535us; 55us; 544us; 59us; 544us; 64us; 544us; 68us; 544us; 72us; 544us; 77us; 544us; 156us; 544us; 184us; 544us; 376us; 544us; 383us; 544us; 387us; 544us; 390us; 544us; 393us; 544us; 395us; 544us; 396us; 544us; 399us; 544us; 409us; 544us; 410us; 544us; 449us; 544us; 461us; 544us; 466us; 544us; 468us; 544us; 482us; 544us; 483us; 544us; 486us; 544us; 488us; 544us; 494us; 544us; 495us; 544us; 501us; 544us; 502us; 544us; 503us; 544us; 506us; 544us; 520us; 544us; 527us; 544us; 534us; 544us; 542us; 544us; 553us; 544us; 560us; 544us; 562us; 544us; 564us; 544us; 566us; 544us; 571us; 544us; 573us; 544us; 582us; 544us; 592us; 544us; 636us; 544us; 3us; 65535us; 395us; 556us; 396us; 556us; 553us; 556us; 2us; 65535us; 562us; 563us; 564us; 565us; 2us; 65535us; 560us; 561us; 566us; 567us; 8us; 65535us; 299us; 308us; 301us; 308us; 302us; 308us; 303us; 308us; 304us; 308us; 305us; 308us; 306us; 308us; 558us; 559us; 10us; 65535us; 299us; 568us; 301us; 568us; 302us; 568us; 303us; 568us; 304us; 568us; 305us; 568us; 306us; 568us; 558us; 568us; 569us; 570us; 644us; 646us; 4us; 65535us; 406us; 554us; 546us; 554us; 550us; 574us; 551us; 583us; 4us; 65535us; 576us; 577us; 579us; 580us; 584us; 585us; 586us; 587us; 47us; 65535us; 55us; 558us; 59us; 558us; 64us; 558us; 68us; 558us; 72us; 558us; 77us; 558us; 156us; 558us; 184us; 558us; 376us; 558us; 383us; 558us; 387us; 558us; 390us; 558us; 393us; 558us; 395us; 558us; 396us; 558us; 399us; 558us; 409us; 558us; 410us; 558us; 449us; 558us; 461us; 558us; 466us; 558us; 468us; 558us; 482us; 558us; 483us; 558us; 486us; 558us; 488us; 558us; 494us; 558us; 495us; 558us; 501us; 558us; 502us; 558us; 503us; 558us; 506us; 558us; 520us; 558us; 527us; 558us; 534us; 558us; 537us; 538us; 542us; 558us; 553us; 558us; 560us; 558us; 562us; 558us; 564us; 558us; 566us; 558us; 571us; 558us; 573us; 558us; 582us; 558us; 592us; 558us; 636us; 558us; 56us; 65535us; 55us; 589us; 59us; 589us; 61us; 431us; 64us; 589us; 68us; 589us; 72us; 589us; 74us; 431us; 77us; 589us; 156us; 589us; 184us; 589us; 376us; 589us; 383us; 589us; 387us; 589us; 390us; 589us; 393us; 589us; 395us; 589us; 396us; 589us; 399us; 589us; 409us; 589us; 410us; 589us; 418us; 419us; 421us; 422us; 432us; 433us; 435us; 436us; 438us; 439us; 449us; 589us; 461us; 589us; 463us; 431us; 466us; 589us; 468us; 589us; 482us; 589us; 483us; 589us; 486us; 589us; 488us; 589us; 494us; 589us; 495us; 589us; 501us; 589us; 502us; 589us; 503us; 589us; 506us; 589us; 520us; 589us; 527us; 589us; 534us; 589us; 537us; 589us; 539us; 431us; 542us; 589us; 553us; 589us; 560us; 589us; 562us; 589us; 564us; 589us; 566us; 589us; 571us; 589us; 573us; 589us; 582us; 589us; 592us; 589us; 636us; 589us; 2us; 65535us; 399us; 594us; 592us; 594us; 4us; 65535us; 51us; 52us; 508us; 509us; 514us; 515us; 531us; 532us; 8us; 65535us; 51us; 595us; 508us; 595us; 514us; 595us; 531us; 595us; 613us; 614us; 625us; 626us; 629us; 630us; 632us; 633us; 8us; 65535us; 51us; 600us; 508us; 600us; 514us; 600us; 531us; 600us; 613us; 600us; 625us; 600us; 629us; 600us; 632us; 600us; 8us; 65535us; 51us; 604us; 508us; 604us; 514us; 604us; 531us; 604us; 613us; 604us; 625us; 604us; 629us; 604us; 632us; 604us; 11us; 65535us; 51us; 601us; 508us; 601us; 514us; 601us; 531us; 601us; 602us; 603us; 605us; 606us; 608us; 609us; 613us; 601us; 625us; 601us; 629us; 601us; 632us; 601us; 12us; 65535us; 51us; 607us; 508us; 607us; 514us; 607us; 531us; 607us; 602us; 607us; 605us; 607us; 608us; 607us; 611us; 612us; 613us; 607us; 625us; 607us; 629us; 607us; 632us; 607us; 12us; 65535us; 51us; 610us; 508us; 610us; 514us; 610us; 531us; 610us; 602us; 610us; 605us; 610us; 608us; 610us; 611us; 610us; 613us; 610us; 625us; 610us; 629us; 610us; 632us; 610us; 12us; 65535us; 51us; 615us; 508us; 615us; 514us; 615us; 531us; 615us; 602us; 615us; 605us; 615us; 608us; 615us; 611us; 615us; 613us; 615us; 625us; 615us; 629us; 615us; 632us; 615us; 12us; 65535us; 51us; 623us; 508us; 623us; 514us; 623us; 531us; 623us; 602us; 623us; 605us; 623us; 608us; 623us; 611us; 623us; 613us; 623us; 625us; 623us; 629us; 623us; 632us; 623us; 1us; 65535us; 596us; 597us; 1us; 65535us; 598us; 599us; 3us; 65535us; 596us; 638us; 598us; 638us; 639us; 638us; 3us; 65535us; 596us; 628us; 598us; 631us; 639us; 640us; 12us; 65535us; 51us; 624us; 508us; 624us; 514us; 624us; 531us; 624us; 602us; 624us; 605us; 624us; 608us; 624us; 611us; 624us; 613us; 624us; 625us; 624us; 629us; 624us; 632us; 624us; 3us; 65535us; 642us; 643us; 644us; 645us; 647us; 648us; 3us; 65535us; 625us; 663us; 660us; 663us; 665us; 666us; 3us; 65535us; 617us; 650us; 649us; 650us; 651us; 652us; 23us; 65535us; 51us; 616us; 508us; 616us; 514us; 616us; 531us; 616us; 602us; 616us; 605us; 616us; 608us; 616us; 611us; 616us; 613us; 616us; 617us; 651us; 618us; 619us; 620us; 621us; 621us; 622us; 625us; 617us; 629us; 616us; 632us; 616us; 642us; 647us; 644us; 647us; 647us; 647us; 649us; 651us; 651us; 651us; 660us; 649us; 665us; 649us; 23us; 65535us; 51us; 653us; 508us; 653us; 514us; 653us; 531us; 653us; 602us; 653us; 605us; 653us; 608us; 653us; 611us; 653us; 613us; 653us; 617us; 653us; 618us; 653us; 620us; 653us; 621us; 653us; 625us; 653us; 629us; 653us; 632us; 653us; 642us; 653us; 644us; 653us; 647us; 653us; 649us; 653us; 651us; 653us; 660us; 653us; 665us; 653us; 23us; 65535us; 51us; 654us; 508us; 654us; 514us; 654us; 531us; 654us; 602us; 654us; 605us; 654us; 608us; 654us; 611us; 654us; 613us; 654us; 617us; 654us; 618us; 654us; 620us; 654us; 621us; 654us; 625us; 654us; 629us; 654us; 632us; 654us; 642us; 654us; 644us; 654us; 647us; 654us; 649us; 654us; 651us; 654us; 660us; 654us; 665us; 654us; 2us; 65535us; 625us; 661us; 660us; 661us; 2us; 65535us; 625us; 664us; 660us; 664us; 201us; 65535us; 8us; 312us; 10us; 671us; 12us; 671us; 16us; 312us; 21us; 312us; 26us; 312us; 28us; 671us; 31us; 37us; 33us; 671us; 35us; 671us; 38us; 671us; 40us; 41us; 43us; 671us; 46us; 671us; 48us; 671us; 50us; 669us; 51us; 671us; 53us; 54us; 55us; 500us; 57us; 58us; 59us; 500us; 64us; 500us; 66us; 671us; 68us; 500us; 70us; 71us; 72us; 500us; 77us; 500us; 79us; 312us; 84us; 671us; 87us; 671us; 93us; 671us; 98us; 312us; 104us; 105us; 106us; 671us; 109us; 671us; 110us; 671us; 111us; 671us; 113us; 671us; 116us; 671us; 118us; 671us; 120us; 671us; 123us; 671us; 131us; 671us; 136us; 671us; 139us; 671us; 141us; 671us; 145us; 671us; 152us; 153us; 155us; 671us; 156us; 500us; 159us; 671us; 160us; 671us; 179us; 312us; 184us; 500us; 187us; 312us; 192us; 312us; 194us; 312us; 196us; 312us; 200us; 312us; 203us; 312us; 234us; 312us; 235us; 312us; 236us; 312us; 237us; 312us; 238us; 312us; 239us; 312us; 240us; 312us; 241us; 312us; 242us; 312us; 243us; 312us; 244us; 312us; 245us; 312us; 246us; 312us; 247us; 312us; 248us; 312us; 249us; 312us; 250us; 312us; 252us; 671us; 254us; 671us; 258us; 671us; 263us; 671us; 265us; 312us; 266us; 312us; 272us; 312us; 274us; 312us; 277us; 312us; 279us; 312us; 282us; 312us; 284us; 312us; 285us; 312us; 286us; 312us; 288us; 312us; 290us; 312us; 292us; 312us; 294us; 312us; 296us; 312us; 297us; 312us; 300us; 312us; 307us; 312us; 309us; 312us; 310us; 319us; 313us; 312us; 317us; 312us; 326us; 312us; 329us; 312us; 334us; 312us; 338us; 312us; 341us; 671us; 343us; 312us; 346us; 671us; 348us; 312us; 352us; 671us; 354us; 312us; 355us; 671us; 357us; 312us; 360us; 671us; 362us; 312us; 366us; 671us; 369us; 671us; 372us; 671us; 376us; 500us; 383us; 408us; 387us; 408us; 390us; 408us; 393us; 408us; 395us; 407us; 396us; 407us; 397us; 398us; 399us; 408us; 409us; 500us; 410us; 500us; 425us; 448us; 428us; 448us; 441us; 448us; 445us; 448us; 449us; 500us; 454us; 459us; 457us; 459us; 461us; 500us; 466us; 500us; 468us; 500us; 470us; 481us; 477us; 480us; 482us; 500us; 483us; 500us; 486us; 500us; 488us; 500us; 494us; 500us; 495us; 500us; 501us; 500us; 502us; 500us; 503us; 500us; 504us; 505us; 506us; 500us; 508us; 671us; 514us; 671us; 518us; 519us; 520us; 500us; 525us; 526us; 527us; 500us; 531us; 671us; 534us; 500us; 537us; 671us; 542us; 500us; 553us; 498us; 560us; 500us; 562us; 499us; 564us; 499us; 566us; 500us; 571us; 500us; 573us; 500us; 576us; 312us; 579us; 312us; 582us; 500us; 584us; 312us; 586us; 312us; 590us; 591us; 592us; 408us; 602us; 671us; 605us; 671us; 608us; 671us; 611us; 671us; 613us; 671us; 617us; 671us; 618us; 671us; 620us; 671us; 621us; 671us; 625us; 671us; 629us; 671us; 632us; 671us; 634us; 635us; 636us; 500us; 641us; 642us; 642us; 671us; 644us; 671us; 647us; 671us; 649us; 671us; 651us; 671us; 660us; 671us; 665us; 671us; 672us; 673us; 1us; 65535us; 50us; 51us; 112us; 65535us; 10us; 11us; 12us; 13us; 28us; 379us; 33us; 34us; 35us; 379us; 38us; 379us; 43us; 379us; 46us; 379us; 48us; 49us; 51us; 644us; 55us; 588us; 59us; 588us; 64us; 588us; 66us; 67us; 68us; 588us; 72us; 588us; 77us; 588us; 84us; 134us; 87us; 134us; 93us; 134us; 106us; 134us; 109us; 134us; 110us; 134us; 111us; 134us; 113us; 134us; 116us; 134us; 118us; 134us; 120us; 135us; 123us; 134us; 131us; 134us; 136us; 134us; 139us; 140us; 141us; 134us; 145us; 134us; 155us; 134us; 156us; 588us; 159us; 134us; 160us; 134us; 184us; 588us; 252us; 134us; 254us; 134us; 258us; 134us; 263us; 134us; 341us; 342us; 346us; 347us; 352us; 353us; 355us; 134us; 360us; 134us; 366us; 134us; 369us; 134us; 372us; 379us; 376us; 588us; 383us; 588us; 387us; 588us; 390us; 588us; 393us; 588us; 395us; 588us; 396us; 588us; 399us; 588us; 409us; 588us; 410us; 588us; 449us; 588us; 461us; 588us; 466us; 588us; 468us; 588us; 482us; 588us; 483us; 588us; 486us; 588us; 488us; 588us; 494us; 588us; 495us; 588us; 501us; 588us; 502us; 588us; 503us; 588us; 506us; 588us; 508us; 644us; 514us; 644us; 520us; 588us; 527us; 588us; 531us; 644us; 534us; 588us; 537us; 588us; 542us; 588us; 553us; 588us; 560us; 588us; 562us; 588us; 564us; 588us; 566us; 588us; 571us; 588us; 573us; 588us; 582us; 588us; 592us; 588us; 602us; 644us; 605us; 644us; 608us; 644us; 611us; 644us; 613us; 644us; 617us; 655us; 618us; 655us; 620us; 655us; 621us; 655us; 625us; 644us; 629us; 644us; 632us; 644us; 636us; 588us; 642us; 655us; 644us; 655us; 647us; 655us; 649us; 655us; 651us; 655us; 660us; 655us; 665us; 655us; 0us; 65535us; 0us; 65535us; 0us; 65535us; 0us; 65535us; 3us; 65535us; 15us; 677us; 19us; 677us; 677us; 677us; 3us; 65535us; 15us; 16us; 19us; 21us; 677us; 678us; 7us; 65535us; 138us; 139us; 144us; 145us; 186us; 187us; 339us; 352us; 344us; 352us; 349us; 352us; 476us; 477us; 1us; 65535us; 459us; 460us; 4us; 65535us; 339us; 340us; 344us; 345us; 349us; 350us; 476us; 479us; 2us; 65535us; 83us; 84us; 86us; 87us; 2us; 65535us; 365us; 366us; 368us; 369us; 1us; 65535us; 379us; 416us; |]
let _fsyacc_sparseGotoTableRowOffsets = [|0us; 1us; 3us; 6us; 9us; 12us; 14us; 18us; 23us; 28us; 30us; 32us; 80us; 128us; 130us; 133us; 136us; 140us; 144us; 151us; 158us; 179us; 186us; 213us; 240us; 242us; 244us; 246us; 248us; 250us; 265us; 280us; 394us; 399us; 400us; 415us; 431us; 434us; 481us; 483us; 486us; 493us; 500us; 502us; 504us; 507us; 510us; 513us; 515us; 562us; 614us; 668us; 671us; 736us; 739us; 741us; 806us; 871us; 936us; 938us; 1026us; 1029us; 1030us; 1033us; 1037us; 1084us; 1131us; 1135us; 1142us; 1149us; 1156us; 1159us; 1168us; 1177us; 1186us; 1195us; 1197us; 1199us; 1204us; 1207us; 1210us; 1214us; 1217us; 1222us; 1224us; 1226us; 1229us; 1232us; 1234us; 1237us; 1240us; 1242us; 1289us; 1292us; 1339us; 1386us; 1433us; 1437us; 1440us; 1443us; 1452us; 1463us; 1468us; 1473us; 1521us; 1578us; 1581us; 1586us; 1595us; 1604us; 1613us; 1625us; 1638us; 1651us; 1664us; 1677us; 1679us; 1681us; 1685us; 1689us; 1702us; 1706us; 1710us; 1714us; 1738us; 1762us; 1786us; 1789us; 1792us; 1994us; 1996us; 2109us; 2110us; 2111us; 2112us; 2113us; 2117us; 2121us; 2129us; 2131us; 2136us; 2139us; 2142us; |]
let _fsyacc_stateToProdIdxsTableElements = [| 1us; 0us; 1us; 0us; 1us; 1us; 1us; 1us; 1us; 1us; 2us; 3us; 4us; 1us; 4us; 1us; 4us; 1us; 5us; 1us; 5us; 2us; 6us; 7us; 3us; 6us; 7us; 353us; 1us; 7us; 2us; 7us; 353us; 1us; 8us; 21us; 10us; 11us; 12us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 2us; 10us; 11us; 1us; 10us; 1us; 13us; 4us; 14us; 15us; 16us; 17us; 1us; 14us; 2us; 15us; 17us; 1us; 15us; 1us; 18us; 2us; 18us; 111us; 1us; 19us; 3us; 19us; 112us; 113us; 2us; 20us; 28us; 1us; 20us; 1us; 20us; 1us; 20us; 6us; 21us; 22us; 23us; 31us; 32us; 33us; 2us; 21us; 31us; 1us; 21us; 2us; 21us; 353us; 1us; 21us; 1us; 21us; 3us; 22us; 32us; 33us; 1us; 22us; 1us; 22us; 1us; 23us; 1us; 23us; 1us; 23us; 1us; 24us; 1us; 24us; 1us; 24us; 1us; 25us; 1us; 25us; 1us; 26us; 2us; 26us; 353us; 1us; 27us; 1us; 27us; 1us; 27us; 1us; 28us; 1us; 28us; 1us; 28us; 2us; 28us; 273us; 2us; 29us; 30us; 2us; 29us; 30us; 1us; 29us; 2us; 29us; 273us; 1us; 30us; 1us; 30us; 1us; 30us; 1us; 30us; 2us; 30us; 273us; 1us; 31us; 2us; 31us; 353us; 1us; 31us; 2us; 31us; 273us; 2us; 32us; 33us; 2us; 32us; 33us; 1us; 32us; 2us; 32us; 273us; 1us; 33us; 1us; 33us; 1us; 33us; 1us; 33us; 2us; 33us; 273us; 1us; 34us; 1us; 34us; 1us; 34us; 1us; 35us; 1us; 37us; 1us; 37us; 1us; 37us; 1us; 38us; 1us; 38us; 1us; 38us; 1us; 38us; 1us; 39us; 1us; 40us; 1us; 40us; 1us; 41us; 1us; 41us; 1us; 41us; 1us; 43us; 1us; 43us; 1us; 43us; 1us; 43us; 4us; 44us; 45us; 46us; 50us; 4us; 45us; 46us; 46us; 50us; 4us; 45us; 46us; 49us; 50us; 4us; 45us; 46us; 50us; 50us; 1us; 45us; 1us; 45us; 1us; 46us; 2us; 47us; 49us; 1us; 48us; 1us; 49us; 1us; 50us; 3us; 51us; 52us; 60us; 1us; 51us; 1us; 52us; 1us; 52us; 1us; 53us; 3us; 54us; 55us; 56us; 1us; 54us; 1us; 55us; 1us; 55us; 1us; 57us; 1us; 57us; 1us; 57us; 1us; 58us; 1us; 58us; 1us; 58us; 1us; 59us; 1us; 60us; 1us; 61us; 1us; 62us; 1us; 63us; 1us; 64us; 1us; 64us; 1us; 64us; 2us; 65us; 353us; 2us; 66us; 353us; 1us; 66us; 5us; 66us; 75us; 77us; 78us; 81us; 2us; 66us; 67us; 1us; 67us; 2us; 67us; 353us; 1us; 67us; 5us; 67us; 75us; 77us; 78us; 81us; 5us; 69us; 75us; 77us; 78us; 81us; 2us; 69us; 71us; 1us; 71us; 5us; 71us; 75us; 77us; 78us; 81us; 5us; 73us; 75us; 77us; 78us; 81us; 5us; 75us; 77us; 77us; 78us; 81us; 5us; 75us; 77us; 78us; 80us; 81us; 5us; 75us; 77us; 78us; 81us; 81us; 5us; 75us; 77us; 78us; 81us; 141us; 1us; 75us; 1us; 75us; 2us; 76us; 80us; 1us; 77us; 1us; 78us; 2us; 78us; 273us; 1us; 79us; 1us; 80us; 1us; 81us; 1us; 82us; 1us; 83us; 1us; 84us; 1us; 85us; 1us; 86us; 1us; 87us; 1us; 88us; 1us; 89us; 1us; 90us; 1us; 91us; 1us; 92us; 1us; 93us; 1us; 94us; 1us; 95us; 1us; 96us; 1us; 97us; 1us; 98us; 1us; 99us; 1us; 100us; 1us; 100us; 1us; 100us; 1us; 101us; 2us; 104us; 105us; 1us; 104us; 2us; 104us; 273us; 21us; 106us; 107us; 108us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 2us; 106us; 107us; 1us; 106us; 1us; 109us; 1us; 110us; 1us; 111us; 1us; 111us; 1us; 111us; 2us; 112us; 113us; 1us; 112us; 1us; 113us; 1us; 113us; 1us; 114us; 1us; 115us; 1us; 116us; 1us; 116us; 1us; 116us; 1us; 117us; 19us; 117us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 1us; 117us; 2us; 118us; 156us; 19us; 119us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 131us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 132us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 133us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 134us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 135us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 144us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 150us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 153us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 155us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 156us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 157us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 195us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 196us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 197us; 19us; 119us; 120us; 121us; 122us; 123us; 124us; 125us; 126us; 127us; 128us; 129us; 130us; 131us; 132us; 133us; 134us; 135us; 157us; 198us; 1us; 119us; 1us; 120us; 1us; 121us; 1us; 122us; 1us; 123us; 1us; 124us; 1us; 125us; 1us; 126us; 1us; 127us; 1us; 128us; 1us; 129us; 1us; 130us; 1us; 131us; 1us; 132us; 1us; 133us; 1us; 134us; 1us; 135us; 1us; 136us; 1us; 137us; 1us; 137us; 1us; 138us; 1us; 138us; 1us; 138us; 1us; 139us; 1us; 140us; 1us; 140us; 1us; 141us; 2us; 142us; 143us; 2us; 142us; 143us; 1us; 143us; 1us; 143us; 1us; 144us; 1us; 146us; 1us; 146us; 2us; 147us; 148us; 1us; 147us; 1us; 149us; 1us; 149us; 1us; 150us; 1us; 151us; 1us; 151us; 1us; 151us; 1us; 151us; 1us; 153us; 1us; 154us; 1us; 154us; 1us; 154us; 1us; 154us; 1us; 155us; 1us; 155us; 1us; 156us; 1us; 157us; 1us; 158us; 1us; 158us; 1us; 159us; 1us; 159us; 1us; 160us; 1us; 160us; 1us; 161us; 1us; 161us; 1us; 162us; 1us; 162us; 2us; 163us; 164us; 4us; 164us; 195us; 196us; 197us; 1us; 164us; 4us; 165us; 168us; 169us; 171us; 1us; 166us; 4us; 166us; 168us; 169us; 171us; 4us; 167us; 168us; 169us; 171us; 4us; 168us; 168us; 169us; 171us; 4us; 168us; 169us; 170us; 171us; 5us; 168us; 169us; 171us; 176us; 177us; 5us; 168us; 169us; 171us; 294us; 295us; 1us; 168us; 1us; 169us; 1us; 170us; 1us; 171us; 1us; 171us; 1us; 172us; 1us; 173us; 1us; 173us; 1us; 173us; 1us; 174us; 1us; 177us; 1us; 177us; 1us; 178us; 1us; 179us; 1us; 180us; 1us; 181us; 1us; 182us; 1us; 183us; 1us; 184us; 1us; 185us; 1us; 185us; 1us; 185us; 2us; 186us; 187us; 2us; 186us; 187us; 1us; 186us; 1us; 187us; 1us; 189us; 1us; 190us; 1us; 190us; 1us; 190us; 1us; 191us; 1us; 195us; 2us; 195us; 198us; 1us; 195us; 1us; 196us; 2us; 196us; 353us; 1us; 196us; 2us; 196us; 198us; 1us; 196us; 1us; 197us; 2us; 197us; 353us; 1us; 197us; 2us; 197us; 198us; 1us; 197us; 1us; 197us; 2us; 198us; 397us; 2us; 198us; 353us; 1us; 198us; 2us; 200us; 201us; 2us; 200us; 201us; 2us; 200us; 201us; 1us; 200us; 1us; 201us; 1us; 202us; 1us; 202us; 1us; 202us; 1us; 202us; 1us; 202us; 1us; 203us; 1us; 203us; 1us; 203us; 1us; 204us; 1us; 204us; 1us; 204us; 1us; 204us; 1us; 205us; 1us; 205us; 1us; 205us; 2us; 207us; 208us; 1us; 208us; 1us; 208us; 1us; 209us; 3us; 210us; 211us; 353us; 1us; 210us; 1us; 211us; 1us; 211us; 1us; 213us; 1us; 213us; 2us; 214us; 215us; 3us; 214us; 215us; 218us; 1us; 215us; 1us; 215us; 1us; 216us; 1us; 216us; 1us; 216us; 1us; 217us; 1us; 217us; 1us; 217us; 4us; 218us; 219us; 274us; 275us; 5us; 218us; 219us; 274us; 275us; 301us; 4us; 218us; 298us; 299us; 300us; 4us; 218us; 298us; 299us; 300us; 3us; 218us; 299us; 300us; 1us; 218us; 1us; 219us; 2us; 219us; 301us; 1us; 219us; 2us; 219us; 301us; 2us; 220us; 273us; 5us; 220us; 273us; 274us; 278us; 279us; 5us; 221us; 262us; 264us; 280us; 352us; 4us; 221us; 262us; 264us; 352us; 3us; 221us; 262us; 264us; 4us; 221us; 262us; 264us; 280us; 3us; 221us; 264us; 273us; 4us; 221us; 264us; 273us; 280us; 1us; 222us; 1us; 223us; 1us; 224us; 1us; 225us; 1us; 225us; 4us; 226us; 227us; 228us; 229us; 3us; 226us; 227us; 228us; 1us; 226us; 1us; 227us; 1us; 227us; 1us; 227us; 1us; 227us; 1us; 228us; 1us; 228us; 1us; 228us; 1us; 229us; 1us; 229us; 1us; 229us; 2us; 230us; 231us; 1us; 231us; 1us; 231us; 1us; 231us; 1us; 233us; 1us; 233us; 1us; 233us; 1us; 234us; 1us; 234us; 1us; 234us; 1us; 235us; 1us; 235us; 2us; 237us; 238us; 1us; 238us; 1us; 240us; 1us; 240us; 1us; 240us; 1us; 241us; 1us; 241us; 2us; 241us; 273us; 2us; 242us; 273us; 1us; 243us; 1us; 244us; 1us; 245us; 1us; 245us; 2us; 246us; 247us; 1us; 246us; 1us; 246us; 4us; 248us; 249us; 250us; 251us; 1us; 248us; 1us; 249us; 2us; 249us; 273us; 1us; 250us; 1us; 250us; 1us; 250us; 1us; 250us; 2us; 250us; 273us; 1us; 251us; 2us; 251us; 273us; 4us; 252us; 253us; 254us; 263us; 2us; 252us; 253us; 1us; 252us; 1us; 253us; 1us; 254us; 1us; 254us; 2us; 255us; 256us; 2us; 255us; 397us; 1us; 255us; 1us; 256us; 1us; 257us; 2us; 257us; 263us; 1us; 257us; 2us; 257us; 263us; 3us; 257us; 263us; 273us; 2us; 257us; 273us; 1us; 259us; 2us; 259us; 273us; 1us; 260us; 2us; 260us; 273us; 3us; 261us; 262us; 266us; 3us; 261us; 265us; 266us; 2us; 261us; 266us; 3us; 261us; 266us; 267us; 1us; 261us; 2us; 261us; 262us; 3us; 261us; 262us; 273us; 2us; 261us; 273us; 4us; 262us; 264us; 280us; 352us; 4us; 262us; 264us; 282us; 352us; 3us; 262us; 264us; 352us; 2us; 262us; 264us; 3us; 262us; 264us; 280us; 3us; 262us; 264us; 282us; 1us; 263us; 1us; 263us; 1us; 263us; 2us; 263us; 273us; 1us; 263us; 1us; 263us; 1us; 263us; 2us; 264us; 273us; 3us; 264us; 273us; 280us; 3us; 264us; 273us; 282us; 1us; 264us; 1us; 264us; 1us; 264us; 1us; 265us; 1us; 265us; 1us; 265us; 1us; 265us; 1us; 265us; 1us; 265us; 1us; 265us; 1us; 267us; 1us; 267us; 1us; 267us; 1us; 267us; 1us; 267us; 1us; 267us; 1us; 267us; 1us; 268us; 1us; 268us; 1us; 268us; 1us; 269us; 2us; 269us; 273us; 1us; 269us; 1us; 270us; 1us; 270us; 1us; 271us; 1us; 271us; 1us; 271us; 1us; 271us; 2us; 271us; 273us; 1us; 272us; 1us; 273us; 4us; 273us; 274us; 278us; 279us; 2us; 273us; 281us; 3us; 273us; 283us; 284us; 4us; 273us; 287us; 288us; 289us; 2us; 273us; 288us; 2us; 273us; 291us; 2us; 273us; 323us; 2us; 274us; 275us; 1us; 274us; 1us; 274us; 1us; 275us; 1us; 275us; 2us; 276us; 277us; 1us; 277us; 1us; 279us; 1us; 279us; 1us; 280us; 1us; 280us; 1us; 282us; 1us; 282us; 1us; 284us; 1us; 284us; 1us; 285us; 1us; 286us; 1us; 286us; 4us; 287us; 288us; 289us; 290us; 1us; 287us; 1us; 288us; 1us; 288us; 1us; 288us; 1us; 289us; 1us; 289us; 1us; 289us; 1us; 290us; 1us; 290us; 1us; 290us; 1us; 291us; 1us; 291us; 1us; 292us; 1us; 292us; 1us; 295us; 1us; 295us; 2us; 296us; 353us; 1us; 297us; 3us; 298us; 299us; 300us; 3us; 298us; 299us; 300us; 2us; 299us; 300us; 1us; 299us; 1us; 300us; 1us; 302us; 1us; 303us; 1us; 303us; 1us; 304us; 1us; 304us; 1us; 305us; 3us; 306us; 308us; 309us; 1us; 306us; 1us; 306us; 1us; 307us; 1us; 308us; 1us; 308us; 2us; 310us; 311us; 1us; 310us; 1us; 310us; 2us; 312us; 313us; 1us; 312us; 1us; 312us; 2us; 314us; 349us; 1us; 314us; 1us; 315us; 1us; 316us; 3us; 316us; 331us; 332us; 1us; 316us; 1us; 316us; 1us; 317us; 1us; 317us; 1us; 317us; 1us; 318us; 1us; 319us; 2us; 320us; 342us; 1us; 320us; 1us; 320us; 1us; 321us; 1us; 321us; 1us; 321us; 1us; 322us; 1us; 322us; 1us; 322us; 1us; 323us; 1us; 323us; 1us; 323us; 1us; 323us; 2us; 324us; 325us; 1us; 324us; 1us; 324us; 1us; 326us; 1us; 326us; 1us; 326us; 4us; 327us; 328us; 337us; 353us; 1us; 327us; 1us; 328us; 1us; 330us; 1us; 330us; 2us; 331us; 332us; 1us; 331us; 2us; 333us; 334us; 1us; 334us; 1us; 335us; 1us; 336us; 2us; 337us; 353us; 1us; 338us; 1us; 339us; 1us; 340us; 1us; 341us; 1us; 342us; 1us; 342us; 1us; 342us; 2us; 344us; 346us; 2us; 345us; 347us; 2us; 345us; 347us; 2us; 345us; 347us; 1us; 348us; 1us; 349us; 1us; 350us; 1us; 350us; 1us; 352us; 1us; 353us; 1us; 353us; 1us; 386us; 1us; 387us; 1us; 388us; 2us; 389us; 390us; 1us; 390us; 2us; 391us; 393us; 2us; 392us; 394us; 1us; 393us; 1us; 394us; 1us; 395us; 1us; 399us; 1us; 401us; 1us; 403us; |]
let _fsyacc_stateToProdIdxsTableRowOffsets = [|0us; 2us; 4us; 6us; 8us; 10us; 13us; 15us; 17us; 19us; 21us; 24us; 28us; 30us; 33us; 35us; 57us; 60us; 62us; 64us; 69us; 71us; 74us; 76us; 78us; 81us; 83us; 87us; 90us; 92us; 94us; 96us; 103us; 106us; 108us; 111us; 113us; 115us; 119us; 121us; 123us; 125us; 127us; 129us; 131us; 133us; 135us; 137us; 139us; 141us; 144us; 146us; 148us; 150us; 152us; 154us; 156us; 159us; 162us; 165us; 167us; 170us; 172us; 174us; 176us; 178us; 181us; 183us; 186us; 188us; 191us; 194us; 197us; 199us; 202us; 204us; 206us; 208us; 210us; 213us; 215us; 217us; 219us; 221us; 223us; 225us; 227us; 229us; 231us; 233us; 235us; 237us; 239us; 241us; 243us; 245us; 247us; 249us; 251us; 253us; 255us; 260us; 265us; 270us; 275us; 277us; 279us; 281us; 284us; 286us; 288us; 290us; 294us; 296us; 298us; 300us; 302us; 306us; 308us; 310us; 312us; 314us; 316us; 318us; 320us; 322us; 324us; 326us; 328us; 330us; 332us; 334us; 336us; 338us; 340us; 343us; 346us; 348us; 354us; 357us; 359us; 362us; 364us; 370us; 376us; 379us; 381us; 387us; 393us; 399us; 405us; 411us; 417us; 419us; 421us; 424us; 426us; 428us; 431us; 433us; 435us; 437us; 439us; 441us; 443us; 445us; 447us; 449us; 451us; 453us; 455us; 457us; 459us; 461us; 463us; 465us; 467us; 469us; 471us; 473us; 475us; 477us; 479us; 481us; 484us; 486us; 489us; 511us; 514us; 516us; 518us; 520us; 522us; 524us; 526us; 529us; 531us; 533us; 535us; 537us; 539us; 541us; 543us; 545us; 547us; 567us; 569us; 572us; 592us; 612us; 632us; 652us; 672us; 692us; 712us; 732us; 752us; 772us; 792us; 812us; 832us; 852us; 872us; 892us; 912us; 932us; 952us; 972us; 992us; 1012us; 1032us; 1052us; 1072us; 1092us; 1112us; 1114us; 1116us; 1118us; 1120us; 1122us; 1124us; 1126us; 1128us; 1130us; 1132us; 1134us; 1136us; 1138us; 1140us; 1142us; 1144us; 1146us; 1148us; 1150us; 1152us; 1154us; 1156us; 1158us; 1160us; 1162us; 1164us; 1166us; 1169us; 1172us; 1174us; 1176us; 1178us; 1180us; 1182us; 1185us; 1187us; 1189us; 1191us; 1193us; 1195us; 1197us; 1199us; 1201us; 1203us; 1205us; 1207us; 1209us; 1211us; 1213us; 1215us; 1217us; 1219us; 1221us; 1223us; 1225us; 1227us; 1229us; 1231us; 1233us; 1235us; 1237us; 1239us; 1242us; 1247us; 1249us; 1254us; 1256us; 1261us; 1266us; 1271us; 1276us; 1282us; 1288us; 1290us; 1292us; 1294us; 1296us; 1298us; 1300us; 1302us; 1304us; 1306us; 1308us; 1310us; 1312us; 1314us; 1316us; 1318us; 1320us; 1322us; 1324us; 1326us; 1328us; 1330us; 1332us; 1335us; 1338us; 1340us; 1342us; 1344us; 1346us; 1348us; 1350us; 1352us; 1354us; 1357us; 1359us; 1361us; 1364us; 1366us; 1369us; 1371us; 1373us; 1376us; 1378us; 1381us; 1383us; 1385us; 1388us; 1391us; 1393us; 1396us; 1399us; 1402us; 1404us; 1406us; 1408us; 1410us; 1412us; 1414us; 1416us; 1418us; 1420us; 1422us; 1424us; 1426us; 1428us; 1430us; 1432us; 1434us; 1436us; 1439us; 1441us; 1443us; 1445us; 1449us; 1451us; 1453us; 1455us; 1457us; 1459us; 1462us; 1466us; 1468us; 1470us; 1472us; 1474us; 1476us; 1478us; 1480us; 1482us; 1487us; 1493us; 1498us; 1503us; 1507us; 1509us; 1511us; 1514us; 1516us; 1519us; 1522us; 1528us; 1534us; 1539us; 1543us; 1548us; 1552us; 1557us; 1559us; 1561us; 1563us; 1565us; 1567us; 1572us; 1576us; 1578us; 1580us; 1582us; 1584us; 1586us; 1588us; 1590us; 1592us; 1594us; 1596us; 1598us; 1601us; 1603us; 1605us; 1607us; 1609us; 1611us; 1613us; 1615us; 1617us; 1619us; 1621us; 1623us; 1626us; 1628us; 1630us; 1632us; 1634us; 1636us; 1638us; 1641us; 1644us; 1646us; 1648us; 1650us; 1652us; 1655us; 1657us; 1659us; 1664us; 1666us; 1668us; 1671us; 1673us; 1675us; 1677us; 1679us; 1682us; 1684us; 1687us; 1692us; 1695us; 1697us; 1699us; 1701us; 1703us; 1706us; 1709us; 1711us; 1713us; 1715us; 1718us; 1720us; 1723us; 1727us; 1730us; 1732us; 1735us; 1737us; 1740us; 1744us; 1748us; 1751us; 1755us; 1757us; 1760us; 1764us; 1767us; 1772us; 1777us; 1781us; 1784us; 1788us; 1792us; 1794us; 1796us; 1798us; 1801us; 1803us; 1805us; 1807us; 1810us; 1814us; 1818us; 1820us; 1822us; 1824us; 1826us; 1828us; 1830us; 1832us; 1834us; 1836us; 1838us; 1840us; 1842us; 1844us; 1846us; 1848us; 1850us; 1852us; 1854us; 1856us; 1858us; 1860us; 1863us; 1865us; 1867us; 1869us; 1871us; 1873us; 1875us; 1877us; 1880us; 1882us; 1884us; 1889us; 1892us; 1896us; 1901us; 1904us; 1907us; 1910us; 1913us; 1915us; 1917us; 1919us; 1921us; 1924us; 1926us; 1928us; 1930us; 1932us; 1934us; 1936us; 1938us; 1940us; 1942us; 1944us; 1946us; 1948us; 1953us; 1955us; 1957us; 1959us; 1961us; 1963us; 1965us; 1967us; 1969us; 1971us; 1973us; 1975us; 1977us; 1979us; 1981us; 1983us; 1985us; 1988us; 1990us; 1994us; 1998us; 2001us; 2003us; 2005us; 2007us; 2009us; 2011us; 2013us; 2015us; 2017us; 2021us; 2023us; 2025us; 2027us; 2029us; 2031us; 2034us; 2036us; 2038us; 2041us; 2043us; 2045us; 2048us; 2050us; 2052us; 2054us; 2058us; 2060us; 2062us; 2064us; 2066us; 2068us; 2070us; 2072us; 2075us; 2077us; 2079us; 2081us; 2083us; 2085us; 2087us; 2089us; 2091us; 2093us; 2095us; 2097us; 2099us; 2102us; 2104us; 2106us; 2108us; 2110us; 2112us; 2117us; 2119us; 2121us; 2123us; 2125us; 2128us; 2130us; 2133us; 2135us; 2137us; 2139us; 2142us; 2144us; 2146us; 2148us; 2150us; 2152us; 2154us; 2156us; 2159us; 2162us; 2165us; 2168us; 2170us; 2172us; 2174us; 2176us; 2178us; 2180us; 2182us; 2184us; 2186us; 2188us; 2191us; 2193us; 2196us; 2199us; 2201us; 2203us; 2205us; 2207us; 2209us; |]
let _fsyacc_action_rows = 687
let _fsyacc_actionTableElements = [|1us; 32768us; 143us; 10us; 0us; 49152us; 1us; 16386us; 134us; 5us; 1us; 32768us; 10us; 4us; 0us; 16385us; 1us; 16387us; 143us; 10us; 1us; 16386us; 134us; 5us; 0us; 16388us; 47us; 16393us; 21us; 360us; 22us; 368us; 29us; 86us; 32us; 31us; 36us; 27us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 96us; 48us; 98us; 46us; 102us; 324us; 104us; 43us; 105us; 57us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 150us; 50us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16389us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 16390us; 70us; 672us; 71us; 12us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16391us; 70us; 672us; 0us; 16392us; 21us; 16396us; 18us; 676us; 53us; 247us; 60us; 675us; 62us; 240us; 72us; 245us; 77us; 674us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 47us; 16395us; 21us; 360us; 22us; 368us; 29us; 86us; 32us; 31us; 36us; 27us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 96us; 48us; 98us; 46us; 102us; 324us; 104us; 43us; 105us; 57us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 150us; 50us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16394us; 0us; 16397us; 12us; 16400us; 18us; 676us; 29us; 86us; 32us; 31us; 36us; 27us; 60us; 675us; 77us; 674us; 96us; 48us; 98us; 46us; 104us; 43us; 105us; 57us; 150us; 50us; 156us; 83us; 0us; 16398us; 47us; 16401us; 21us; 360us; 22us; 368us; 29us; 86us; 32us; 31us; 36us; 27us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 96us; 48us; 98us; 46us; 102us; 324us; 104us; 43us; 105us; 57us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 150us; 50us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16399us; 0us; 16402us; 1us; 16402us; 121us; 192us; 0us; 16403us; 40us; 16403us; 19us; 179us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 32768us; 104us; 28us; 105us; 53us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16590us; 124us; 372us; 0us; 16404us; 4us; 32768us; 33us; 40us; 149us; 668us; 186us; 32us; 188us; 667us; 2us; 16420us; 105us; 66us; 141us; 82us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 70us; 672us; 104us; 35us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16405us; 2us; 32768us; 104us; 38us; 105us; 70us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16406us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 66us; 79us; 0us; 16407us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16590us; 124us; 372us; 0us; 16408us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16409us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16410us; 70us; 672us; 2us; 16735us; 149us; 668us; 188us; 667us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16411us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 71us; 55us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16412us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 71us; 59us; 137us; 61us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16413us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 1us; 32768us; 84us; 590us; 1us; 32768us; 138us; 63us; 1us; 32768us; 71us; 64us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16414us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 70us; 672us; 71us; 68us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16415us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 71us; 72us; 137us; 74us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16416us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 1us; 32768us; 84us; 590us; 1us; 32768us; 138us; 76us; 1us; 32768us; 71us; 77us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16417us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 27us; 16578us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 52us; 81us; 0us; 16418us; 0us; 16419us; 1us; 16784us; 99us; 684us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16421us; 1us; 16784us; 99us; 684us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 16us; 90us; 0us; 16422us; 0us; 16423us; 1us; 16426us; 124us; 93us; 0us; 16424us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 16426us; 124us; 93us; 0us; 16425us; 1us; 16642us; 71us; 486us; 1us; 32768us; 62us; 98us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16427us; 3us; 16428us; 72us; 106us; 88us; 110us; 125us; 104us; 1us; 16430us; 72us; 106us; 1us; 16433us; 72us; 106us; 1us; 16434us; 72us; 106us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16429us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 16431us; 88us; 109us; 0us; 16432us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 27us; 16444us; 31us; 113us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16435us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16436us; 0us; 16437us; 27us; 16440us; 31us; 118us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16438us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16439us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 52us; 122us; 0us; 16441us; 26us; 16454us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 51us; 125us; 0us; 16442us; 0us; 16443us; 0us; 16444us; 0us; 16445us; 0us; 16446us; 0us; 16447us; 26us; 16458us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 86us; 133us; 0us; 16448us; 1us; 16449us; 70us; 672us; 2us; 32768us; 62us; 136us; 70us; 672us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 4us; 16452us; 71us; 156us; 72us; 155us; 88us; 160us; 125us; 152us; 2us; 16450us; 18us; 679us; 77us; 680us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 62us; 141us; 70us; 672us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 4us; 16451us; 71us; 156us; 72us; 155us; 88us; 160us; 125us; 152us; 4us; 16456us; 71us; 156us; 72us; 155us; 88us; 160us; 125us; 152us; 2us; 16453us; 18us; 679us; 77us; 680us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 4us; 16455us; 71us; 156us; 72us; 155us; 88us; 160us; 125us; 152us; 4us; 16457us; 71us; 156us; 72us; 155us; 88us; 160us; 125us; 152us; 2us; 16461us; 71us; 156us; 72us; 155us; 2us; 16464us; 71us; 156us; 72us; 155us; 2us; 16465us; 71us; 156us; 72us; 155us; 5us; 16529us; 71us; 156us; 72us; 155us; 78us; 265us; 88us; 160us; 125us; 152us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16459us; 1us; 16460us; 88us; 159us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16462us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 0us; 16463us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16466us; 0us; 16467us; 0us; 16468us; 0us; 16469us; 0us; 16470us; 0us; 16471us; 0us; 16472us; 0us; 16473us; 0us; 16474us; 0us; 16475us; 0us; 16476us; 0us; 16477us; 0us; 16478us; 0us; 16479us; 0us; 16480us; 0us; 16481us; 0us; 16482us; 0us; 16483us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 14us; 181us; 0us; 16484us; 0us; 16485us; 1us; 16489us; 71us; 184us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16488us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 20us; 16492us; 18us; 679us; 53us; 247us; 62us; 240us; 72us; 245us; 77us; 680us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 40us; 16491us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16490us; 0us; 16493us; 0us; 16494us; 1us; 32768us; 121us; 192us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16495us; 42us; 32768us; 18us; 196us; 19us; 179us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16496us; 41us; 32768us; 19us; 179us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16497us; 0us; 16498us; 0us; 16499us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 32768us; 23us; 254us; 80us; 252us; 0us; 16500us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 20us; 32768us; 25us; 273us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 100us; 272us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 0us; 16501us; 1us; 16502us; 88us; 284us; 15us; 16503us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 7us; 16504us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 15us; 16505us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 13us; 16506us; 53us; 247us; 62us; 240us; 72us; 245us; 87us; 248us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 13us; 16507us; 53us; 247us; 62us; 240us; 72us; 245us; 87us; 248us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 7us; 16508us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 7us; 16509us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 7us; 16510us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 7us; 16511us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 7us; 16512us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 1us; 16513us; 185us; 250us; 7us; 16514us; 53us; 247us; 72us; 245us; 87us; 248us; 154us; 244us; 178us; 246us; 180us; 249us; 185us; 250us; 4us; 16515us; 87us; 248us; 154us; 244us; 180us; 249us; 185us; 250us; 4us; 16516us; 87us; 248us; 154us; 244us; 180us; 249us; 185us; 250us; 1us; 16517us; 185us; 250us; 1us; 16518us; 185us; 250us; 1us; 16519us; 185us; 250us; 18us; 16528us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16534us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16537us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 20us; 32768us; 25us; 273us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 100us; 272us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 17us; 16540us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 17us; 16541us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16583us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16583us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16583us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 18us; 16582us; 53us; 247us; 62us; 240us; 72us; 245us; 82us; 237us; 83us; 238us; 87us; 248us; 88us; 285us; 97us; 236us; 136us; 234us; 137us; 242us; 138us; 243us; 154us; 244us; 178us; 246us; 179us; 239us; 180us; 249us; 182us; 235us; 184us; 241us; 185us; 250us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16520us; 27us; 32768us; 50us; 258us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16521us; 27us; 32768us; 50us; 258us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 17us; 256us; 0us; 16522us; 0us; 16523us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16524us; 0us; 16525us; 1us; 32768us; 89us; 266us; 1us; 16526us; 50us; 263us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16527us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16530us; 1us; 16532us; 15us; 269us; 0us; 16531us; 3us; 16536us; 24us; 278us; 132us; 277us; 133us; 282us; 0us; 16533us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 19us; 274us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 14us; 276us; 0us; 16535us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 19us; 279us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 14us; 281us; 0us; 16538us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16539us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 32us; 32768us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16542us; 32us; 32768us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16543us; 32us; 32768us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16544us; 32us; 32768us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16545us; 32us; 32768us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16546us; 28us; 16547us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 300us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 31us; 32768us; 23us; 346us; 62us; 338us; 63us; 313us; 66us; 334us; 80us; 341us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 300us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16548us; 4us; 16549us; 30us; 569us; 31us; 307us; 70us; 310us; 137us; 571us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 3us; 16550us; 30us; 569us; 31us; 307us; 70us; 310us; 4us; 16551us; 30us; 569us; 31us; 307us; 70us; 310us; 137us; 571us; 1us; 16552us; 30us; 569us; 3us; 16554us; 30us; 569us; 31us; 307us; 70us; 310us; 5us; 16560us; 30us; 569us; 31us; 307us; 70us; 310us; 77us; 317us; 137us; 571us; 5us; 16678us; 30us; 569us; 31us; 307us; 70us; 310us; 88us; 586us; 137us; 571us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16553us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16555us; 0us; 16556us; 27us; 16559us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 51us; 315us; 0us; 16557us; 0us; 16558us; 27us; 16559us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16561us; 0us; 16562us; 0us; 16563us; 0us; 16564us; 0us; 16565us; 0us; 16566us; 0us; 16567us; 0us; 16568us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 134us; 328us; 0us; 16569us; 40us; 16572us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 3us; 32768us; 10us; 190us; 86us; 331us; 191us; 189us; 0us; 16570us; 0us; 16571us; 0us; 16573us; 27us; 16578us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 52us; 336us; 0us; 16574us; 0us; 16575us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 16782us; 18us; 679us; 77us; 680us; 0us; 16579us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 62us; 343us; 70us; 672us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 16782us; 18us; 679us; 77us; 680us; 0us; 16580us; 2us; 32768us; 149us; 668us; 188us; 667us; 2us; 32768us; 62us; 348us; 70us; 672us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 2us; 16782us; 18us; 679us; 77us; 680us; 1us; 32768us; 17us; 351us; 0us; 16581us; 2us; 16781us; 149us; 668us; 188us; 667us; 2us; 32768us; 62us; 354us; 70us; 672us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 89us; 357us; 42us; 32768us; 19us; 179us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 191us; 359us; 0us; 16584us; 0us; 16585us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 89us; 362us; 40us; 32768us; 21us; 360us; 22us; 368us; 29us; 86us; 53us; 286us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 110us; 200us; 116us; 323us; 118us; 355us; 119us; 365us; 120us; 203us; 128us; 326us; 149us; 668us; 153us; 292us; 154us; 294us; 156us; 83us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 177us; 290us; 178us; 288us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 17us; 364us; 0us; 16586us; 1us; 16786us; 50us; 685us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16587us; 1us; 16786us; 50us; 685us; 26us; 32768us; 58us; 126us; 63us; 123us; 66us; 120us; 85us; 131us; 102us; 130us; 116us; 129us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 17us; 371us; 0us; 16588us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16590us; 124us; 372us; 0us; 16589us; 1us; 16591us; 62us; 376us; 9us; 32768us; 50us; 454us; 59us; 488us; 66us; 470us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16592us; 0us; 16593us; 4us; 16596us; 30us; 686us; 70us; 672us; 72us; 383us; 137us; 16788us; 0us; 16594us; 1us; 16596us; 72us; 383us; 0us; 16595us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16597us; 1us; 16598us; 91us; 387us; 2us; 16598us; 86us; 400us; 91us; 387us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16599us; 1us; 32768us; 91us; 390us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16600us; 1us; 32768us; 91us; 393us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16601us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 397us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 397us; 85us; 395us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16682us; 72us; 399us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 396us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16602us; 1us; 32768us; 86us; 403us; 1us; 32768us; 86us; 404us; 0us; 16603us; 59us; 16603us; 10us; 16685us; 14us; 16685us; 15us; 16685us; 16us; 16685us; 17us; 16685us; 18us; 16685us; 23us; 16685us; 24us; 16685us; 25us; 16685us; 29us; 16685us; 30us; 16685us; 32us; 16685us; 36us; 16685us; 50us; 16685us; 51us; 16685us; 52us; 16685us; 53us; 16685us; 60us; 16685us; 62us; 16685us; 66us; 16685us; 71us; 16685us; 72us; 16685us; 77us; 16685us; 78us; 16685us; 80us; 16685us; 82us; 16685us; 83us; 16685us; 84us; 16685us; 86us; 16685us; 87us; 16685us; 88us; 16685us; 89us; 16685us; 96us; 16685us; 97us; 16685us; 98us; 16685us; 100us; 16685us; 104us; 16685us; 105us; 16685us; 121us; 16685us; 124us; 16685us; 125us; 16685us; 132us; 16685us; 133us; 16685us; 134us; 16685us; 136us; 16685us; 137us; 16685us; 138us; 16685us; 149us; 16685us; 150us; 16685us; 154us; 16685us; 156us; 16685us; 178us; 16685us; 179us; 16685us; 180us; 16685us; 182us; 16685us; 184us; 16685us; 185us; 16685us; 188us; 16685us; 191us; 16685us; 3us; 16604us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 7us; 16604us; 77us; 584us; 84us; 16657us; 86us; 16662us; 87us; 560us; 88us; 582us; 149us; 16657us; 188us; 16657us; 1us; 16736us; 71us; 410us; 1us; 16736us; 71us; 409us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 4us; 16605us; 66us; 514us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 5us; 16605us; 66us; 514us; 84us; 16657us; 87us; 562us; 149us; 16657us; 188us; 16657us; 0us; 16606us; 0us; 16607us; 0us; 16608us; 1us; 32768us; 137us; 418us; 0us; 16609us; 2us; 32768us; 77us; 428us; 84us; 590us; 3us; 32768us; 77us; 425us; 88us; 421us; 138us; 420us; 0us; 16610us; 1us; 32768us; 84us; 590us; 2us; 16620us; 77us; 441us; 88us; 438us; 1us; 32768us; 138us; 424us; 0us; 16611us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 138us; 427us; 0us; 16612us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 138us; 430us; 0us; 16613us; 1us; 16614us; 88us; 432us; 1us; 32768us; 84us; 590us; 1us; 16616us; 88us; 435us; 0us; 16615us; 1us; 32768us; 84us; 590us; 1us; 16616us; 88us; 435us; 0us; 16617us; 1us; 32768us; 84us; 590us; 2us; 16620us; 77us; 441us; 88us; 438us; 0us; 16618us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16619us; 1us; 16621us; 88us; 445us; 0us; 16622us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16623us; 88us; 445us; 0us; 16624us; 1us; 32768us; 71us; 449us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16625us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 3us; 16626us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 0us; 16627us; 0us; 16628us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16629us; 1us; 16631us; 50us; 457us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16630us; 4us; 16780us; 18us; 683us; 71us; 461us; 114us; 468us; 137us; 463us; 0us; 16632us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16633us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 1us; 32768us; 84us; 590us; 1us; 32768us; 138us; 465us; 1us; 32768us; 71us; 466us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16634us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16635us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 3us; 32768us; 149us; 668us; 188us; 667us; 191us; 474us; 3us; 32768us; 10us; 190us; 52us; 472us; 191us; 189us; 0us; 16636us; 0us; 16637us; 1us; 32768us; 52us; 475us; 0us; 16638us; 2us; 16782us; 18us; 679us; 77us; 680us; 2us; 16781us; 149us; 668us; 188us; 667us; 0us; 16639us; 0us; 16640us; 1us; 32768us; 71us; 482us; 1us; 32768us; 71us; 483us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 4us; 16641us; 50us; 508us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 3us; 16641us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 3us; 16643us; 84us; 16657us; 149us; 16657us; 188us; 16657us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16644us; 1us; 16650us; 89us; 495us; 2us; 16650us; 86us; 521us; 89us; 494us; 1us; 16650us; 89us; 494us; 2us; 16650us; 86us; 528us; 89us; 494us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 53us; 16645us; 10us; 16646us; 14us; 16646us; 15us; 16646us; 16us; 16646us; 17us; 16646us; 18us; 16646us; 23us; 16646us; 24us; 16646us; 25us; 16646us; 29us; 16646us; 32us; 16646us; 36us; 16646us; 50us; 16646us; 51us; 16646us; 52us; 16646us; 53us; 16646us; 60us; 16646us; 62us; 16646us; 71us; 16646us; 72us; 16646us; 77us; 16646us; 78us; 16646us; 80us; 16646us; 82us; 16646us; 83us; 16646us; 86us; 16646us; 88us; 16646us; 89us; 16646us; 96us; 16646us; 97us; 16646us; 98us; 16646us; 100us; 16646us; 104us; 16646us; 105us; 16646us; 121us; 16646us; 124us; 16646us; 125us; 16646us; 132us; 16646us; 133us; 16646us; 134us; 16646us; 136us; 16646us; 137us; 16646us; 138us; 16646us; 150us; 16646us; 154us; 16646us; 156us; 16646us; 178us; 16646us; 179us; 16646us; 180us; 16646us; 182us; 16646us; 184us; 16646us; 185us; 16646us; 191us; 16646us; 0us; 16645us; 1us; 16736us; 71us; 502us; 1us; 16736us; 71us; 503us; 1us; 16736us; 71us; 501us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 71us; 506us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 16657us; 50us; 508us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 52us; 510us; 0us; 16647us; 1us; 16657us; 66us; 514us; 2us; 16657us; 66us; 514us; 87us; 562us; 2us; 16657us; 66us; 514us; 87us; 564us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 52us; 516us; 0us; 16648us; 1us; 32768us; 85us; 518us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 71us; 520us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 32768us; 89us; 522us; 3us; 32768us; 63us; 534us; 66us; 531us; 118us; 524us; 0us; 16649us; 1us; 32768us; 85us; 525us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 71us; 527us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 32768us; 89us; 529us; 3us; 32768us; 63us; 534us; 66us; 531us; 118us; 524us; 0us; 16651us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 52us; 533us; 0us; 16652us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 16657us; 51us; 536us; 0us; 16653us; 3us; 32768us; 84us; 590us; 149us; 668us; 188us; 667us; 0us; 16654us; 1us; 32768us; 84us; 590us; 1us; 32768us; 138us; 541us; 1us; 32768us; 70us; 542us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16655us; 0us; 16656us; 0us; 16657us; 4us; 16657us; 77us; 584us; 86us; 16662us; 87us; 560us; 88us; 582us; 1us; 16657us; 86us; 16665us; 2us; 16657us; 86us; 16667us; 87us; 566us; 3us; 16657us; 77us; 576us; 88us; 573us; 138us; 572us; 3us; 16657us; 77us; 584us; 88us; 582us; 138us; 16677us; 4us; 16657us; 77us; 584us; 86us; 16677us; 88us; 582us; 138us; 16677us; 1us; 16657us; 86us; 637us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 32768us; 86us; 555us; 0us; 16658us; 1us; 32768us; 86us; 557us; 0us; 16659us; 2us; 16660us; 30us; 569us; 137us; 571us; 0us; 16661us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16663us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16664us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16666us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16668us; 0us; 16669us; 1us; 32768us; 137us; 571us; 0us; 16670us; 9us; 32768us; 59us; 488us; 66us; 504us; 77us; 579us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16671us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 1us; 32768us; 138us; 575us; 0us; 16672us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 138us; 578us; 0us; 16673us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 138us; 581us; 0us; 16674us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16675us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16676us; 27us; 32768us; 63us; 313us; 66us; 334us; 85us; 329us; 102us; 324us; 116us; 323us; 128us; 326us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 181us; 309us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16679us; 1us; 16680us; 70us; 672us; 0us; 16681us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 16682us; 72us; 592us; 11us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 396us; 87us; 413us; 118us; 517us; 137us; 539us; 149us; 668us; 178us; 414us; 187us; 415us; 188us; 667us; 0us; 16683us; 0us; 16684us; 0us; 16686us; 1us; 32768us; 85us; 634us; 0us; 16687us; 1us; 32768us; 85us; 634us; 0us; 16688us; 0us; 16689us; 2us; 16693us; 91us; 602us; 92us; 605us; 26us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16690us; 0us; 16691us; 26us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16692us; 1us; 16695us; 136us; 608us; 26us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16694us; 1us; 16697us; 83us; 611us; 26us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16696us; 28us; 16733us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16698us; 0us; 16699us; 1us; 32768us; 62us; 618us; 25us; 16716us; 62us; 618us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 24us; 32768us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16700us; 24us; 32768us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 24us; 32768us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16701us; 0us; 16702us; 0us; 16703us; 28us; 16727us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 86us; 627us; 0us; 16704us; 1us; 32768us; 70us; 629us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16705us; 1us; 32768us; 70us; 632us; 28us; 32768us; 62us; 620us; 66us; 334us; 84us; 641us; 85us; 625us; 102us; 659us; 116us; 658us; 147us; 596us; 148us; 598us; 149us; 613us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16706us; 2us; 32768us; 149us; 668us; 188us; 667us; 1us; 32768us; 71us; 636us; 8us; 32768us; 59us; 488us; 66us; 504us; 84us; 590us; 85us; 553us; 118us; 517us; 137us; 539us; 149us; 668us; 188us; 667us; 0us; 16707us; 1us; 16709us; 88us; 639us; 1us; 32768us; 85us; 634us; 0us; 16708us; 2us; 32768us; 149us; 668us; 188us; 667us; 24us; 16713us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16710us; 28us; 16713us; 62us; 16721us; 66us; 334us; 70us; 672us; 85us; 660us; 88us; 16721us; 102us; 659us; 116us; 658us; 137us; 571us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16711us; 0us; 16712us; 24us; 16713us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16714us; 24us; 16716us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16715us; 24us; 16717us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 0us; 16718us; 0us; 16719us; 0us; 16720us; 1us; 16721us; 70us; 672us; 0us; 16722us; 0us; 16723us; 0us; 16724us; 0us; 16725us; 24us; 16727us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 32768us; 86us; 662us; 0us; 16726us; 1us; 16728us; 88us; 16730us; 1us; 32768us; 88us; 665us; 24us; 32768us; 66us; 334us; 85us; 660us; 102us; 659us; 116us; 658us; 149us; 668us; 159us; 174us; 160us; 176us; 161us; 175us; 162us; 173us; 163us; 172us; 164us; 171us; 165us; 169us; 166us; 170us; 167us; 168us; 168us; 166us; 169us; 164us; 170us; 162us; 171us; 167us; 172us; 165us; 174us; 163us; 175us; 161us; 188us; 667us; 189us; 177us; 190us; 178us; 1us; 16729us; 88us; 16731us; 0us; 16732us; 0us; 16733us; 1us; 32768us; 71us; 670us; 0us; 16734us; 0us; 16736us; 2us; 32768us; 149us; 668us; 188us; 667us; 0us; 16737us; 0us; 16770us; 0us; 16771us; 0us; 16772us; 3us; 16773us; 18us; 676us; 60us; 675us; 77us; 674us; 0us; 16774us; 1us; 16775us; 77us; 681us; 1us; 16776us; 18us; 682us; 0us; 16777us; 0us; 16778us; 0us; 16779us; 0us; 16783us; 0us; 16785us; 0us; 16787us; |]
let _fsyacc_actionTableRowOffsets = [|0us; 2us; 3us; 5us; 7us; 8us; 10us; 12us; 13us; 61us; 62us; 65us; 68us; 71us; 73us; 74us; 96us; 144us; 145us; 146us; 159us; 160us; 208us; 209us; 210us; 212us; 213us; 254us; 257us; 260us; 262us; 263us; 268us; 271us; 274us; 277us; 280us; 281us; 284us; 287us; 288us; 291us; 293us; 294us; 297us; 299us; 300us; 303us; 304us; 307us; 309us; 312us; 341us; 342us; 345us; 347us; 356us; 360us; 363us; 366us; 375us; 379us; 381us; 383us; 385us; 394us; 398us; 401us; 404us; 413us; 417us; 420us; 423us; 432us; 436us; 438us; 440us; 442us; 451us; 455us; 483us; 485us; 486us; 487us; 489us; 516us; 517us; 519us; 546us; 548us; 549us; 550us; 552us; 553us; 580us; 582us; 583us; 585us; 587us; 628us; 629us; 633us; 635us; 637us; 639us; 642us; 643us; 670us; 672us; 673us; 700us; 727us; 755us; 756us; 783us; 784us; 785us; 813us; 814us; 841us; 842us; 845us; 847us; 848us; 875us; 877us; 878us; 879us; 880us; 881us; 882us; 883us; 910us; 912us; 913us; 915us; 918us; 945us; 950us; 953us; 956us; 959us; 986us; 991us; 996us; 999us; 1026us; 1031us; 1036us; 1039us; 1042us; 1045us; 1051us; 1054us; 1055us; 1057us; 1084us; 1093us; 1097us; 1098us; 1125us; 1152us; 1153us; 1154us; 1155us; 1156us; 1157us; 1158us; 1159us; 1160us; 1161us; 1162us; 1163us; 1164us; 1165us; 1166us; 1167us; 1168us; 1169us; 1170us; 1211us; 1213us; 1214us; 1215us; 1217us; 1226us; 1230us; 1251us; 1292us; 1293us; 1294us; 1295us; 1297us; 1338us; 1339us; 1382us; 1383us; 1425us; 1426us; 1427us; 1428us; 1469us; 1472us; 1473us; 1514us; 1535us; 1536us; 1538us; 1554us; 1562us; 1578us; 1592us; 1606us; 1614us; 1622us; 1630us; 1638us; 1646us; 1648us; 1656us; 1661us; 1666us; 1668us; 1670us; 1672us; 1691us; 1710us; 1729us; 1750us; 1768us; 1786us; 1805us; 1824us; 1843us; 1862us; 1903us; 1944us; 1985us; 2026us; 2067us; 2108us; 2149us; 2190us; 2231us; 2272us; 2313us; 2354us; 2395us; 2436us; 2477us; 2518us; 2559us; 2560us; 2588us; 2589us; 2617us; 2619us; 2620us; 2621us; 2648us; 2649us; 2650us; 2652us; 2654us; 2681us; 2682us; 2723us; 2764us; 2765us; 2767us; 2768us; 2772us; 2773us; 2814us; 2816us; 2857us; 2859us; 2860us; 2901us; 2903us; 2944us; 2946us; 2947us; 2988us; 2989us; 3030us; 3071us; 3104us; 3105us; 3138us; 3139us; 3172us; 3173us; 3206us; 3207us; 3240us; 3241us; 3270us; 3302us; 3303us; 3308us; 3336us; 3340us; 3345us; 3347us; 3351us; 3357us; 3363us; 3391us; 3392us; 3420us; 3423us; 3424us; 3425us; 3453us; 3455us; 3456us; 3457us; 3485us; 3486us; 3487us; 3488us; 3489us; 3490us; 3491us; 3492us; 3493us; 3534us; 3536us; 3537us; 3578us; 3582us; 3583us; 3584us; 3585us; 3613us; 3615us; 3616us; 3617us; 3658us; 3661us; 3662us; 3665us; 3668us; 3709us; 3712us; 3713us; 3716us; 3719us; 3760us; 3763us; 3765us; 3766us; 3769us; 3772us; 3813us; 3840us; 3842us; 3885us; 3886us; 3887us; 3914us; 3916us; 3957us; 3959us; 3960us; 3962us; 3989us; 3990us; 3992us; 4019us; 4021us; 4022us; 4025us; 4027us; 4028us; 4030us; 4040us; 4041us; 4042us; 4047us; 4048us; 4050us; 4051us; 4063us; 4064us; 4066us; 4069us; 4081us; 4082us; 4084us; 4096us; 4097us; 4099us; 4111us; 4112us; 4124us; 4136us; 4139us; 4141us; 4153us; 4154us; 4156us; 4158us; 4159us; 4219us; 4223us; 4231us; 4233us; 4235us; 4244us; 4253us; 4258us; 4264us; 4265us; 4266us; 4267us; 4269us; 4270us; 4273us; 4277us; 4278us; 4280us; 4283us; 4285us; 4286us; 4289us; 4291us; 4292us; 4295us; 4297us; 4298us; 4300us; 4302us; 4304us; 4305us; 4307us; 4309us; 4310us; 4312us; 4315us; 4316us; 4319us; 4320us; 4322us; 4323us; 4326us; 4328us; 4329us; 4331us; 4340us; 4344us; 4348us; 4349us; 4350us; 4353us; 4354us; 4356us; 4359us; 4360us; 4365us; 4366us; 4375us; 4379us; 4381us; 4383us; 4385us; 4394us; 4398us; 4407us; 4411us; 4415us; 4419us; 4420us; 4421us; 4423us; 4424us; 4427us; 4430us; 4431us; 4432us; 4434us; 4436us; 4445us; 4454us; 4459us; 4463us; 4472us; 4476us; 4485us; 4486us; 4488us; 4491us; 4493us; 4496us; 4505us; 4514us; 4568us; 4569us; 4571us; 4573us; 4575us; 4584us; 4593us; 4602us; 4605us; 4607us; 4616us; 4618us; 4647us; 4649us; 4650us; 4652us; 4655us; 4658us; 4687us; 4689us; 4690us; 4692us; 4695us; 4697us; 4706us; 4708us; 4712us; 4713us; 4715us; 4718us; 4720us; 4729us; 4731us; 4735us; 4736us; 4765us; 4767us; 4768us; 4777us; 4779us; 4780us; 4784us; 4785us; 4787us; 4789us; 4791us; 4800us; 4801us; 4802us; 4803us; 4808us; 4810us; 4813us; 4817us; 4821us; 4826us; 4828us; 4837us; 4839us; 4840us; 4842us; 4843us; 4846us; 4847us; 4856us; 4857us; 4866us; 4867us; 4876us; 4877us; 4886us; 4887us; 4888us; 4890us; 4891us; 4901us; 4902us; 4911us; 4913us; 4914us; 4942us; 4944us; 4945us; 4973us; 4975us; 4976us; 4985us; 4986us; 5014us; 5015us; 5043us; 5044us; 5046us; 5047us; 5050us; 5052us; 5064us; 5065us; 5066us; 5067us; 5069us; 5070us; 5072us; 5073us; 5074us; 5077us; 5104us; 5105us; 5106us; 5133us; 5134us; 5136us; 5163us; 5164us; 5166us; 5193us; 5194us; 5223us; 5224us; 5225us; 5227us; 5253us; 5278us; 5279us; 5304us; 5329us; 5330us; 5331us; 5332us; 5361us; 5363us; 5364us; 5366us; 5395us; 5396us; 5398us; 5427us; 5428us; 5431us; 5433us; 5442us; 5443us; 5445us; 5447us; 5448us; 5451us; 5476us; 5477us; 5506us; 5507us; 5508us; 5533us; 5534us; 5559us; 5560us; 5585us; 5586us; 5587us; 5588us; 5590us; 5591us; 5592us; 5593us; 5594us; 5619us; 5621us; 5622us; 5624us; 5626us; 5651us; 5653us; 5654us; 5655us; 5657us; 5658us; 5659us; 5662us; 5663us; 5664us; 5665us; 5666us; 5670us; 5671us; 5673us; 5675us; 5676us; 5677us; 5678us; 5679us; 5680us; |]
let _fsyacc_reductionSymbolCounts = [|1us; 3us; 0us; 1us; 3us; 2us; 2us; 4us; 1us; 0us; 3us; 2us; 1us; 1us; 2us; 3us; 1us; 2us; 1us; 1us; 4us; 6us; 4us; 4us; 3us; 2us; 2us; 3us; 5us; 4us; 7us; 6us; 6us; 9us; 3us; 1us; 0us; 3us; 4us; 1us; 2us; 3us; 0us; 4us; 1us; 3us; 3us; 1us; 1us; 3us; 3us; 2us; 3us; 1us; 2us; 3us; 1us; 3us; 3us; 1us; 1us; 1us; 1us; 1us; 3us; 1us; 4us; 5us; 0us; 2us; 0us; 3us; 0us; 1us; 0us; 3us; 1us; 3us; 3us; 1us; 3us; 3us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 3us; 1us; 3us; 1us; 3us; 1us; 3us; 2us; 1us; 1us; 1us; 3us; 2us; 3us; 1us; 1us; 3us; 3us; 1us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 3us; 1us; 2us; 3us; 1us; 2us; 2us; 2us; 4us; 2us; 0us; 2us; 2us; 1us; 2us; 2us; 4us; 0us; 2us; 4us; 3us; 3us; 3us; 2us; 2us; 2us; 2us; 2us; 1us; 2us; 1us; 2us; 1us; 3us; 2us; 2us; 3us; 1us; 3us; 1us; 0us; 1us; 3us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 3us; 3us; 3us; 0us; 1us; 3us; 1us; 0us; 1us; 0us; 5us; 7us; 8us; 5us; 0us; 4us; 4us; 5us; 3us; 4us; 3us; 0us; 1us; 3us; 1us; 2us; 3us; 0us; 2us; 1us; 3us; 3us; 3us; 6us; 3us; 1us; 3us; 1us; 1us; 1us; 2us; 3us; 6us; 5us; 4us; 1us; 4us; 0us; 3us; 3us; 2us; 0us; 1us; 2us; 0us; 3us; 3us; 1us; 1us; 1us; 2us; 3us; 1us; 2us; 3us; 6us; 3us; 3us; 3us; 3us; 3us; 2us; 3us; 0us; 2us; 2us; 3us; 5us; 7us; 6us; 8us; 1us; 8us; 3us; 3us; 2us; 5us; 1us; 1us; 4us; 3us; 1us; 2us; 1us; 3us; 5us; 1us; 5us; 1us; 3us; 1us; 2us; 3us; 6us; 5us; 4us; 3us; 2us; 0us; 1us; 3us; 1us; 1us; 2us; 4us; 4us; 3us; 1us; 2us; 2us; 1us; 3us; 1us; 3us; 1us; 3us; 1us; 3us; 1us; 2us; 1us; 3us; 3us; 1us; 1us; 3us; 3us; 3us; 5us; 3us; 1us; 3us; 2us; 2us; 0us; 2us; 2us; 1us; 1us; 2us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 3us; 0us; 1us; 3us; 1us; 3us; 1us; 1us; 2us; 0us; 1us; 3us; 3us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 1us; 2us; 1us; 1us; 1us; 2us; 2us; 1us; 2us; 3us; 1us; 1us; 1us; 1us; 2us; 1us; 1us; 2us; 2us; 1us; 0us; 1us; 0us; 1us; 0us; 1us; 0us; 1us; 0us; |]
let _fsyacc_productionToNonTerminalTable = [|0us; 1us; 2us; 2us; 2us; 3us; 4us; 4us; 5us; 5us; 6us; 6us; 6us; 6us; 7us; 7us; 7us; 7us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 8us; 9us; 10us; 10us; 11us; 12us; 13us; 14us; 15us; 15us; 16us; 17us; 18us; 18us; 18us; 18us; 19us; 19us; 20us; 20us; 20us; 21us; 21us; 21us; 22us; 22us; 22us; 22us; 22us; 22us; 22us; 22us; 23us; 24us; 25us; 25us; 26us; 26us; 27us; 27us; 28us; 28us; 29us; 29us; 29us; 29us; 29us; 30us; 30us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 31us; 32us; 32us; 33us; 33us; 34us; 34us; 35us; 35us; 35us; 36us; 36us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 37us; 38us; 38us; 39us; 39us; 40us; 41us; 41us; 42us; 42us; 43us; 44us; 44us; 45us; 46us; 46us; 47us; 47us; 47us; 47us; 48us; 48us; 49us; 49us; 49us; 49us; 49us; 49us; 50us; 50us; 51us; 51us; 52us; 52us; 52us; 52us; 52us; 52us; 52us; 53us; 53us; 53us; 54us; 55us; 55us; 55us; 55us; 55us; 55us; 56us; 57us; 57us; 58us; 58us; 59us; 60us; 61us; 61us; 62us; 62us; 62us; 62us; 63us; 63us; 64us; 64us; 64us; 65us; 65us; 66us; 66us; 67us; 67us; 68us; 69us; 69us; 70us; 70us; 71us; 71us; 71us; 71us; 72us; 72us; 73us; 73us; 74us; 74us; 74us; 75us; 76us; 76us; 76us; 76us; 77us; 77us; 78us; 78us; 79us; 79us; 79us; 80us; 80us; 81us; 81us; 82us; 83us; 83us; 83us; 84us; 85us; 85us; 86us; 86us; 86us; 86us; 87us; 87us; 87us; 88us; 88us; 89us; 90us; 90us; 91us; 91us; 91us; 91us; 91us; 91us; 91us; 92us; 92us; 92us; 93us; 93us; 93us; 94us; 94us; 95us; 95us; 95us; 96us; 96us; 96us; 97us; 97us; 98us; 98us; 99us; 99us; 100us; 100us; 100us; 100us; 101us; 101us; 101us; 102us; 102us; 103us; 103us; 104us; 104us; 104us; 105us; 106us; 107us; 107us; 107us; 108us; 108us; 109us; 109us; 110us; 110us; 111us; 111us; 112us; 112us; 113us; 113us; 113us; 114us; 114us; 115us; 116us; 117us; 118us; 118us; 119us; 119us; 119us; 120us; 120us; 121us; 121us; 122us; 122us; 123us; 123us; 124us; 124us; 124us; 124us; 124us; 125us; 126us; 126us; 126us; 127us; 127us; 128us; 128us; 129us; 129us; 130us; 130us; 131us; 131us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 132us; 133us; 134us; 134us; 135us; 135us; 135us; 136us; 136us; 137us; 137us; 137us; 137us; 138us; 138us; 139us; 139us; 140us; 140us; 141us; 141us; 142us; 142us; |]
let _fsyacc_immediateActions = [|65535us; 49152us; 65535us; 65535us; 16385us; 65535us; 65535us; 16388us; 65535us; 16389us; 65535us; 65535us; 65535us; 65535us; 16392us; 65535us; 65535us; 16394us; 16397us; 65535us; 16398us; 65535us; 16399us; 16402us; 65535us; 16403us; 65535us; 65535us; 65535us; 65535us; 16404us; 65535us; 65535us; 65535us; 65535us; 65535us; 16405us; 65535us; 65535us; 16406us; 65535us; 65535us; 16407us; 65535us; 65535us; 16408us; 65535us; 16409us; 65535us; 65535us; 65535us; 65535us; 16411us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16418us; 16419us; 65535us; 65535us; 16421us; 65535us; 65535us; 65535us; 16422us; 16423us; 65535us; 16424us; 65535us; 65535us; 16425us; 65535us; 65535us; 65535us; 16427us; 65535us; 65535us; 65535us; 65535us; 65535us; 16429us; 65535us; 65535us; 16432us; 65535us; 65535us; 65535us; 16435us; 65535us; 16436us; 16437us; 65535us; 16438us; 65535us; 16439us; 65535us; 65535us; 16441us; 65535us; 65535us; 16442us; 16443us; 16444us; 16445us; 16446us; 16447us; 65535us; 65535us; 16448us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16459us; 65535us; 65535us; 65535us; 65535us; 16463us; 65535us; 65535us; 16466us; 16467us; 16468us; 16469us; 16470us; 16471us; 16472us; 16473us; 16474us; 16475us; 16476us; 16477us; 16478us; 16479us; 16480us; 16481us; 16482us; 16483us; 65535us; 65535us; 16484us; 16485us; 65535us; 65535us; 65535us; 65535us; 65535us; 16490us; 16493us; 16494us; 65535us; 65535us; 16495us; 65535us; 16496us; 65535us; 16497us; 16498us; 16499us; 65535us; 65535us; 16500us; 65535us; 65535us; 16501us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16520us; 65535us; 16521us; 65535us; 65535us; 16522us; 16523us; 65535us; 16524us; 16525us; 65535us; 65535us; 65535us; 16527us; 65535us; 65535us; 16530us; 65535us; 16531us; 65535us; 16533us; 65535us; 65535us; 65535us; 65535us; 16535us; 65535us; 65535us; 65535us; 65535us; 16538us; 65535us; 16539us; 65535us; 65535us; 65535us; 16542us; 65535us; 16543us; 65535us; 16544us; 65535us; 16545us; 65535us; 16546us; 65535us; 65535us; 16548us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16553us; 65535us; 65535us; 16555us; 16556us; 65535us; 65535us; 16557us; 16558us; 65535us; 16561us; 16562us; 16563us; 16564us; 16565us; 16566us; 16567us; 16568us; 65535us; 65535us; 16569us; 65535us; 65535us; 16570us; 16571us; 16573us; 65535us; 65535us; 16574us; 16575us; 65535us; 65535us; 16579us; 65535us; 65535us; 65535us; 65535us; 16580us; 65535us; 65535us; 65535us; 65535us; 65535us; 16581us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16584us; 16585us; 65535us; 65535us; 65535us; 65535us; 16586us; 65535us; 65535us; 16587us; 65535us; 65535us; 65535us; 16588us; 65535us; 65535us; 16589us; 65535us; 65535us; 16592us; 16593us; 65535us; 16594us; 65535us; 16595us; 65535us; 16597us; 65535us; 65535us; 65535us; 16599us; 65535us; 65535us; 16600us; 65535us; 65535us; 16601us; 65535us; 65535us; 65535us; 65535us; 65535us; 16602us; 65535us; 65535us; 16603us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16606us; 16607us; 16608us; 65535us; 16609us; 65535us; 65535us; 16610us; 65535us; 65535us; 65535us; 16611us; 65535us; 65535us; 16612us; 65535us; 65535us; 16613us; 65535us; 65535us; 65535us; 16615us; 65535us; 65535us; 16617us; 65535us; 65535us; 16618us; 65535us; 16619us; 65535us; 16622us; 65535us; 65535us; 16624us; 65535us; 65535us; 65535us; 65535us; 16627us; 16628us; 65535us; 16629us; 65535us; 65535us; 16630us; 65535us; 16632us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16636us; 16637us; 65535us; 16638us; 65535us; 65535us; 16639us; 16640us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16647us; 65535us; 65535us; 65535us; 65535us; 65535us; 16648us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16649us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16651us; 65535us; 65535us; 16652us; 65535us; 65535us; 16653us; 65535us; 16654us; 65535us; 65535us; 65535us; 65535us; 65535us; 16656us; 16657us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 65535us; 16658us; 65535us; 16659us; 65535us; 16661us; 65535us; 16663us; 65535us; 16664us; 65535us; 16666us; 65535us; 16668us; 16669us; 65535us; 16670us; 65535us; 16671us; 65535us; 65535us; 16672us; 65535us; 65535us; 16673us; 65535us; 65535us; 16674us; 65535us; 16675us; 65535us; 16676us; 65535us; 16679us; 65535us; 16681us; 65535us; 65535us; 65535us; 16683us; 16684us; 16686us; 65535us; 16687us; 65535us; 16688us; 16689us; 65535us; 65535us; 16690us; 16691us; 65535us; 16692us; 65535us; 65535us; 16694us; 65535us; 65535us; 16696us; 65535us; 16698us; 16699us; 65535us; 65535us; 65535us; 16700us; 65535us; 65535us; 16701us; 16702us; 16703us; 65535us; 65535us; 16704us; 65535us; 65535us; 16705us; 65535us; 65535us; 16706us; 65535us; 65535us; 65535us; 16707us; 65535us; 65535us; 16708us; 65535us; 65535us; 16710us; 65535us; 16711us; 16712us; 65535us; 16714us; 65535us; 16715us; 65535us; 16718us; 16719us; 16720us; 65535us; 16722us; 16723us; 16724us; 16725us; 65535us; 65535us; 16726us; 65535us; 65535us; 65535us; 65535us; 16732us; 16733us; 65535us; 16734us; 16736us; 65535us; 16737us; 16770us; 16771us; 16772us; 65535us; 16774us; 65535us; 65535us; 16777us; 16778us; 16779us; 16783us; 16785us; 16787us; |]
let _fsyacc_reductions ()  =    [| 
# 1838 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ParsedImplFile)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
                      raise (Microsoft.FSharp.Text.Parsing.Accept(Microsoft.FSharp.Core.Operators.box _1))
                   )
                 : '_startimplementationFile));
# 1847 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'fileModuleImpl)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'fileModuleImpls)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.lexcont)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 397 "parser/pars.fsy"
                             
                             let firstModule = _1 [] in
                             let allModules = firstModule::_2 in
                             let fmi = ParsedImplFile(allModules) in
                               check_eof_error_USE_ONLY_IF_NOT_IN_ERROR_RECOVERY fmi _3
                           
                   )
# 397 "parser/pars.fsy"
                 : Sugar.ParsedImplFile));
# 1865 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 416 "parser/pars.fsy"
                                [] 
                   )
# 416 "parser/pars.fsy"
                 : 'fileModuleImpls));
# 1875 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 417 "parser/pars.fsy"
                                [] 
                   )
# 417 "parser/pars.fsy"
                 : 'fileModuleImpls));
# 1885 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'fileModuleImpl)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'fileModuleImpls)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 418 "parser/pars.fsy"
                                                              let fmi = _2 [] in fmi::_3 
                   )
# 418 "parser/pars.fsy"
                 : 'fileModuleImpls));
# 1897 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleIntro)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefnsOrExprPossiblyEmpty)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 422 "parser/pars.fsy"
                           
                           let m = (rhs2 parseState 1 2) in 
                             (fun path -> 
                                let path2, extendsOpt = _1 in 
                                let lid = path@path2 in 
                                  NamedTopModuleImpl(ModuleOrNamespaceImpl(lid, extendsOpt, _2, m))) 
                         
                   )
# 422 "parser/pars.fsy"
                 : 'fileModuleImpl));
# 1915 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 442 "parser/pars.fsy"
                                       _2,None 
                   )
# 442 "parser/pars.fsy"
                 : 'moduleIntro));
# 1926 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 443 "parser/pars.fsy"
                                                  _2, Some _4 
                   )
# 443 "parser/pars.fsy"
                 : 'moduleIntro));
# 1938 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefnsOrExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 448 "parser/pars.fsy"
                            _1 
                   )
# 448 "parser/pars.fsy"
                 : 'moduleDefnsOrExprPossiblyEmpty));
# 1949 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 450 "parser/pars.fsy"
                            [] 
                   )
# 450 "parser/pars.fsy"
                 : 'moduleDefnsOrExprPossiblyEmpty));
# 1959 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefnsOrExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 455 "parser/pars.fsy"
                             Def_expr (_1,rhs parseState 3) :: _3 
                   )
# 455 "parser/pars.fsy"
                 : 'moduleDefnsOrExpr));
# 1972 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 458 "parser/pars.fsy"
                             [Def_expr(_1,rhs parseState 3)] 
                   )
# 458 "parser/pars.fsy"
                 : 'moduleDefnsOrExpr));
# 1984 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 461 "parser/pars.fsy"
                             [ Def_expr(_1,rhs parseState 3) ] 
                   )
# 461 "parser/pars.fsy"
                 : 'moduleDefnsOrExpr));
# 1995 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 464 "parser/pars.fsy"
                             _1 
                   )
# 464 "parser/pars.fsy"
                 : 'moduleDefnsOrExpr));
# 2006 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefn)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 473 "parser/pars.fsy"
                              _1 @ _2 
                   )
# 473 "parser/pars.fsy"
                 : 'moduleDefns));
# 2018 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefn)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefnsOrExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 475 "parser/pars.fsy"
                              _1 @ _3 
                   )
# 475 "parser/pars.fsy"
                 : 'moduleDefns));
# 2031 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefn)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 477 "parser/pars.fsy"
                             _1 
                   )
# 477 "parser/pars.fsy"
                 : 'moduleDefns));
# 2042 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moduleDefn)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 479 "parser/pars.fsy"
                             _1 
                   )
# 479 "parser/pars.fsy"
                 : 'moduleDefns));
# 2054 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'defnBindings)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 489 "parser/pars.fsy"
                             
                             new_arg_uniq_ref := 0;
                             mkDefnBindings (rhs parseState 1) _1  
                           
                   )
# 489 "parser/pars.fsy"
                 : 'moduleDefn));
# 2068 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'hardwhiteLetBindings)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 495 "parser/pars.fsy"
                             
                             new_arg_uniq_ref := 0;
                             mkDefnBindings (rhs parseState 1) _1  
                           
                   )
# 495 "parser/pars.fsy"
                 : 'moduleDefn));
# 2082 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefnList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 501 "parser/pars.fsy"
                             
                             [ Def_tycon(_3 :: _4, false, true, rhs2 parseState 1 4) ] 
                           
                   )
# 501 "parser/pars.fsy"
                 : 'moduleDefn));
# 2096 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'maybenullary)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 506 "parser/pars.fsy"
                            
                             let lang = string_to_lang _2 (rhs parseState 2) in
                             let nullary = _3 in
                             let dll, className = match _4 with 
                               | hd::tl -> hd.idText, tl 
                               | _ -> failExtern (rhs parseState 3) in
                             let tdef = _6 in
                               [Def_extern_typ(ExternRef (mk_extern_ref lang dll [] className None nullary), tdef, rhs2 parseState 1 6)]
                           
                   )
# 506 "parser/pars.fsy"
                 : 'moduleDefn));
# 2118 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 517 "parser/pars.fsy"
                            
                             let tdef = _4 in
                               [Def_extern_typ(ExternRefId(_2), tdef, rhs2 parseState 1 4)]
                           
                   )
# 517 "parser/pars.fsy"
                 : 'moduleDefn));
# 2133 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'extern_ref)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 523 "parser/pars.fsy"
                            
                             [Def_extern_ref (_3, _4, rhs2 parseState 1 4)]
                           
                   )
# 523 "parser/pars.fsy"
                 : 'moduleDefn));
# 2147 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefnList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 528 "parser/pars.fsy"
                             
                             [ Def_tycon(_2 :: _3, false, false, rhs2 parseState 1 3) ] 
                           
                   )
# 528 "parser/pars.fsy"
                 : 'moduleDefn));
# 2161 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 533 "parser/pars.fsy"
                             
                             [ Def_tycon([_2], true, false, rhs2 parseState 1 3) ] 
                           
                   )
# 533 "parser/pars.fsy"
                 : 'moduleDefn));
# 2174 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 538 "parser/pars.fsy"
                             
                             [Def_open(_2,rhs parseState 1)] 
                           
                   )
# 538 "parser/pars.fsy"
                 : 'moduleDefn));
# 2187 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_ident_colon)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'formula)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 543 "parser/pars.fsy"
                            
                             [Def_assume(_2, _3, false, rhs2 parseState 1 3)]
                           
                   )
# 543 "parser/pars.fsy"
                 : 'moduleDefn));
# 2201 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 548 "parser/pars.fsy"
                             
                             let name = _3 in
                             let typ = _5 in
                               [Def_val (name, typ, rhs2 parseState 1 4)]
                           
                   )
# 548 "parser/pars.fsy"
                 : 'moduleDefn));
# 2217 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 555 "parser/pars.fsy"
                             
                             let name = _2 in
                             let typ = _4 in
                               [Def_val (name, typ, rhs2 parseState 1 4)]
                           
                   )
# 555 "parser/pars.fsy"
                 : 'moduleDefn));
# 2233 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTypars)) in
            let _7 = (let data = parseState.GetInput(7) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 562 "parser/pars.fsy"
                            
                             let loc1 = rhs2 parseState 1 4 in
                             let loc2 = rhs parseState 7 in
                             let loc = rhs2 parseState 1 7 in
                             let name = _2 in
                             let tpars = _4 in
                             let typ = _7 in
                               [Def_val(name, Type_qtyp(tpars, typ, loc2), loc)]
                           
                   )
# 562 "parser/pars.fsy"
                 : 'moduleDefn));
# 2254 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 573 "parser/pars.fsy"
                            
                             let lang = string_to_lang _2 (rhs parseState 2) in
                             let dll, className, methodName = 
                               match _4 with 
                                 | dll::rest -> 
                                     let className, methodName = Util.pfx rest in
                                       if className = [] then failExtern (rhs parseState 4)
                                       else 
                                         dll.idText, className, methodName
                                 | _ -> failExtern (rhs parseState 4) in
                             let t = _6 in
                               [Def_extern_val(ExternRef (mk_extern_ref lang dll [] className None None), methodName, t, rhs2 parseState 1 6)]
                           
                   )
# 573 "parser/pars.fsy"
                 : 'moduleDefn));
# 2279 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 588 "parser/pars.fsy"
                            
                             [Def_extern_val(ExternRefId _2, _4, _6, rhs2 parseState 1 6)]
                           
                   )
# 588 "parser/pars.fsy"
                 : 'moduleDefn));
# 2294 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTypars)) in
            let _9 = (let data = parseState.GetInput(9) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 595 "parser/pars.fsy"
                            
                             let loc2 = rhs parseState 9 in
                             let eref = _2 in
                             let name = _4 in
                             let tpars = _6 in
                             let typ = _9 in
                               [Def_extern_val(ExternRefId eref, 
                                               name, 
                                               Type_qtyp(tpars, typ, loc2), 
                                               rhs2 parseState 1 9)]
                           
                   )
# 595 "parser/pars.fsy"
                 : 'moduleDefn));
# 2318 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'braceExprBody)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 610 "parser/pars.fsy"
                            
                             let pos = rhs parseState 2 in
                               match _2 pos with 
                                 | Expr_recd(None, fields, _) -> 
                                     let find_opt id = 
                                       match List.tryFind (fun ((_, (id':Sugar.ident)), _) -> id'.idText = id) fields with
                                         | Some(_, Expr_const(Sugar.Const_string(f, _), _)) -> 
                                             Some (Util.unicodeEncoding.GetString(f)) 
                                         | _ -> None in 
                                     let find id = match find_opt id with 
                                       | Some s -> s
                                       | _ -> failExtern pos in
                                     let lang = find "language" in
                                     let dll = find "dll" in
                                     let ns = (lid_of_path -<- path_of_text) (find "namespace") pos in
                                     let cn = (lid_of_path -<- path_of_text) (find "classname") pos in
                                     let icn = find_opt "innerclass" in 
                                     let eq = match (find_opt "externqual") with 
                                         None -> None
                                       | Some "nullary" -> Some Nullary in
                                     let lang = match lang with 
                                       | "F#" -> FSharp
                                       | "C#" -> CSharp
                                       | "Fine" -> Fine
                                       | "F7" -> F7
                                       | "F*" -> FStar
                                       | _ -> failExtern pos in
                                       (mk_extern_ref lang dll ns cn icn eq) 
                                 | _ -> failExtern pos 
                           
                   )
# 610 "parser/pars.fsy"
                 : 'extern_ref));
# 2358 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 642 "parser/pars.fsy"
                                Some Nullary 
                   )
# 642 "parser/pars.fsy"
                 : 'maybenullary));
# 2368 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 643 "parser/pars.fsy"
                                None 
                   )
# 643 "parser/pars.fsy"
                 : 'maybenullary));
# 2378 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : bool)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_rec)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'localBindings)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 650 "parser/pars.fsy"
                             
                             let letm = rhs parseState 1 in 
                               let isUse,isRec,bindings = _1,_2,_3 in 
                               BindingSet(rhs parseState 1, isRec, isUse, bindings letm)
                               
                   )
# 650 "parser/pars.fsy"
                 : 'defnBindings));
# 2395 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : bool)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_rec)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'localBindings)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'hardwhiteDefnBindingsTerminator)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 658 "parser/pars.fsy"
                             _4 (rhs parseState 1);  (* report unterminated error *)
                             let letm = rhs parseState 1 in 
                             let isUse,isRec,bindings = _1,_2,_3 in 
                             BindingSet(letm,isRec,isUse, bindings letm) 
                   )
# 658 "parser/pars.fsy"
                 : 'hardwhiteLetBindings));
# 2412 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 666 "parser/pars.fsy"
                            (fun m -> ()) 
                   )
# 666 "parser/pars.fsy"
                 : 'hardwhiteDefnBindingsTerminator));
# 2422 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'localBinding)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'more_localBindings)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 674 "parser/pars.fsy"
                             (fun letm -> 
                                match _1 with 
                                | Some f -> (f letm ::  _2) 
                                | None -> _2) 
                   )
# 674 "parser/pars.fsy"
                 : 'localBindings));
# 2437 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'localBinding)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'more_localBindings)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 681 "parser/pars.fsy"
                             let letm = (rhs parseState 1) in
                             (match _2 with 
                              | Some f -> f letm :: _3 
                              | None -> _3) 
                   )
# 681 "parser/pars.fsy"
                 : 'more_localBindings));
# 2452 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 687 "parser/pars.fsy"
                             [] 
                   )
# 687 "parser/pars.fsy"
                 : 'more_localBindings));
# 2462 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'bindingPattern)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_topReturnType)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 691 "parser/pars.fsy"
                               let bindm = rhs parseState 1 in 
                               let expr = _4 in
                               let eqm = rhs parseState 3 in 
                               let rhsm = range_of_synexpr expr in (* From sugar *)
                               let optReturnType = _2 in 
                               let bindingBuilder = _1 in 
                                 Some (fun letm -> 
                                    let wholem = union_ranges letm rhsm in (* From range *)
                                      bindingBuilder (bindm,wholem,optReturnType,expr,rhsm))
                             
                   )
# 691 "parser/pars.fsy"
                 : 'localBinding));
# 2484 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 715 "parser/pars.fsy"
                              mksyn_binding (_1) (* from Sugar *) 
                   )
# 715 "parser/pars.fsy"
                 : 'bindingPattern));
# 2495 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 719 "parser/pars.fsy"
                             StartName parseState (rhs parseState 3); 
                             Pat_as (_1,_3,false,rhs2 parseState 1 3) 
                   )
# 719 "parser/pars.fsy"
                 : 'headBindingPattern));
# 2508 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 726 "parser/pars.fsy"
                             Pat_lid (mksyn_constr (rhs2 parseState 1 3) Const.cons_str, [_1;_3], lhs(parseState)) 
                   )
# 726 "parser/pars.fsy"
                 : 'headBindingPattern));
# 2520 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tuplePatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 729 "parser/pars.fsy"
                             Pat_tuple(List.rev _1, lhs(parseState)) 
                   )
# 729 "parser/pars.fsy"
                 : 'headBindingPattern));
# 2531 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'constrPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 732 "parser/pars.fsy"
                             _1 
                   )
# 732 "parser/pars.fsy"
                 : 'headBindingPattern));
# 2542 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tuplePatternElements)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 735 "parser/pars.fsy"
                                                                         _3 :: _1 
                   )
# 735 "parser/pars.fsy"
                 : 'tuplePatternElements));
# 2554 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'headBindingPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 736 "parser/pars.fsy"
                                                                       _3 :: _1 :: [] 
                   )
# 736 "parser/pars.fsy"
                 : 'tuplePatternElements));
# 2566 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatternLongIdent)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 740 "parser/pars.fsy"
                             Pat_lid (_1,_2,lhs(parseState)) 
                   )
# 740 "parser/pars.fsy"
                 : 'constrPattern));
# 2578 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatternLongIdent)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 743 "parser/pars.fsy"
                             let lid = _1 in Pat_lid (lid,_3,lhs(parseState)) 
                   )
# 743 "parser/pars.fsy"
                 : 'constrPattern));
# 2590 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 745 "parser/pars.fsy"
                                         _1 
                   )
# 745 "parser/pars.fsy"
                 : 'constrPattern));
# 2601 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPattern)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 749 "parser/pars.fsy"
                             _1 :: _2 
                   )
# 749 "parser/pars.fsy"
                 : 'atomicPatterns));
# 2613 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 751 "parser/pars.fsy"
                             report_parse_error_at (rhs parseState 1) "Successive patterns should be separated by spaces or tupled";
                             _1 :: _3 
                   )
# 751 "parser/pars.fsy"
                 : 'atomicPatterns));
# 2626 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 754 "parser/pars.fsy"
                             [_1] 
                   )
# 754 "parser/pars.fsy"
                 : 'atomicPatterns));
# 2637 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'recordPatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 758 "parser/pars.fsy"
                             _2 
                   )
# 758 "parser/pars.fsy"
                 : 'atomicPattern));
# 2648 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'listPatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 761 "parser/pars.fsy"
                               MatchPair parseState 1 3; Pat_array_or_list(false,_2,lhs(parseState)) 
                   )
# 761 "parser/pars.fsy"
                 : 'atomicPattern));
# 2659 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 763 "parser/pars.fsy"
                                      Pat_wild (lhs(parseState)) 
                   )
# 763 "parser/pars.fsy"
                 : 'atomicPattern));
# 2669 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatternLongIdent)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 766 "parser/pars.fsy"
                             let lid = _1 in 
                             if List.length lid > 1 || (let c = (List.hd lid).idText.[0] in Char.IsUpper(c) && not (Char.IsLower c)) 
                             then mksyn_pat_maybe_var lid (lhs(parseState)) (* from Sugar *)
                             else mksyn_pat_var (List.hd lid) 
                   )
# 766 "parser/pars.fsy"
                 : 'atomicPattern));
# 2683 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'constant)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 770 "parser/pars.fsy"
                                    Pat_const (_1,range_of_synconst _1 (lhs(parseState))) 
                   )
# 770 "parser/pars.fsy"
                 : 'atomicPattern));
# 2694 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 771 "parser/pars.fsy"
                                  Pat_const(Const_bool false,lhs(parseState)) 
                   )
# 771 "parser/pars.fsy"
                 : 'atomicPattern));
# 2704 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 772 "parser/pars.fsy"
                                 Pat_const(Const_bool true,lhs(parseState)) 
                   )
# 772 "parser/pars.fsy"
                 : 'atomicPattern));
# 2714 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPatternBody)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 773 "parser/pars.fsy"
                                                           MatchPair parseState 1 3; let m = (lhs(parseState)) in Pat_paren(_2 m,m) 
                   )
# 773 "parser/pars.fsy"
                 : 'atomicPattern));
# 2725 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 781 "parser/pars.fsy"
                                _1 
                   )
# 781 "parser/pars.fsy"
                 : 'atomicPatternLongIdent));
# 2736 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreRecordPatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 785 "parser/pars.fsy"
                             Pat_recd ((FrontAndBack _1,_3) :: List.rev _4,lhs(parseState)) 
                   )
# 785 "parser/pars.fsy"
                 : 'recordPatternElements));
# 2749 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreRecordPatternElements)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 789 "parser/pars.fsy"
                            (FrontAndBack _3,_5)::_1 
                   )
# 789 "parser/pars.fsy"
                 : 'moreRecordPatternElements));
# 2763 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 791 "parser/pars.fsy"
                            [] 
                   )
# 791 "parser/pars.fsy"
                 : 'moreRecordPatternElements));
# 2773 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreListPatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 795 "parser/pars.fsy"
                           _1 :: List.rev _2 
                   )
# 795 "parser/pars.fsy"
                 : 'listPatternElements));
# 2785 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 797 "parser/pars.fsy"
                           [] 
                   )
# 797 "parser/pars.fsy"
                 : 'listPatternElements));
# 2795 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreListPatternElements)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 801 "parser/pars.fsy"
                            _3 :: _1 
                   )
# 801 "parser/pars.fsy"
                 : 'moreListPatternElements));
# 2808 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 803 "parser/pars.fsy"
                            [] 
                   )
# 803 "parser/pars.fsy"
                 : 'moreListPatternElements));
# 2818 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 807 "parser/pars.fsy"
                             (fun m -> _1) 
                   )
# 807 "parser/pars.fsy"
                 : 'parenPatternBody));
# 2829 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 809 "parser/pars.fsy"
                             (fun m -> Pat_const(Const_unit,m)) 
                   )
# 809 "parser/pars.fsy"
                 : 'parenPatternBody));
# 2839 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 838 "parser/pars.fsy"
                             StartName parseState (rhs parseState 3); Pat_as (_1,_3,false,rhs2 parseState 1 3) 
                   )
# 838 "parser/pars.fsy"
                 : 'parenPattern));
# 2851 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleParenPatternElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 842 "parser/pars.fsy"
                             Pat_tuple(List.rev _1,lhs(parseState)) 
                   )
# 842 "parser/pars.fsy"
                 : 'parenPattern));
# 2862 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 844 "parser/pars.fsy"
                             Pat_lid (mksyn_constr (rhs2 parseState 1 3) Const.cons_str, [_1;_3], lhs(parseState)) 
                   )
# 844 "parser/pars.fsy"
                 : 'parenPattern));
# 2874 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 847 "parser/pars.fsy"
                             let lhsm = lhs(parseState) in 
                             Pat_typed(_1,_3,lhsm) 
                   )
# 847 "parser/pars.fsy"
                 : 'parenPattern));
# 2887 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'constrPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 849 "parser/pars.fsy"
                                         _1 
                   )
# 849 "parser/pars.fsy"
                 : 'parenPattern));
# 2898 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleParenPatternElements)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 853 "parser/pars.fsy"
                                                                         _3 :: _1 
                   )
# 853 "parser/pars.fsy"
                 : 'tupleParenPatternElements));
# 2910 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 854 "parser/pars.fsy"
                                                            _3 :: _1 :: [] 
                   )
# 854 "parser/pars.fsy"
                 : 'tupleParenPatternElements));
# 2922 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : sbyte * bool)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 863 "parser/pars.fsy"
                             
                             if snd _1 then
                               errorR(Error("This number is outside the allowable range for 8-bit signed integers", lhs(parseState)));
                             Const_int8 (fst _1) 
                           
                   )
# 863 "parser/pars.fsy"
                 : 'constant));
# 2937 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : byte)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 868 "parser/pars.fsy"
                                 Const_uint8 _1 
                   )
# 868 "parser/pars.fsy"
                 : 'constant));
# 2948 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : int16 * bool)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 870 "parser/pars.fsy"
                             
                             if snd _1 then 
                               errorR(Error("This number is outside the allowable range for 16-bit signed integers", lhs(parseState)));
                             Const_int16 (fst _1) 
                           
                   )
# 870 "parser/pars.fsy"
                 : 'constant));
# 2963 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : uint16)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 875 "parser/pars.fsy"
                                  Const_uint16 _1 
                   )
# 875 "parser/pars.fsy"
                 : 'constant));
# 2974 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : int32 * bool)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 877 "parser/pars.fsy"
                             
                             if snd _1 then 
                               errorR(Error("This number is outside the allowable range for 32-bit signed integers", lhs(parseState)));
                             Const_int32 (fst _1) 
                           
                   )
# 877 "parser/pars.fsy"
                 : 'constant));
# 2989 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : uint32)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 882 "parser/pars.fsy"
                                  Const_uint32 _1 
                   )
# 882 "parser/pars.fsy"
                 : 'constant));
# 3000 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : int64 * bool)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 884 "parser/pars.fsy"
                             
                             if snd _1 then 
                               errorR(Error("This number is outside the allowable range for 64-bit signed integers", lhs(parseState)));
                             Const_int64 (fst _1) 
                           
                   )
# 884 "parser/pars.fsy"
                 : 'constant));
# 3015 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : uint64)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 889 "parser/pars.fsy"
                                  Const_uint64 _1 
                   )
# 889 "parser/pars.fsy"
                 : 'constant));
# 3026 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : int64)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 890 "parser/pars.fsy"
                                     Const_nativeint _1 
                   )
# 890 "parser/pars.fsy"
                 : 'constant));
# 3037 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : uint64)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 891 "parser/pars.fsy"
                                      Const_unativeint _1 
                   )
# 891 "parser/pars.fsy"
                 : 'constant));
# 3048 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : single)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 892 "parser/pars.fsy"
                                  Const_float32 _1 
                   )
# 892 "parser/pars.fsy"
                 : 'constant));
# 3059 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : double)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 893 "parser/pars.fsy"
                                  Const_float _1 
                   )
# 893 "parser/pars.fsy"
                 : 'constant));
# 3070 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : char)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 894 "parser/pars.fsy"
                                Const_char _1 
                   )
# 894 "parser/pars.fsy"
                 : 'constant));
# 3081 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : byte[])) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 895 "parser/pars.fsy"
                                  Const_bigint _1 
                   )
# 895 "parser/pars.fsy"
                 : 'constant));
# 3092 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : System.Decimal)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 896 "parser/pars.fsy"
                                   Const_decimal _1 
                   )
# 896 "parser/pars.fsy"
                 : 'constant));
# 3103 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : byte[])) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 897 "parser/pars.fsy"
                                  Const_bignum _1 
                   )
# 897 "parser/pars.fsy"
                 : 'constant));
# 3114 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : byte[])) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 898 "parser/pars.fsy"
                                  Const_string (_1,lhs(parseState)) 
                   )
# 898 "parser/pars.fsy"
                 : 'constant));
# 3125 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : byte[])) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 899 "parser/pars.fsy"
                                     Const_bytearray (_1,lhs(parseState)) 
                   )
# 899 "parser/pars.fsy"
                 : 'constant));
# 3136 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 903 "parser/pars.fsy"
                                                              _2 
                   )
# 903 "parser/pars.fsy"
                 : 'typedSeqExprBlock));
# 3147 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 904 "parser/pars.fsy"
                                        _1 
                   )
# 904 "parser/pars.fsy"
                 : 'typedSeqExprBlock));
# 3158 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 908 "parser/pars.fsy"
                                                              _2 
                   )
# 908 "parser/pars.fsy"
                 : 'declExprBlock));
# 3169 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 909 "parser/pars.fsy"
                                    _1 
                   )
# 909 "parser/pars.fsy"
                 : 'declExprBlock));
# 3180 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'seqExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 912 "parser/pars.fsy"
                                                   Expr_typed (_1,_3, union_ranges (range_of_synexpr _1) (range_of_syntype _3)) 
                   )
# 912 "parser/pars.fsy"
                 : 'typedSeqExpr));
# 3192 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'seqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 913 "parser/pars.fsy"
                                   _1 
                   )
# 913 "parser/pars.fsy"
                 : 'typedSeqExpr));
# 3203 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'seqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 916 "parser/pars.fsy"
                                                                 Expr_seq(true,_1,_3,union_ranges (range_of_synexpr _1) (range_of_synexpr _3)) 
                   )
# 916 "parser/pars.fsy"
                 : 'seqExpr));
# 3216 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 917 "parser/pars.fsy"
                                                                 _1 
                   )
# 917 "parser/pars.fsy"
                 : 'seqExpr));
# 3228 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 918 "parser/pars.fsy"
                                                                 _1 
                   )
# 918 "parser/pars.fsy"
                 : 'seqExpr));
# 3239 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 924 "parser/pars.fsy"
                                  
                   )
# 924 "parser/pars.fsy"
                 : 'recover));
# 3249 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.lexcont)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 925 "parser/pars.fsy"
                                
                   )
# 925 "parser/pars.fsy"
                 : 'recover));
# 3260 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'defnBindings)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 929 "parser/pars.fsy"
                            mkLocalBindings (rhs2 parseState 1 3) _1 _3 
                   )
# 929 "parser/pars.fsy"
                 : 'declExpr));
# 3272 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'hardwhiteLetBindings)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExprBlock)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 938 "parser/pars.fsy"
                            mkLocalBindings (rhs2 parseState 1 2) _1 _2 
                   )
# 938 "parser/pars.fsy"
                 : 'declExpr));
# 3284 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'hardwhiteLetBindings)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExprBlock)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 945 "parser/pars.fsy"
                            mkLocalBindings (rhs2 parseState 1 3) _1 _3 
                   )
# 945 "parser/pars.fsy"
                 : 'declExpr));
# 3296 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'anonMatchingExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 952 "parser/pars.fsy"
                             _1 
                   )
# 952 "parser/pars.fsy"
                 : 'declExpr));
# 3307 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'anonLambdaExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 954 "parser/pars.fsy"
                                                          _1 
                   )
# 954 "parser/pars.fsy"
                 : 'declExpr));
# 3318 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'withClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 957 "parser/pars.fsy"
                             let mMatch = (rhs parseState 1) in
                             let mWith,(clauses,mLast) = _3 in 
                             Expr_match(_2,clauses,false,union_ranges mMatch mLast) 
                   )
# 957 "parser/pars.fsy"
                 : 'declExpr));
# 3332 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'ifExprCases)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 962 "parser/pars.fsy"
                             let mIf = (rhs parseState 1) in
                             _3 _2 mIf 
                   )
# 962 "parser/pars.fsy"
                 : 'declExpr));
# 3345 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 965 "parser/pars.fsy"
                                                        Expr_tuple( List.rev _1,lhs(parseState)) 
                   )
# 965 "parser/pars.fsy"
                 : 'declExpr));
# 3356 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 967 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "||" _3 
                   )
# 967 "parser/pars.fsy"
                 : 'declExpr));
# 3368 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 968 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 968 "parser/pars.fsy"
                 : 'declExpr));
# 3381 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 969 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "or" _3 
                   )
# 969 "parser/pars.fsy"
                 : 'declExpr));
# 3393 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 970 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "&" _3 
                   )
# 970 "parser/pars.fsy"
                 : 'declExpr));
# 3405 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 971 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "&&" _3 
                   )
# 971 "parser/pars.fsy"
                 : 'declExpr));
# 3417 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 972 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 972 "parser/pars.fsy"
                 : 'declExpr));
# 3430 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 973 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (union_ranges (range_of_synexpr _1) (range_of_synexpr _3)) _1 "=" _3 
                   )
# 973 "parser/pars.fsy"
                 : 'declExpr));
# 3442 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 974 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 974 "parser/pars.fsy"
                 : 'declExpr));
# 3455 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 975 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "<" _3 
                   )
# 975 "parser/pars.fsy"
                 : 'declExpr));
# 3467 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 976 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 ">" _3 
                   )
# 976 "parser/pars.fsy"
                 : 'declExpr));
# 3479 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 977 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 977 "parser/pars.fsy"
                 : 'declExpr));
# 3492 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 978 "parser/pars.fsy"
                                                                    Expr_app(Expr_app (mksyn_item (lhs(parseState)) Const.cons_str, _1, lhs parseState), _3, lhs(parseState)) 
                   )
# 978 "parser/pars.fsy"
                 : 'declExpr));
# 3504 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 979 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2)  (lhs(parseState)) _1 _2 _3 
                   )
# 979 "parser/pars.fsy"
                 : 'declExpr));
# 3517 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 980 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "-" _3 
                   )
# 980 "parser/pars.fsy"
                 : 'declExpr));
# 3529 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 981 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 "*" _3 
                   )
# 981 "parser/pars.fsy"
                 : 'declExpr));
# 3541 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 982 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 982 "parser/pars.fsy"
                 : 'declExpr));
# 3554 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 983 "parser/pars.fsy"
                                                                    mksyn_infix (rhs parseState 2) (lhs(parseState)) _1 _2 _3 
                   )
# 983 "parser/pars.fsy"
                 : 'declExpr));
# 3567 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 984 "parser/pars.fsy"
                                                                  _1 
                   )
# 984 "parser/pars.fsy"
                 : 'declExpr));
# 3578 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'withPatternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 987 "parser/pars.fsy"
                                                         rhs parseState 1, _2 
                   )
# 987 "parser/pars.fsy"
                 : 'withClauses));
# 3589 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'withPatternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 988 "parser/pars.fsy"
                                                         rhs parseState 1, _2 
                   )
# 988 "parser/pars.fsy"
                 : 'withClauses));
# 3600 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 995 "parser/pars.fsy"
                             _1 
                   )
# 995 "parser/pars.fsy"
                 : 'withPatternClauses));
# 3611 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 997 "parser/pars.fsy"
                              _2 
                   )
# 997 "parser/pars.fsy"
                 : 'withPatternClauses));
# 3622 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenPattern)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternGuard)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1011 "parser/pars.fsy"
                             _1, _2, rhs parseState 1 
                   )
# 1011 "parser/pars.fsy"
                 : 'patternAndGuard));
# 3634 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternAndGuard)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternResult)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1015 "parser/pars.fsy"
                            let pat,guard,patm = _1 in 
                            let mLast = range_of_synexpr _2 in 
                            [Clause(pat,guard,_2,patm)], mLast  
                   )
# 1015 "parser/pars.fsy"
                 : 'patternClauses));
# 3648 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternAndGuard)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternResult)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1019 "parser/pars.fsy"
                            let pat,guard,patm = _1 in 
                            let clauses,mLast = _4 in 
                            (Clause(pat,guard,_2,patm) :: clauses), mLast 
                   )
# 1019 "parser/pars.fsy"
                 : 'patternClauses));
# 3663 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1037 "parser/pars.fsy"
                            Some _2 
                   )
# 1037 "parser/pars.fsy"
                 : 'patternGuard));
# 3674 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1039 "parser/pars.fsy"
                            None 
                   )
# 1039 "parser/pars.fsy"
                 : 'patternGuard));
# 3684 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExprBlockR)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1043 "parser/pars.fsy"
                            _2 
                   )
# 1043 "parser/pars.fsy"
                 : 'patternResult));
# 3695 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1048 "parser/pars.fsy"
                                                         _1 
                   )
# 1048 "parser/pars.fsy"
                 : 'typedSeqExprBlockR));
# 3706 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1049 "parser/pars.fsy"
                                        _1 
                   )
# 1049 "parser/pars.fsy"
                 : 'typedSeqExprBlockR));
# 3717 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'ifExprThen)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'ifExprElifs)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1054 "parser/pars.fsy"
                             let exprThen,mThen = _1 in 
                             (fun exprGuard mIf -> 
                                 let mIfToThen = union_ranges mIf mThen in
                                 let mIfToEndOfElseBranch = union_ranges mIf (range_of_synexpr (match _2 with None -> exprThen | Some e -> e)) in
                                 Expr_cond(exprGuard,exprThen,_2,mIfToThen,mIfToEndOfElseBranch)) 
                   )
# 1054 "parser/pars.fsy"
                 : 'ifExprCases));
# 3733 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1061 "parser/pars.fsy"
                                                             _2, rhs parseState 1 
                   )
# 1061 "parser/pars.fsy"
                 : 'ifExprThen));
# 3744 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1062 "parser/pars.fsy"
                                                                                        _3,rhs parseState 1 
                   )
# 1062 "parser/pars.fsy"
                 : 'ifExprThen));
# 3755 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1066 "parser/pars.fsy"
                             None 
                   )
# 1066 "parser/pars.fsy"
                 : 'ifExprElifs));
# 3765 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1068 "parser/pars.fsy"
                             Some _2 
                   )
# 1068 "parser/pars.fsy"
                 : 'ifExprElifs));
# 3776 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1070 "parser/pars.fsy"
                             Some _3 
                   )
# 1070 "parser/pars.fsy"
                 : 'ifExprElifs));
# 3787 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'ifExprCases)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1072 "parser/pars.fsy"
                             let mElif = rhs parseState 1 in 
                             Some (_3 _2 mElif) 
                   )
# 1072 "parser/pars.fsy"
                 : 'ifExprElifs));
# 3800 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1077 "parser/pars.fsy"
                             NextParameter parseState (rhspos parseState 2); _3 :: _1 
                   )
# 1077 "parser/pars.fsy"
                 : 'tupleExpr));
# 3812 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1079 "parser/pars.fsy"
                             NextParameter parseState (rhspos parseState 2); _3 :: _1 :: [] 
                   )
# 1079 "parser/pars.fsy"
                 : 'tupleExpr));
# 3824 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1083 "parser/pars.fsy"
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr _2)) "~-" _2 
                   )
# 1083 "parser/pars.fsy"
                 : 'minusExpr));
# 3835 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1085 "parser/pars.fsy"
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr _2)) ("~"^(_1)) _2 
                   )
# 1085 "parser/pars.fsy"
                 : 'minusExpr));
# 3847 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1087 "parser/pars.fsy"
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr _2)) ("~"^(_1)) _2 
                   )
# 1087 "parser/pars.fsy"
                 : 'minusExpr));
# 3859 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1089 "parser/pars.fsy"
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr _2)) _1 _2 
                   )
# 1089 "parser/pars.fsy"
                 : 'minusExpr));
# 3871 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'minusExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1091 "parser/pars.fsy"
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr _2)) ("~"^(_1)) _2 
                   )
# 1091 "parser/pars.fsy"
                 : 'minusExpr));
# 3883 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1093 "parser/pars.fsy"
                             _1 
                   )
# 1093 "parser/pars.fsy"
                 : 'minusExpr));
# 3894 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'argExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1097 "parser/pars.fsy"
                             Expr_app (_1,_2,union_ranges (range_of_synexpr _1) (range_of_synexpr _2))  
                   )
# 1097 "parser/pars.fsy"
                 : 'appExpr));
# 3906 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1099 "parser/pars.fsy"
                             let arg,_ = _1 in 
                             arg 
                   )
# 1099 "parser/pars.fsy"
                 : 'appExpr));
# 3918 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1104 "parser/pars.fsy"
                             
                             let arg2,hpa2 = _2 in 
                               begin
                                 if hpa2 then 
                                   (report_parse_error_at (rhs parseState 1) "Successive arguments should be separated by spaces or tupled, and arguments involving function or method applications should be parenthesized");
                               end;
                               mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr arg2)) ("~"^(_1)) arg2 
                           
                   )
# 1104 "parser/pars.fsy"
                 : 'argExpr));
# 3937 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1113 "parser/pars.fsy"
                             
                             let arg,hpa = _1 in 
                               begin
                                 if hpa then 
                                   report_parse_error_at (range_of_synexpr arg) "Successive arguments should be separated by spaces or tupled, and arguments involving function or method applications should be parenthesized";
                               end;
                               arg 
                           
                   )
# 1113 "parser/pars.fsy"
                 : 'argExpr));
# 3955 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1124 "parser/pars.fsy"
                             let arg1,_ = _1 in 
                             let arg2,_ = _3 in 
                             Expr_app (arg1,arg2,union_ranges (range_of_synexpr arg1) (range_of_synexpr arg2)),true  
                   )
# 1124 "parser/pars.fsy"
                 : 'atomicExpr));
# 3969 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeArgs)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1129 "parser/pars.fsy"
                             let arg1,_ = _1 in 
                             Expr_tyapp (arg1,_2,lhs(parseState)),true  
                   )
# 1129 "parser/pars.fsy"
                 : 'atomicExpr));
# 3982 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1133 "parser/pars.fsy"
                             let arg2,hpa2 = _2 in 
                             mksyn_prefix (union_ranges (rhs parseState 1) (range_of_synexpr arg2)) _1 arg2,hpa2 
                   )
# 1133 "parser/pars.fsy"
                 : 'atomicExpr));
# 3995 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExprQualification)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1137 "parser/pars.fsy"
                             let arg1,hpa1 = _1 in 
                             _3 arg1 (lhs(parseState)) (rhs parseState 2),hpa1 
                   )
# 1137 "parser/pars.fsy"
                 : 'atomicExpr));
# 4008 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1141 "parser/pars.fsy"
                             Expr_id_get (_1),false 
                   )
# 1141 "parser/pars.fsy"
                 : 'atomicExpr));
# 4019 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'listExprElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1144 "parser/pars.fsy"
                             Expr_list(_2, rhs2 parseState 1 3), false 
                   )
# 1144 "parser/pars.fsy"
                 : 'atomicExpr));
# 4030 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExprAfterType)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1154 "parser/pars.fsy"
                             _1,false 
                   )
# 1154 "parser/pars.fsy"
                 : 'atomicExpr));
# 4041 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1158 "parser/pars.fsy"
                             [] 
                   )
# 1158 "parser/pars.fsy"
                 : 'listExprElements));
# 4051 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1161 "parser/pars.fsy"
                             let x, _ = _1 in [x] 
                   )
# 1161 "parser/pars.fsy"
                 : 'listExprElements));
# 4062 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'listExprElements)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1164 "parser/pars.fsy"
                             let x, _ = _1 in x::_3 
                   )
# 1164 "parser/pars.fsy"
                 : 'listExprElements));
# 4074 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1168 "parser/pars.fsy"
                             let idm = rhs parseState 1 in 
                             (fun e lhsm dotm -> 
                                qualifyNameIfAlongside parseState dotm idm; 
                                let res = mksyn_dot lhsm e _1 in
                                (* let _ = Printf.printf "Parser: making dot expression from %A\n got %A\n" (e, _1) res in *)
                                  res)
                   )
# 1168 "parser/pars.fsy"
                 : 'atomicExprQualification));
# 4090 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'constant)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1182 "parser/pars.fsy"
                             Expr_const (_1,range_of_synconst _1 (lhs(parseState))) 
                   )
# 1182 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4101 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1184 "parser/pars.fsy"
                             _1 
                   )
# 1184 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4112 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'braceExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1186 "parser/pars.fsy"
                             _1 
                   )
# 1186 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4123 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1188 "parser/pars.fsy"
                             Expr_const(Const_bool false,lhs(parseState)) 
                   )
# 1188 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4133 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1190 "parser/pars.fsy"
                             Expr_const(Const_bool true,lhs(parseState)) 
                   )
# 1190 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4143 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'beginEndExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1192 "parser/pars.fsy"
                             _1 
                   )
# 1192 "parser/pars.fsy"
                 : 'atomicExprAfterType));
# 4154 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1196 "parser/pars.fsy"
                             Expr_paren(_2,rhs2 parseState 1 3) 
                   )
# 1196 "parser/pars.fsy"
                 : 'beginEndExpr));
# 4165 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenExprBody)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1208 "parser/pars.fsy"
                             StartParameters parseState (rhspos parseState 1); EndParameters parseState (rhspos parseState 3); MatchPair parseState 1 3; _2 (rhs2 parseState 1 3) 
                   )
# 1208 "parser/pars.fsy"
                 : 'parenExpr));
# 4176 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenExprBody)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'recover)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1210 "parser/pars.fsy"
                           StartParameters parseState (rhspos parseState 1); EndParameters parseState (rhspos parseState 3); report_parse_error_at (rhs parseState 1) "unmatched '('"; let lhsm = rhs2 parseState 1 2 in Expr_paren(_2 lhsm,lhsm) 
                   )
# 1210 "parser/pars.fsy"
                 : 'parenExpr));
# 4188 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1220 "parser/pars.fsy"
                              (fun m -> Expr_const(Const_unit,m)) 
                   )
# 1220 "parser/pars.fsy"
                 : 'parenExprBody));
# 4198 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1222 "parser/pars.fsy"
                             (fun m -> Expr_paren(_1,m)) 
                   )
# 1222 "parser/pars.fsy"
                 : 'parenExprBody));
# 4209 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'braceExprBody)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1226 "parser/pars.fsy"
                             MatchPair parseState 1 3; _2 (lhs(parseState)) 
                   )
# 1226 "parser/pars.fsy"
                 : 'braceExpr));
# 4220 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1236 "parser/pars.fsy"
                             (fun m -> let b,c = _1 in Expr_recd(b,c,m)) 
                   )
# 1236 "parser/pars.fsy"
                 : 'braceExprBody));
# 4231 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1239 "parser/pars.fsy"
                            None 
                   )
# 1239 "parser/pars.fsy"
                 : 'opt_atomicExprAfterType));
# 4241 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExprAfterType)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1240 "parser/pars.fsy"
                                                Some(_1) 
                   )
# 1240 "parser/pars.fsy"
                 : 'opt_atomicExprAfterType));
# 4252 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1245 "parser/pars.fsy"
                            (None, []) 
                   )
# 1245 "parser/pars.fsy"
                 : 'recdExpr));
# 4262 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdExprBindings)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1249 "parser/pars.fsy"
                            match _1 with 
                              | Expr_id_get id -> 
                                  (None, (FrontAndBack [id],_3) :: List.rev _4) 
                              | _ -> raise_parse_error_at (rhs parseState 2) "field bindings must have the form 'id = expr;'" 
                   )
# 1249 "parser/pars.fsy"
                 : 'recdExpr));
# 4279 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdExprBindings)) in
            let _7 = (let data = parseState.GetInput(7) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1255 "parser/pars.fsy"
                             (Some _1,(FrontAndBack _3,_5):: List.rev _6) 
                   )
# 1255 "parser/pars.fsy"
                 : 'recdExpr));
# 4294 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdExprBindings)) in
            let _7 = (let data = parseState.GetInput(7) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1257 "parser/pars.fsy"
                             (Some _1,(FrontAndBack _3,_5):: List.rev _6) 
                   )
# 1257 "parser/pars.fsy"
                 : 'recdExpr));
# 4309 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdExprBindings)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'declExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1260 "parser/pars.fsy"
                                                                      (FrontAndBack _3,_5) :: _1 
                   )
# 1260 "parser/pars.fsy"
                 : 'recdExprBindings));
# 4323 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1261 "parser/pars.fsy"
                                                                      [] 
                   )
# 1261 "parser/pars.fsy"
                 : 'recdExprBindings));
# 4333 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExprBlock)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1265 "parser/pars.fsy"
                            let mAll = union_ranges (rhs parseState 1) (range_of_synexpr _4) in
                            mksyn_match_lambdas false mAll _2 _4 
                   )
# 1265 "parser/pars.fsy"
                 : 'anonLambdaExpr));
# 4346 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1268 "parser/pars.fsy"
                            let mAll = rhs2 parseState 1 3 in
                            mksyn_match_lambdas false mAll _2 (arbExpr(parseState)) 
                   )
# 1268 "parser/pars.fsy"
                 : 'anonLambdaExpr));
# 4358 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicPatterns)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'typedSeqExprBlockR)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1271 "parser/pars.fsy"
                            let mAll = union_ranges (rhs parseState 1) (range_of_synexpr _4) in
                            mksyn_match_lambdas false mAll _2 _4 
                   )
# 1271 "parser/pars.fsy"
                 : 'anonLambdaExpr));
# 4371 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_bar)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1280 "parser/pars.fsy"
                             let clauses,mLast = _3 in
                             let mAll = union_ranges (rhs parseState 1) mLast in
                             mksyn_match_lambda(false,false,mAll,clauses) 
                   )
# 1280 "parser/pars.fsy"
                 : 'anonMatchingExpr));
# 4385 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_bar)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'patternClauses)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1285 "parser/pars.fsy"
                             let clauses,mLast = _3 in
                             let mAll = union_ranges (rhs parseState 1) mLast in
                             mksyn_match_lambda(false,false,mAll,clauses) 
                   )
# 1285 "parser/pars.fsy"
                 : 'anonMatchingExpr));
# 4399 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefn)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefnList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1302 "parser/pars.fsy"
                            _2 :: _3 
                   )
# 1302 "parser/pars.fsy"
                 : 'tyconDefnList));
# 4411 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1304 "parser/pars.fsy"
                             [] 
                   )
# 1304 "parser/pars.fsy"
                 : 'tyconDefnList));
# 4421 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeNameInfo)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1308 "parser/pars.fsy"
                            
                            TyconDefn(_1, TyconCore_repr_hidden(lhs(parseState)),lhs(parseState))
                          
                   )
# 1308 "parser/pars.fsy"
                 : 'tyconDefn));
# 4434 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeNameInfo)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconDefnRhs)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1313 "parser/pars.fsy"
                            
                            let tcDefRepr= _3 in
                            TyconDefn(_1,tcDefRepr,lhs(parseState)) (* <-- mark changed *) 
                          
                   )
# 1313 "parser/pars.fsy"
                 : 'tyconDefn));
# 4449 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tyconNameAndTyparDecls)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1321 "parser/pars.fsy"
                             StartName parseState (rhs parseState 1);
                             let tkind, tparams, tpath, preferPostfix = _1 in 
                               Tycon(tpath, tkind, tparams, preferPostfix, rhs2 parseState 1 1)  
                   )
# 1321 "parser/pars.fsy"
                 : 'typeNameInfo));
# 4462 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1327 "parser/pars.fsy"
                             
                             StartName parseState (rhs parseState 1); 
                             _2, [], _1, false 
                           
                   )
# 1327 "parser/pars.fsy"
                 : 'tyconNameAndTyparDecls));
# 4477 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'postfixTyparDecls)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1333 "parser/pars.fsy"
                             
                             StartName parseState (rhs parseState 1); 
                             _3, _2, _1,true 
                           
                   )
# 1333 "parser/pars.fsy"
                 : 'tyconNameAndTyparDecls));
# 4493 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1339 "parser/pars.fsy"
                                            Kind_star 
                   )
# 1339 "parser/pars.fsy"
                 : 'opt_kind));
# 4503 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1340 "parser/pars.fsy"
                                            _2 
                   )
# 1340 "parser/pars.fsy"
                 : 'opt_kind));
# 4514 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'base_kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1343 "parser/pars.fsy"
                                     _1 
                   )
# 1343 "parser/pars.fsy"
                 : 'kind));
# 4525 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'base_kind)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1344 "parser/pars.fsy"
                                                  Kind_tcon(None, _1, _3) 
                   )
# 1344 "parser/pars.fsy"
                 : 'kind));
# 4537 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typ_with_binding)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1346 "parser/pars.fsy"
                            
                             let (bopt, t) = _1 in
                             let k' = _3 in 
                               Kind_dcon(bopt, t, k') 
                           
                   )
# 1346 "parser/pars.fsy"
                 : 'kind));
# 4553 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomic_kind_with_binding)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1352 "parser/pars.fsy"
                            
                             let (bopt, k) = _1 in
                             let k' = _3 in 
                               Kind_tcon(bopt, k, k') 
                           
                   )
# 1352 "parser/pars.fsy"
                 : 'kind));
# 4569 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'base_kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1360 "parser/pars.fsy"
                                                                           (Some(_3), _5) 
                   )
# 1360 "parser/pars.fsy"
                 : 'atomic_kind_with_binding));
# 4581 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1361 "parser/pars.fsy"
                                              (None, _2) 
                   )
# 1361 "parser/pars.fsy"
                 : 'atomic_kind_with_binding));
# 4592 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1364 "parser/pars.fsy"
                                (None, _1) 
                   )
# 1364 "parser/pars.fsy"
                 : 'typ_with_binding));
# 4603 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1365 "parser/pars.fsy"
                                           (Some _1, _3) 
                   )
# 1365 "parser/pars.fsy"
                 : 'typ_with_binding));
# 4615 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1368 "parser/pars.fsy"
                                              Kind_star 
                   )
# 1368 "parser/pars.fsy"
                 : 'base_kind));
# 4625 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1369 "parser/pars.fsy"
                                              Kind_affine 
                   )
# 1369 "parser/pars.fsy"
                 : 'base_kind));
# 4636 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1370 "parser/pars.fsy"
                                     
                           match _1 with 
                             | "*" -> Kind_star 
                             | "A" -> Kind_affine
                             | "E" -> Kind_erasable 
                             | "P" -> Kind_prop 
                             | _ -> error (Error("Unrecognized kind", lhs(parseState)));
                         
                   )
# 1370 "parser/pars.fsy"
                 : 'base_kind));
# 4654 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_HIGH_PRECEDENCE_TYAPP)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeparDeclsActual)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1381 "parser/pars.fsy"
                             _2 
                   )
# 1381 "parser/pars.fsy"
                 : 'postfixTyparDecls));
# 4666 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1385 "parser/pars.fsy"
                            [TyparDecl _2] 
                   )
# 1385 "parser/pars.fsy"
                 : 'typeparDeclsActual));
# 4677 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'typarIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1387 "parser/pars.fsy"
                            
                            (TyparDecl _2)::(TyparDecl _4)::_5
                          
                   )
# 1387 "parser/pars.fsy"
                 : 'typeparDeclsActual));
# 4692 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1391 "parser/pars.fsy"
                            
                            (TyparDecl _2)::_4 
                          
                   )
# 1391 "parser/pars.fsy"
                 : 'typeparDeclsActual));
# 4706 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1395 "parser/pars.fsy"
                             _3 
                   )
# 1395 "parser/pars.fsy"
                 : 'typeparDeclsActual));
# 4717 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1398 "parser/pars.fsy"
                                 [_1] 
                   )
# 1398 "parser/pars.fsy"
                 : 'onlyTypars));
# 4728 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTyparList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1399 "parser/pars.fsy"
                                                            _1::_3::_4 
                   )
# 1399 "parser/pars.fsy"
                 : 'onlyTypars));
# 4741 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1402 "parser/pars.fsy"
                                 [] 
                   )
# 1402 "parser/pars.fsy"
                 : 'onlyTyparList));
# 4751 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTyparList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1403 "parser/pars.fsy"
                                                     _2::_3 
                   )
# 1403 "parser/pars.fsy"
                 : 'onlyTyparList));
# 4763 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typarIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1407 "parser/pars.fsy"
                                                    TyparDecl _2 :: _3 
                   )
# 1407 "parser/pars.fsy"
                 : 'typarIndices));
# 4775 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1408 "parser/pars.fsy"
                                                     _2 
                   )
# 1408 "parser/pars.fsy"
                 : 'typarIndices));
# 4786 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1409 "parser/pars.fsy"
                                                  [] 
                   )
# 1409 "parser/pars.fsy"
                 : 'typarIndices));
# 4796 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparDecl)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1412 "parser/pars.fsy"
                                      [_1]
                   )
# 1412 "parser/pars.fsy"
                 : 'termparIndices));
# 4807 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparDecl)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreTermParIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1413 "parser/pars.fsy"
                                                          _1::_2 
                   )
# 1413 "parser/pars.fsy"
                 : 'termparIndices));
# 4819 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1416 "parser/pars.fsy"
                            [] 
                   )
# 1416 "parser/pars.fsy"
                 : 'moreTermParIndices));
# 4829 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'termparDecl)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'moreTermParIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1417 "parser/pars.fsy"
                                                                _2::_3 
                   )
# 1417 "parser/pars.fsy"
                 : 'moreTermParIndices));
# 4841 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1420 "parser/pars.fsy"
                                             TermparDecl(_1, _3) 
                   )
# 1420 "parser/pars.fsy"
                 : 'termparDecl));
# 4853 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1426 "parser/pars.fsy"
                            
                            TyconCore_abbrev (_1, lhs(parseState)) 
                          
                   )
# 1426 "parser/pars.fsy"
                 : 'tyconDefnRhs));
# 4866 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'unionRepr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1431 "parser/pars.fsy"
                            
                            TyconCore_union (_1, lhs(parseState)) 
                          
                   )
# 1431 "parser/pars.fsy"
                 : 'tyconDefnRhs));
# 4879 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'braceFieldDeclList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1436 "parser/pars.fsy"
                            
                            TyconCore_recd (_1,lhs(parseState)) 
                          
                   )
# 1436 "parser/pars.fsy"
                 : 'tyconDefnRhs));
# 4892 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'unionCaseDecls)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1443 "parser/pars.fsy"
                            _2 
                   )
# 1443 "parser/pars.fsy"
                 : 'unionRepr));
# 4903 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'unionCaseDecl)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'unionCaseDecls)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1446 "parser/pars.fsy"
                                                             _1::_3 
                   )
# 1446 "parser/pars.fsy"
                 : 'unionCaseDecls));
# 4915 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'unionCaseDecl)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1447 "parser/pars.fsy"
                                         ([ _1 ]) 
                   )
# 1447 "parser/pars.fsy"
                 : 'unionCaseDecls));
# 4926 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_OBLOCKSEP)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1451 "parser/pars.fsy"
                             
                             let mDecl = rhs parseState 1 in
                               StartName parseState (rhs parseState 1); 
                               UnionConstr (_1, ConstrFields [], mDecl)
                           
                   )
# 1451 "parser/pars.fsy"
                 : 'unionCaseDecl));
# 4942 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1458 "parser/pars.fsy"
                            
                             UnionConstr (_1, ConstrType _3, rhs2 parseState 1 3)
                           
                   )
# 1458 "parser/pars.fsy"
                 : 'unionCaseDecl));
# 4956 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTypars)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1463 "parser/pars.fsy"
                            
                             UnionConstr (_1, ConstrType (Type_qtyp(_3, _6, rhs2 parseState 3 6)), 
                                         rhs2 parseState 1 6)
                           
                   )
# 1463 "parser/pars.fsy"
                 : 'unionCaseDecl));
# 4972 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1469 "parser/pars.fsy"
                            
                             UnionConstr(_1, ConstrMLType _3, rhs2 parseState 1 3)
                           
                   )
# 1469 "parser/pars.fsy"
                 : 'unionCaseDecl));
# 4986 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdFieldDeclList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1476 "parser/pars.fsy"
                            MatchPair parseState 1 3;   _2 
                   )
# 1476 "parser/pars.fsy"
                 : 'braceFieldDeclList));
# 4997 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdFieldDeclList)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'recover)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1478 "parser/pars.fsy"
                            _2 
                   )
# 1478 "parser/pars.fsy"
                 : 'braceFieldDeclList));
# 5009 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1480 "parser/pars.fsy"
                            MatchPair parseState 1 3;   [] 
                   )
# 1480 "parser/pars.fsy"
                 : 'braceFieldDeclList));
# 5019 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdFieldDecl)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdFieldDeclList)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1484 "parser/pars.fsy"
                            _1 :: _3 
                   )
# 1484 "parser/pars.fsy"
                 : 'recdFieldDeclList));
# 5032 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'recdFieldDecl)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'opt_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1486 "parser/pars.fsy"
                            [_1] 
                   )
# 1486 "parser/pars.fsy"
                 : 'recdFieldDeclList));
# 5044 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1490 "parser/pars.fsy"
                             
                             StartName parseState (rhs parseState 1); 
                             let rhsm = rhs2 parseState 1 3 in 
                               Field(_1, _3, rhsm) 
                           
                   )
# 1490 "parser/pars.fsy"
                 : 'recdFieldDecl));
# 5060 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1501 "parser/pars.fsy"
                             None 
                   )
# 1501 "parser/pars.fsy"
                 : 'opt_topReturnType));
# 5070 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1503 "parser/pars.fsy"
                             
                             let ty = _2 in 
                               Some (ty, rhs parseState 2)
                           
                   )
# 1503 "parser/pars.fsy"
                 : 'opt_topReturnType));
# 5084 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1510 "parser/pars.fsy"
                            
                             Type_affine(_2, lhs parseState)
                           
                   )
# 1510 "parser/pars.fsy"
                 : Sugar.typ));
# 5097 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appType)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1515 "parser/pars.fsy"
                             
                             let tt = _1 in 
                               match tt with 
                                 | Type_refinement(n, _, _, _, _) -> 
                                     Type_fun((Some n, tt), _3, rhs2 parseState 1 4)
                                 | _ -> Type_fun((None, tt), _3, lhs(parseState)) 
                           
                   )
# 1515 "parser/pars.fsy"
                 : Sugar.typ));
# 5115 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'appType)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1524 "parser/pars.fsy"
                             
                             let nopt = Some _1 in
                             let tt = _3 in 
                               Type_fun((nopt, tt), _5, lhs(parseState)) 
                           
                   )
# 1524 "parser/pars.fsy"
                 : Sugar.typ));
# 5132 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _6 = (let data = parseState.GetInput(6) in (Microsoft.FSharp.Core.Operators.unbox data : 'formula)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1541 "parser/pars.fsy"
                            
                             Type_refinement(_2, _4, _6, false, rhs2 parseState 1 7)
                           
                   )
# 1541 "parser/pars.fsy"
                 : Sugar.typ));
# 5147 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'formula)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1546 "parser/pars.fsy"
                            
                             Type_refinement(_1, _3, _5, true, rhs2 parseState 1 6)
                           
                   )
# 1546 "parser/pars.fsy"
                 : Sugar.typ));
# 5162 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'appType)) in
            let _8 = (let data = parseState.GetInput(8) in (Microsoft.FSharp.Core.Operators.unbox data : 'tfun)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1551 "parser/pars.fsy"
                            
                             Type_lam (_3, _5, _8, rhs2 parseState 1 8)
                           
                   )
# 1551 "parser/pars.fsy"
                 : Sugar.typ));
# 5177 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appType)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1556 "parser/pars.fsy"
                             _1 
                   )
# 1556 "parser/pars.fsy"
                 : Sugar.typ));
# 5188 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'appType)) in
            let _8 = (let data = parseState.GetInput(8) in (Microsoft.FSharp.Core.Operators.unbox data : 'tfun)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1561 "parser/pars.fsy"
                             
                             Type_lam (_3, _5, _8, rhs2 parseState 1 8)
                           
                   )
# 1561 "parser/pars.fsy"
                 : 'tfun));
# 5203 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'formula)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1566 "parser/pars.fsy"
                            
                             Type_formula (_2, rhs parseState 2)
                           
                   )
# 1566 "parser/pars.fsy"
                 : 'tfun));
# 5216 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1571 "parser/pars.fsy"
                             _2 
                   )
# 1571 "parser/pars.fsy"
                 : 'tfun));
# 5227 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typars)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'appTypeCon)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1575 "parser/pars.fsy"
                             
                             let tt = _1 in 
                               Type_app(_2,tt,lhs(parseState)) 
                           
                   )
# 1575 "parser/pars.fsy"
                 : 'appType));
# 5242 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'onlyTypars)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1580 "parser/pars.fsy"
                             Type_qtyp(_2, _5, rhs2 parseState 1 3) 
                   )
# 1580 "parser/pars.fsy"
                 : 'appType));
# 5254 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomType)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1582 "parser/pars.fsy"
                             _1 
                   )
# 1582 "parser/pars.fsy"
                 : 'appType));
# 5265 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1585 "parser/pars.fsy"
                               [_1] 
                   )
# 1585 "parser/pars.fsy"
                 : 'typars));
# 5276 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1586 "parser/pars.fsy"
                                                         _2::_3 
                   )
# 1586 "parser/pars.fsy"
                 : 'typars));
# 5288 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleType)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1595 "parser/pars.fsy"
                              
                             let t = _2 in 
                               match t with 
                                   [(None, t)] -> t
                                 | _ ->  Type_tuple(_2, rhs2 parseState 1 3) 
                           
                   )
# 1595 "parser/pars.fsy"
                 : 'atomType));
# 5304 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appTypeCon)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1610 "parser/pars.fsy"
                            _1 
                   )
# 1610 "parser/pars.fsy"
                 : 'atomType));
# 5315 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'appTypeCon)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeArgs)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1612 "parser/pars.fsy"
                            Type_app(_1,_2,lhs(parseState)) 
                   )
# 1612 "parser/pars.fsy"
                 : 'atomType));
# 5327 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1616 "parser/pars.fsy"
                                [(None, _1)] 
                   )
# 1616 "parser/pars.fsy"
                 : 'tupleType));
# 5338 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleTypeTail)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1617 "parser/pars.fsy"
                                                  (None, _1)::_3 
                   )
# 1617 "parser/pars.fsy"
                 : 'tupleType));
# 5350 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'dtupleTypeTail)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1618 "parser/pars.fsy"
                                                               (Some _1, _3)::_5 
                   )
# 1618 "parser/pars.fsy"
                 : 'tupleType));
# 5363 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1621 "parser/pars.fsy"
                               [(None, _1)] 
                   )
# 1621 "parser/pars.fsy"
                 : 'dtupleTypeTail));
# 5374 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'dtupleTypeTail)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1622 "parser/pars.fsy"
                                                               (Some _1, _3)::_5 
                   )
# 1622 "parser/pars.fsy"
                 : 'dtupleTypeTail));
# 5387 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1625 "parser/pars.fsy"
                               [(None, _1)] 
                   )
# 1625 "parser/pars.fsy"
                 : 'tupleTypeTail));
# 5398 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleTypeTail)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1626 "parser/pars.fsy"
                                                  (None, _1)::_3 
                   )
# 1626 "parser/pars.fsy"
                 : 'tupleTypeTail));
# 5410 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeArgsActual)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1632 "parser/pars.fsy"
                            _1 
                   )
# 1632 "parser/pars.fsy"
                 : 'typeArgs));
# 5421 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeArgsActual)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1634 "parser/pars.fsy"
                            _2 
                   )
# 1634 "parser/pars.fsy"
                 : 'typeArgs));
# 5432 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1639 "parser/pars.fsy"
                            [_2] 
                   )
# 1639 "parser/pars.fsy"
                 : 'typeArgsActual));
# 5443 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _5 = (let data = parseState.GetInput(5) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1641 "parser/pars.fsy"
                            _2::_4::_5 
                   )
# 1641 "parser/pars.fsy"
                 : 'typeArgsActual));
# 5456 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1643 "parser/pars.fsy"
                            _2::_4 
                   )
# 1643 "parser/pars.fsy"
                 : 'typeArgsActual));
# 5468 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1645 "parser/pars.fsy"
                             _3 
                   )
# 1645 "parser/pars.fsy"
                 : 'typeArgsActual));
# 5479 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1648 "parser/pars.fsy"
                                                 _2 :: _3 
                   )
# 1648 "parser/pars.fsy"
                 : 'typeIndices));
# 5491 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1649 "parser/pars.fsy"
                                                  _2 
                   )
# 1649 "parser/pars.fsy"
                 : 'typeIndices));
# 5502 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1650 "parser/pars.fsy"
                                                  [] 
                   )
# 1650 "parser/pars.fsy"
                 : 'typeIndices));
# 5512 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1653 "parser/pars.fsy"
                                     [Type_term_index (fst _1, lhs parseState)]
                   )
# 1653 "parser/pars.fsy"
                 : 'termIndices));
# 5523 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicExpr)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndices)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1655 "parser/pars.fsy"
                             
                             let ti = Type_term_index (fst _1, lhs parseState) in
                               ti::_3 
                           
                   )
# 1655 "parser/pars.fsy"
                 : 'termIndices));
# 5538 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1662 "parser/pars.fsy"
                             Type_lid(_1, lhs(parseState)) 
                   )
# 1662 "parser/pars.fsy"
                 : 'appTypeCon));
# 5549 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'typar)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1665 "parser/pars.fsy"
                            Type_var(_1, lhs(parseState)) 
                   )
# 1665 "parser/pars.fsy"
                 : 'appTypeCon));
# 5560 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1669 "parser/pars.fsy"
                           
                            let id = mksyn_id (lhs(parseState)) (_2).idText in
                             Typar(id, None) 
                          
                   )
# 1669 "parser/pars.fsy"
                 : 'typar));
# 5574 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1675 "parser/pars.fsy"
                             
                            let id = mksyn_id (lhs(parseState)) (_2).idText in
                             Typar(id, Some _4) 
                          
                   )
# 1675 "parser/pars.fsy"
                 : 'typar));
# 5589 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomic_kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1681 "parser/pars.fsy"
                             
                            let id = mksyn_id (lhs(parseState)) (_2).idText in
                             Typar(id, Some _4) 
                          
                   )
# 1681 "parser/pars.fsy"
                 : 'typar));
# 5604 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'kind)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1687 "parser/pars.fsy"
                                              _2 
                   )
# 1687 "parser/pars.fsy"
                 : 'atomic_kind));
# 5615 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1693 "parser/pars.fsy"
                                    _1
                   )
# 1693 "parser/pars.fsy"
                 : 'formula));
# 5626 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'Forall)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1695 "parser/pars.fsy"
                                                 _2 
                   )
# 1695 "parser/pars.fsy"
                 : 'Form));
# 5637 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'Exists)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1696 "parser/pars.fsy"
                                                 _2 
                   )
# 1696 "parser/pars.fsy"
                 : 'Form));
# 5648 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form0)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1697 "parser/pars.fsy"
                                          _1 
                   )
# 1697 "parser/pars.fsy"
                 : 'Form));
# 5659 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1699 "parser/pars.fsy"
                                                      Form_impl(_1,_3, rhs2 parseState 1 3) 
                   )
# 1699 "parser/pars.fsy"
                 : 'Form0));
# 5671 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form0a)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1700 "parser/pars.fsy"
                                           _1 
                   )
# 1700 "parser/pars.fsy"
                 : 'Form0));
# 5682 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1703 "parser/pars.fsy"
                               
                                let f1, f2 = _1, _3 in
                                let m = rhs2 parseState 1 3 in
                                  Form_conj(Form_impl(f1, f2, m), Form_impl(f2, f1, m), m) 
                              
                   )
# 1703 "parser/pars.fsy"
                 : 'Form0a));
# 5698 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1708 "parser/pars.fsy"
                                          _1 
                   )
# 1708 "parser/pars.fsy"
                 : 'Form0a));
# 5709 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form2)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form1)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1710 "parser/pars.fsy"
                                                                      Form_disj(_1, _3, rhs2 parseState 1 3) 
                   )
# 1710 "parser/pars.fsy"
                 : 'Form1));
# 5721 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form2)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1711 "parser/pars.fsy"
                                          _1 
                   )
# 1711 "parser/pars.fsy"
                 : 'Form1));
# 5732 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form3)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form2)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1713 "parser/pars.fsy"
                                                                    Form_conj(_1, _3, rhs2 parseState 1 3) 
                   )
# 1713 "parser/pars.fsy"
                 : 'Form2));
# 5744 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form3)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1714 "parser/pars.fsy"
                                          _1 
                   )
# 1714 "parser/pars.fsy"
                 : 'Form2));
# 5755 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1716 "parser/pars.fsy"
                                                           Form_not(_2, lhs parseState) 
                   )
# 1716 "parser/pars.fsy"
                 : 'Form3));
# 5766 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form3a)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1717 "parser/pars.fsy"
                                           _1 
                   )
# 1717 "parser/pars.fsy"
                 : 'Form3));
# 5777 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1721 "parser/pars.fsy"
                               
                                let opeq = Type_lid(eq_lid, lhs parseState) in
                                let v1 = Type_term_index (_1, lhs parseState) in
                                let v2 = Type_term_index (_3, lhs parseState) in
                                let eqpred = Type_app(opeq, [Type_anon (lhs parseState); v1;v2], rhs2 parseState 1 3) in
                                  Form_atom(Atom_pred(eqpred), rhs2 parseState 1 3)
                              
                   )
# 1721 "parser/pars.fsy"
                 : 'Form3a));
# 5795 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1729 "parser/pars.fsy"
                               
                                let opeq = Type_lid(eq_lid, lhs parseState) in
                                let v1 = Type_term_index (_2, lhs parseState) in
                                let v2 = Type_term_index (_3, lhs parseState) in
                                let eqpred = Type_app(opeq, [Type_anon (lhs parseState); v1;v2], rhs2 parseState 1 3) in
                                  Form_atom(Atom_pred(eqpred), rhs2 parseState 1 3)
                              
                   )
# 1729 "parser/pars.fsy"
                 : 'Form3a));
# 5813 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form4)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1736 "parser/pars.fsy"
                                                                                       _1 
                   )
# 1736 "parser/pars.fsy"
                 : 'Form3a));
# 5824 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atom)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1738 "parser/pars.fsy"
                                                                                       Form_atom(_1, lhs parseState) 
                   )
# 1738 "parser/pars.fsy"
                 : 'Form4));
# 5835 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1739 "parser/pars.fsy"
                                                     _2 
                   )
# 1739 "parser/pars.fsy"
                 : 'Form4));
# 5846 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'fbinders)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1742 "parser/pars.fsy"
                            
                             let binders = _1 in 
                             let form = _3 in 
                               List.fold_right 
                                 (fun (x,t) form -> 
                                    Form_all (x, t, form, union_ranges x.idRange (rhs parseState 4), [], false))
                                 binders form
                           
                   )
# 1742 "parser/pars.fsy"
                 : 'Forall));
# 5865 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'fbinders)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'Form)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1752 "parser/pars.fsy"
                            
                             let binders = _1 in 
                             let form = _3 in 
                               List.fold_right 
                                 (fun (x,t) form -> 
                                    Form_exists (x, t, form, union_ranges x.idRange (rhs parseState 4), [], false))
                                 binders form
                           
                   )
# 1752 "parser/pars.fsy"
                 : 'Exists));
# 5884 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _4 = (let data = parseState.GetInput(4) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.typ)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1763 "parser/pars.fsy"
                             (_2, _4) 
                   )
# 1763 "parser/pars.fsy"
                 : 'fbinder));
# 5896 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'fbinder)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'fbinders)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1767 "parser/pars.fsy"
                             _1::_3 
                   )
# 1767 "parser/pars.fsy"
                 : 'fbinders));
# 5908 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'fbinder)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1769 "parser/pars.fsy"
                             [_1] 
                   )
# 1769 "parser/pars.fsy"
                 : 'fbinders));
# 5919 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndexSeq)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1773 "parser/pars.fsy"
                            
                             let r = (rhs parseState 2) in
                             let id = mksyn_id r (_2).idText in
                             let t = Type_var(Typar(id, None), r)  in
                             let indices = _3 in
                             let pred =  Type_app(t, indices, rhs2 parseState 1 2) in
                               Atom_pred(pred)
                           
                   )
# 1773 "parser/pars.fsy"
                 : 'atom));
# 5938 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndexSeq)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1782 "parser/pars.fsy"
                            
                             let t = Type_lid(_1, lhs(parseState)) in
                             let indices = _2 in
                             let pred =  Type_app(t, indices, rhs2 parseState 1 2) in
                               Atom_pred(pred)
                           
                   )
# 1782 "parser/pars.fsy"
                 : 'atom));
# 5955 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'typeArgsActual)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1790 "parser/pars.fsy"
                            
                             let t = Type_lid(_1, lhs(parseState)) in
                             let indices = _2 in
                             let pred = Type_app(t, indices, rhs2 parseState 1 2) in
                               Atom_pred(pred)
                           
                   )
# 1790 "parser/pars.fsy"
                 : 'atom));
# 5972 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1798 "parser/pars.fsy"
                            [] 
                   )
# 1798 "parser/pars.fsy"
                 : 'termIndexSeq));
# 5982 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'termIndexSeq)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1802 "parser/pars.fsy"
                            
                             let ti = Type_term_index (_1, lhs parseState) in
                               ti::_2
                           
                   )
# 1802 "parser/pars.fsy"
                 : 'termIndexSeq));
# 5997 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'valueSeq)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1809 "parser/pars.fsy"
                             
                             let args = _2 in
                               List.fold_left 
                                 (fun out v -> Expr_app(out, v, union_ranges (range_of_synexpr out) (range_of_synexpr v))) 
                                 _1 args
                           
                   )
# 1809 "parser/pars.fsy"
                 : 'value));
# 6014 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1816 "parser/pars.fsy"
                                       _1 
                   )
# 1816 "parser/pars.fsy"
                 : 'value));
# 6025 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1820 "parser/pars.fsy"
                                       [_1] 
                   )
# 1820 "parser/pars.fsy"
                 : 'valueSeq));
# 6036 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'atomicValue)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'valueSeq)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1821 "parser/pars.fsy"
                                                _1::_2 
                   )
# 1821 "parser/pars.fsy"
                 : 'valueSeq));
# 6048 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'valueConst)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1824 "parser/pars.fsy"
                                      _1 
                   )
# 1824 "parser/pars.fsy"
                 : 'atomicValue));
# 6059 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenValue)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1825 "parser/pars.fsy"
                                      _1 
                   )
# 1825 "parser/pars.fsy"
                 : 'atomicValue));
# 6070 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1829 "parser/pars.fsy"
                             
                             Expr_lid_get(false, _1, lhs parseState)
                           
                   )
# 1829 "parser/pars.fsy"
                 : 'valueConst));
# 6083 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'constant)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1834 "parser/pars.fsy"
                            
                             Expr_const(_1, lhs parseState)
                           
                   )
# 1834 "parser/pars.fsy"
                 : 'valueConst));
# 6096 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'braceExpr)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1838 "parser/pars.fsy"
                             _1 
                   )
# 1838 "parser/pars.fsy"
                 : 'valueConst));
# 6107 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1840 "parser/pars.fsy"
                             Expr_const(Const_bool false,lhs(parseState)) 
                   )
# 1840 "parser/pars.fsy"
                 : 'valueConst));
# 6117 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1842 "parser/pars.fsy"
                             Expr_const(Const_bool true,lhs(parseState)) 
                   )
# 1842 "parser/pars.fsy"
                 : 'valueConst));
# 6127 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'parenValBody)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1847 "parser/pars.fsy"
                              _2 (rhs2 parseState 1 3) 
                   )
# 1847 "parser/pars.fsy"
                 : 'parenValue));
# 6138 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1851 "parser/pars.fsy"
                              (fun m -> Expr_const(Const_unit,m)) 
                   )
# 1851 "parser/pars.fsy"
                 : 'parenValBody));
# 6148 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'value)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1854 "parser/pars.fsy"
                             (fun m -> Expr_paren(_1,m)) 
                   )
# 1854 "parser/pars.fsy"
                 : 'parenValBody));
# 6159 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleVal)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'value)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1857 "parser/pars.fsy"
                             (fun m -> Expr_paren(Expr_tuple(((List.rev _1) @ [_3]),m), m)) 
                   )
# 1857 "parser/pars.fsy"
                 : 'parenValBody));
# 6171 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'value)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1861 "parser/pars.fsy"
                                 [_1] 
                   )
# 1861 "parser/pars.fsy"
                 : 'tupleVal));
# 6182 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'tupleVal)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'value)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1862 "parser/pars.fsy"
                                                     _3 :: _1 
                   )
# 1862 "parser/pars.fsy"
                 : 'tupleVal));
# 6194 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1871 "parser/pars.fsy"
                            ident(_1,rhs parseState 1) 
                   )
# 1871 "parser/pars.fsy"
                 : Sugar.ident));
# 6205 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1873 "parser/pars.fsy"
                            ident("not", rhs parseState 1) 
                   )
# 1873 "parser/pars.fsy"
                 : Sugar.ident));
# 6215 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1877 "parser/pars.fsy"
                             Some _1 
                   )
# 1877 "parser/pars.fsy"
                 : 'opt_ident_colon));
# 6226 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1878 "parser/pars.fsy"
                             None 
                   )
# 1878 "parser/pars.fsy"
                 : 'opt_ident_colon));
# 6236 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1882 "parser/pars.fsy"
                           (* StartName parseState (rhs parseState 1);  *)[_1] 
                   )
# 1882 "parser/pars.fsy"
                 : 'path));
# 6247 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'path)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : Sugar.ident)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1884 "parser/pars.fsy"
                            qualifyNameIfAlongside parseState (rhs parseState 2) (rhs parseState 3); (* silent recovery  *)_1 @ [_3] 
                   )
# 1884 "parser/pars.fsy"
                 : 'path));
# 6259 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'operatorName)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1892 "parser/pars.fsy"
                             MatchPair parseState 1 3; 
                             ident(CompileOpName _2,rhs parseState 2) 
                   )
# 1892 "parser/pars.fsy"
                 : 'opname));
# 6271 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1895 "parser/pars.fsy"
                             MatchPair parseState 1 1; 
                             ident(CompileOpName "*",rhs parseState 1) 
                   )
# 1895 "parser/pars.fsy"
                 : 'opname));
# 6282 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1899 "parser/pars.fsy"
                                     _1 
                   )
# 1899 "parser/pars.fsy"
                 : 'operatorName));
# 6293 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1900 "parser/pars.fsy"
                                               _1 
                   )
# 1900 "parser/pars.fsy"
                 : 'operatorName));
# 6304 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1901 "parser/pars.fsy"
                                            _1 
                   )
# 1901 "parser/pars.fsy"
                 : 'operatorName));
# 6315 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1902 "parser/pars.fsy"
                                            _1 
                   )
# 1902 "parser/pars.fsy"
                 : 'operatorName));
# 6326 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1903 "parser/pars.fsy"
                                         _1 
                   )
# 1903 "parser/pars.fsy"
                 : 'operatorName));
# 6337 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1904 "parser/pars.fsy"
                                        _1 
                   )
# 1904 "parser/pars.fsy"
                 : 'operatorName));
# 6348 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1905 "parser/pars.fsy"
                                          _1 
                   )
# 1905 "parser/pars.fsy"
                 : 'operatorName));
# 6359 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1906 "parser/pars.fsy"
                                                 _1 
                   )
# 1906 "parser/pars.fsy"
                 : 'operatorName));
# 6370 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1907 "parser/pars.fsy"
                                  "$" 
                   )
# 1907 "parser/pars.fsy"
                 : 'operatorName));
# 6380 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1908 "parser/pars.fsy"
                                                         _1 
                   )
# 1908 "parser/pars.fsy"
                 : 'operatorName));
# 6391 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1909 "parser/pars.fsy"
                                 "-" 
                   )
# 1909 "parser/pars.fsy"
                 : 'operatorName));
# 6401 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1910 "parser/pars.fsy"
                                "*" 
                   )
# 1910 "parser/pars.fsy"
                 : 'operatorName));
# 6411 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1911 "parser/pars.fsy"
                                  "=" 
                   )
# 1911 "parser/pars.fsy"
                 : 'operatorName));
# 6421 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1912 "parser/pars.fsy"
                              "or" 
                   )
# 1912 "parser/pars.fsy"
                 : 'operatorName));
# 6431 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1913 "parser/pars.fsy"
                                "<" 
                   )
# 1913 "parser/pars.fsy"
                 : 'operatorName));
# 6441 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1914 "parser/pars.fsy"
                                   ">" 
                   )
# 1914 "parser/pars.fsy"
                 : 'operatorName));
# 6451 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1915 "parser/pars.fsy"
                               "&" 
                   )
# 1915 "parser/pars.fsy"
                 : 'operatorName));
# 6461 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1916 "parser/pars.fsy"
                                   "&&" 
                   )
# 1916 "parser/pars.fsy"
                 : 'operatorName));
# 6471 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1917 "parser/pars.fsy"
                                   "||" 
                   )
# 1917 "parser/pars.fsy"
                 : 'operatorName));
# 6481 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1918 "parser/pars.fsy"
                                        ":=" 
                   )
# 1918 "parser/pars.fsy"
                 : 'operatorName));
# 6491 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1919 "parser/pars.fsy"
                                               deprecatedOperator (lhs(parseState)); _1 
                   )
# 1919 "parser/pars.fsy"
                 : 'operatorName));
# 6502 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1920 "parser/pars.fsy"
                                     deprecatedOperator (lhs(parseState)); "."^ _2 
                   )
# 1920 "parser/pars.fsy"
                 : 'operatorName));
# 6513 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1921 "parser/pars.fsy"
                                         _1 
                   )
# 1921 "parser/pars.fsy"
                 : 'operatorName));
# 6524 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1922 "parser/pars.fsy"
                                      _1 
                   )
# 1922 "parser/pars.fsy"
                 : 'operatorName));
# 6535 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1923 "parser/pars.fsy"
                                   deprecatedOperator (lhs(parseState)); ".." 
                   )
# 1923 "parser/pars.fsy"
                 : 'operatorName));
# 6545 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1924 "parser/pars.fsy"
                                           deprecatedOperator (lhs(parseState)); ".. .." 
                   )
# 1924 "parser/pars.fsy"
                 : 'operatorName));
# 6555 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string * bool)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : string * bool)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1926 "parser/pars.fsy"
                             
                             if _1 <> _2 then 
                               report_parse_error_at (rhs parseState 1) ("mismatched quotation operator name, beginning with '"^fst _1^"'");  
                             fst _1 
                           
                   )
# 1926 "parser/pars.fsy"
                 : 'operatorName));
# 6571 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : string)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1934 "parser/pars.fsy"
                             
                             if not (is_upper_string _1) then 
                               report_parse_error_at (rhs parseState 1) ("active pattern case identifiers must begin with an uppercase letter");  
                             _1 
                           
                   )
# 1934 "parser/pars.fsy"
                 : 'barName));
# 6586 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'barName)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1942 "parser/pars.fsy"
                             [_2] 
                   )
# 1942 "parser/pars.fsy"
                 : 'barNames));
# 6597 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'barNames)) in
            let _3 = (let data = parseState.GetInput(3) in (Microsoft.FSharp.Core.Operators.unbox data : 'barName)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1944 "parser/pars.fsy"
                             _3 :: _1 
                   )
# 1944 "parser/pars.fsy"
                 : 'barNames));
# 6609 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1947 "parser/pars.fsy"
                                     
                   )
# 1947 "parser/pars.fsy"
                 : 'top_sep));
# 6619 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1948 "parser/pars.fsy"
                                               
                   )
# 1948 "parser/pars.fsy"
                 : 'top_sep));
# 6629 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1949 "parser/pars.fsy"
                                     
                   )
# 1949 "parser/pars.fsy"
                 : 'top_sep));
# 6639 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_sep)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1952 "parser/pars.fsy"
                                                       
                   )
# 1952 "parser/pars.fsy"
                 : 'top_seps));
# 6650 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_sep)) in
            let _2 = (let data = parseState.GetInput(2) in (Microsoft.FSharp.Core.Operators.unbox data : 'top_seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1953 "parser/pars.fsy"
                                            
                   )
# 1953 "parser/pars.fsy"
                 : 'top_seps));
# 6662 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1956 "parser/pars.fsy"
                                     
                   )
# 1956 "parser/pars.fsy"
                 : 'seps));
# 6672 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1957 "parser/pars.fsy"
                                     
                   )
# 1957 "parser/pars.fsy"
                 : 'seps));
# 6682 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1958 "parser/pars.fsy"
                                               
                   )
# 1958 "parser/pars.fsy"
                 : 'seps));
# 6692 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1959 "parser/pars.fsy"
                                               
                   )
# 1959 "parser/pars.fsy"
                 : 'seps));
# 6702 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1962 "parser/pars.fsy"
                                     
                   )
# 1962 "parser/pars.fsy"
                 : 'opt_OBLOCKSEP));
# 6712 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1963 "parser/pars.fsy"
                                    
                   )
# 1963 "parser/pars.fsy"
                 : 'opt_OBLOCKSEP));
# 6722 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            let _1 = (let data = parseState.GetInput(1) in (Microsoft.FSharp.Core.Operators.unbox data : 'seps)) in
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1966 "parser/pars.fsy"
                                
                   )
# 1966 "parser/pars.fsy"
                 : 'opt_seps));
# 6733 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1967 "parser/pars.fsy"
                                
                   )
# 1967 "parser/pars.fsy"
                 : 'opt_seps));
# 6743 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1970 "parser/pars.fsy"
                               true 
                   )
# 1970 "parser/pars.fsy"
                 : 'opt_rec));
# 6753 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1971 "parser/pars.fsy"
                               false 
                   )
# 1971 "parser/pars.fsy"
                 : 'opt_rec));
# 6763 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1974 "parser/pars.fsy"
                               
                   )
# 1974 "parser/pars.fsy"
                 : 'opt_bar));
# 6773 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1975 "parser/pars.fsy"
                               
                   )
# 1975 "parser/pars.fsy"
                 : 'opt_bar));
# 6783 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1978 "parser/pars.fsy"
                                                 
                   )
# 1978 "parser/pars.fsy"
                 : 'opt_HIGH_PRECEDENCE_TYAPP));
# 6793 "parser\pars.fs"
        (fun (parseState : Microsoft.FSharp.Text.Parsing.IParseState) ->
            Microsoft.FSharp.Core.Operators.box
                (
                   (
# 1979 "parser/pars.fsy"
                              
                   )
# 1979 "parser/pars.fsy"
                 : 'opt_HIGH_PRECEDENCE_TYAPP));
|]
# 6804 "parser\pars.fs"
let tables () : Microsoft.FSharp.Text.Parsing.Tables<_> = 
  { reductions= _fsyacc_reductions ();
    endOfInputTag = _fsyacc_endOfInputTag;
    tagOfToken = tagOfToken;
    dataOfToken = _fsyacc_dataOfToken; 
    actionTableElements = _fsyacc_actionTableElements;
    actionTableRowOffsets = _fsyacc_actionTableRowOffsets;
    stateToProdIdxsTableElements = _fsyacc_stateToProdIdxsTableElements;
    stateToProdIdxsTableRowOffsets = _fsyacc_stateToProdIdxsTableRowOffsets;
    reductionSymbolCounts = _fsyacc_reductionSymbolCounts;
    immediateActions = _fsyacc_immediateActions;
    gotos = _fsyacc_gotos;
    sparseGotoTableRowOffsets = _fsyacc_sparseGotoTableRowOffsets;
    tagOfErrorTerminal = _fsyacc_tagOfErrorTerminal;
    parseError = (fun (ctxt:Microsoft.FSharp.Text.Parsing.ParseErrorContext<_>) -> 
                              match parse_error_rich with 
                              | Some f -> f ctxt
                              | None -> parse_error ctxt.Message);
    numTerminals = 194;
    productionToNonTerminalTable = _fsyacc_productionToNonTerminalTable  }
let engine lexer lexbuf startState = (tables ()).Interpret(lexer, lexbuf, startState)
let implementationFile lexer lexbuf : Sugar.ParsedImplFile =
    Microsoft.FSharp.Core.Operators.unbox ((tables ()).Interpret(lexer, lexbuf, 0))
